/*
 * Title:	Parallel Equation Solver: PES
 * Author:	Aleksandar Milenkovic
 * Date:	March, 1998.
 * Note:	
 * 
 *------------------------------------------------------------
 *    Command Line Options
 *     -pP: P = number of processors; must be a power of 2.
 *     -nN: N = Simulate NxN ocean: N must be (power of 2) + 2.
 *     -tD: D = Difference.
 *     -h : Print out command line options.
 *------------------------------------------------------------ 
 * */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PREF(X)  	asm(INSTRUCTION_PREF " %0" : "=m" (X));
#define PREFEX(X)	asm(INSTRUCTION_PREFEX " %0" : "=m" (X));
#define LPREF(X)	asm(INSTRUCTION_LPREF " %0" : "=m" (X));
#define WRBACK(X)	asm(INSTRUCTION_WRBACK " %0" : "=m" (X));


#define DEFAULT_P 1  


#define MAXPROCS 16
#define DEFAULT_N 258
#define DEFAULT_TOL 20

int numProcs=DEFAULT_P;   
int N=DEFAULT_N;
/*int numIteration=DEFAULT_I;*/
double TOL=DEFAULT_TOL;

/* truly shared vars */
LOCKDEC(indexLock);
int Index; /* Used by processors to determine which input to take */

LOCKDEC(diffLock);
double diff;

BARDEC(sBarrier1);
BARDEC(sBarrier2); 
double **grid;

/* private array */
double **scratch;
 
void SlaveStart();
int log_2(int);

void main(int argc, char*argv[]) {
    
    int i, j, m1;
    unsigned time;
    int c;
    extern char *optarg;
    
    while((c=getopt(argc, argv, "p:n:t:h")) != -1)
       {
         switch(c) {
           case 'p': numProcs=atoi(optarg);
                     if(numProcs<1) {
                       printf("P must be >=1\n");
                       exit(-1);
                     };
                     if(log_2(numProcs) == -1) {
                       printf("P must be a power of 2\n");
                       exit(-1);
                     };          	
                     break;       
           
           case 'n': N=atoi(optarg);
                     if(log_2(N-2) == -1) {
                       printf("N must be (power of 2) + 2\n");
                       exit(-1);
                     }
                     break;
                     
           case 't': TOL=atof(optarg);
                     if(TOL<=0) {
                       printf("TOL must be positive\n");
                       exit(-1);
                     }
                     break;
                              	       
           case 'h': printf("Usage mytestp <options>\n\n");
                     printf("options:\n");
                     printf("    -pP: P = number of processors; must be a power of 2.\n");
                     printf("    -nN: N = NxN Ocean.\n");
                     printf("    -tF: F = difference to check the convergence\n");
                     printf("    -h : Print out command line options.\n");
                     printf("\nDefault: pes -p%d -n%d -t%f\n", DEFAULT_P, DEFAULT_N, DEFAULT_TOL);   
                     exit(0);
                     break;
        }
    }

    /* Define shared matrix grid */
    grid = (double **) G_MALLOC(N*sizeof(double *));
    scratch = (double **) malloc(N*sizeof(double *));
    for(i=0; i<N; i++) {
       grid[i] = (double *) G_MALLOC(N*sizeof(double));
       scratch[i] = (double *) malloc(N*sizeof(double));
    } 
     
    /* Initialize grid */
    for(i=0; i<N; i++)
       for(j=0; j<N; j++) {
          grid[i][j] = (double) (rand()/1000);
          scratch[i][j] = grid[i][j];
          /*printf("grid[%d][%d]=%f\n", i,j,grid[i][j]); */      
       }
    /* Initialize the Index */
    Index = 0;

    /* Initialize the barriers and the lock */
    LOCKINIT(indexLock) 
    LOCKINIT(diffLock)   
    BARINIT(sBarrier1)
    BARINIT(sBarrier2)
        
    CLOCK(time);

    /* Create the slave processes. */
    for (i = 0; i < numProcs-1; i++)
        CREATE(SlaveStart)
        
    /* Make the master do slave work so we don't waste a processor */
    SlaveStart();        

    /* wait for all the slaves to complete */
    WAIT_FOR_END(numProcs-1)

    CLOCK(time);
    printf("Elapsed time (in cycles): %d\n", time); 
    /*
    for (i=0; i<M; i++)
      for (j=0; j<N; j++)
         printf("ADDR(grid[%d][%d])=%x\tADDR(scratch[%d][%d])=%x\n",i,j,&grid[i][j],i,j,&scratch[i][j]);
    */
       
    }


/*====================== SlaveStart =====================================*/
/* This is the routine that each processor will be executing in parallel */
void SlaveStart() {
    int myIndex, i, j, k, begin, end;
    
    int done=0;
    double mydiff;
    
    /**************************** 
    double **scratch;
    scratch = (double **) malloc((N/numProcs)*sizeof(double));
    for(i=0; i<N/numProcs; i++) {
      scratch[i]= (double *) malloc(N*sizeof(double));
      for(j=0; j<N; j++) scratch[i][j]=grid[i][j];
    }; 
    **********************/

    LOCK(indexLock);        /* enter the critical section */    
    myIndex = Index;        /* read your ID */
    ++Index;                /* increment it, so the next will operate on ID+1 */
    UNLOCK(indexLock);      /* leave the critical section */
      
    /* Initialize begin and end */
    begin = 1+(N-2)/numProcs*myIndex;
    end = 1+(N-2)/numProcs*(myIndex+1); 
    i=0;
    /* the main body of a thread */
    while (!done) {   
        i=i+1; mydiff=0; diff=0;
        for(j=begin; j<end; j++) {            
            for(k=1; k<N-1; k++) {           
                      scratch[j][k] = 0.2*(grid[j][k] + 
                      grid[j-1][k]+grid[j+1][k]+
                      grid[j][k-1]+grid[j][k+1]);
                      /*
                      printf("\n<I%d,P%d>: s[%d][%d]=g[%d][%d],[%d][%d],[%d][%d],[%d][%d]", 
                             i, myIndex, j, k, (j-1), k, j+1, k, j, (k-1), j, (k+1));
                      */
                      mydiff=mydiff+abs(scratch[j][k]-grid[j][k]);
            } /* end of k */
        } /* end of j */
        /*printf("-P%d--I%d: mydiff=%f\n", myIndex, i, mydiff);*/
        
        LOCK(diffLock);
          diff=diff+mydiff;
        UNLOCK(diffLock);

        BARRIER(sBarrier1, numProcs)
        /*printf(">>>P%d,I%d, diff=%f\n", myIndex, i, diff); */
         
        if ( (i>16) || (diff/(N*N) < TOL)) done = 1;
        else {
           for(j=begin; j<end; j++) {
             for(k=1; k<N-1; k++) {
               grid[j][k]=scratch[j][k];                          
            }
           }
       }
       BARRIER(sBarrier2, numProcs);
       
    } /* end while*/

 }

/*=============== log_2 function ===============================*/
int log_2(int number)

{
  int cumulative = 1;
  int out = 0;
  int done = 0;

  while ((cumulative < number) && (!done) && (out < 50)) {
    if (cumulative == number) {
      done = 1;
    } else {
      cumulative = cumulative * 2;
      out ++;
    }
  }

  if (cumulative == number) {
    return(out);
  } else {
    return(-1);
  }
}
