/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: augment.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The module for static augmentation of the compiled (and optimized)
 * parallel application code.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define extern
#include "augment.h"
#include "types.h"
#undef extern
#ifndef REQUEST_INTERFACE_ROUTINE
    #define REQUEST_INTERFACE_ROUTINE  "_sys_memory_interface"  /* read/write/lock */
#endif
#ifndef STRINGOP_INTERFACE_ROUTINE
    #define STRINGOP_INTERFACE_ROUTINE "_sys_memory_stringop"   /* rep movsl */
#endif
#define exiterror(ARGS...) { fprintf(stderr, ARGS); exit(1); }
int preserve_old_value_on_write = 1;    /* see below for explanations */

/*
 * Level of augmentation
 */
enum {  
    AUG_LEVEL_SYNC      = 0,        /* only locks are instrumented */
    AUG_LEVEL_SHARED    = 1,        /* only code/data references */
    AUG_LEVEL_ALL       = 2,        /* all data references (even the stack) */
    } aug_level = AUG_LEVEL_SHARED; /* default */

/*
 * input, output assembler source files 
 */
FILE *infile = stdin, *outfile = stdout;

/*
 * Usage: AUG [-0 | -1 | -2]  [infile [outfile]]. Typical: AUG lu.s lu.S.
 */
main(int argc, char *argv[]) {
    #define OPTIONS "012ph?"            /* -(0,1,2): augmentation levels */
    extern char *optarg;
    extern int optind, opterr, optopt;
    optind = 1;
    while(1) {
        int c = getopt(argc, argv, OPTIONS);
        if( c==-1 ) break;
        if( c=='?') 
            usage:
            exiterror("Try %s [-0|-1|-2] [-p] [input_file]\n",argv[0]);
        switch(c) {
            case '0': aug_level = AUG_LEVEL_SYNC;   break;
            case '1': aug_level = AUG_LEVEL_SHARED; break;
            case '2': aug_level = AUG_LEVEL_ALL;    break;
            case 'p': preserve_old_value_on_write = 0; break;
            case '?':
            case 'h': goto usage;
            default : break;
            }
        }
    if(optind < argc)   {
        infile =  fopen(argv[optind],"r");
        if (!infile) 
            exiterror("\aError: can't open input file %s!\n", argv[optind]);
        ++optind;
        }
    if(optind < argc)   {
        outfile =  fopen(argv[optind],"w");
        if (!outfile) 
            exiterror("\aError: can't create output file %s!\n", argv[optind]);
        ++optind;
        }
    set_yyinout(infile, outfile);

    islabel = 0;
    while(yylex()) 
        process_instruction(yyinstr);

    fclose(infile);
    }

/*
 * To be honest, what follows in an example of badly written code.
 * While much of it can safely be ignored if you just want to add
 * your own instructions, rewriting it to be more structural can't hurt.
 * It will be probably done in future.
 * 
 * Note:
 * ALL the code that follows refer to the processing of the CURRENT
 * instruction ONLY. Nearly all the variables here are reset to their
 * default values when processing of a new instruction begins.
 */

/*
 *  +------------------------------------------------------------------+
 *  |             Constants, types, and auxiliary routines             |
 *  +------------------------------------------------------------------+
 */ 

/* 
 * Macro that outputs the given string to the resulting asm file.
 */
#define asm_output(format, args...) {               \
    fprintf(outfile, ("\t" format "\n") , ##args);  \
    }

/*
 * "Global" placeholder for the instruction being processed. "Global"
 * in the sense that procedures copy it to their local counterparts.
 */
struct Instruction gin;
char *iname;                /* instruction name (e.g. "movl") */

/*
 * Now be aware of the form the instructions generated by the GCC take.
 * Statements taking operands usually have <operand size> suffix;
 * for example, "incl" means increment long (4 bytes), "incw" means
 * increment word (2 bytes), "incb" increment byte. So this suffix
 * is used to determine the size of the operands.
 */
char csuf;                  /* suffix character */

/*
 * However, there are exceptions: the 'l' suffix in the floating-point
 * statement means 8 bytes (long double), 's' is 4 bytes (float),
 * while FPU statements operating over integers use suffixes as the
 * ordinary statements. This structures carries the information
 * what is the size of the operands for such "extra" instructions.
 */
struct {
    int fpu;    /* is a FPU instruction ? */
    int fpuint; /* is a FPU instruction operating on integers? */
    int size;   /* manual correction. "-1" means "read from the suffix". */
    } extra;

/*
 * This is a set of constants and primitives used to determine whether
 * the "current" instructions matches the given pattern, and to determine
 * the size of the operands from the instruction name.
 * Use it as follows: match("ret") will return true if the instructions
 * is "ret" and no other. match("movS") will return true if the instruction
 * is one of the {movl, movw, movb} (also {movq, movs, movt} but these will
 * never be generated). In addition, it sets the `csuf' global var to be
 * the suffix (matched by the 'S') of the current instruction.
 * Also '?' and '*' are supported ("j*" will match any of the jump instructions).
 */
    
#define SIZECHAR    'S'
#define SIZELONG    'l'
#define SIZEWORD    'w'
#define SIZEBYTE    'b'
#define SIZETEN     't'
#define SIZEQUAD    'q' /* 8 bytes for most of the FPU instrs (double) */
#define SIZESINGLE  's' /* 4 bytes for float FPU instructions */
#define ANYCHAR     '?'
#define ANYSTRING   '*'
    
#define NOMATCH  0
#define YESMATCH 1
int match (char *pattern) {
    int i=0; char pc,ic;
    while(pc = pattern[i]) {            /* loop while pattern matches iname */
        ic = iname[i];
        switch (pc) {
            case SIZECHAR:
                switch (ic) {
                    case SIZELONG:
                    case SIZEWORD:
                    case SIZEBYTE:
                        csuf = ic;
                        break; /*return YESMATCH;*/
                    case SIZESINGLE:    /* 's' as a size is for floats only */
                    case SIZEQUAD:
                    case SIZETEN:
                        csuf = ic;
                        if (extra.fpu) break; /*return YESMATCH;*/
                        else return NOMATCH;
                    default:
                        return NOMATCH;
                    }
            case ANYCHAR:
                if (!ic) return NOMATCH;    
                break;
            case ANYSTRING:
                return YESMATCH;
            default:
                if (ic != pc) return NOMATCH;
                break;
            }
        ++i;
        }
    if (iname[i]) return NOMATCH;   /* if the pattern is exausted and iname isn't */
    return YESMATCH;
    }           

/*
 * Now we get more into details. Here we define the types of instructions
 * that need augmentation (unlike "ret" for example). These are:
 * - AUG_RW_NORMAL      : such as "movl $0,(%eax)". The first operand is
 *                        always to be read, and the second to be written to.
 * - AUG_RW_ALWAYS_R    : always read. An example is "cmpl" - none of the
 *                        operands is written to.
 * - AUG_RW_ALWAYS_W    : always write, as in "fstpl"
 * - AUG_ARITHMETIC     : need extra attention. Example: addl $2,(%eax).
 *      This means that value (%eax) is first read, then added with 2,
 *      and then written back to (%eax). Another example is incl (%eax);
 *      it is also a read and a write. addl (%eax), %edx, on the other hand,
 *      is only a read. That is further elaborated in the aug() routine.
 */
#define AUG_RW_NORMAL   0   /* as mov */
#define AUG_RW_ALWAYS_R 1   /* as cmp */
#define AUG_RW_ALWAYS_W 2   /* as set or fstp */
#define AUG_ARITHMETIC  3   /* as add, inc etc. */
#define AUG_PREF        4   /* prefetch instruction */
#define AUG_PREFEX      5   /* prefetch-exclusive instr. */
#define AUG_LPREF       6   /* lazy prefetch instruction */
#define AUG_WRBACK      7   /* write-back instruction */
#define AUG_OWL         8   /* open window: the lowest address */
#define AUG_OWH         9   /* open window: the highest address */
#define AUG_CWL        10   /* clese window: the lowest address */

/*
 * do_print tells whether to print the line after processing.
 * If you augment a read instruction, you will add what it takes
 * in the resulting asm, and leave do_print=1, so it will be printed.
 * If you augment a write instruction, you will have to "print" the
 * instruction yourself (to the asm output) and add what is needed
 * after it, so you'll set do_print to 0.
 * If an artificial instruction is augmented (such as "cache_flush"),
 * do_print will, of course, have to be set to 0.
 */

int do_print;

/*
 * Code and vars related to the counting of instructions within 
 * basic blocks and updating the global "timer". Asume, for example,
 * that the basic block that has just ended consists of five instructions.
 * Upon the beginning of a new basic block, we have to increase the
 * global counter by five, and we do it with
 * "pushfl; addl $5, sys_IC; popfl".
 * Therefore, the IC counter refers to the instruction count of the
 * CURRENT BASIC BLOCK, where the sys_IC is the global timer (for the
 * currently executing process) where the IC's of the basic blocks
 * the threads flows through accumulate.
 */

int IC = 0;
int update_sys_IC() {
    char asm_IC[1000];
    if (!IC) return 0;
    sprintf(asm_IC, "pushfl");
    if (IC == 1) 
        sprintf(asm_IC+strlen(asm_IC),"; incl _sys_IC  ");
    else    
        sprintf(asm_IC+strlen(asm_IC),"; addl $%d,_sys_IC", IC);
    sprintf(asm_IC+strlen(asm_IC),"; popfl    # sys_IC+=%d", IC);
    asm_output("%s", asm_IC);
    IC = 0;
    }

/*
 * For string operations, define instructions that are to be handled:
 * The only case that can happen in a preprocessed Limes application
 * is the "cld -- set ecx/esi/edi -- rep -- movsl" combination.
 * "cld" is ignored, "rep" is omitted, "movs" is replaced with a kernel call.
 * GCC does not seem to ever generate movsl alone. 
 * "scasl" and "cmpsl" actually do get generated, but only
 * as speed-ups for strlen() and strcmp() if -O2 is given. Since
 * str*() functions are under kernel's control, all we really care
 * about is "rep movs{l|w|b}". It is used when one structure is
 * copied to another (as in struct s1=s2).
 */
enum String_Op {STRINGOP_REP, STRINGOP_MOVS};
void StringOperation(enum String_Op);

/*
 *  +------------------------------------------------------------------+
 *  |          Main CASE structure for processing instructions         |
 *  +------------------------------------------------------------------+
 *
 *  Only here the processing of the current istruction begins. main()
 *  supplies this routine with the current instruction, and this routine
 *  initializes the data relevant for processing.
 */

process_instruction(struct Instruction in) {
    int cycles = 1;     /* the default execution time */
    /* 
     * if this "instruction" was merely a label, we know that 
     * the basic block has ended, so we update the global timer and return.
     */
    if (islabel)    {
        update_sys_IC();
        fprintf(outfile, "%s", in.line);    /* output label */
        islabel = 0;
        return;
        }

    /*
     * copy the instruction to global vars. This is silly, since communicating
     * through procedure arguments wouldn't slow down the whole process much,
     * and it's lightning fast anyway (since its job is relatively simple).
     */
    in.argc = (in.argv[0][0] !=0 ) + (in.argv[1][0] !=0 );
    in.memory = in.memo[0] | in.memo[1];
    gin = in;
    iname = gin.name;

    /* prepare the defaults */
    do_print = 1;       /* default: the instruction is to be put out */
    extra.fpu = 0;      
    extra.fpuint = 0;
    extra.size = -1;    /* -1 means read the size from csuf - b,w,l,q,s,t */

    /* 
     * Now in the good ol' C style we define macros to make the code
     * more synoptic. :-)
     * in.memory is true if the statement operates with memory.
     * For example, there's no need to augment "incl %eax".
     *
     * The argument to all these macros is an estimated duration in
     * _pipelined_ instruction processing. It is assummed that ordinary
     * instructions can be processed in one cycle, while most FPU instructions
     * cause a delay of two or more cycles. Of course, this is a dynamic
     * matter; but on the average, simulated execution times for many basic 
     * blocks should match those of the P6 processor. If the P6 had a perfect
     * pipeline, you could say that the average execution time is one 
     * instruction per cycle.
     */

    #define MATCH(PAT)      (match(PAT))
    #define SKIP(DURATION)  {   /* when a pattern is matched */ \
        /*                                                      \
         * if the instr was instrumented (do_print==0),         \
         * one cycle has been calculated in. IC will be         \
         * increased by the value of 'cycles'.                  \
         */                                                     \
        cycles = do_print ? DURATION : DURATION - 1;            \
        goto endmatchtest;                                      \
        }
    #define AUG(D)      {if (in.memory) aug(AUG_RW_NORMAL);      SKIP(D) }
    #define AUGR(D)     {if (in.memory) aug(AUG_RW_ALWAYS_R);    SKIP(D) }
    #define AUGW(D)     {if (in.memory) aug(AUG_RW_ALWAYS_W);    SKIP(D) }
    #define AUGA(D)     {if (in.memory) aug(AUG_ARITHMETIC);     SKIP(D) }
    #define AUGP(D)     {if (in.memory) aug(AUG_PREF);           SKIP(D) }
    #define AUGX(D)     {if (in.memory) aug(AUG_PREFEX);	 SKIP(D) }
    #define AUGF(D)     {if (in.memory) aug(AUG_LPREF);	  	 SKIP(D) }
    #define AUGB(D)     {if (in.memory) aug(AUG_WRBACK);	 SKIP(D) }
    #define AUGL(D)     {if (in.memory) aug(AUG_OWL);	         SKIP(D) }
    #define AUGH(D)     {if (in.memory) aug(AUG_OWH);	  	 SKIP(D) }
    #define AUGC(D)     {if (in.memory) aug(AUG_CWL);     	 SKIP(D) }
       
    /* a branch is the end of a basic block */
    #define BRNC(D)     {++IC; update_sys_IC(); SKIP(D-1); } 

    #define STACK_PUSH 0
    #define STACK_POP  1
    #define STACK(TYPE) {instrument_stack(TYPE); SKIP(1) }

    /*
     * This is the list of i386/i387 instructions *that can appear in the
     * asm code generated by GCC*.
     * The vast majority of "missing" instructions are system instructions
     * (like "cli") or instructions never generated by GCC (like "enter").
     *
     * The order of testing for pattern matching may be very important.
     * For example, we first test whether the instructon matches "jmp",
     * and, if not, if matches "j*". (The example is dumm, but if they required
     * different actions, reverse order of testing would be a mistake.)
     *
     * Remember that testing for matching also extracts the size suffix (csuf)
     * from the instruction to be used later.
     *
     * Again, the number used as the macro argument is the delay in 
     * pipelined processing for the instruction, measured in cycles.
     * If (1) is used for all the instructions, that would represent
     * a perfect Pentium. Using values greater than 1 (mainly for FPU
     * instructions) makes timing more realistic; the greater the value,
     * the slower the processor. Given values model a P6 or higher CPU.
     * 
     * So, here they are, in alphabetical order. Comments are added to more
     * obscure ones. You will also see that some instructions that do not
     * fit the rule are patched (like movsll).
     */

    if MATCH("addS")    AUGA(1)
    if MATCH("adcS")    AUGA(1)
    if MATCH("andS")    AUGA(1)
    if MATCH("call")    BRNC(1)
    if MATCH("clt?")    SKIP(1) 
    if MATCH("cmpS")    AUGR(1)
    if MATCH("decS")    AUGA(1)
    if MATCH("divS")    AUGR(1)

    /* floating point instructions */
    extra.fpu = 1;  extra.fpuint = 0;
    if MATCH("fabs")    SKIP(1)
    if MATCH("fadd")    SKIP(1)
    if MATCH("faddS")   AUGR(1)
    if MATCH("faddp")   SKIP(1)
    if MATCH("fchs")    SKIP(1)
    if MATCH("fcom")    SKIP(1)
    if MATCH("fcomS")   AUGR(1)
    if MATCH("fcomp")   SKIP(1)
    if MATCH("fcompS")  AUGR(1)
    if MATCH("fcompp")  SKIP(1)
    if MATCH("fcos")    SKIP(1)
    if MATCH("fdiv")    SKIP(1)
    if MATCH("fdivp")   SKIP(1)
    if MATCH("fdivS")   AUGR(1)
    if MATCH("fdivr")   SKIP(1)
    if MATCH("fdivrS")  AUGR(1)
    if MATCH("fdivrp")  SKIP(1)
    extra.fpuint = 1;           /* Integer FPU insts: they have w=2, l=4, q=8 */
    if MATCH("fiaddS")  AUGR(1) 
    if MATCH("ficomS")  AUGR(1) 
    if MATCH("ficompS") AUGR(1) 
    if MATCH("fidivS")  AUGR(1)     
    if MATCH("fidivrS") AUGR(1)     
    if MATCH("fildS")   AUGR(1)
    if MATCH("fimulS")  AUGR(1)
    if MATCH("fistS")   AUGW(1)
    if MATCH("fistpS")  AUGW(1)
    if MATCH("fisubS")  AUGR(1)
    if MATCH("fisubrS") AUGR(1)
    extra.fpuint = 0;           /* other Real FPU instrs: s=4, l=8, t=10 */
    if MATCH("fld")     SKIP(1)
    if MATCH("fld1")    SKIP(1)
    if MATCH("fldcw")   AUGR(1)
    if MATCH("fldS")    AUGR(1)
    if MATCH("fldz")    SKIP(1)
    if MATCH("fmul")    SKIP(1)
    if MATCH("fmulS")   AUGR(1)
    if MATCH("fmulp")   SKIP(1)
    if MATCH("fnop")    SKIP(1)
    if MATCH("fnstcw")  {extra.size = 2; AUGW(1)}
    if MATCH("fnstsw")  {extra.size = 2; AUGW(1)}
    if MATCH("fsin")    SKIP(1)
    if MATCH("fsincos") SKIP(1)
    if MATCH("fsqrt")   SKIP(1)
    if MATCH("fst")     SKIP(1)
    if MATCH("fstS")    AUGW(1)
    if MATCH("fstp")    SKIP(1)
    if MATCH("fstpS")   AUGW(1)
    if MATCH("fsub")    SKIP(1)
    if MATCH("fsubS")   AUGR(1)
    if MATCH("fsubp")   SKIP(1)
    if MATCH("fsubr")   SKIP(1)
    if MATCH("fsubrS")  AUGR(1)
    if MATCH("fsubrp")  SKIP(1)
    if MATCH("fucom")   SKIP(1)
    if MATCH("fucomp")  SKIP(1)
    if MATCH("fucompp") SKIP(1)
    if MATCH("fxch")    SKIP(1)
    extra.fpu = 0;
    
    if MATCH("idivS")   AUGR(1)
    if MATCH("imulS")   AUGR(1)
    if MATCH("incS")    AUGA(1)

    if MATCH("jmp")     BRNC(1) 
    if MATCH("j*")      BRNC(1) /* short jumps: j* that are not jmp */
    if MATCH("leaS")    SKIP(1)
    if MATCH("loop*")   BRNC(1) /* loop, loope etc. */
    if MATCH("movS")    AUG(1)
    if MATCH("mov?SS")  {csuf=in.name[4]; AUGR(1)}  /* mov{s|z}{bl|bw|lw} */
    if MATCH("negS")    AUGA(1)
    if MATCH("nop")     SKIP(1)
    if MATCH("notS")    AUGA(1)
    if MATCH("orS")     AUGA(1)
    if MATCH("popS")    STACK(STACK_POP)
    if MATCH("pushS")   STACK(STACK_PUSH)
    if MATCH("ret")     BRNC(1)
    if MATCH("rolS")    AUGA(1)
    if MATCH("rorS")    AUGA(1)
    if MATCH("salS")    AUGA(1)
    if MATCH("sarS")    AUGA(1)
    if MATCH("shlS")    AUGA(1)
    if MATCH("shrS")    AUGA(1)
    if MATCH("set*")    {extra.size=1; AUGW(1)}
    if MATCH("shrS")    AUGA(1)
    if MATCH("subS")    AUGA(1)
    if MATCH("sbbS")    AUGA(1)
    if MATCH("testS")   AUGR(1)
    if MATCH("xorS")    AUGA(1)
    
    /* user-defined instructions */
    if MATCH("pref")    {do_print=0; AUGP(1)}   
    if MATCH("prefex")  {do_print=0; AUGX(1)}
    if MATCH("lpref")   {do_print=0; AUGF(1)}
    if MATCH("wrback")  {do_print=0; AUGB(1)}
    if MATCH("owl")     {do_print=0; AUGL(1)}
    if MATCH("owh")     {do_print=0; AUGH(1)}
    if MATCH("cwl")     {do_print=0; AUGC(1)}
    
    /*
     * String operations: rep followed by a movsl with cld ignored.
     */
    if MATCH("cld")     SKIP(1)
    if MATCH("rep")     { StringOperation(STRINGOP_REP) ; SKIP(1) }
    if MATCH("movsS")   { StringOperation(STRINGOP_MOVS); SKIP(1) }

    /*
     * Now you can add the list of your own instructions.
     * Currently only "lock" and "unlock" are here.
     */
    if MATCH(INSTRUCTION_LOCK)    {artificial(ARTIF_LOCK)  ; SKIP(1)}
    if MATCH(INSTRUCTION_UNLOCK)  {artificial(ARTIF_UNLOCK); SKIP(1)}
   /*
    * List of added instructions includes:
    *	PREF 	- prefetch instruction
    *	PREFEX  - prefetch-exclusive instruction
    *	LPREF	- lazy prefetch instruction
    *	WRBACK  - write-back instruction
    *   OWL	- open-window instruction
    *   OWH	- open window instruction (defines the highest address)
    *   CWL	- clese window instruction
    */
    if MATCH(INSTRUCTION_PREF)    {artificial(ARTIF_PREF); SKIP(1)}
    if MATCH(INSTRUCTION_PREFEX)  {artificial(ARTIF_PREFEX); SKIP(1)}
    if MATCH(INSTRUCTION_LPREF)	  {artificial(ARTIF_LPREF); SKIP(1)}
    if MATCH(INSTRUCTION_WRBACK)  {artificial(ARTIF_WRBACK); SKIP(1)}
    if MATCH(INSTRUCTION_OWL)     {artificial(ARTIF_OWL); SKIP(1)}
    if MATCH(INSTRUCTION_OWH)	  {artificial(ARTIF_OWH); SKIP(1)}
    if MATCH(INSTRUCTION_CWL)     {artificial(ARTIF_CWL); SKIP(1)}
     
    /*
     * No match.
     */
    fprintf(stderr,"\aNo match for %s (%s, line %d)\n", in.name, in.line, 
        in.linenumber);
        
endmatchtest:
    if (do_print)   asm_output("%s", gin.line);
    IC += cycles;
    }   

/*
 *  +------------------------------------------------------------------+
 *  |           Macros and functions needed for augmentation           |
 *  +------------------------------------------------------------------+
 *
 *  What follows is a set of macros and routines used for inserting
 *  new assembler statements between the existing ones, namely for
 *  saving registers, pushing parameters, calling the routine in the
 *  kernel etc. You will see the best what they do by running the
 *  program over the excerpt of the real .s generated by cc.
 */

#define SRC_ARG 0           /* source is always arg 0 */
#define DST_ARG 1           /* and destination is always arg 1 in GCC */

int asm_aug_getsize() {     /* determines the size of the operand in bytes */
    int size;
    if (extra.size < 0) {
        switch (csuf) {     /* on the basis of csuf */
            case SIZEBYTE:  size = 1; break;
            case SIZEWORD:  size = 2; break;
            case SIZELONG:  size = 4; if (extra.fpu && !extra.fpuint) size = 8; 
                            break;
            case SIZEQUAD:  size = 8; break;
            case SIZETEN :  size = 10; break;
            default      :  size = 4; break;
            }
        }
    else    size = extra.size;
    return size;
    }

/* 
 * we need to know the size of the stack created when calling
 * sys_memory_interface, so we can increment the ESP by its size upon return.
 * aug_stack_frame_size is incremented by 4 for each argument pushed.
 */
int aug_stack_frame_size;       /* 4 * number of pushed parameters */
/*
 * Some extra effort for giving correct (%esp) related address is needed
 * (as in movl $2, (%esp)).
 * Machine status is saved with 
 * "pushfl"     -- pushes the extended flags on the stack (4 bytes)
 * "pushal"     -- pushes EAX, ECX, EDX, EBX, ESP, EBP, ESI, EDI
 * which is the total of AUG_STACK_SIZE bytes.
 */
#define AUG_STACK_SIZE 36

/*
 * Enter the sequence of instructions for sending parameters of the
 * current memory reference to the simulator. Common for all types
 * of instructions.
 * Consists of pushing flags and registers, updating sys_IC,
 * and resetting the IC (Instruction counter), since an instruction 
 * with a memory reference is the end of a basic block as well.
 */
void asm_aug_enter(char *title) {                   
    asm_output("############ AUG_BEGIN: %s", title);    
    asm_output("pushfl; pushal");                           
    if (IC == 1) 
        asm_output("incl _sys_IC")
    else if (IC > 1)
        asm_output("addl $%d,_sys_IC", IC)
    aug_stack_frame_size = 0;
    IC = 0;
    }

/*
 * PUSH the argument for the simulator on the stack.
 */
#define asm_aug_push(format, args...) {     \
    asm_output("pushl " format, ##args);    \
    aug_stack_frame_size += 4;              \
    }

/*
 * PUSH the memory address being referenced as the argument.
 * Store the address to the EDI register and push it.
 * For those that are not (%esp) dependent, address is computed as
 *      leal <memory address>, %edi
 * For (%esp) related (very rare), do
 *      leal <AUG_STACK_SIZE>+<original %esp related address>
 * Then do 
 *      pushl %edi
 * and that will be the <address> argument of the memory reference
 * for the simulator.
 * Note: an asm programmer could write a program that would rule out
 * all the assumptions implied here; but GCC never does so.
 */
void asm_aug_pushaddr(char *addr) {
    if (!strstr(addr, "%esp"))              /* if "%esp" is not mentioned */
        asm_output("leal %s,%%edi", addr)
    else    {
        if (addr[0]=='(')   /* form: (%esp,...) --> 36(%esp,...) */
            asm_output("leal %d%s,%%edi", AUG_STACK_SIZE, addr)
        else                /* form: -4(%esp,...) --> 36+-4(%esp,...) */
            asm_output("leal %d+%s,%%edi", AUG_STACK_SIZE, addr)
        }
    asm_aug_push("%%edi     # addr");
    }   

/*
 * Support for PRESERVE_OLD_VALUE_ON_WRITE (<==> !MEMSIM_HAS_DATA_COPY).
 * Stores the old contents of the location being written to to the
 * sysvar_old_value global variable. This method is not recommended,
 * however (see kernel/preferences.h).
 * It is done only for non-stack writes, if at all.
 */
void asm_aug_writenonstack_enter(char *addr, char *instr) {
    #define OLDVALUEHOLDER "_sysvar_old_value"
    int wordcount, size;
    if (!preserve_old_value_on_write) return;

    wordcount = 1; size = asm_aug_getsize();
    if (size >8)        wordcount = 3;
    else if (size>4)    wordcount = 2;
    else                wordcount = 1;
    asm_output("############ AUG_store_old_value: %s", instr);
    if (wordcount == 1) {           /* the most common case */
        asm_output("pushl %%edi");
        asm_output("movl %s, %%edi", addr);
        asm_output("movl %%edi, %s", OLDVALUEHOLDER);
        asm_output("popl %%edi");   /* flags are not affected by mov/push/lea */
        }
    else    {
        asm_output("pushl %%edi");
        asm_output("pushl %%esi");
        asm_output("leal %s, %%edi", addr);
        asm_output("movl (%%edi), %%esi");
        asm_output("movl %%esi, %s", OLDVALUEHOLDER);
        if (wordcount >= 2) {
            asm_output("movl 4(%%edi), %%esi");
            asm_output("movl %%esi, %s+4", OLDVALUEHOLDER);
            }
        if (wordcount > 2)  {
            asm_output("movl 8(%%edi), %%esi");
            asm_output("movl %%esi, %s+8", OLDVALUEHOLDER);
            }
        asm_output("popl %%esi");
        asm_output("popl %%edi");
        }
    asm_output("######");
    }

/*
 * Pushes the size of the operand.
 */
void asm_aug_pushsize() {
    asm_aug_push("$%d       # size", asm_aug_getsize());
    }

/*
 * and the type. May require some intervention on your side
 * for some of your artificial instructions (if you have them)
 * that are neither load/stores nor lock/unlocks.
 */
void asm_aug_pushtype(int type) {
    char *stype;
    switch (type) {
        case REQUEST_READ   : stype = "READ";   break;
        case REQUEST_WRITE  : stype = "WRITE";  break;
        case REQUEST_LOCK   : stype = "LOCK";   break;
        case REQUEST_UNLOCK : stype = "UNLOCK"; break;
        case REQUEST_PREF   : stype = "PREF";	break;
        case REQUEST_PREFEX : stype = "PREFEX"; break;
        case REQUEST_LPREF  : stype = "LPREF";	break;
        case REQUEST_WRBACK : stype = "WRBACK";  break;
        case REQUEST_OWL    : stype = "OWL"; break;
        case REQUEST_OWH    : stype = "OWH"; break;
        case REQUEST_CWL    : stype = "CWL"; break; 
        default             : stype = "CUSTOM"; break;
        }
    asm_aug_push("$%d       # type - %s ", type, stype)
    }

/*
 * Leave the augmentation sequence. Restores the ESP, the registers and
 * the flag.
 */
void asm_aug_leave() {
    asm_output("addl $%d,%%esp",aug_stack_frame_size);
    asm_output("popal; popfl");
    asm_output("############ AUG_END");
    }

/*
 * Call the specified kernel routine.
 */
#define asm_aug_call(ROUTINE) asm_output("call " ROUTINE)

/*
 *  +------------------------------------------------------------------+
 *  |          The decision of the augmentation type (READ/WRITE...)   |
 *  +------------------------------------------------------------------+
 *
 *  This is it. Now that we have defined a bunch of macros and functions,
 *  we can actually do the augmentation for the current instruction.
 *  The "normality" indicates whether this is an ordinary SRC, DST
 *  instruction, or "ALWAYS READ" like cmp, or "ALWAYS WRITE" like
 *  many of those with one argument, or "ARITHMETIC", which require
 *  special handling.
 */
aug(int normality) {
    struct Instruction AUGin = gin;
    int in_type; char *addr;

    if (!AUGin.memory) return;  /* if it doesn't work with memory, I'm not interested. */

    /* but if it does, the address is in the first or the second argument */
    addr = ( AUGin.memo[SRC_ARG] ? AUGin.argv[SRC_ARG] : AUGin.argv[DST_ARG] );

    /* check whether to instrument this instruction: */
    if (aug_level < AUG_LEVEL_SHARED) return;
    if (aug_level < AUG_LEVEL_ALL)  {
        /* 
         * If stack references are not of interest, exclude the statements
         * who have the address operand related to ESP or EBP.
         */
        if (strstr(addr,"%ebp") || strstr(addr,"%esp")) return;
        }

    /*
     * If the instruction is of type AUG_ARITHMETIC:
     * if there are two arguments and the SRC is memory, than it is
     * an ordinary READ instruction.
     * If not, it will be first augmented as a READ, then executed,
     * and then augmented as a WRITE. (In fact, a READ is simulated,
     * then the instruction is attributed "always write" since the rest
     * of it is a write.               .-- READ  (%eax) // fetch argument
     * example:         incl (%eax) ---+   incl  (%eax) // read and write
     *                                 `-- WRITE (%eax) // store argument
     */
    if (normality == AUG_ARITHMETIC) {
        if (AUGin.argc == 2 && AUGin.memo[SRC_ARG]) 
            normality = AUG_RW_ALWAYS_R;
        else    {                           /* simulate a READ */
            asm_aug_enter(AUGin.line);      /* enter sequence */
            asm_aug_pushaddr(addr);         /* push the address */
            asm_aug_pushsize();             /* push the size of the operand */
            asm_aug_pushtype(REQUEST_READ); /* push type - READ */
            asm_aug_call(REQUEST_INTERFACE_ROUTINE);    /* call the kernel */
            asm_aug_leave();                /* leave the sequence */
            normality = AUG_RW_ALWAYS_W;    /* the rest is a WRITE */
            }
        }

    /*
     * Determine whether it is a read or a write.
     */
    switch (normality) {
        case AUG_RW_ALWAYS_R:   in_type = REQUEST_READ;  break;
        case AUG_RW_ALWAYS_W:   in_type = REQUEST_WRITE; break;
        case AUG_PREF:          in_type = REQUEST_PREF;  break;
        case AUG_PREFEX:	in_type = REQUEST_PREFEX; break;
        case AUG_LPREF:	       	in_type = REQUEST_LPREF; break;
        case AUG_WRBACK:	in_type = REQUEST_WRBACK; break;
        case AUG_OWL:		in_type = REQUEST_OWL; break;
        case AUG_OWH:		in_type = REQUEST_OWH; break;
        case AUG_CWL:		in_type = REQUEST_CWL; break;
        default:
            if (AUGin.argc < 2) {
                fprintf(stderr,"\aError: AUG_RW_NORMAL for <2 ops! (%s)\n",
                    AUGin.line); return;
                    }
            if (AUGin.memo[SRC_ARG]) in_type = REQUEST_READ; else
            if (AUGin.memo[DST_ARG]) in_type = REQUEST_WRITE;
            break;
        }

    /*
     * If it is a write, output the original instruction first, and then
     * do the augmentation.
     * This has the efect that the host iCPU precomputes for us the value
     * to be written to the address.
     * Example: movl $2, (%eax): Assume %eax == 0x4000.
     * This instruction will be replaced with
     *          movl $2, (%eax)
     *          call sys_memory_interface WITH addr=%eax, size=4, type=WRITE
     * The CPU will store $2 to 0x4000, and we will call the kernel with
     * the address 0x4000. It will know that the CPU has just written something
     * sized 4 at that address, and will read the new contents of 0x4000.
     * The value found there will be the argument to the actual memory 
     * simulator. (This value will be ignored by that simulator if the address
     * does not belong to the shared area, since no one else will write 
     * anything to this address.)
     */
    if (in_type == REQUEST_WRITE)   {
        /* 
         * Output the line to be written. But first remember the old value,
         * if PRESERVE_OLD_VALUE_ON_WRITE is used, if ESP and EBP are not
         * mentioned in the address (there's no need to keep the copy of the
         * old value for a reference to stack segment data).
         */
        if (!strstr(addr,"%ebp") && !strstr(addr,"%esp")) 
            asm_aug_writenonstack_enter(addr, AUGin.line);  
        asm_output("%s", AUGin.line);
        }

    /*
     * Do the augmentation.
     */
    asm_aug_enter(AUGin.line);                  /* enter sequence */
    asm_aug_pushaddr(addr);                     /* push memory address */
    asm_aug_pushsize();                         /* push the size of the operand */
    asm_aug_pushtype(in_type);                  /* push the type - READ/WRITE */
    asm_aug_call(REQUEST_INTERFACE_ROUTINE);    /* call the kernel */
    asm_aug_leave();                            /* leave the sequence */
    
    /*
     * If the instruction is of the READ type, then it will come
     * AFTER the augmentation. This way the host iCPU will do what
     * is needed with the value supplied by the memory simulator.
     * Example: cmpl %edx,(%eax)    will be replaced with
     *          call sys_memory_interface WITH addr=%eax, size=4, type=READ
     *          cmpl %edx,(%eax)
     */
    if (in_type == REQUEST_READ)    asm_output("%s", AUGin.line);

    do_print = 0;   /* and tell the calling routine that this one need not be printed */
    }

/*
 *  +------------------------------------------------------------------+
 *  |            Special case: stack instructions (push, pop)          |
 *  +------------------------------------------------------------------+
 *
 *  Instrumenting stack references (if it is desired):
 *      pushl $0, for example, is a WRITE. So it is replaced with
 *      pushl $0
 *      call sys_memory_interface WITH addr=old %esp, size=4, type=WRITE
 *  The order could be reversed - we are not interested with the contents
 *  of the old (%esp) since it contains a private data and can be written by
 *  no one else; all we really need is the address, for correct simulation.
 *  A pop, on the other hand, is a READ.
 *
 *  Stack is a special case on a CISC. It may seem that stack would not be
 *  of interest if only shared references are to be instrumented, but it
 *  is not so. Consider the following example:
 *      pushl (%eax)
 *  Now this is first a READ (from the (%eax) location) and then a WRITE
 *  (to the (%esp) location); (%eax) may be the address of a shared location.
 *
 *  In this version, pushes and pops are handled. Implicit pushes and pops
 *  like CALL and RET are not handled (were once), since they really have 
 *  a little impact on the data traffic.
 */
instrument_stack(int instr_type) {
    struct Instruction STin;
    STin = gin; 
    /* determine whether to perform augmentation or not */
    switch (aug_level) {
        case AUG_LEVEL_SYNC :           /* locks */
            return;                     /* have nothing to do with stack */
        case AUG_LEVEL_SHARED :         /* shared data */
            /* 
             * continue only if the arg of the push/pop 
             * is a memaddr not related to (%ebp)
             */
            if ( STin.memory && !strstr(STin.argv[SRC_ARG],"%ebp") ) break;
            else    return;
        case AUG_LEVEL_ALL :            /* stack */
            break;                      /* is why we are here */
        default : exiterror("\aInvalid aug_level!\n")
        }
    if (instr_type == STACK_PUSH) {
        if (STin.memo[SRC_ARG]) {   /* first: are we pushing from a location? */
            asm_aug_enter(STin.line);           /* yes */
            asm_aug_pushaddr(STin.argv[SRC_ARG]);
            asm_aug_pushsize();
            asm_aug_pushtype(REQUEST_READ);     /* we have to READ from it */
            asm_aug_call(REQUEST_INTERFACE_ROUTINE);
            asm_aug_leave();
            }
        if (aug_level < AUG_LEVEL_ALL) return;  /* what follows is strictly private */
        asm_output("%s", STin.line);            /* output the original push instr */
        asm_aug_enter(STin.line);               /* then augment it as usual */
        asm_aug_pushaddr("(%esp)");
        asm_aug_pushsize();
        asm_aug_pushtype(REQUEST_WRITE);
        asm_aug_call(REQUEST_INTERFACE_ROUTINE);
        asm_aug_leave();
        }
    if (instr_type == STACK_POP) {
        if (aug_level > AUG_LEVEL_SHARED) {
            /* a pop is a READ */
            asm_aug_enter(STin.line);
            asm_aug_pushaddr("(%esp)");
            asm_aug_pushsize();
            asm_aug_pushtype(REQUEST_READ);
            asm_aug_call(REQUEST_INTERFACE_ROUTINE);
            asm_aug_leave();
            }
        if (STin.memo[SRC_ARG]) {   
            /* but a pop to a location is also a WRITE (like pop (%eax)) */
            /* 
             * for PRESERVE_OLD_..., if we are popping to a location, we are
             * doing a WRITE, so we have to remember the old contents - if
             * this address is not EBP/ESP based.
             */
            if (!strstr(STin.argv[SRC_ARG],"%ebp") && !strstr(STin.argv[SRC_ARG],"%esp")) 
                asm_aug_writenonstack_enter(STin.argv[SRC_ARG], STin.line);
            /* the rest is as for the usual WRITE */
            asm_output("%s", STin.line);
            asm_aug_enter(STin.line);
            asm_aug_pushaddr(STin.argv[SRC_ARG]);
            asm_aug_pushsize();
            asm_aug_pushtype(REQUEST_WRITE);
            asm_aug_call(REQUEST_INTERFACE_ROUTINE);
            asm_aug_leave();
            }
        else    {
            /* but a pop to a register deserves to be put out, too */
            asm_output("%s", STin.line);
            }
        }
    do_print = 0;
    }

/*
 *  +------------------------------------------------------------------+
 *  |  Special case: non-existing instructions (lock/unlock/cflush...) |
 *  +------------------------------------------------------------------+
 *
 *  Artificial instructions. LOCK and UNLOCK are here as examples.
 *  If you want to add your own, mimic these.
 */
artificial (enum Artificial_Instruction incode) {
    struct Instruction AIin = gin;
    switch( incode ) {
        case ARTIF_LOCK:
        case ARTIF_UNLOCK:
            asm_aug_enter(AIin.line);
            asm_aug_pushaddr(AIin.argv[SRC_ARG]);
            asm_aug_push("$4        # size ");
            asm_aug_pushtype(incode==ARTIF_LOCK ? REQUEST_LOCK : REQUEST_UNLOCK);
            asm_aug_call(REQUEST_INTERFACE_ROUTINE);
            asm_aug_leave();
            do_print = 0;   /* don't forget this or your AI will go straight to the code! */
            return; 
        /* prefetch-like instructions +  */
        case ARTIF_PREF:
        case ARTIF_PREFEX:
        case ARTIF_LPREF:
        case ARTIF_WRBACK:
        case ARTIF_OWL:
        case ARTIF_OWH:
        case ARTIF_CWL:
            asm_aug_enter(AIin.line);
            asm_aug_pushaddr(AIin.argv[SRC_ARG]);
            asm_aug_push("$4        # size ");
            if(incode==ARTIF_PREF)
              asm_aug_pushtype(REQUEST_PREF);
            else if(incode==ARTIF_PREFEX)
              asm_aug_pushtype(REQUEST_PREFEX);
            else if(incode==ARTIF_LPREF)
              asm_aug_pushtype(REQUEST_LPREF);
            else if(incode==ARTIF_WRBACK)
              asm_aug_pushtype(REQUEST_WRBACK);
            else if(incode==ARTIF_OWL)
              asm_aug_pushtype(REQUEST_OWL);
            else if(incode==ARTIF_OWH)
              asm_aug_pushtype(REQUEST_OWH);
            else if(incode==ARTIF_CWL)
              asm_aug_pushtype(REQUEST_CWL);
              
            asm_aug_call(REQUEST_INTERFACE_ROUTINE);
            asm_aug_leave();
            do_print = 0;   /* don't forget this or your AI will go straight to the code! */
            return;
            
        default:
            break;
        }
    }

/*
 *  +------------------------------------------------------------------+
 *  |           Special case: REP MOVS string operation                |
 *  +------------------------------------------------------------------+
 *
 *  Handling MOVS operation: rep movs{l|w|b} is replaced with a kernel call
 *  with count (ECX), source addr (ESI), destination addr (EDI), and
 *  size (4 for movsl, 2 for movsw, 1 for movsb).
 */
void StringOperation(enum String_Op string_op) {
    static REP_prefix = 0;  char title[30];
    struct Instruction MOVSin = gin;

    if (aug_level < AUG_LEVEL_SHARED) return;

    do_print = 0;
    if (string_op == STRINGOP_REP)  {
        REP_prefix = 1;
        return;
        }
    if (! REP_prefix )  {
        fprintf(stderr, "\aError: can't handle MOVS not preceded by a REP [line %d]\n",
            MOVSin.linenumber); 
        return;
        }
    REP_prefix = 0;
    /* 
     * Now we know that the previous "instruction" was "rep",
     * and that this one is "movsS" (S=b|w|l)
     */
    sprintf(title, "rep %s", MOVSin.name);
    asm_aug_enter(title);
    asm_aug_pushsize();
    asm_aug_push("%%ecx     # count");
    asm_aug_push("%%esi     # source");
    asm_aug_push("%%edi     # destination");
    asm_aug_call(STRINGOP_INTERFACE_ROUTINE);
    #if 0   /* paranoia: new values ECX, ESI, EDI may be used after REP MOVS */
        asm_aug_leave();
    #else   /* though I've never seen such a case in GCC generated code */
        asm_output("addl $%d,%%esp",aug_stack_frame_size);
        asm_output("popal");
        asm_output("addl %%ecx,%%esi; addl %%ecx,%%edi; xorl %%ecx,%%ecx");
        asm_output("popfl");
        asm_output("############ AUG_END");
    #endif
    return;
    }
