%option noyywrap

%s fetch_name fetch_arg

%{
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: parser.lex
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This scanner parses assembler instructions into units relevant for
 * augmentation: statement and operands, indicating further whether an
 * operand is a memory operand or not.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <string.h>
#include <stdio.h>
#include "augment.h"
%}

	int linecount = 0;
	int do_aug = 1;

FPREG		st(\([0-9]+\))?
REG			%(e?(([abcd]x)|([bs]p)|([sd]i))|([abcd][lh])|{FPREG})
INTEGER		(-|\+)?[0-9]+
LABEL 		[_a-zA-Z][_a-zA-Z0-9]*(\.{INTEGER})?
WS			[ \t]
WWS			{WS}*

ABSADDR		{LABEL}((-|\+)[0-9]+)?

IMMEDIATE	\$({INTEGER}|{ABSADDR})
OFFSET		{INTEGER}|{ABSADDR}

ADDRESS		{OFFSET}?\({WWS}{REG}?{WWS}(,{WWS}{REG}{WWS}(,{WWS}[1248]{WWS})?)?\)

%%

^{WWS}\.file.*\n	{
	/* 
	 * .file indicates the name of the source .C file. Will be used for 
	 * debugging. 
	 */
	char sourcefilename[100];
	sscanf(yytext,"\t.file %s", sourcefilename);
	ECHO;
	fprintf(yyout, "SOURCE_FILENAME:\n\t.ascii %s, \"\\0\"\n", sourcefilename);
	}
	
^{WWS}\.stabd.*\n	{
	/*
	 * OK. This is extremely simple, yet invaluable for debugging. Programs
	 * compiled with "-g" switch have ".stabd <something>" directive embedded
	 * in their .asm. While I'm not acquainted with the exact format, 
	 * .stabd <number>,<number>,<line> indicate the line number of the 
	 * originating source file. A sequence of instructions is generated to
	 * call the kernel routine with <line>, <file> as arguments, and the
	 * kernel will record these numbers (and optionally print appropriate
	 * lines) and dump the lines of code the threads were executing at the
	 * time of the application crash.
	 */
	int d1, d2, line;
	sscanf(yytext,"\t.stabd %d,%d,%d", &d1, &d2, &line);
	ECHO;
	fprintf(yyout, "\tpushfl; pushal # DEBUG_LINE\n");
	fprintf(yyout, "\tpushl $%d; pushl $SOURCE_FILENAME; call _sys_print_source_line; addl $8, %%esp\n", line);
	fprintf(yyout, "\tpopal; popfl   # DEBUG_LINE#\n");
	}

^{WWS}#{WWS}AUG_MSG{WWS}:.*\n	{
	/* 
	 * Message for the parser: DO_AUG or DONT_AUG
	 */
	if (strstr(yytext,"DONT_AUG")) do_aug = 0;
	else if (strstr(yytext,"DO_AUG")) do_aug = 1;
	ECHO;
	++linecount;
	}

^{WWS}{LABEL}:.*\n	{
	/* 
	 * Remember the label and return (it is the beggining of a basic block). 
	 * Note: this implies no statement follows a label. GCC always does so.
	 */
	strcpy(yyinstr.line, yytext);
	islabel = 1;
	++linecount;
	return;
	}


^{WWS}(({LABEL}.*:{WWS})|(\..*)|(#.*)|{WWS})\n 	{
	/* 
	 * asm directives and comments are left unchanged.
	 */
	ECHO;
	++linecount;
	}

^.*$	{	
	/* 
	 * This line is a statement. Remeber the original line,
	 * cut the white space at the beginning off, and switch to
	 * "statement decoding" state.
	 */
	int i;
	if (do_aug) {
		strcpy(yyinstr.line, yytext);
		yyinstr.linenumber = linecount + 1;
		yyinstr.argc = 0;
		yyinstr.name[0] = 0;
		for(i=0; i< MAX_INSTR_ARGS; ++i) {
			yyinstr.argv[i][0] = '\0';
			yyinstr.memo[i]    = 0;
			}
		i = 0;
		while( (yytext[i] == '\t' || yytext[i] == ' ') && yytext[i]) ++i;
		if (!i) {fprintf(stderr,"\aStatement begins at BOL at line %d!\n",linecount+1); exit(1);}
		yyless(i);		/* A line MUST begin with WS. GCC always does so. */
		BEGIN(fetch_name);
		}
	else	{
		ECHO; input(); fprintf(yyout, "\n"); ++linecount;
		}
	}

{WS}	/* cut off all the white space *that remains* */

<fetch_name>[a-z][a-z1]+ 	{	
	/* 
	 * Remember the name of the statement, and switch to
	 * "argument decoding" state. The rest of the line is in the form
	 * <arg>[,<arg>]...
	 */
	strcpy(yyinstr.name, yytext);
	BEGIN(fetch_arg);
	}

<fetch_arg>{ABSADDR}|{ADDRESS}	{
	/* 
	 * Here we come if the operand is an address. Remember the operand
	 * and note that it is an address, so the augmentator can have this
	 * information readily available.
	 */
	strcpy(yyinstr.argv[yyinstr.argc], yytext);
	yyinstr.memo[yyinstr.argc] = 1;
	}

<fetch_arg>{IMMEDIATE}|{REG}	strcpy(yyinstr.argv[yyinstr.argc], yytext);

<fetch_arg>,	{
	++yyinstr.argc;
	if (yyinstr.argc >= MAX_INSTR_ARGS)	{
		fprintf(stderr,"\aStatement \"%s\" at line %d has too many arguments! \n",
		yyinstr.line, linecount+1); exit(1);
		}
	}

<fetch_arg>\n	{
	/* 
	 * Operand decoding complete, now return.
	 */
	++linecount;
	return;
	}

<fetch_arg>.	{
	fprintf(yyout,"# AUG warning: can't handle '%s' in %s (%d)\n", 
		yytext, yyinstr.line, linecount);
	}
	
%%

set_yyinout(FILE *fin, FILE *fout) {
	yyin = fin;
	yyout = fout;
	}
	