/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: debug.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This module contains functions needed for debugging.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <unistd.h>
#include <sys/wait.h>
#include <signal.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#include "kernel.h"
#include "funcdecl.h"

void sys_dump_thread_list(FILE *fout, int entering, int highlite);

static void show_details(int full);

/* monitor: catches SIGQUIT (Ctrl-\) and dumps the system state to the terminak */
static void catch_signal_quit(int signum) { 
    /*
     * TODO: it should have a menu, 
     * like "press this to do that ... or just <Enter> to continue."
     */
    fprintf(sys_ttyout,"\nMonitor: threads' states are:\n");
        { 
            FILE * tempfd = sys_output; sys_output = sys_ttyout;
            show_details(0); 
            sys_output = tempfd;
        }
    fprintf(sys_ttyout,"Simulation continued.\n");
    }       

static void catch_signal(int signum) {  /* fatal error handler */
    if (!isatty(fileno(sys_output))) /* this goes to the redirected file */
        fprintf(sys_ttyout, "\aApplication terminated abnormally. (Reason: %s).\n", strsignal(signum));
    if (signum == SIGINT) 
        PRINTF("Simulation terminated violently.\n");
    else    {
        PRINTF("\n\aApplication terminated: thread %d incurred a %s. Details:\n", 
            sys_thread_ID, strsignal(signum));
        show_details(1);
        }
    sys_simulation_end(-1);
    }

void sys_signal_setup() {

    struct sigaction sigquithandle = {catch_signal_quit, 0, SA_NOMASK, NULL};
    struct sigaction sighandle     = {catch_signal,      0, SA_NOMASK, NULL};

    sigaction(SIGQUIT, &sigquithandle, NULL);       /* catch ^\: */
    sigaction(SIGUSR2, &sigquithandle, NULL);       /* alternatively: kill -12 */

    sigaction(SIGINT,  &sighandle, NULL);           
    sigaction(SIGSEGV, &sighandle, NULL);           
    sigaction(SIGILL,  &sighandle, NULL);           
    sigaction(SIGFPE,  &sighandle, NULL);           
    sigaction(SIGBUS,  &sighandle, NULL);           
    sigaction(SIGIOT,  &sighandle, NULL);           
    sigaction(SIGABRT, &sighandle, NULL);           
    sigaction(SIGINT,  &sighandle, NULL);           
    
    signal(SIGHUP,  SIG_IGN); 
    signal(SIGTERM, SIG_IGN); 
    signal(SIGALRM, SIG_IGN); 
    signal(SIGCHLD, SIG_IGN); 
    signal(SIGPIPE, SIG_IGN);

    /* get a terminal */
    sys_ttyout = fdopen(open(ctermid(NULL), O_WRONLY), "w");
    
    fprintf(sys_ttyout, "Monitor activated. Use Ctrl-\\ (or kill -%d %d if piped) for state dump.\n",
        SIGUSR2, getpid()); 
    }

void sys_debug_source_toggle(int on) {
    sys_current_thread->debug.printsourcelines = on;
    }

void sys_debug_events_toggle(int on) {
    if (sys_thread_ID == 0) sys_debug_events = on;
    }

void sys_debug_memsim_toggle(int on) {
    if (sys_thread_ID == 0) memsim_debug_toggle(on);
    }

static void show_source_line(int thread_ID, char *file, int line);

/* 
 * this one is called directly from the augmented application,
 * if it was compiled with the "-g" switch
 */
void sys_print_source_line(char *filename, int line) {
    sys_current_thread->debug.sourcefile = filename;
    sys_current_thread->debug.sourceline = line;
    if (!sys_current_thread->debug.printsourcelines) return;
    if (sys_debug_events) return;   /* if event dumping is on, the source will be printed later */
    show_source_line(sys_thread_ID, filename, line);
    }

static void show_source_line(int thread_ID, char *file, int line) {
    char linestr[1000]; int l = line;
    FILE *f;
    f = fopen(file, "r");
    do {
        fgets(linestr, 1000, f); --l;
        } while (l);
    fclose(f);
    PRINTF("__%d:%s:%4d: %s", thread_ID, file, line, linestr);
    }

static void show_details(int full) {
    int i; char dumbuff[100];
    sys_dump_thread_list(sys_output, 1,0);
    PRINTF("Data segment size = %d\n", (int)sbrk(0)-DATA_SEGMENT_START);
    if (isatty(fileno(sys_output))) {
        fprintf(sys_ttyout, "Press enter to continue. "); gets(dumbuff);
        }
    if (!full) return;
    for(i=0; i< sys_thread_count; ++i) {
        char * mode_desc = NULL;
        switch( sys_threads[i].debug.mode ) {
            case THMODE_USER:    mode_desc = "USER"; break;
            case THMODE_KERNEL:  mode_desc = "KERNEL"; break;
            case THMODE_MEMSIM:  mode_desc = "MEMSIM"; break;
            }
        PRINTF("Thread %d: in %s\tmode", i, mode_desc);
        if (sys_threads[i].debug.sourcefile) {
            PRINTF("; file %s, line (.stabd #,#,) %d",
                sys_threads[i].debug.sourcefile, sys_threads[i].debug.sourceline);
            }
        PRINTF("\n");
        }
    }

/*
 * This code dumps the state of all threads, and their currently executing
 * request (if not in READY or FINISHED state). It does that in fancy 
 * ANSI colors, which do not look so fancy in a redirected file.
 * But you can always use more -f if you need.
 */
void sys_dump_thread_list(FILE *fout, int entering, int high) {
    int i; char col, *s, *sp=""; unsigned addr, size; char *seg_type, seg_org[100];
    if (entering) fprintf(fout,"\n");
    for(i=0; i<sys_thread_count; ++i) {
        if (i == high) 
            if (entering)
                fprintf(fout,"\x1b[1;41mT%d: [%2d]\x1b[0m\t",i, sys_threads[i].thread_time);
            else 
                fprintf(fout,"\x1b[1;42mT%d: [%2d]\x1b[0m\t",i, sys_threads[i].thread_time);
        else
            fprintf(fout,"T%d: [%2d]\t",i, sys_threads[i].thread_time);
        switch( sys_threads[i].thread_state ) {
            case THREAD_READY:      
                fprintf(fout,"\x1b[1;32m""READY""\x1b[0m"); 
                break;
            case THREAD_FINISHED:   
                fprintf(fout,"\x1b[34m""FINISHED""\x1b[0m"); 
                break;
            case THREAD_BLOCKED:    
                fprintf(fout,"\x1b[34m""BLOCKED""\x1b[0m"); 
                break;
            case THREAD_STALLED:    
            case THREAD_IO:         
                if (sys_threads[i].thread_state == THREAD_IO) 
                    s = "\x1b[1mI/O";
                else
                    s = "STALLED";
                col = '7';
                switch (sys_threads[i].memory_request.type) {
                    case REQUEST_READ  : col = '5'; sp = "(READ)"; break;
                    case REQUEST_WRITE : col = '6'; sp = "(WRIT)"; break;
                    case REQUEST_LOCK  : col = '1'; sp = "(LOCK)"; break;
                    case REQUEST_UNLOCK: col = '3'; sp = "(UNLK)"; break;
                    default : break;
                    }
                /* - no request descriptions anymore */
                fprintf(fout,"\x1b[3%cm%s %s\x1b[0m\t",col, s, sp);
                /**/

                /* now figure out which segment does it belong */
                addr = (unsigned)sys_threads[i].memory_request.virtual_addr;
                seg_org[0] = 0;
                if (ISSHARED(addr)) {   
                    seg_type = "\x1b[1mShared \x1b[0m";
                    sprintf(seg_org, "   (%d)", addr);
                    addr -= (unsigned)DATA_SEGMENT_START;
                    }
                else seg_type = "Private";
                size = sys_threads[i].memory_request.size;

                switch (sys_threads[i].memory_request.type) {
                    case REQUEST_READ: 
                    case REQUEST_WRITE: 
                        fprintf(fout,"%s [&%10u,%d]%s", seg_type, addr, size, seg_org);
                        break;
                    case REQUEST_LOCK: 
                    case REQUEST_UNLOCK: 
                        fprintf(fout,"%s [&%10u]  %s", seg_type, addr, seg_org);
                        break;
                    default: break;
                    }
                break;
            }
            fprintf(fout,"\n");
        }
    /* print source lines, if compiled with -g */
    for(i=0; i< sys_thread_count; ++i)  
            if (sys_threads[i].debug.sourcefile) 
                show_source_line(i, sys_threads[i].debug.sourcefile, sys_threads[i].debug.sourceline);
    fflush(fout);
    }
