/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: interface.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Here are the routines entered into from the augmented assembly code.
 * They make the interface between the application and the kernel.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "kernel.h" 
#include "funcdecl.h"

int sysvar_old_value[3];        /* old contents of the location written onto */
                                /* (copied immediately before the write) */

/*
 *  +------------------------------------------------------------------+
 *  |         Interface for read/write/lock/unlock instructions        |
 *  +------------------------------------------------------------------+
 */ 
void sys_memory_interface(enum Memory_Request_Type request_type, 
    size_t total_size, void *virtual_addr) {
    int isshared;

    ENTER_KERNEL_MODE
    
    isshared = ISSHARED(virtual_addr);

    switch (request_type)   {
        /*
         * Memory references are either simulated (the default), or not
         * simulated, if ALGEVALONLY == 1. Obviously, one should use the
         * second strategy if he does not care about hardware, but instead
         * wants only to evaluate parallel algorithms. In that case, no
         * memory simulator will ever be called.
         */
        #if ALGEVALONLY != 1    
            /* if the user does want to simulate references */
            case REQUEST_WRITE : 
                if (isshared) { /* if shared, copy the new data into the buffer */
                    fastcpy(sys_current_thread->local_buffer,virtual_addr, total_size);
                    /* and restore the old value */
                    fastcpy(virtual_addr, (void*)sysvar_old_value, total_size);
                    }           
                sys_memory_loadstore(REQUEST_WRITE, virtual_addr, total_size, isshared);
                break;
            case REQUEST_READ  : 
                sys_memory_loadstore(REQUEST_READ,  virtual_addr, total_size, isshared);
                if (isshared) { /* if shared, give the data from the buffer */
                    fastcpy(virtual_addr, sys_current_thread->local_buffer, total_size);
                    fastcpy((void*)sysvar_old_value, virtual_addr, total_size);
                    }
                break;
        #else       
            /* fast code for parallel algorithms evaluation only */
            case REQUEST_WRITE : 
                ++sys_current_thread->thread_time;  /* this WRITE will last one cycle */
                if (isshared) {                     /* if shared, copy the new data into the buffer */
                    int sysvar_new_value[3];        /* placeholder for the new value */
                    fastcpy((void *)sysvar_new_value,virtual_addr, total_size);
                    /* and restore the old value */
                    fastcpy(virtual_addr, (void*)sysvar_old_value, total_size);
                    sys_scheduler();                /* will return when becomes the earliest */
                    /* now, put back the new value -- do the actual write */
                    fastcpy(virtual_addr, (void*)sysvar_new_value, total_size);
                    }
                break;
            case REQUEST_READ  : 
                ++sys_current_thread->thread_time;  /* this READ will last one cycle */
                sys_scheduler();
                break;
        #endif ALGEVALONLY != 1
        case REQUEST_LOCK  : 
        case REQUEST_UNLOCK: 
            sys_memory_lock(request_type, virtual_addr); 
            break;
        default : 
            fprintf(stderr, "\aError! Unknown request type %d!\n", request_type);
            exit(1);
        }           

    EXIT_KERNEL_MODE
    }

/*
 * sys_memory_loadstore deals specifically with reads and writes.
 * Apart from the above routine, it is called from some sys_stdlib.c functions 
 * where a string, for example, is accessed byte by byte, an array word by word, 
 * etc.
 */
void sys_memory_loadstore(enum Memory_Request_Type request_type, 
    void *virtual_addr, size_t total_size, int isshared) {
    Thread_Time time_now=0;
    int offset = 0;
    sys_current_thread->memory_request.type = request_type;
    sys_current_thread->memory_request.virtual_addr = (bitsADDR) virtual_addr;
    sys_current_thread->memory_request.isshared = isshared;
    sys_current_thread->memory_request.size = DATA_WIDTH;   /* default, 32bit */
    STATS( time_now = sys_current_thread->thread_time;)

    do  {
        sys_current_thread->thread_state = THREAD_IO;
        if (request_type == REQUEST_WRITE) {
            sys_current_thread->memory_request.data = 0;    /* zero extend */
            fastcpy(&sys_current_thread->memory_request.data,
                    sys_current_thread->local_buffer + offset, DATA_WIDTH);
            }
        if (offset + DATA_WIDTH > total_size)
            sys_current_thread->memory_request.size = total_size - offset;

        sys_scheduler();

        if (request_type == REQUEST_READ)
            fastcpy(sys_current_thread->local_buffer + offset,
                    &sys_current_thread->memory_request.data, DATA_WIDTH);
        offset += DATA_WIDTH;
        if (offset >= LOCAL_BUFFER_SIZE) {
            PRINTF("Error: LOCAL_BUFFER_SIZE in kernel.h too small, 
                should be at least %d\n", total_size); sys_simulation_end(-1);
            }
        sys_current_thread->memory_request.virtual_addr += DATA_WIDTH;
        } while (offset < total_size);

    STATS( 
        int chunks = (total_size-1)/DATA_WIDTH+1;   /* DATA_WIDTH chunks */
        if (request_type == REQUEST_READ) 
            if (isshared)   sys_current_thread->stat.shared_reads   += chunks;
            else            sys_current_thread->stat.private_reads  += chunks;
        else if(isshared)   sys_current_thread->stat.shared_writes  += chunks;
            else            sys_current_thread->stat.private_writes += chunks;
        sys_current_thread->stat.memop_time +=
        sys_current_thread->thread_time - time_now; 
        )

    }

/*
 * Locks are simple.
 */
void sys_memory_lock(enum Memory_Request_Type request_type, void *lock_addr) {
    sys_current_thread->thread_state                = THREAD_IO;
    sys_current_thread->memory_request.type         = request_type;
    sys_current_thread->memory_request.virtual_addr = (bitsADDR)lock_addr;
    sys_current_thread->memory_request.size         = sizeof(bitsLOCK);
    sys_current_thread->memory_request.isshared     = 1;
    sys_scheduler();
    }

/*
 *  +------------------------------------------------------------------+
 *  |           Interface for string operations (REP MOVSL)            |
 *  +------------------------------------------------------------------+
 */ 
void sys_memory_stringop(bitsADDR dst, bitsADDR src, int count, int size) {
    int issharedsrc, isshareddst;

    ENTER_KERNEL_MODE

    issharedsrc = ISSHARED(src);
    isshareddst = ISSHARED(dst);
    sys_current_thread->memory_request.size = size;     /* 4 for l, 2 for w, 1 for b */

    /* now copy SIZE longs/words/bytes from SRC to DST, COUNT times */
    while(count > 0) {
        bitsMAX buffer; 

        /* first a READ, to pass through the simulator */
        sys_current_thread->thread_state = THREAD_IO;
        sys_current_thread->memory_request.type = REQUEST_READ;
        sys_current_thread->memory_request.virtual_addr = src;
        sys_current_thread->memory_request.isshared = issharedsrc;
        sys_scheduler();    
        if (issharedsrc)        
            fastcpy(&buffer, &sys_current_thread->memory_request.data, size);
        else    /* if not shared, memsim returns nothing. Ignore it and take the real data. */
            fastcpy(&buffer, (void *)src, size);

        /* then a WRITE, to pass through the simulator */
        sys_current_thread->thread_state = THREAD_IO;
        sys_current_thread->memory_request.type = REQUEST_WRITE;
        sys_current_thread->memory_request.virtual_addr = dst;
        sys_current_thread->memory_request.isshared = isshareddst;
        sys_current_thread->memory_request.data = 0;    /* zero extend */
        fastcpy(&sys_current_thread->memory_request.data, &buffer, size);
        sys_scheduler();
        if (!isshareddst)   /* if not shared, memsim did nothing so we have to. */
            fastcpy((void *)dst, &buffer, size);

        src += size; dst += size; --count;
        }   

    EXIT_KERNEL_MODE
    }


/*
 *  +------------------------------------------------------------------+
 *  |                 User-defined interface routines                  |
 *  +------------------------------------------------------------------+
 *
 *  You can add your own interfacing routines here (or in a separate
 *  module, which is actually a better idea).
 */
