/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: kernel.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The main module of the execution-driven simulation. Comprises of sections 
 * for startup & cleanup, scheduling, and context switching.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#define extern
#include "kernel.h"
#undef extern
#include "funcdecl.h"

char *INIfile = "limes.ini";  /* default ini file name */
char **startup_argv = NULL;   /* the list of limes args */

#include "utilities.h"

/*
 *  +------------------------------------------------------------------+
 *  |                 INITIALIZATION AND TERMINATION                   |
 *  +------------------------------------------------------------------+
 */ 

static void system_init(int *argc, char *argv[]);
static void system_cleanup();

/* 
 * application's main() gets renamed to application_main() 
 */
int application_main(int argc, char *argv[]);

/* 
 * and this is the real main 
 */
int main(int argc, char *argv[]) {
    void sys_signal_setup();    /* from debug.c */
    system_init(&argc, argv);
    sys_signal_setup();         
    sys_create_thread(NULL);    /* registers current process as thread #0 */
    START_TIMER;
    application_main(argc, argv);
    STOP_TIMER;
    PRINTF("Thread 0 exited happily.\n");
    system_cleanup();
    exit(0);
}

/* 
 * while this is the replacement for application's exit() 
 */
void sys_simulation_end(int exitcode) {
    system_cleanup();
    exit(exitcode);
}

/*
 * initialize system
 */
static time_t simulation_start; 
static int beep_at_end = 1;
static int kernel_is_responsible_for_consistency = 1;

static int Trace_file_type = 0;  /* 1 for text, 2 for binary */
static char *TRACEfilename;
static union Trace_file TRACEfile;

static void system_init(int *argc, char *argv[]) {
    int i;
    /* set up defaults */
    sys_output = stdout;        /* by default, output goes to stdout */
    sys_debug_events = 0;
    sys_gather_statistics = 0;
    
    /* parse command line options: "--" option means the rest is for Limes */
    for(i=1; i < *argc; ++i) 
        if (!strcmp(argv[i],"--"))  {
            int appargc = i++;
            startup_argv = &argv[i];    /* remember the rest of the array */
            /* Analyze Limes command line options */
            while(i < *argc) {
                if (strcmp (argv[i],(char*)&("-b")) == 0)
                    beep_at_end = 0;
                else if ((argv[i][0] == '-') && (argv[i][1] == 'd')) /* initialization values in form -dITEM=VALUE */
                    ;
                else if (strcmp (argv[i],(char*)&("-e")) == 0) /* force output to stderr */
                    sys_output = stderr;
                else if ((strcmp (argv[i],(char*)&("-h")) == 0) ||
                         (strcmp (argv[i],(char*)&("--help")) == 0))
                    goto PRINT_OPTIONS;
                else if (strcmp (argv[i],(char*)&("-i")) == 0) /* INI file other than limes.ini */
                    INIfile = argv[i+++1];
                else if (strcmp (argv[i],(char*)&("-s")) == 0) /* statistics on */
                    sys_gather_statistics = 1;
                else if (strcmp (argv[i],(char*)&("-t")) == 0) { /* output text trace to stdout */
                    Trace_file_type = 1;
                    TRACEfilename = "";
                    TRACEfile.txt = stdout;
                }
                else if (strcmp (argv[i],(char*)&("-tx")) == 0) { /* produce text trace file */
                    Trace_file_type = 1;
                    if ((i < (*argc-1)) && (argv[i+1][0] != '-'))
                        TRACEfilename = argv[i+++1];
                    else 
                        TRACEfilename = "TRACE.txt";
                    if ((TRACEfile.txt = fopen (TRACEfilename,"w")) == NULL) {
                        fprintf (stderr, "Error opening trace file %s\n", TRACEfilename);
                        exit (1);
                    }
                }
                else if (strcmp (argv[i],(char*)&("-tb")) == 0) { /* produce binary trace file */
                    Trace_file_type = 2;
                    if ((i < (*argc-1)) && (argv[i+1][0] != '-'))
                        TRACEfilename = argv[i+++1];
                    else 
                        TRACEfilename = "TRACE.bin";
                    creat (TRACEfilename,0666); /* brrrr */
                    if ((TRACEfile.bin = open (TRACEfilename,1)) == -1) {
                        fprintf (stderr, "Error opening trace file %s\n", TRACEfilename);
                        exit (1);
                    }
                }
                else {
                    fprintf(stderr, "\nError - Unknown option %s",argv[i]);
                    PRINT_OPTIONS :
                    fprintf(stdout, "\n");
                    fprintf(stdout, "Syntax:\n");
                    fprintf(stdout, "<application> <app args> [ -- [-b][-ditem=value][-e][-i][-s] [-t]|[-t[x|b] outfile] [-l[y|n]] [-h]|[--help] ]\n");
                    fprintf(stdout, "Options:\n");
                    fprintf(stdout, "\t-b   beep at end of simulation\n");
                    fprintf(stdout, "\t-d   specifies item's value (overrides one in .ini file)\n");
                    fprintf(stdout, "\t-e   force output to stderr instead of stdout\n");
                    fprintf(stdout, "\t-h   print this help message\n");
                    fprintf(stdout, "\t-i   specifies initialization file other than limes.ini\n");
                    fprintf(stdout, "\t-s   dump threads' statistics\n");
                    fprintf(stdout, "\t-t   produce trace in text format and send it to stdout\n");
                    fprintf(stdout, "\t-tx  produce trace file in text format\n");
                    fprintf(stdout, "\t-tb  produce trace file in binary format\n");
                    fprintf(stdout, "\t--help  print this help message\n");
                    exit (1);
                }
                ++i;
            }            
            *argc = appargc; argv[appargc] = NULL;  /* the app won't see "--" options */
            break;
        }

    PRINTF("\n****************************************************************************\n");
    PRINTF("Simulation start: ");
    for(i=0; argv[i]; ++i) PRINTF("%s ",argv[i]); 
    if (startup_argv) {
        PRINTF("-- "); for(i=0; startup_argv[i]; ++i) PRINTF("%s ",startup_argv[i]); 
    }
    PRINTF("\n");

    if (!sys_output) sys_output = stdout;
    sys_thread_count = 0;
    if (memsim_init() == Memsim_Maintains_Memory_Consistency)
        kernel_is_responsible_for_consistency = 0;
    simulation_start = time(NULL);

    PRINTF("MAX_CPU_COUNT=%d, BUSY_LOCK_TIME=%d, %sBUSY_LOCK_OPTIMIZATION\n",
        MAX_CPU_COUNT, BUSY_LOCK_SLEEP_TIME, BUSY_LOCK_OPTIMIZATION ? "" : "NO ");
    PRINTF("----------------------------------------------------------------------------\n");
    FFLUSH;
}

/*
 * clean up
 */
static void system_cleanup() {
    PRINTF("----------------------------------------------------------------------------\n");
    STATS(sys_dump_statistics();)
    memsim_cleanup();
    PRINTF("Simulation end. Elapsed time: %lu seconds.\n", 
        time(NULL)-simulation_start);
    if (beep_at_end) fputc('\a', sys_ttyout);   /* wake up the one who simulates */
}

/*
 * Dump threads' statistics
 */
void sys_dump_statistics() {
    int i;
    PRINTF("**** THREADS STATISTICS: \n");
    for (i=0; i<sys_thread_count; ++i) {
        Thread_Time sync_time;
        PRINTF("Thread %d:\n", i);
        PRINTF("\tshared  reads: %10d, \tshared  writes: %10d\n",
            sys_threads[i].stat.shared_reads, sys_threads[i].stat.shared_writes);
        PRINTF("\tprivate reads: %10d, \tprivate writes: %10d\n",
            sys_threads[i].stat.private_reads, sys_threads[i].stat.private_writes);
        PRINTF("\txplicit locks: %10d, \tbarriers      : %10d\n",       
            sys_threads[i].stat.locks, sys_threads[i].stat.barriers);
        PRINTF("\tTIMES: total: %d, mem: %d, xpl.locks: %d, barriers: %d\n",
            sys_threads[i].thread_time,     sys_threads[i].stat.memop_time,
            sys_threads[i].stat.lock_time,  sys_threads[i].stat.barrier_time);
        sync_time = sys_threads[i].stat.lock_time + 
                    sys_threads[i].stat.barrier_time;
        if (sync_time >0)
            PRINTF("\tTime spent in synchronization: %5.2f%%\n",
                100.0*sync_time / (double)sys_threads[i].thread_time);
    }
    PRINTF("**** END OF STATISTICS\n\n");
}
    
/*
 *  +------------------------------------------------------------------+
 *  |                       CONTEXT SWITCHING                          |
 *  +------------------------------------------------------------------+
 */ 

static volatile int saved_esp[MAX_THREAD_COUNT];    /* threads' stack pointers */
static volatile int thread_to_wakeup_ID;

/*
 * This function stores, for the current thread, all registers on thread's
 * stack, and his stack pointer to saved_esp[his id]. Then it restores
 * the stack pointer from the thread to be woken up, and pops his registers
 * off the stack.
 */
void sys_swap();
asm  (  "_sys_swap:\n\t"
                "pushal\n\t"
                "movl _sys_thread_ID,%eax\n\t"
                "movl %esp, _saved_esp(,%eax,4)\n\t"
                "movl _thread_to_wakeup_ID,%eax\n\t"
                "movl _saved_esp(,%eax,4), %esp\n\t"
                "popal\n\t"
                "ret\n"
            );
/*
 * Macros for saving and restoring FPU (80387) status.
 */
#define SAVE_FPU_REGS(ThrID)    \
    asm("fnsave %0" : "=m" (sys_threads[ThrID].fpu_state))
#define RESTORE_FPU_REGS(ThrID) \
    asm("frstor %0" : "=m" (sys_threads[ThrID].fpu_state))

/*
 * This one creates the stack for a new child. It will be 64K*child_ID
 * bytes below the parent's.
 */
static void sys_create_childs_stack(int new_thread_ID, void (*func)()) {
    volatile int temp_sp, new_sp;   
    asm __volatile__ ("nop; movl %%esp, %0" : "=g" (temp_sp) );
    new_sp = temp_sp - new_thread_ID * MAX_STACK_PER_THREAD; /* preferences.h */
    asm __volatile__ (
        "movl %0, %%esp\n\t"
        # if 0
            "pushl %1\n\t"
        # else                          /* -O2 screws %1 */
            "pushl $_sys_create_thread_conted\n\t"
        # endif
        "pushal\n\t"
        "movl %%esp, %2\n\t"            
        "movl %3, %%esp\n\t"            
        : "=g" (new_sp) , "=g" (func) , "=g" (new_sp) , "=g" (temp_sp)
        );
    saved_esp[new_thread_ID] = new_sp;  
    SAVE_FPU_REGS(new_thread_ID);
}

/*
 * This is actually an extension for sys_swap(); sys_swap() had to be
 * pure asm, so only a minimum of the work is put there. 
 * TODO: AUG could supply the information whether the FPU is used in
 * the next basic block so that we do not save/restore all the FPU
 * regs and states in vain.
 */
static void sys_schedule(int given_thread_to_wakeup_ID) {
    if (sys_thread_ID == given_thread_to_wakeup_ID) return;
    thread_to_wakeup_ID = given_thread_to_wakeup_ID;
    SAVE_FPU_REGS(sys_thread_ID);
    sys_swap();     /* another thread comes here */
    sys_thread_ID = thread_to_wakeup_ID;
    RESTORE_FPU_REGS(sys_thread_ID);
    sys_current_thread = &sys_threads[sys_thread_ID];
}

/*
 *  +------------------------------------------------------------------+
 *  |                 THREADS CREATION AND TERMINATION                 |
 *  +------------------------------------------------------------------+
 */ 
/* 
 * This is where the new threads bear and where the Parent registers himself
 * as thread #0. It is called through ANL's CREATE(function) macro.
 */
void sys_create_thread( void (* ThreadFunction)() )  {
    Thread_Time parent_time;
    int new_thread_ID, threads_parent_ID;
    struct Thread_Struct * new_current_thread;
    static void sys_create_thread_conted();
    
    threads_parent_ID = sys_thread_ID;
    new_thread_ID = sys_thread_count++;     
    parent_time = ( new_thread_ID == 0 ? 0 : sys_current_thread->thread_time );

    if (new_thread_ID >= MAX_THREAD_COUNT) {
        fprintf(stderr, "CREATE failed: too many threads.\n"
            "Current limit is MAX_THREAD_COUNT=%d. "
            "Please edit kernel/preferences.h\n", MAX_THREAD_COUNT);
        sys_thread_count = MAX_THREAD_COUNT-1;
        sys_simulation_end(-1);
    }
    
    new_current_thread = &sys_threads[new_thread_ID];
    new_current_thread->thread_ID = new_thread_ID;
    new_current_thread->proc_ID = new_thread_ID;        /* processor ID */
                            /* (opened for enhancements in the future) */
    new_current_thread->ThreadFunction = ThreadFunction;

    new_current_thread->thread_time = 
        initial_thread_time(parent_time, new_thread_ID); 
    /* 
     * this normally yields parent_time, but the user can adjust it to reflect
     * a real system behavior, since subsequent creations occur after
     * many cycles in real systems.
     * (initial_thread_time is defined in memsim.cc)
     */  
           
    new_current_thread->thread_state = THREAD_READY;
    new_current_thread->memory_request.pid = new_thread_ID;
    new_current_thread->memory_request.type = REQUEST_NONE;
    new_current_thread->global_operation.type = GLOBAL_NONE;
    
    DEBUG(new_current_thread->debug.mode = THMODE_KERNEL;)
    DEBUG(new_current_thread->debug.sourcefile = NULL;)
    DEBUG(new_current_thread->debug.printsourcelines = 0;)

    if (new_thread_ID == 0) {   /* Thread#0 does not call any function now */
        sys_thread_ID = 0;
        sys_current_thread = &sys_threads[0];
        return; 
    }
    
    /* 
     * A DSM memsim must be aware that the new process is created.
     * This function is empty by default.
     */
    memsim_fork(threads_parent_ID, new_thread_ID);

    /* Add global CREATED operation to the thread structure */
    SYS_CREATED(threads_parent_ID,new_thread_ID)

    /*
     * Now create the stack for the child. This stack will contain all
     * the registers, and the return point will be sys_create_thread_conted
     * function below.
     */
    sys_create_childs_stack(new_thread_ID, sys_create_thread_conted);
    return;
}

static int finished_thread_count = 0;   /* number of threads that have completed */

/*
 * OK, here is where the newly born really becomes live, 
 * and where it will quietly die when its job is finished.
 */
static void sys_create_thread_conted() {
    sys_thread_ID = thread_to_wakeup_ID;
    sys_current_thread = &sys_threads[sys_thread_ID];
    RESTORE_FPU_REGS(sys_thread_ID);

    DEBUG(sys_current_thread->debug.mode = THMODE_USER;)
    START_TIMER;                                    /* fasten your seatbelt */
    (* sys_current_thread->ThreadFunction)();       /* GO! */
    STOP_TIMER;                                     /* OK, our job is over */
    DEBUG(sys_current_thread->debug.mode = THMODE_KERNEL;)

    sys_current_thread->thread_state                 = THREAD_FINISHED;
    sys_current_thread->memory_request.type          = REQUEST_NONE;
    sys_current_thread->memory_request.pid           = sys_thread_ID;
    sys_current_thread->memory_request.requested_now = 0;

    if( ++finished_thread_count == sys_thread_count -1) {
        /*
         * If you are the last child, then the parent is either waiting for
         * me to finish, or is waiting to continue his work.
         */
        if (sys_threads[0].thread_state == THREAD_FINISHED) {
            /* the parent has been waiting for me. Set his time to mine. */
            sys_threads[0].thread_time  = sys_current_thread->thread_time;
            sys_schedule(0);                          /* schedule parent */
            /* never gets here */
        }
        else
            sys_scheduler();    /* he still has some memory operations to do */
    }
    else                        /* I'm not the last child */
        sys_scheduler();        /* so I schedule the others for execution */
    /* never gets here (hopefully) */
}   

void sys_wait_for_end() {   /* called by the parent only with WAIT_FOR_END */
    DEBUG(sys_current_thread->debug.mode = THMODE_KERNEL;)
    if (finished_thread_count < sys_thread_count - 1) {
        sys_current_thread->thread_state = THREAD_FINISHED; 
        sys_scheduler();                /* activate the others */
        sys_current_thread->thread_state = THREAD_READY;
    }
    DEBUG(sys_current_thread->debug.mode = THMODE_USER;)
}       

/*
 *  +------------------------------------------------------------------+
 *  |                     The Core:  SCHEDULER                         |
 *  +------------------------------------------------------------------+
 */ 
/*
 * This code selects the thread(s) with the smallest time (there may be
 * more than one with minimial time). If there is a READY thread within
 * them, it is scheduled for exectuion. Otherwise, these threads are 
 * waiting for the memory subsystem and their requests are scheduled
 * for simulation (by calling the memory simulator).
 * We have two time quantities to think about: the first one is `time_now',
 * which equals the time of the earliest (unfinished) thread.
 * The second is `last_simulation_time', which represent the last cycle
 * the memory simulator simulated, plus one. If time_now is greater than
 * last_simulation_time, that means that at the instant last_simulation_time
 * all the processors' requests were satisfied; but this does not
 * necessarily mean that at that point of time the memsim returned to
 * its initial state. Consider, for example, the following situation:
 * Thread#0 has a WRITE at time=100, and thread#1, which just entered the
 * scheduler, has a READ request and his current time is 120. So we pick
 * thread#0, for his time is smaller, and schedule his WRITE by calling
 * memsim with time_now == 100. The memsim always simulate only up to
 * the current cycle, so it returns, saying that the thread#0's WRITE 
 * is satisfied. The current time is now 101, which equals the last_
 * _simulation_time. Thread#0 is now ready, and is scheduled for execution.
 * It may return after, say, 30 cycles, waiting for another operation.
 * In this next iteration of scheduling, time_now will be 120, i.e. the
 * time of thread#1, i.e. the minimal time. Now we call the memsim with
 * last_simulation_time=101, time_now=120. Why? Because the memsim might
 * have returned OK for the write just to let the processor go on, but in
 * deed he might have had to do some other things, like updating other
 * caches or memory, which last more than one cycle. So it will do the rest
 * of the WRITE now, at time=101, and will simulate up to time=120, where
 * another request exists.
 * (Of course, this gives memsim the oportunity to optimize himself: if,
 * while doing the loop from 101 to 120, he detects at time=105 that 
 * all the memory elements (caches etc.) are in initial state, it will
 * know that nothing can happen before time=120, and will set his time to
 * 120, simulate this one cycle and return.)
 */
static Thread_Time last_simulation_time = 0;    

void sys_scheduler() {
    void sys_dump_thread_list(FILE * fout, int entering, int highlite);  /* from debug.c */
    Thread_Time time_now=0; int j, any_unfinished = 0;
    register int i;
    int ready_thread_ID = -1;       /* index of the earliest ready thread */
    int busy_lockers_count, requesters_count;

    if (sys_debug_events) sys_dump_thread_list(sys_output, 1, sys_thread_ID);

    /* find the time of the first unfinished, non-blocked thread */
    for(j=0; j < sys_thread_count; ++j) 
        if ((sys_threads[j].thread_state != THREAD_FINISHED) &&
            (sys_threads[j].thread_state != THREAD_BLOCKED)) {
            time_now = sys_threads[j].thread_time;
            any_unfinished = 1;
            if (sys_threads[j].thread_state == THREAD_READY) 
                ready_thread_ID = j;
            break;
        }
    if (!any_unfinished) return;    /* no such! But there should be. */ 

    /* then find the earliest's time (among the unfinished, non-blocked) */
    for (i=j+1; i < sys_thread_count; ++i)
        if (sys_threads[i].thread_time <= time_now          && 
           (sys_threads[i].thread_state != THREAD_FINISHED) &&
           (sys_threads[i].thread_state != THREAD_BLOCKED)) {
            if (sys_threads[i].thread_state == THREAD_READY) 
                ready_thread_ID = i;
            else if (sys_threads[i].thread_time < time_now)
                ready_thread_ID = -1;
            time_now = sys_threads[i].thread_time;
        }
    
    /* 
     * if there is a ready thread with the time equal to the earliest time,
     * then schedule him and return.
     */
    if ( ready_thread_ID > -1 ) {
        sys_schedule(ready_thread_ID);
        return; /* this thread will return here after his request is satisfied */
    }

    /* 
     * OK, now we enter the main scheduling loop. 
     */
    do {
        /* 
         * adjust the memory request structures for each thread, and check 
         * if a ready thread (which has been waiting till now) appeared
         */
        for (i=0; i<sys_thread_count; ++i) {
            if (sys_threads[i].thread_time == time_now) {
                if (sys_threads[i].thread_state == THREAD_READY) {
                    ready_thread_ID = i;    /* one ready found */
                    sys_threads[i].memory_request.requested_now = 0;
                }
                else if (sys_threads[i].thread_state == THREAD_IO) {
                    sys_threads[i].memory_request.requested_now = 1;
                }
                else
                    sys_threads[i].memory_request.requested_now = 0;
            }
            else {
                sys_threads[i].memory_request.requested_now = 0;
            }
        }

        if ( ready_thread_ID > -1 ) {   /* if there is a READY, activate him */
            sys_schedule(ready_thread_ID);
            return; 
        }

        for (i=0; i<sys_thread_count; i++)
            if (sys_threads[i].memory_request.requested_now) {
                /* Update information about address area */
                sys_threads[i].memory_request.area = 
                  sys_threads[i].memory_request.virtual_addr < DATA_SEGMENT_START ? 'c' : 
                  (sys_threads[i].memory_request.virtual_addr > DATA_SEGMENT_START + DATA_SEGMENT_MAXSIZE ? 's' : 'd');
                /* Update CREATE time */
                if (sys_threads[i].global_operation.type == GLOBAL_CREATED)
                    sys_threads[i].global_operation.params.created.time = time_now;
            }

        /*
         * Now the memory simulator has all the necessary information:
         * whether a thread on a processor is issuing a request,
         * and what are the parameters of that request.
         */

        /* Produce trace file if such option was given */
        if (Trace_file_type)
            produce_trace(
                TRACEfilename,
                TRACEfile,
                Trace_file_type,
                time_now
            ); /* from utilities.h */

        DEBUG(sys_current_thread->debug.mode = THMODE_MEMSIM;)
        memory_simulate(last_simulation_time, time_now);
        DEBUG(sys_current_thread->debug.mode = THMODE_KERNEL;)

        last_simulation_time = time_now + 1;

        /* 
         * This lasted only one virtual cycle. Now we collect the information
         * which of the requests were satisfied (the candidates are either
         * those who issued a request in this cycle, or who were stalled).
         */
        busy_lockers_count = requesters_count = 0;
        /*
         * Now: if the kernel maintains consistency, it will perform all the
         * shared WRITEs finished in *this* cycle to the real memory.
         * Do it before other requests (like READ!) are processed.
         */
        if (kernel_is_responsible_for_consistency)  
            for (i=0; i<sys_thread_count; ++i) 
                if( (sys_threads[i].memory_request.requested_now ||
                        (sys_threads[i].thread_state == THREAD_STALLED) )  &&
                    (sys_threads[i].memory_request.satisfied 
                        == PROCESSOR_SATISFIED) &&
                    (sys_threads[i].memory_request.type == REQUEST_WRITE) &&
                    sys_threads[i].memory_request.isshared
                    )
                    fastcpy((void *)sys_threads[i].memory_request.virtual_addr, 
                            &sys_threads[i].memory_request.data,
                            sys_threads[i].memory_request.size);
        for (i=0; i<sys_thread_count; ++i) {
            if ( (sys_threads[i].memory_request.requested_now) ||
                 (sys_threads[i].thread_state == THREAD_STALLED) ) {
                /* 
                 * if there was a request, check if it is satisfied 
                 */
                sys_threads[i].memory_request.requested_now = 0;
                sys_threads[i].global_operation.type = GLOBAL_NONE;
                ++requesters_count;
                switch (sys_threads[i].memory_request.satisfied) {
                    case PROCESSOR_SATISFIED:
                        sys_threads[i].thread_state = THREAD_READY;
                        ready_thread_ID = i;
                        /*
                         * If the kernel maintains consitency, if this
                         * was a READ and if it was shared, read the actual
                         * contents from the main, real memory, and put it
                         * in thread's read buffer.
                         * The memsim should only maintain lock values 
                         * consistent (if the REALISTIC sync model is used);
                         * kernel takes care of WRITEs and READs.
                         */
                        if (kernel_is_responsible_for_consistency &&
                            (sys_threads[i].memory_request.type == REQUEST_READ) &&
                            sys_threads[i].memory_request.isshared
                            ) 
                            fastcpy(&sys_threads[i].memory_request.data,
                                (void *)sys_threads[i].memory_request.virtual_addr, 
                                sys_threads[i].memory_request.size);
                        break;
                    case PROCESSOR_LOCK_BUSY:
                        sys_threads[i].thread_state = THREAD_IO;
                        /* test&set satisfied, but the lock was busy: */
                        /* increase his time by [one] (loop: test&set; jr nz, loop) */
                        +sys_threads[i].thread_time += BUSY_LOCK_SLEEP_TIME;
                        busy_lockers_count++;
                        break;
                    case PROCESSOR_STALLED:
                        sys_threads[i].thread_state = THREAD_STALLED;
                        break;
                    default:
                        break;
                    }
                ++sys_threads[i].thread_time;
                }
            }
        ++time_now;         

#if BUSY_LOCK_OPTIMIZATION
        if (busy_lockers_count == requesters_count) {
            /*
             * Of N threads, K have the minimal time and have been scheduled.
             * But all of them have requested a busy lock. Now let tm be
             * the minimal time for the other (N-K) threads, somewhen
             * in the future. Since all the locks requested by the K threads
             * will remain busy at least by the time tm, we can safely
             * set current time (and the time for the K threads) to tm.
             * This optimization may greatly speed up the simulation, and,
             * in additition, is capable of detecting a deadlock. 
             * See also preferences.h.
             */
            Thread_Time min_nonlock_time, inf_time = (unsigned)(-1);    /* set inital to +infinite */
            min_nonlock_time = inf_time;
            for(j=0; j < sys_thread_count; ++j) {
                if ( (sys_threads[j].thread_state != THREAD_FINISHED)   &&
                     (sys_threads[j].thread_state != THREAD_BLOCKED)    &&
                    !( (sys_threads[j].thread_state == THREAD_IO) &&
                       (sys_threads[j].memory_request.type == REQUEST_LOCK) &&
                       (sys_threads[j].thread_time - BUSY_LOCK_SLEEP_TIME == \
                            time_now) )
                     ) 
                {
                     if (sys_threads[j].thread_time < min_nonlock_time) 
                        min_nonlock_time = sys_threads[j].thread_time;
                }
            }
            if (min_nonlock_time == inf_time) {         /* it could be a deadlock */
                if (busy_lockers_count < (sys_thread_count - finished_thread_count) )
                    goto end_of_busy_lock_optimization; /* but only if not all live threads were unsuccessful lockers */
                fprintf(sys_ttyout, "\aDeadlock detected! Press ^C.\n"); fflush(stdout); 
                while(1);
            }
            time_now = min_nonlock_time;
            for(j=0; j < sys_thread_count; ++j) 
                if ( (sys_threads[j].thread_state == THREAD_IO) &&
                     (sys_threads[j].memory_request.type == REQUEST_LOCK) &&
                     (sys_threads[j].thread_time - BUSY_LOCK_SLEEP_TIME <= time_now) )
                    sys_threads[j].thread_time = time_now;
        }
            end_of_busy_lock_optimization:
#endif BUSY_LOCK_OPTIMIZATION 

    } while (ready_thread_ID < 0);  /* repeat calls to memsim until one gets ready */
    sys_schedule(ready_thread_ID);
}


/*
 *  +------------------------------------------------------------------+
 *  |                          MISCELLANEOUS                           |
 *  +------------------------------------------------------------------+
 */ 

void sys_update_thread_time(Thread_Time elapsed_thread_time) {
    sys_current_thread->thread_time += elapsed_thread_time;
}
    
Thread_Time sys_get_thread_time() {
    return sys_current_thread->thread_time + sys_IC;
}   

int sys_get_thread_ID() {
    return sys_thread_ID;
}

/*
 *  +------------------------------------------------------------------+
 *  |                   DEALING WITH BLOCKED THREADS                   |
 *  +------------------------------------------------------------------+
 */ 
/*
 * BLOCKED threads are not realistic if the number of threads is equal
 * to the number of processors; however, for the purpose of various 
 * experiments with synchronization, it may be useful to have a thread
 * blocked when a lock is busy, for example, and to wake him up when
 * the lock gets freed. BLOCKED threads will be ignored by the kernel
 * until they become READY.
 * There are two functions for blocking and unblocking threads, which
 * resemble those used by any UNIX kernel:
 * 1) sys_thread_sleep() puts current thread to sleep (by marking its state
 *    BLOCKED). It will return when another thread wakes him up (by marking
 *    its state READY and when the scheduler picks him up for execution
 *    thereafter).
 * 2) sys_thread_wakeup(thread_ID) will set the state of the thread whose
 *    thread ID equals 'thread_ID' to READY and update its time, but will not 
 *    reschedule him immediately; this allows for multiple blocked threads
 *    to be awaken at once, as in the case of a barrier.
 */

void sys_thread_sleep() {
    sys_current_thread->thread_state = THREAD_BLOCKED;
    sys_scheduler();    
    /* will return when it becomes READY */
}
                    
void sys_thread_wakeup(int thread_to_wakeup_ID) {
    sys_threads[thread_to_wakeup_ID].thread_state = THREAD_READY;
    sys_threads[thread_to_wakeup_ID].thread_time = 
        sys_current_thread->thread_time;    /* update the time of the awaken */
}
