#ifndef _KERNEL_H_
#define _KERNEL_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: kernel.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This header file is #included by most of the kernel modules. 
 * It defines structures for description of memory requests, threads,
 * some global variables, macros, and declarations for various functions.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <stdio.h>

#include "preferences.h"
#include "types.h"
#include "timing.h"

/*
 *  +------------------------------------------------------------------+
 *  |                    MEMORY REQUEST DATA TYPES                     |
 *  +------------------------------------------------------------------+
 */ 
enum PE_response {          /* MemSim's PE[ID] response to a request */
    PROCESSOR_NORESPONSE,   /* no response due to no request */
    PROCESSOR_STALLED,      /* processor stalled due to a read miss etc. */
    PROCESSOR_SATISFIED,    /* memory operation completed */
    PROCESSOR_LOCK_BUSY     /* memory reference complete but the lock is busy */
};

struct Memory_Request {     /* structure describing the request, filled by the thread upon a memory reference */
    enum Memory_Request_Type type;  /* READ/WRITE/LOCK/UNLOCK/... */
    bits32 virtual_addr;    /* set by the kernel */
    bits32 addr;            /* computed by the memsim, if TLBs are simulated */
    int    pid;             /* process ID (== sys_thread_ID) -- for TLBs */
    int    size;            /* size of one data chunk - 1..MAX_DATA_WIDTH */
    char   area;            /* c=code, s=stack, d=data - written by the scheduler */
    int    isshared;        /* 1 if it is, 0 if it is not */
    bitsMAX data;           /* data buffer, for reads and writes */
    int requested_now;      /* this one is set by the scheduler when appropriate */
    enum PE_response satisfied; /* and this one is set by the memory simulator */
};

struct Global_Operation {   /* structure of global synchronization operations */
    enum Global_Synchro_Type type;
    union {
        struct {
            bitsADDR virtual_addr;
            char     order;
            char     total;
        } barrier;
        struct {
            int         parent;
            Thread_Time time;
        } created;
        /* here should be parameters of other global synchro operations */
    } params;
};

/* handy fast memcpy function */
static inline void fastcpy(void *dst, void *src, int size) {
    switch (size) {
        case 4: *(bits32 *)dst = *(bits32 *)src; break;
        case 2: *(bits16 *)dst = *(bits16 *)src; break;
        case 1: *(bits8  *)dst = *(bits8  *)src; break;
        case 8: *(bits64 *)dst = *(bits64 *)src; break;
        default: memcpy(dst, src, size); break;
    }
}

/*
 *  +------------------------------------------------------------------+
 *  |                           THREADS                                |
 *  +------------------------------------------------------------------+
 */ 
enum Thread_State {         /* possible states a thread can be in */
    THREAD_READY,           /* ready to be scheduled for execution */
    THREAD_IO,              /* issuing a read/write/lock/unlock request */
    THREAD_STALLED,         /* waiting for a memory operation */
    THREAD_FINISHED,        /* completed its execution */
    THREAD_BLOCKED          /* artificial: blocked: not ready until awakened */
};

enum Thread_Mode {          /* used for debugging. Extend them if you want */
    THMODE_USER, THMODE_KERNEL, THMODE_MEMSIM
};

struct Thread_Struct {      /* structure describing a thread */
    int thread_ID;          /* thread's logical number */
    int proc_ID;            /* thread's processor ID. Currently proc_ID == thread_ID */

    void (*ThreadFunction)();   /* function to be called upon creation */
    
    Thread_Time thread_time;        /* simulated thread time */
    enum Thread_State thread_state; /* state of the tread */
    int fpu_state[50];              /* FPU state */
    
    struct Memory_Request   memory_request;   /* request for the thread's processor */
    struct Global_Operation global_operation; /* global operation that memory requests are realizing */

    #define LOCAL_BUFFER_SIZE 32          /* actually, 10 is sufficient */
    char local_buffer[LOCAL_BUFFER_SIZE]; /* intermediate buffer for load/store */

    struct {                        /* debugging */
        enum Thread_Mode mode;      /* mode: USER/KERNEL/MEMSIM... */
        char *sourcefile;       
        int sourceline;
        int printsourcelines;
    } debug;

    /* statistics */
    struct {
        int shared_reads, shared_writes, private_reads, private_writes;
        int locks, barriers;
        Thread_Time memop_time, lock_time, barrier_time;
    } stat;
};

extern struct Thread_Struct sys_threads[MAX_THREAD_COUNT]; /* the table of threads */
extern int sys_thread_count;                        /* number of threads */
extern int sys_thread_ID;   /* number of the currently executing thread */
extern struct Thread_Struct * sys_current_thread;   /* == sys_threads[sys_thread_ID] */

/*
 * The following pair of macros is to be used at the beginning and the end
 * of each kernel function that is called directly from the "user mode".
 * sys_IC is contains the information how many cycles a thread has spent
 * during its last trip to user mode.
 */
#define ENTER_KERNEL_MODE                                       \
    DEBUG(sys_current_thread->debug.mode = THMODE_KERNEL;)      \
    sys_update_thread_time(sys_IC);

#define EXIT_KERNEL_MODE                                        \
    sys_IC = 0; /* reset the instruction counter and return to user mode */\
    DEBUG(sys_current_thread->debug.mode = THMODE_USER;)

/* 
 * Definitions regarding trace files 
 */
#define MAX_FILENAME_LENGTH 64

/* Trace files can be of text or binary type */
union Trace_file {
    FILE* txt;
    int   bin;
};

/* This is internal structure of one processor request in a 'per processor' trace file 
   Attention: Don't change the order of elements in a structure, as this order
   gives size of 12B, and some other order may occupy 16B or even 20B! 
   Explanation: Compiler gets elements of a structure one by one and tries to fit 
   them together in 1 word (4B). If the next element can't fit in the rest of the 
   word, it will occupy next 4 bytes, while the remaining space in the previous 
   word will remain unused. It all results in an effect that structure occupies 
   more memory then necessary. This is important when a file consists of great
   number of such elements */
struct P_event {
    Thread_Time time;
    bitsADDR address;
    char     request;
    char     area;
    char     size;
    char     global;
};

/* This is internal structure of an "additional" data needed to describe a 
   global operation that current processor's request belongs to. In a 'per
   processor' trace file this data is placed after P_event type of structure,
   only if global!=GLOBAL_NONE */
union P_global {
    struct {
        bitsADDR address;
        char     order;
        char     total;
    } barrier;
    struct {
        int         parent;
        Thread_Time time;
    } created;
    /* here should be parameters of other global synchro operations */
};

/*
 *  +------------------------------------------------------------------+
 *  |                 DEALING WITH THE DATA SEGMENT                    |
 *  +------------------------------------------------------------------+
 */ 

#ifdef extern           /* make sure _etext is really extern */
    #undef extern
    extern int _etext;  /* it is set by the linker, and denotes the end of the text segment */
    extern int _end;    /* _end is at the end of the data segment */
    #define extern      /* of the text segment, whereafter the data segment follows */
#else
    extern int _etext;
    extern int _end;
#endif
/* 
 * data segment start: references above this address and below
 * DATA_SEGMENT_MAXSIZE (512MB) will be simulated to the level of contents, 
 * where the others will be simulated to the level of address only, if at all.
 * (see ISSHARED macro).
 * _etext would save some memory for the memsim's copy if it were page aligned.
 */
#define DATA_SEGMENT_START (bitsADDR)&_etext 
#define DATA_SEGMENT_MAXSIZE (512*1024*1024)
#define DATA_HEAP_START (bitsADDR)&_end /* some simulators may make use of this info */

/*
 * Macro determinining whether an adress is shared - in the sense that it could
 * be both read and written by all. Locations other than those in the data
 * segment could be shared as well (like constants in the code segment), 
 * but they are read only and need not be simulated to the level of contents.
 * Actually, the second line should read "START+MAXSIZE", but since MAXSIZE
 * is 512MB, DATA_SEGMENT_START is much less than DATA_SEGMENT_MAXSIZE.
 */
#define ISSHARED(ADDR)                                              \
    ((bitsADDR)(ADDR) >  DATA_SEGMENT_START &&                      \
     (bitsADDR)(ADDR) < DATA_SEGMENT_MAXSIZE)

/*
 * system-memory allocation policy: kernel and memory simulator heap should
 * be separated from the application's; this is what mmap() is for, called 
 * by the sys_malloc() that kernel and memsim should use instead of malloc().
 * if mmap() doesn't work on your system, uncomment this and the ordinary 
 * malloc() will be used; but there will be no much difference anyway.
 */
#define USE_MMAP_FOR_SYS_MALLOC

/*
 *  +------------------------------------------------------------------+
 *  |                        MEMSIM FUNCTIONS                          |
 *  +------------------------------------------------------------------+
 */ 
/* describe memory_simulator-to-kernel interface */
enum Memsim_Memory_Strategy {
    Kernel_Maintains_Memory_Consistency,    /* -- this is the default */
    Memsim_Maintains_Memory_Consistency     /* -- if you are brave */
};
enum Memsim_Memory_Strategy memsim_init();
Thread_Time initial_thread_time(Thread_Time parent_time, int thread_number);
void memsim_fork(int parent_ID, int child_ID);
void memory_simulate(Thread_Time time_before, Thread_Time time_now);
void memsim_dump_state();
void memsim_cleanup();
void memsim_debug_toggle(int on);

/*
 *  +------------------------------------------------------------------+
 *  |                       DEBUGGING CONTROL                          |
 *  +------------------------------------------------------------------+
 */ 
extern  FILE * sys_output;          /* std file for kernel/memsim messages */
extern  FILE * sys_ttyout;          /* fd for the user's terminal */
#define PRINTF(args...) fprintf(sys_output, args)
#define FFLUSH fflush(sys_output)

extern int sys_debug_events;        /* turned on/off by the user */

/*
 *  +------------------------------------------------------------------+
 *  |  memory operations inside kernel modules that need be simulated  |
 *  +------------------------------------------------------------------+
 */ 
#define SYS_STORE(VAR,DATA)  ({                                         \
    typeof (VAR) _st_retval = (DATA);                                   \
    *(typeof (VAR) *)sys_current_thread->local_buffer = _st_retval;     \
    sys_memory_loadstore(REQUEST_WRITE,                                 \
        &(VAR), sizeof(DATA), ISSHARED(&(VAR)));                        \
    _st_retval;                                                         \
})
#define SYS_LOAD(VAR)     ({                                            \
    typeof (VAR) _ld_retval;                                            \
    sys_memory_loadstore(REQUEST_READ,                                  \
        &(VAR), sizeof(VAR), ISSHARED(&(VAR)));                         \
    _ld_retval = *( typeof (VAR)* )sys_current_thread->local_buffer;    \
    _ld_retval;                                                         \
})
#define SYS_LOCK(LOCK)      sys_memory_lock(REQUEST_LOCK, &(LOCK));
#define SYS_UNLOCK(LOCK)    sys_memory_lock(REQUEST_UNLOCK, &(LOCK));
/* this should be changed */
#define SYS_PREF(VAR)        sys_memory_lock(REQUEST_PREF, &(VAR));
#define SYS_PREFEX(VAR)      sys_memory_lock(REQUEST_PREFEX, &(VAR));
#define SYS_LPREF(VAR)	     sys_memory_lock(REQUEST_LPREF, &(VAR));
#define SYS_WRBACK(VAR)      sys_memory_lock(REQUEST_WRBACK, &(VAR)); 
#define SYS_OWL(VAR)         sys_memory_lock(REQUEST_OWL, &(VAR));
#define SYS_OWH(VAR)	     sys_memory_lock(REQUEST_OWH, &(VAR));
#define SYS_CWL(VAR)         sys_memory_lock(REQUEST_CWL, &(VAR)); 

/* Created */
#define SYS_CREATED(PARENT,CHILD) {                                       \
    sys_threads[CHILD].global_operation.type = GLOBAL_CREATED;            \
    sys_threads[CHILD].global_operation.params.created.parent = (PARENT); \
}

/* Barriers */
#define SYS_BARRIER_ENTER(ADDR,TOTAL) sys_memory_barrier_enter(GLOBAL_BARRIER_ENTER, &(ADDR), TOTAL);
#define SYS_BARRIER_IN(ORDER) sys_memory_barrier_in(GLOBAL_BARRIER_IN, ORDER);
#define SYS_BARRIER_WAIT sys_memory_barrier_wait(GLOBAL_BARRIER_WAIT);
#define SYS_BARRIER_OUT(ORDER) sys_memory_barrier_out(GLOBAL_BARRIER_OUT, ORDER);
#define SYS_BARRIER_EXIT sys_memory_barrier_exit(GLOBAL_BARRIER_EXIT);

#endif _KERNEL_H_
