#ifndef _MEMSIM_INTERFACE_H_
#define _MEMSIM_INTERFACE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header:  memsim_interface.h
 * Authors: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *          Igor Ikodinovic, igi@eunet.yu
 *
 * The memory simulator communicates with the kernel through the set of
 * macros. A memory simulator can be either detached, trace-driven, or 
 * execution-driven (default). 
 *    A detached simulator is linked with the simple scheduler in detached.c, 
 * but still includes this file. Therefore a different set of macros is 
 * defined for a detached mode. One will probably first want to test a memory 
 * simulator in a detached mode and debug it on simple examples before 
 * connecting it to a real application in execution-driven mode, or using a 
 * trace-driven approach. 
 *    In an execution-driven approach memory simulator is linked directly to
 * the application and scheduler in kernel.c file, and requires a use of 
 * different definitions of macros then in detached mode.
 *    In a trace-driven simulation memory simulator is linked with the 
 * scheduler in tracer.c file, which has its own definitions of macros.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "kernel.h"
#include "funcdecl.h"

/*-----------------------------------------------------------------------*/

#ifdef DETACHED

#undef  DATA_SEGMENT_START
#undef  ISSHARED(addr)
#define DATA_SEGMENT_START 0
#define MAX_SIMULATED_MEMORY_SIZE 65536
#define ISSHARED(addr) 1

#define sys_malloc(SIZE) malloc(SIZE)

extern struct Memory_Request sm_memory_request[MAX_CPU_COUNT];

#define CPU_COUNT 4         /* change this if testing more than 4 elements */
#define REQUEST_EXISTS(I)               \
        sm_memory_request[I].requested_now
#define REQUEST_RESPOND(I, RESPONSE)    \
        sm_memory_request[I].satisfied = (RESPONSE)
#define REQUEST_FILL(I, REQUEST_PTR)    \
        REQUEST_PTR = &sm_memory_request[I]

#define COUT cout
#undef PRINTF
#define PRINTF(args...) printf(args)
#undef FFLUSH
#define FFLUSH fflush(stdout)

/* The scheduler can handle only one lock at a time */
extern bitsLOCK one_lock_for_testing;   
#define MEMSIM_READ_LOCK(lock_addr) one_lock_for_testing
#define MEMSIM_WRITE_LOCK(lock_addr, value) one_lock_for_testing = value

/*-----------------------------------------------------------------------*/

#else
#ifdef TRACE_DRIVEN

#define CPU_COUNT sys_thread_count      /* number of "live" processors */

/* the following use I as the CPU index */
#define REQUEST_EXISTS(I)               \
    sys_threads[(I)].memory_request.requested_now
#define REQUEST_RESPOND(I, RESPONSE)    \
    sys_threads[(I)].memory_request.satisfied = (RESPONSE)
#define REQUEST_FILL(I, REQUEST_PTR)    \
    REQUEST_PTR = &sys_threads[(I)].memory_request
#define REQUEST_TYPE(I)                 \
    sys_threads[(I)].memory_request.type
#define REQUEST_VIRTUAL_ADDR(I)         \
    sys_threads[(I)].memory_request.virtual_addr
#define REQUEST_DATA(I)                 \
    sys_threads[(I)].memory_request.data
#define REQUEST_SIZE(I)                 \
    sys_threads[(I)].memory_request.size
#define REQUEST_THREAD_ID(I)            \
    sys_threads[(I)].memory_request.pid

#define COUT cout
#undef PRINTF
#define PRINTF(args...) printf(args)
#undef FFLUSH
#define FFLUSH fflush(stdout)

/* The scheduler must take care about the locks "manually" */
#define MEMSIM_READ_LOCK(lock_addr) read_lock (lock_addr)
#define MEMSIM_WRITE_LOCK(lock_addr, value) write_lock (lock_addr,value)

/*-----------------------------------------------------------------------*/

#else  /* execution driven */

#define CPU_COUNT sys_thread_count      /* number of "live" processors */

/* the following use I as the CPU index */
#define REQUEST_EXISTS(I)               \
    sys_threads[(I)].memory_request.requested_now
#define REQUEST_RESPOND(I, RESPONSE)    \
    sys_threads[(I)].memory_request.satisfied = (RESPONSE)
#define REQUEST_FILL(I, REQUEST_PTR)    \
    REQUEST_PTR = &sys_threads[(I)].memory_request
#define REQUEST_TYPE(I)                 \
    sys_threads[(I)].memory_request.type
#define REQUEST_VIRTUAL_ADDR(I)         \
    sys_threads[(I)].memory_request.virtual_addr
#define REQUEST_DATA(I)                 \
    sys_threads[(I)].memory_request.data
#define REQUEST_SIZE(I)                 \
    sys_threads[(I)].memory_request.size
#define REQUEST_THREAD_ID(I)            \
    sys_threads[(I)].memory_request.pid

/* The simulator must take care about the locks "manually" */
#define MEMSIM_READ_LOCK(lock_addr) *(bitsLOCK *)lock_addr
#define MEMSIM_WRITE_LOCK(lock_addr, value) *(bitsLOCK *)lock_addr = value

#define COUT ostream((streambuf *)sys_output)   /* debugging output */
#define exit(C) sys_simulation_end(C)

/*-----------------------------------------------------------------------*/

#endif  TRACE_DRIVEN

#endif  DETACHED

#endif  _MEMSIM_INTERFACE_H_
