#ifndef _PREFERENCES_H_
#define _PREFERENCES_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Header: preferences.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * You can tune some global settings here, like what is the maximum number
 * of processors/threads, how the locks will behave, what is the default
 * size of the memory copy etc.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/* 
 * Maximum number of processors in the system.
 */
#define MAX_CPU_COUNT 16

/* 
 * Maximum number of threads in the system.
 * Current version of the kernel supports only one thread per processor.
 */
#define MAX_THREAD_COUNT MAX_CPU_COUNT

/*
 * The size of the stack for each thread. 256K by default.
 */
#define MAX_STACK_PER_THREAD (256*1024)

/*-----------------------------------------------------------------------*/
/* 
 * LOCK-ing semantics.
 *
 * When simulating locks in memsim, one can either choose that memsim
 * repeats lock acquire attempts until the lock gets free, thus stalling the
 * processor until then, or to return PROCESSOR_LOCK_BUSY (see kernel.h). 
 * In the second case, the kernel will adjust the time of the thread who
 * requested a lock by BUSY_LOCK_SLEEP_TIME. It is normally one, since
 * this is what would happen for
 * loop: lock <addr>
 *       jnz  loop          # 1 cycle before reattempting
 * You can choose, however, a greater value for less aggresive locking.
 *
 * The second parameter is more interesting for those who examine
 * parallel algorithms only: if you set BUSY_LOCK_OPTIMIZATION to 1,
 * when the kernel sees that all the request at the present time are
 * requests for locks that are busy, then it realizes that the sequence
 * of attempting to obtain busy locks will repeat until an event in the
 * future which occurs on some other processor. Then the kernel raises
 * the simulation time to the one of that future event, which can speed
 * up the whole simulation process significantly.
 * (More formally: of N threads, if K threads have reached time t1,
 * and N-K threads have times greater than t1, where t2 is minimum of
 * those times, and the global simulation time is t1, and all the K
 * threads are requesting a busy lock, then the global simulation time
 * (as well as the time for the K threads) shifts to t2, if 
 * BUSY_LOCK_OPTIMIZATION is set.)
 */ 
#define BUSY_LOCK_SLEEP_TIME    1   /* 1 za jrz, loop, N>1 za fair wait */
#define BUSY_LOCK_OPTIMIZATION  0   /* snoopies usually ne koriste ovo  */

/*-----------------------------------------------------------------------*/
/* 
 * Data types, reflecting the architecture of the simulated system. 
 * If you don't care about architecture, say that bitsMAX = bits64.
 * That way all the FPU memory operations will be handled as a single
 * request.
 */
typedef unsigned char       bits8;
typedef unsigned short int  bits16;
typedef unsigned int        bits32;
typedef unsigned long long  bits64;

typedef bits32              bitsMAX;  /* 32-bit bus. Use bits64 for 64-bit bus. */
typedef bits32              bitsLOCK;
typedef bits32              bitsADDR;

#define DATA_WIDTH sizeof(bitsMAX)

/*-----------------------------------------------------------------------*/
/*
 * Maintaining debug info, such as the mode (USER/KERNEL/MEMSIM) a thread
 * is in, introduces a slight overhead. You can turn it off here if you are
 * sure your code is flawless. :-)
 */
#if 1
#define DEBUG(statements...) {statements}
#else
#define DEBUG(statements...)
#endif

/*-----------------------------------------------------------------------*/
/*
 * Gathering statistics introduces a slight overhead; but, after all,
 * statistics is what this simulation is all about. Turn it off if you want.
 * Collecting statistics at run time must be explicitely enabled with -- -s
 * option anyway.
 * (Some numbers: certain simulation lasts 121 secs with null STATS macro,
 * 123 secs with non-null macro but without -s, and 128 secs with -s switch,
 * which collects all the statistics and in addition dumps the data after 
 * the simulation finishes.)
 */
extern int sys_gather_statistics;
#if 1
#define STATS(statements...) if (sys_gather_statistics) {statements}
#else
#define STATS(statements...)
#endif

#endif /* _PREFERENCES_H_ */