#ifndef _DRA_CACHE_H_
#define _DRA_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: 	dra_cache.h
 * Author: 	Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 * Description:	Ideal Base Cache Controller is extended to support 
 *       	prefetch, prefetch-ex, lazy prefetch and write-back
 *
 * Note: This simulatar simulates an ideal memory system
 *	 in the sense that all read/write/unlock/pref/pref-ex/wb/...
 *       requests are served with zero latency. 
 *       About lock???????
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"


// states a line in the Dra protocol can be found in
class Dra_Line_State {
  public:
    enum State { UNMOD_SHD, UNMOD_EXC, MOD_SHD, MOD_EXC } state;
    Dra_Line_State()            { state = UNMOD_SHD; }
    bool  isvalid()             { return 1; }
    State get()                 { return state; }
    void  set(State new_state)  { state = new_state; }
    char  symbol() {
        switch (state) {
            case UNMOD_SHD: return 's';     // Unmodified, Shared
            case UNMOD_EXC: return 'e';     // Unmodifed, Exclusive
            case MOD_SHD  : return 'S';     // Modified, Shared
            case MOD_EXC  : return 'E';     // Modified, Exclusive
            }
        }
    };


// the Dra cache controler module declaration
class Dra_Cache {
    int index;
    int i; 
    int shared; 	// temporal variable for snooping
    bool exist_inv;	// temporal variable to find if the inv cycle is needed
    bitsADDR tadr;
    bitsLOCK lock_value;
    // instantiate Cache Memory with the Dra_Line_State as its Line_State
    Cache_Memory<Dra_Line_State> cache_memory;  // the whole cache memory
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        } out;
    // Processor Cache Conntoller finite state 
    enum PCC_State {     // processor cache controller
        PCC_INITIAL,                        // INITIAL -- 
        PCC_IDLE
       } pcc_state;

    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK, FIL_UNLOCK, FIL_PREF, FIL_PREFEX, FIL_LPREF, FIL_WRBACK} FIL_action;

    struct Stat {       // statistics:
        int req;        // requests
        int rep;        // replacements
        int read_hit;   // read hits
        int read_miss;  // read misses
        int rm_nowb;
        int rm_wb;	// read miss: write-back  
        int write_hit;  // write hits
        int write_miss; // write misses
        int write_upd;	// write update
        int wm_nowb;
        int wm_wb;
        int lock;	 // locks
        int lock_hit;
        int lock_miss;
        int lock_upd;
        int lm_nowb;
        int lm_wb;
        int unlock;	 // unlocks
        int unlock_hit;
        int unlock_miss;
        int unlock_upd;
        int um_nowb;
        int um_wb;
        int pref_hit;    // prefetch hits
        int pref_miss;   // prefetch misses
        Stat()  { req=rep=read_hit=read_miss=rm_wb=rm_nowb=0;
                  write_hit=write_miss=write_upd=0;
                  lock=lock_hit=lock_miss=lock_upd=lm_nowb=lm_wb=0;
                  unlock=unlock_hit=unlock_miss=unlock_upd=um_nowb=um_wb=0;
                  pref_hit=pref_miss=0;
                }
        void dump(int index);
        } stat;
        
    struct BCU_Stat { // statistics of bus transactions
       int rdc;    	// bus read cycle
       int updc;	// bus update cycle (update word)
       int swbc;    	// bus flush (snoop wb)
       int rwbc;        // write-back (replacement)
       BCU_Stat() { rdc=updc=swbc=rwbc=0;}
       void dump(int index);
       } bcu_stat;
       
   struct Snoop_Stat {
      int sctosc;
      int smtosc;
      int mtosm;
      int smtosm;
      int etosc;
      Snoop_Stat() {sctosc=smtosc=mtosm=smtosm=etosc=0;}
      void dump(int index);
   } snoop;
       
  public:
    Dra_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {
       return 1;
    }
    void dump_state();
    void dump_stats() {
       stat.dump(index); bcu_stat.dump(index); snoop.dump(index);}
    TOPOLOGY_FRIENDS
    };


#ifdef PRAGMA_IMPLEMENTATION_DRA_CACHE

// the Dra cache controller module implementation

#include <iostream.h>
    
Dra_Cache::Dra_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    pcc_state = PCC_IDLE;                            // PCC will be 
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    };

void Dra_Cache::Introduce() {
    COUT << "Protocol: Dra (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << "LOCK_line aware.";
    #endif
    #ifdef DUMM_LOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    }

void Dra_Cache::Stat::dump(int index) {
    if (req) COUT << "Dra"<<index<<": " << \
        "requests="<<req << ", replacements="<<rep
        << "\n       rh=" << read_hit << ", rm=" << read_miss
        << ", rm_nowb=" << rm_nowb << ", rm_wb=" << rm_wb
        << "\n       wh=" << write_hit << ", wm=" << write_miss
        << ", wupd=" << write_upd 
        << ", wm_nowb=" << wm_nowb << ", wm_wb=" << wm_wb 
        << "\n       lock=" << lock << ", lh=" << lock_hit 
        << ", lm=" << lock_miss << ", lu=" << lock_upd
        << ", lm_nowb=" << lm_nowb << ", lm=" << lm_wb
        << "\n        unlock=" << unlock << ", uh=" << unlock_hit
        << ", um=" << unlock_miss << ", uu=" << unlock_upd 
        << ", um_nowb=" << um_nowb << ", um_wb=" << um_wb   
        << "\n       ph=" << pref_hit << ", pm=" << pref_miss 
        << "\n"; 
    }

void Dra_Cache::BCU_Stat::dump(int index) {
    if (rdc || rwbc || swbc || updc) COUT << "Dra"<<index<<": Bus Traffic\n"
         << "  BusReadCy=" << rdc 
         << ", BusUpdateCy=" << updc
         << ", BusWBCy(Repl)=" << rwbc
         << ", BusWBCy(Snoop)=" << swbc 
         << "\n";
  }

void Dra_Cache::Snoop_Stat::dump(int index) {
    if(sctosc || smtosc || mtosm || smtosm || etosc) 
         COUT << "Dra"<<index<<": Snoop Transitions\n"
         << "  sctosc=" << sctosc 
         << ", smtosc=" << smtosc
         << ", mtosm=" << mtosm
         << ", smtosm=" << smtosm 
         << ", etosc=" << etosc << "\n================\n";
  }
   
/*
* cycle method
*/
// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
    
void Dra_Cache::cycle() {
    
    cache_memory.set_current_frame(0);    
    if (!Topology::cache_read_request(index)) {
         pcc_state=PCC_IDLE;
         return; }
    pcc_state=PCC_INITIAL;
    ++stat.req;                                         // statistics
    Topology::pstats.req++;
    switch (in.request.type) {                          // select where to go
        case REQUEST_READ   : Topology::pstats.rd_req++;
        		      FIL_action=FIL_READ; 
        		      goto L_READ;
        case REQUEST_LOCK   : Topology::pstats.lo_req++;
        		      FIL_action=FIL_LOCK; goto L_LOCK;
        case REQUEST_UNLOCK : Topology::pstats.un_req++;
                              FIL_action=FIL_UNLOCK; goto L_UNLOCK;
        case REQUEST_WRITE  : Topology::pstats.wr_req++;
        	              FIL_action=FIL_WRITE; 
        	              goto L_WRITE;
        case REQUEST_PREF   : FIL_action=FIL_PREF; goto L_PREF;
        case REQUEST_PREFEX : FIL_action=FIL_PREFEX; goto L_PREFEX;
        case REQUEST_LPREF  : FIL_action=FIL_LPREF; goto L_LPREF;
        case REQUEST_WRBACK : FIL_action=FIL_WRBACK; goto L_WRBACK;
        default : break;
        }
//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    if (cache_memory.isvalid(in.request.addr)) {
        stat.read_hit++;
        Topology::pstats.rd_hit++;
        goto L_READ_end;
    }
    //it's a read miss
    stat.read_miss++;
    Topology::pstats.rd_miss++;
    goto L_BUS_READ_CYCLE;
L_READ_end:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;  
//---------------------------------------------------------------
L_WRITE:
    if(!cache_memory.isvalid(in.request.addr)) { //write miss
      Topology::pstats.wr_miss++;
      stat.write_miss++;
      goto L_BUS_READ_CYCLE;
    }   
    stat.write_hit++;
    Topology::pstats.wr_hit++;
 L_WRITE_start:   
    if(cache_memory.getstate()==Dra_Line_State::UNMOD_EXC) {
       cache_memory.setstate(Dra_Line_State::MOD_EXC);
    }
    if((cache_memory.getstate()==Dra_Line_State::UNMOD_SHD) ||
       (cache_memory.getstate()==Dra_Line_State::MOD_SHD)) {
       stat.write_upd++;
       Topology::pstats.wr_upd++;
       goto L_BUS_UPDATE_CYCLE;
    } // end of if  
L_WRITE_end:    
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;
//---------------------------------------------------------------
L_LOCK:
    stat.lock++;
    if(!cache_memory.isvalid(in.request.addr)) { //miss
       Topology::pstats.lo_miss++;
       stat.lock_miss++;
       goto L_BUS_READ_CYCLE;
    }
    stat.lock_hit++;
    Topology::pstats.lo_hit++;
L_LOCK_check_value:   
   lock_value = MEMSIM_READ_LOCK(in.request.addr);
   // a dumm one does test&set and let the kernel spin for him
   MEMSIM_WRITE_LOCK(in.request.addr, 1); 
   if(!lock_value) { // lock is free
      if(cache_memory.getstate()==Dra_Line_State::UNMOD_EXC) {
         cache_memory.setstate(Dra_Line_State::MOD_EXC);
      }
      if((cache_memory.getstate()==Dra_Line_State::UNMOD_SHD) ||
         (cache_memory.getstate()==Dra_Line_State::MOD_SHD)) {
        stat.lock_upd++;
        Topology::pstats.lo_upd++;
        goto L_BUS_UPDATE_CYCLE;
      } // end of if  
   }//end of if(!lock_value)           
L_LOCK_end:   
   out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
   out.response.STROBE = 1;
   return;
//---------------------------------------------------------------
L_UNLOCK:
   stat.unlock++;
   if(!cache_memory.isvalid(in.request.addr)) { //unlock miss
     Topology::pstats.un_miss++;
     stat.unlock_miss++;
     goto L_BUS_READ_CYCLE;
   }
   stat.unlock_hit++;
   Topology::pstats.un_hit++;
L_UNLOCK_start:
   if(cache_memory.getstate()==Dra_Line_State::UNMOD_EXC) {
      cache_memory.setstate(Dra_Line_State::MOD_EXC);
   }
   if((cache_memory.getstate()==Dra_Line_State::UNMOD_SHD) ||
     (cache_memory.getstate()==Dra_Line_State::MOD_SHD)) {
      stat.unlock_upd++;
      Topology::pstats.un_upd++;
      goto L_BUS_UPDATE_CYCLE;
   } // end of if  
L_UNLOCK_value:   
   MEMSIM_WRITE_LOCK( in.request.addr, 0);
   out.response.signal = PROCESSOR_SATISFIED; 
   out.response.STROBE = 1;
   return;
    
//---------------------------------------------------------------
L_PREF:
    if(cache_memory.isvalid(in.request.addr)) {
       stat.pref_hit++;
       goto L_PREF_end;
    }
    // read miss
    stat.pref_miss++;
    goto L_BUS_READ_CYCLE;
L_PREF_end:    
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;  

//---------------------------------------------------------------
L_PREFEX: 
    return;
//----------------------------------------------------------------
L_LPREF:
   return;
//----------------------------------------------------------------
L_WRBACK:
//-------------------end of PCC----------------------------------------
   return;
//====================================================================
// Bus Read Cycle
//====================================================================
L_BUS_READ_CYCLE:
   Topology::bstats.rdc++;
   bcu_stat.rdc++; //read bus cycle is initiated
   // replacement
   switch(Topology::cache[index]->cache_memory.getstate()) {
        case Dra_Line_State::UNMOD_EXC:
        case Dra_Line_State::UNMOD_SHD:
           {//statistics
            if(FIL_action==FIL_READ)
              stat.rm_nowb++; 
            else if(FIL_action==FIL_WRITE)
              stat.wm_nowb++;
            else if(FIL_action==FIL_LOCK)
              stat.lm_nowb++;
            else if(FIL_action==FIL_UNLOCK)
              stat.um_nowb++;
            break;}
        case Dra_Line_State::MOD_EXC:
        case Dra_Line_State::MOD_SHD:
           {
            if(FIL_action==FIL_READ)
              stat.rm_wb++; 
            else if(FIL_action==FIL_WRITE)
              stat.wm_wb++;
            else if(FIL_action==FIL_LOCK)
              stat.lm_wb++;
            else if(FIL_action==FIL_UNLOCK)
              stat.um_wb++;
            bcu_stat.rwbc++;
            Topology::bstats.rwbc++;
            stat.rep++; 
            break;}
   } // end of check block for replacement
   // snooping
   shared=0;
   for(i=0; i<CPU_COUNT; i++) {
      if(i!=index) {
        if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
          switch(Topology::cache[i]->cache_memory.getstate()) {
               case Dra_Line_State::MOD_EXC:
                  {shared=1;
                   //processor i has requested block modified; WriteBack + change state 
                   bcu_stat.swbc++; //flush
                   Topology::bstats.swbc++;
                   Topology::cache[i]->snoop.mtosm++;
                   Topology::cache[i]->cache_memory.setstate(Dra_Line_State::MOD_SHD);
                   break;}
               case Dra_Line_State::MOD_SHD:
                  {shared=1;
                   bcu_stat.swbc++; //flush ????????????????
                   Topology::bstats.swbc++;
                   Topology::cache[i]->snoop.smtosm++;
                   break;} 
               case Dra_Line_State::UNMOD_EXC:
                   shared=1;
                   Topology::cache[i]->snoop.etosc++;
                   Topology::cache[i]->cache_memory.setstate(Dra_Line_State::UNMOD_SHD);
                   break;
               case Dra_Line_State::UNMOD_SHD:
                   default: break; //memory suply
          } // end of switch                   
        } // end of if
      } // end of if
   } //end-for snooping
   if(shared) 
     cache_memory.setstate(Dra_Line_State::UNMOD_SHD); //set new state   
   else
     cache_memory.setstate(Dra_Line_State::UNMOD_EXC); //set new state   
   
   // simulate write of the cache line
   while(cache_memory.line_fetch_putnext(&tadr));
   if(FIL_action==FIL_LOCK) goto L_LOCK_check_value;
   if(FIL_action==FIL_READ) goto L_READ_end;
   if(FIL_action==FIL_WRITE) goto L_WRITE_start; 
   if(FIL_action==FIL_UNLOCK) goto L_UNLOCK_start;
   if(FIL_action==FIL_PREF) goto L_PREF_end; 
//=======================================================================
//   Bus Word Update Cycle
//=======================================================================         
L_BUS_UPDATE_CYCLE:
   bcu_stat.updc++;
   Topology::bstats.updc++;
   shared=0;
   for(i=0; i<CPU_COUNT; i++) {
      if(i!=index) {
        if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
          switch(Topology::cache[i]->cache_memory.getstate()) {
             case Dra_Line_State::UNMOD_SHD: 
                 {Topology::cache[i]->snoop.sctosc++;//update
                  shared=1;
                  break; }
             case Dra_Line_State::MOD_SHD:
                 {Topology::cache[i]->snoop.smtosc++;
                  shared=1;
                  Topology::cache[i]->cache_memory.setstate(Dra_Line_State::UNMOD_SHD);
                  break;}
             default: break;                        
          } 
        } // end of if
      } // end of if
   } //end of for
   //change state (it depends on previous state and shared line)
   if(cache_memory.getstate()==Dra_Line_State::UNMOD_SHD) {
     if (shared) cache_memory.setstate(Dra_Line_State::MOD_SHD);
     else cache_memory.setstate(Dra_Line_State::MOD_EXC);}
   if(cache_memory.getstate()==Dra_Line_State::MOD_SHD)
     if (!shared) cache_memory.setstate(Dra_Line_State::MOD_EXC);
   //        
   if(FIL_action==FIL_WRITE) goto L_WRITE_end; 
   if(FIL_action==FIL_LOCK) goto L_LOCK_end;
   if(FIL_action==FIL_UNLOCK) goto L_UNLOCK_value;
//======================================================================
//  Flush Cycle
//======================================================================


}

// debugging...

void Dra_Cache::dump_state() {
    COUT << "CC" << index << ": PCC_State: ";
        switch (pcc_state) {
            case PCC_INITIAL         : COUT << "PCC_INITIAL"; pcc_state=PCC_IDLE;break;
            case PCC_IDLE	     : COUT << "PCC_IDLE"; break;	
        }
     if(pcc_state==PCC_INITIAL) {
        COUT << "\t\t[";
        switch (in.request.type) {      
            case REQUEST_READ  : COUT << "READ " << hex <<in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << hex <<in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << hex <<in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << hex <<in.request.addr; break;
            case REQUEST_PREF  : COUT << "PREF " << hex <<in.request.addr; break;
            case REQUEST_PREFEX: COUT << "PREFEX"<< hex <<in.request.addr; break;
            case REQUEST_LPREF : COUT << "LPREF" << hex <<in.request.addr; break;
            case REQUEST_WRBACK: COUT << "WRBACK"<< hex <<in.request.addr; break;
            default : break;
        }
       COUT << "]"; 
     }   
     COUT << "\n";    
     cache_memory.dump_state(index);
    }
                

#endif PRAGMA_IMPLEMENTATION_DRA_CACHE
#endif _DRA_CACHE_H_