#ifndef _MESI_CACHE_H_
#define _MESI_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: 	mesi_cache.h
 * Author: 	Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 * Description:	Ideal Base Cache Controller is extended to support 
 *       	prefetch, prefetch-ex, lazy prefetch and write-back
 *
 * Note: This simulatar simulates an ideal memory system
 *	 in the sense that all read/write/unlock/pref/pref-ex/wb/...
 *       requests are served with zero latency. 
 *       About lock???????
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "iat.h"

// states a line in the MESI protocol can be found in
class MESI_Line_State {
  public:
    enum State { INVALID, SHARED, EXCLUSIVE, MODIFIED } state;
    MESI_Line_State()               {state = INVALID;}
    bool  isvalid()		    {return state != INVALID;}
    State get()                     {return state;}
    void  set(State new_state)      {state = new_state; }
    char  symbol() { // for debuging
    	             switch (state) {
    	               case INVALID:   return 'I';
    		       case SHARED:    return 'S';
    		       case EXCLUSIVE: return 'E';
    		       case MODIFIED:  return 'M' ;
    	             }
    		   }
    };

// the MESI cache controler module declaration
class MESI_Cache {
    int index;
    int i; 
    int shared; 	// temporal variable for snooping
    bool exist_inv;	// temporal variable to find if the inv cycle is needed
    bitsADDR tadr;
    bitsLOCK lock_value;
    // instantiate Cache Memory with the MESI_Line_State as its Line_State
    Cache_Memory<MESI_Line_State> cache_memory;  // the whole cache memory
    // instantiate IAT (Injection Anticipation Table)
    IAT iat;
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        } out;
    // Processor Cache Conntoller finite state 
    enum PCC_State {     // processor cache controller
        PCC_INITIAL,                        // INITIAL -- 
        PCC_IDLE
       } pcc_state;

    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK, FIL_UNLOCK, FIL_PREF, FIL_PREFEX, FIL_LPREF, FIL_WRBACK, FIL_OWL, FIL_OWH, FIL_CWL} FIL_action;

    struct Stat {       // statistics:
        int req;        // requests
        int rep;        // replacements
        int read_hit;   // read hits
        int read_miss;  // read misses
        int rm_nowb;    // read miss: no write-back
        int rm_wb;	// read miss: write-back  
        int write_hit;  // write hits
        int write_miss; // write misses
        int wm_nowb;	// write miss: no write-back
        int wm_wb;	// write miss: write-back
        int write_inv;   // write invalidate
        int write_inv_true; // 
        int lock;	 // locks
        int lock_hit;
        int lock_miss;
        int lockm_nowb;
        int lockm_wb;
        int lock_inv;
        int lock_inv_true;
        int unlock;	 // unlocks
        int unlock_hit;
        int unlock_miss;
        int unlockm_nowb;
        int unlockm_wb;
        int unlock_inv;
        int unlock_inv_true;
        int pref_hit;    // prefetch hits
        int pref_miss;   // prefetch misses
        int pm_nowb;	 // prefetch miss: no wb
        int pm_wb;	 // prefetch miss: wb
        int prefex_hit;  // prefetch-ex hits
        int prefex_miss;   	// prefetch-ex misses
        int pxm_nowb;		// prefetch-ex miss: no wb
        int pxm_wb;		// prefetch-ex miss: wb
        int prefex_inv;		// invalidate caused by prefex
        int prefex_inv_true;	// 
        //write back statistics
        int iwb_hit;            // write-back hit
        int iwb_miss;		// write-back miss (real miss + hit in umnod state)
        int inj;		// number of injections (hit in IAT)
        int inj_norepl;		// injection no repl.
        int inj_repl;           // injection with repl.

        Stat()  { req=rep=read_hit=read_miss=rm_nowb=rm_wb=0;
                  write_hit=write_miss=write_inv=write_inv_true=wm_nowb=wm_wb=0;
                  lock=lock_hit=lock_miss=lockm_nowb=lockm_wb=lock_inv=lock_inv_true=0;
                  unlock=unlock_hit=unlock_miss=unlockm_nowb=unlockm_wb=unlock_inv=unlock_inv_true=0;
                  pref_hit=pref_miss=pm_nowb=pm_wb=0;
                  prefex_hit=prefex_miss=prefex_inv=prefex_inv_true=pxm_nowb=pxm_wb=0;
                  iwb_hit=iwb_miss=inj=inj_norepl=inj_repl=0;
                }
        void dump(int index);
        } stat;
        
    struct BCU_Stat { // statistics of bus transactions
       int rdc;    // bus read cycle
       int rdxc;   // bus read exclusive cycle
       int repl_wbc;    // bus write-back cycle caused by replacement
       int snoop_wbc;   // bus write-back cycle caused by snoop-hit
       int invc;   // bus invalidate cycle
       int iwbc;   // instruction initiated wb cycle
       int irpc;    // injection initiated write back cycle (replacement)
       BCU_Stat() { rdc=rdxc=repl_wbc=snoop_wbc=invc=iwbc=irpc=0;}
       void dump(int index);
       } bcu_stat;
       
    struct SNOOP_Stat { //statistics of snoop actions
       int etos;
       int mtos;
       int stoi;
       int etoi;
       int mtoi;
       SNOOP_Stat() {etos=mtos=stoi=etoi=mtoi=0;}
       void dump(int index);
       } snoop;
   
  public:
    MESI_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {
       return 1;
    }
    void dump_state();
    void dump_stats() {
       stat.dump(index); bcu_stat.dump(index); snoop.dump(index);}
    TOPOLOGY_FRIENDS

    };


#ifdef PRAGMA_IMPLEMENTATION_MESI_CACHE

// the MESI cache controller module implementation

#include <iostream.h>
    
MESI_Cache::MESI_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    iat.Initialize();
    pcc_state = PCC_IDLE;                            // PCC will be 
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    };

void MESI_Cache::Introduce() {
    COUT << "Protocol: MESI (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << "LOCK_line aware.";
    #endif
    #ifdef DUMM_LOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    iat.Introduce();
    }

void MESI_Cache::Stat::dump(int index) {
    if (req) COUT << "MESI"<<index<<": " << \
        "requests="<<req << ", replacements="<<rep
        << "\n       rh=" << read_hit << ", rm=" << read_miss
        << ", rm_nowb=" << rm_nowb << ", rm_wb=" << rm_wb
        << "\n       wh=" << write_hit << ", wm=" << write_miss
        << ", wi=" << write_inv << ", witrue=" << write_inv_true 
        << ", wm_nowb=" << wm_nowb << ", wm_wb=" << wm_wb      
        << "\n       lock=" << lock << ", lh=" << lock_hit << ", lm=" << lock_miss
        << ", lm_nowb=" << lockm_nowb << ", lm_wb=" << lockm_wb
        << ", li=" << lock_inv << ", litrue=" << lock_inv_true
        << "\n       unlock=" << unlock << ", uh=" << unlock_hit << ", um=" << unlock_miss
        << ", um_nowb=" << unlockm_nowb << ", um_wb=" << unlockm_wb 
        << ", ui=" << unlock_inv << ", uitrue="<< unlock_inv_true    
        << "\n       ph=" << pref_hit << ", pm=" << pref_miss 
        << ", pm_nowb=" << pm_nowb << ", pm_wb=" << pm_wb
        << "\n       pxh=" << prefex_hit << ", pxm=" << prefex_miss
        <<  ", pxm_nowb=" << pxm_nowb << ", pxm_wb=" << pxm_wb 
        << ", pxi=" << prefex_inv << ", pxitrue=" << prefex_inv_true 
        << "\n       iwbh=" << iwb_hit << ", iwbm=" << iwb_miss 
        << "\n       inj=" << inj << ", inj_norepl=" << inj_norepl 
        << ", inj_repl=" << inj_repl << "\n"; 
    }

void MESI_Cache::BCU_Stat::dump(int index) {
    if (rdc || rdxc || invc || repl_wbc || snoop_wbc || iwbc || irpc) COUT << "MESI"<<index<<": Bus Traffic\n"
         << "  BusReadReqs=" << rdc 
         << ", BusRead-exReqs=" << rdxc
         << ", BusInvReqs=" << invc 
         << "\n  BusReplWBReqs=" << repl_wbc
         << ", BusSnoopWBReqs=" << snoop_wbc 
         << ", BusInsWBReqs=" << iwbc 
         << ", BusInjRpReqs=" << irpc << "\n";
  }

void MESI_Cache::SNOOP_Stat::dump(int index) {
   if (etos || mtos || stoi || mtoi || etoi) COUT << "MESI"<<index<<": SnoopCC Transitions\n"
        << "  etos=" << etos << ", mtos=" << mtos
        << ", stoi=" << stoi << ", mtoi=" << mtoi 
        << ", etoi=" << etoi << "\n";
  }      
              
/*
* cycle method
*/
// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
    
void MESI_Cache::cycle() {
    
    cache_memory.set_current_frame(0);    
    if (!Topology::cache_read_request(index)) {
         pcc_state=PCC_IDLE;
         return; }
    pcc_state=PCC_INITIAL;
    ++stat.req;                                         // statistics
    Topology::pstats.req++;  
    switch (in.request.type) {                          // select where to go
        case REQUEST_READ   : FIL_action=FIL_READ; goto L_READ;
        case REQUEST_LOCK   : FIL_action=FIL_LOCK; goto L_LOCK;
        case REQUEST_UNLOCK : FIL_action=FIL_UNLOCK; goto L_UNLOCK;
        case REQUEST_WRITE  : FIL_action=FIL_WRITE; goto L_WRITE;
        case REQUEST_PREF   : FIL_action=FIL_PREF; goto L_PREF;
        case REQUEST_PREFEX : FIL_action=FIL_PREFEX; goto L_PREFEX;
        case REQUEST_LPREF  : FIL_action=FIL_LPREF; goto L_LPREF;
        case REQUEST_WRBACK : FIL_action=FIL_WRBACK; goto L_WRBACK;
        case REQUEST_OWL    : FIL_action=FIL_OWL; goto L_OWL;
        case REQUEST_OWH    : FIL_action=FIL_OWH; goto L_OWH;
        case REQUEST_CWL    : FIL_action=FIL_CWL; goto L_CWL;
        default : break;
        }
//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    Topology::pstats.rd_req++;
    if (cache_memory.isvalid(in.request.addr)) {
        stat.read_hit++;
        Topology::pstats.rd_hit++;
        goto L_READ_end;
        }
    // it's a miss
    stat.read_miss++;
    Topology::pstats.rd_miss++;
    goto L_BUS_READ_CYCLE;
L_READ_end:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;  
//---------------------------------------------------------------
L_WRITE:
    Topology::pstats.wr_req++;
    if(cache_memory.isvalid(in.request.addr)){ // write hit
       stat.write_hit++;
       Topology::pstats.wr_hit++;
       if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
       }
       if(cache_memory.getstate()==MESI_Line_State::SHARED) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         stat.write_inv++;
         Topology::pstats.wr_inv++;
         goto L_BUS_INVALIDATE_CYCLE;
       } // end of if  
       goto L_WRITE_end;    
    }// end of if
    // write miss
    Topology::pstats.wr_miss++;
    stat.write_miss++;
    goto L_BUS_READEX_CYCLE;
L_WRITE_end:
    if(FIL_action==FIL_UNLOCK) goto L_UNLOCK_value;
    if(FIL_action==FIL_LOCK) goto L_LOCK_end;
    
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;
//---------------------------------------------------------------
L_LOCK:
    Topology::pstats.lo_req++;
    stat.lock++;
    if(!cache_memory.isvalid(in.request.addr)) { //miss
       stat.lock_miss++;
       Topology::pstats.lo_miss++;
       goto L_BUS_READ_CYCLE;
    }
    stat.lock_hit++;
    Topology::pstats.lo_hit++;
L_LOCK_check_value:   
   lock_value = MEMSIM_READ_LOCK(in.request.addr);
   // a dumm one does test&set and let the kernel spin for him
   MEMSIM_WRITE_LOCK(in.request.addr, 1); 
   if (!lock_value) { // lock is free
       if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
       }
       if(cache_memory.getstate()==MESI_Line_State::SHARED) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         stat.lock_inv++;
         Topology::pstats.lo_inv++;
         goto L_BUS_INVALIDATE_CYCLE;
       }   
   }//end of if(!lock_value)           
L_LOCK_end:
   out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
   out.response.STROBE = 1;
   return;
//---------------------------------------------------------------
L_UNLOCK:
   Topology::pstats.un_req++;
   stat.unlock++;
   if(!cache_memory.isvalid(in.request.addr)) {//it's a miss
      stat.unlock_miss++;
      Topology::pstats.un_miss++;
      goto L_BUS_READEX_CYCLE;
   }
   //else it's a hit
   stat.unlock_hit++;
   Topology::pstats.un_hit++;
   if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
     cache_memory.setstate(MESI_Line_State::MODIFIED);
   }
   if(cache_memory.getstate()==MESI_Line_State::SHARED) {
      cache_memory.setstate(MESI_Line_State::MODIFIED);
      stat.unlock_inv++;
      Topology::pstats.un_inv++;
      goto L_BUS_INVALIDATE_CYCLE;
   } // end if(SHARED)  
L_UNLOCK_value:   
   MEMSIM_WRITE_LOCK( in.request.addr, 0);
   out.response.signal = PROCESSOR_SATISFIED; 
   out.response.STROBE = 1;
   return;
    
//---------------------------------------------------------------
L_PREF:
    Topology::pstats.pr_req++;
    if (cache_memory.isvalid(in.request.addr)) {
        stat.pref_hit++;
        Topology::pstats.pr_hit++;
        goto L_PREF_end;
        }
    // it's a miss
    stat.pref_miss++;
    Topology::pstats.pr_miss++;
    goto L_BUS_READ_CYCLE;
L_PREF_end:    
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;  
//---------------------------------------------------------------
L_PREFEX:
    Topology::pstats.px_req++;
    if (cache_memory.isvalid(in.request.addr)) {
       stat.prefex_hit++;
       Topology::pstats.px_hit++;
       if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
       }
       if(cache_memory.getstate()==MESI_Line_State::SHARED) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         bcu_stat.invc++;
         Topology::pstats.px_inv++;
         stat.prefex_inv++;
         goto L_BUS_INVALIDATE_CYCLE;         
       } // end of if  
       goto L_PREFEX_end;
    }// end of if
    // it's a miss
    stat.prefex_miss++;
    Topology::pstats.px_miss++;
    goto L_BUS_READEX_CYCLE;
L_PREFEX_end:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;  

//----------------------------------------------------------------
L_LPREF:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;
//----------------------------------------------------------------
L_WRBACK:
    // software initiated write-back
    if(cache_memory.isvalid(in.request.addr)) {
      if(cache_memory.getstate()==MESI_Line_State::MODIFIED) { //wrback hit
        stat.iwb_hit++;
        cache_memory.setstate(MESI_Line_State::SHARED);
        goto L_BUS_WRBACK_CYCLE;
      };
    };
    stat.iwb_miss++;
L_WRBACK_end:    
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;    
    return;
//----------------------------------------------------------------
L_OWL:
    //this instruction open an address window
    //??????????????put the cache line address into the corespoding entry
    iat.openWinLAddr(in.request.addr);
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;    
    return;
//----------------------------------------------------------------
L_OWH:
    iat.openWinHAddr(in.request.addr);
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;
//----------------------------------------------------------------
L_CWL:
    iat.closeWin(in.request.addr);
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    return;    
//-------------------end of PCC----------------------------------------
     
//   BCU Unit
//===============================================================
// Read Cycle
//===============================================================
L_BUS_READ_CYCLE:    
    Topology::bstats.rdc++;
    bcu_stat.rdc++; //read bus cycle is initiated
    // replacement
    switch(Topology::cache[index]->cache_memory.getstate()) {
       case MESI_Line_State::INVALID:
       case MESI_Line_State::SHARED:
       case MESI_Line_State::EXCLUSIVE:
           {
            if(FIL_action==FIL_READ)
               stat.rm_nowb++; 
            else if(FIL_action==FIL_LOCK)
               stat.lockm_nowb++;
            else if(FIL_action==FIL_PREF)
               stat.pm_nowb++;
            break;} 
       case MESI_Line_State::MODIFIED:
           {
            if(FIL_action==FIL_READ)
               stat.rm_wb++; 
            else if(FIL_action==FIL_LOCK)
               stat.lockm_wb++;
            else if(FIL_action==FIL_PREF)
               stat.pm_wb++;
            bcu_stat.repl_wbc++;
            Topology::bstats.rwbc++;
            stat.rep++; 
            break;}
    } // end of chack block for replacement
    // snooping
    shared=0;
    for(i=0; i<CPU_COUNT; i++) {
       if(i!=index) {
         if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
           shared=1;
           switch(Topology::cache[i]->cache_memory.getstate()) {
              case MESI_Line_State::EXCLUSIVE:
                  {Topology::cache[i]->snoop.etos++;
                   Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
                   break;}
              case MESI_Line_State::MODIFIED:
                  {Topology::cache[i]->snoop.mtos++;
                   Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
                   Topology::cache[i]->bcu_stat.snoop_wbc++;
                   Topology::bstats.swbc++;
                   break;} 
                   default: break;
           } // end of switch                   
         } // end of if
       } // end of if
    } //end-for snooping
    if(shared) 
      cache_memory.setstate(MESI_Line_State::SHARED); //set new state   
    else
      cache_memory.setstate(MESI_Line_State::EXCLUSIVE); //set new state   
    // simulate write of the cache line
    while(cache_memory.line_fetch_putnext(&tadr));

    if(FIL_action==FIL_LOCK) goto L_LOCK_check_value;
    if(FIL_action==FIL_PREF) goto L_PREF_end;
    if(FIL_action==FIL_READ) goto L_READ_end;
//=======================================================================
// Invalidate Cycle   
//=======================================================================
L_BUS_INVALIDATE_CYCLE:
    Topology::bstats.invc++;
    bcu_stat.invc++;
    exist_inv=0;
    for(i=0; i<CPU_COUNT; i++) {
       if(i!=index) {
         if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
           if(Topology::cache[i]->cache_memory.getstate()==MESI_Line_State::SHARED)
             {Topology::cache[i]->snoop.stoi++;
              Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
              exist_inv=1;                    
             } 
         } // end of if
       } // end of if
    } //end of for
    //unlock moze takodje
    if(FIL_action==FIL_WRITE) { 
      if(exist_inv) stat.write_inv_true++;
      goto L_WRITE_end;
    }
    if(FIL_action==FIL_LOCK) {
       if(exist_inv) stat.lock_inv_true++;
       goto L_LOCK_end;
    }
    if(FIL_action==FIL_UNLOCK) {
       if(exist_inv) stat.unlock_inv_true++;
       goto L_WRITE_end;
    }
//========================================================================
//  Read_Ex Cycle
//========================================================================    
L_BUS_READEX_CYCLE:     
    bcu_stat.rdxc++;
    Topology::bstats.rdxc++;
    // replacement
    switch(cache_memory.getstate()) {
         case MESI_Line_State::INVALID:
         case MESI_Line_State::SHARED:
         case MESI_Line_State::EXCLUSIVE:
            {//statistics
             if(FIL_action==FIL_WRITE)
               stat.wm_nowb++; 
             else if(FIL_action==FIL_PREFEX)
               stat.pxm_nowb++;
             else if(FIL_action==FIL_UNLOCK)
               stat.unlockm_nowb++;
             break;} 
         case MESI_Line_State::MODIFIED:
            {//statistics depending on operation
             if(FIL_action==FIL_WRITE)
               stat.wm_wb++;
             else if(FIL_action==FIL_PREFEX)
               stat.pxm_wb++;
             else if(FIL_action==FIL_UNLOCK)
               stat.unlockm_wb++;
             //  
             bcu_stat.repl_wbc++;
             Topology::bstats.rwbc++;
             stat.rep++; 
             break;}
    } // end of switch
    // simulate snooping
    for(i=0; i<CPU_COUNT; i++) {
       if (i!=index) {
         if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
           switch(Topology::cache[i]->cache_memory.getstate()) {
                case MESI_Line_State::EXCLUSIVE:
                    {Topology::cache[i]->snoop.etoi++;
                     Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                     break;}
                case MESI_Line_State::MODIFIED:
                    {Topology::cache[i]->snoop.mtoi++;
                     Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                     Topology::cache[i]->bcu_stat.snoop_wbc++;
                     Topology::bstats.swbc++;
                     break;}
                case MESI_Line_State::SHARED:
                    {Topology::cache[i]->snoop.stoi++;
                     Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                     break;}      
                default: break;
           } // end of switch                   
         } // end of if
       } // end of if
    } //end of for
    // set new state
    cache_memory.setstate(MESI_Line_State::MODIFIED);
    // simulate line fetch
    while(cache_memory.line_fetch_putnext(&tadr));
    
    if(FIL_action==FIL_UNLOCK) goto L_UNLOCK_value;
    if(FIL_action==FIL_WRITE) goto L_WRITE_end;
    if(FIL_action==FIL_PREFEX) goto L_PREFEX_end;
//========================================================================
//  Read_Ex Cycle
//========================================================================    
L_BUS_WRBACK_CYCLE:     
    bcu_stat.iwbc++;
    Topology::bstats.iwbc++;
    // write-back cycle
    for(i=0; i<CPU_COUNT; i++) {
       if (i!=index) {
         if(Topology::cache[i]->iat.isvalid(in.request.addr)) {
           Topology::cache[i]->stat.inj++; //injection hit
           // make a room for incoming data
           // replacement
           // dummy if to set pointers
           if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)); 
           switch(Topology::cache[i]->cache_memory.getstate()) {
              case MESI_Line_State::INVALID:
              case MESI_Line_State::SHARED:
              case MESI_Line_State::EXCLUSIVE:
                 {Topology::cache[i]->stat.inj_norepl++; //injection no repl.
                  break;} 
              case MESI_Line_State::MODIFIED:
                 {Topology::cache[i]->stat.inj_repl++;
                  Topology::cache[i]->bcu_stat.irpc++;
                  Topology::bstats.irpc++;
                  Topology::cache[i]->stat.rep++; 
                  break;}
           } // end of switch
           // put the data into the cache
           // set new state
           Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
           // simulate line injection
           while(Topology::cache[i]->cache_memory.line_fetch_putnext(&tadr));
                   
         }
       }   
    } //end of for
    goto L_WRBACK_end;
//--------------------------------------------------------------------------

 }

// debugging...

void MESI_Cache::dump_state() {
    COUT << "CC" << index << ": PCC_State: ";
        switch (pcc_state) {
            case PCC_INITIAL         : COUT << "PCC_INITIAL"; pcc_state=PCC_IDLE;break;
            case PCC_IDLE	     : COUT << "PCC_IDLE"; break;	
        }
     if(pcc_state==PCC_INITIAL) {
        COUT << "\t\t[";
        switch (in.request.type) {      
            case REQUEST_READ  : COUT << "READ " << hex <<in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << hex <<in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << hex <<in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << hex <<in.request.addr; break;
            case REQUEST_PREF  : COUT << "PREF " << hex <<in.request.addr; break;
            case REQUEST_PREFEX: COUT << "PREFEX"<< hex <<in.request.addr; break;
            case REQUEST_LPREF : COUT << "LPREF" << hex <<in.request.addr; break;
            case REQUEST_WRBACK: COUT << "WRBACK"<< hex <<in.request.addr; break;
            case REQUEST_OWL   : COUT << "OWL"   << hex <<in.request.addr; break;
            case REQUEST_OWH   : COUT << "OWH"   << hex <<in.request.addr; break;
            case REQUEST_CWL   : COUT << "CWL"   << hex <<in.request.addr; break;            
            default : break;
        }
       COUT << "]"; 
     }   
     COUT << "\n";    
     cache_memory.dump_state(index);
    }
                

#endif PRAGMA_IMPLEMENTATION_MESI_CACHE
#endif _MESI_CACHE_H_