#ifndef _MESI_CACHE_H_
#define _MESI_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: 	mesi_cache.h
 * Author: 	Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 * Description:	Ideal Base Cache Controller is extended to support 
 *       	prefetch, prefetch-ex, lazy prefetch and write-back
 *
 * Note: This simulatar simulates an ideal memory system
 *	 in the sense that all read/write/unlock/pref/pref-ex/wb/...
 *       requests are served with zero latency. 
 *       About lock???????
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"


// states a line in the MESI protocol can be found in
class MESI_Line_State {
  public:
    enum State { INVALID, SHARED, EXCLUSIVE, MODIFIED } state;
    MESI_Line_State()               {state = INVALID;}
    bool  isvalid()		    {return state != INVALID;}
    State get()                     {return state;}
    void  set(State new_state)      {state = new_state; }
    char  symbol() { // for debuging
    	             switch (state) {
    	               case INVALID:   return 'I';
    		       case SHARED:    return 'S';
    		       case EXCLUSIVE: return 'E';
    		       case MODIFIED:  return 'M' ;
    	             }
    		   }
    };

// the MESI cache controler module declaration
class MESI_Cache {
    int index;
    int i; 
    int shared; 	// temporal variable for snooping
    bool exist_inv;	// temporal variable to find if the inv cycle is needed
    bitsADDR tadr;
    bitsLOCK lock_value;
    // instantiate Cache Memory with the MESI_Line_State as its Line_State
    Cache_Memory<MESI_Line_State> cache_memory;  // the whole cache memory
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        } out;
    // Processor Cache Conntoller finite state 
    enum PCC_State {     // processor cache controller
        PCC_INITIAL,                        // INITIAL -- 
        PCC_IDLE
       } pcc_state;

    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK, FIL_UNLOCK, FIL_PREF, FIL_PREFEX, FIL_LPREF, FIL_WRBACK} FIL_action;
    /*
    enum WB_ACTIONS {NOWB, WB_REPL, WB_INSTR} WB_action;
    bool WB_flag; 	// write-back flag
    enum INV_ACTIONS {INV_WR, INV_PREFEX} INV_action;
    */
    struct Stat {       // statistics:
        int req;        // requests
        int rep;        // replacements
        int read_hit;   // read hits
        int read_miss;  // read misses
        int rm_nowb_i;    // read miss: no write-back (invalid)
        int rm_nowb_s;   // ...(shared)
        int rm_nowb_e;   // ...(exclusive)
        int rm_wb;	// read miss: write-back  
        int write_hit;  // write hits
        int write_miss; // write misses
        int wm_nowb;	// write miss: no write-back
        int wm_wb;	// write miss: write-back
        int write_inv;   // write invalidate
        int write_inv_true; // 
        int lock;	 // locks
        int lock_inv;
        int lock_inv_true;
        int unlock;	 // unlocks
        int unlock_inv;
        int unlock_inv_true;
        int pref_hit;    // prefetch hits
        int pref_miss;   // prefetch misses
        int pm_nowb;	 // prefetch miss: no wb
        int pm_wb;	 // prefetch miss: wb
        int prefex_hit;  // prefetch-ex hits
        int prefex_miss;   	// prefetch-ex misses
        int pxm_nowb;		// prefetch-ex miss: no wb
        int pxm_wb;		// prefetch-ex miss: wb
        int prefex_inv;		// invalidate caused by prefex
        int prefex_inv_true;	// 
        Stat()  { req=rep=read_hit=read_miss=rm_nowb_i=rm_nowb_s=rm_nowb_e=rm_wb=0;
                  write_hit=write_miss=write_inv=write_inv_true=wm_nowb=wm_wb=0;
                  lock=lock_inv=lock_inv_true=unlock=unlock_inv=unlock_inv_true=0;
                  pref_hit=pref_miss=pm_nowb=pm_wb=0;
                  prefex_hit=prefex_miss=prefex_inv=prefex_inv_true=pxm_nowb=pxm_wb=0;
                }
        void dump(int index);
        } stat;
        
    struct BCU_Stat { // statistics of bus transactions
       int rdc;    // bus read cycle
       int rdxc;   // bus read exclusive cycle
       int repl_wbc;    // bus write-back cycle caused by replacement
       int snoop_wbc;   // bus write-back cycle caused by snoop-hit
       int invc;   // bus invalidate cycle
       BCU_Stat() { rdc=rdxc=repl_wbc=snoop_wbc=invc=0;}
       void dump(int index);
       } bcu_stat;
       
    struct SNOOP_Stat { //statistics of snoop actions
       int etos;
       int mtos;
       int stoi;
       int etoi;
       int mtoi;
       SNOOP_Stat() {etos=mtos=stoi=etoi=mtoi=0;}
       void dump(int index);
       } snoop;
   
  public:
    MESI_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {
       return 1;
    }
    void dump_state();
    void dump_stats() {
       stat.dump(index); bcu_stat.dump(index); snoop.dump(index);}
    TOPOLOGY_FRIENDS
    };


#ifdef PRAGMA_IMPLEMENTATION_MESI_CACHE

// the MESI cache controller module implementation

#include <iostream.h>
    
MESI_Cache::MESI_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    pcc_state = PCC_IDLE;                            // PCC will be 
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    };

void MESI_Cache::Introduce() {
    COUT << "Protocol: MESI (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << "LOCK_line aware.";
    #endif
    #ifdef DUMM_LOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    }

void MESI_Cache::Stat::dump(int index) {
    if (req) COUT << "MESI"<<index<<": " << \
        "requests="<<req << ", replacements="<<rep
        << "\n       rh=" << read_hit << ", rm=" << read_miss
        << ", rm_nowb_i=" << rm_nowb_i << ", rm_nowb_s=" << rm_nowb_s 
        << ", rm_nowb_e=" << rm_nowb_e <<", rm_wb=" << rm_wb
        << "\n       wh=" << write_hit << ", wm=" << write_miss
        << ", wi=" << write_inv << ", witrue=" << write_inv_true 
        << ", wm_nowb=" << wm_nowb << ", wm_wb=" << wm_wb      
        << "\n       lock=" << lock << ", li=" << lock_inv << ", litrue=" << lock_inv_true
        << ", unlock=" << unlock << ", ui=" << unlock_inv << ", uitrue="<< unlock_inv_true    
        << "\n       ph=" << pref_hit << ", pm=" << pref_miss 
        << ", pm_nowb=" << pm_nowb << ", pm_wb=" << pm_wb
        << "\n       pxh=" << prefex_hit << ", pxm=" << prefex_miss
        <<  ", pxm_nowb=" << pxm_nowb << ", pxm_wb=" << pxm_wb 
        << ", pxi=" << prefex_inv << ", pxitrue=" << prefex_inv_true << "\n"; 
    }

void MESI_Cache::BCU_Stat::dump(int index) {
    if (rdc || rdxc || invc || repl_wbc || snoop_wbc) COUT << "MESI"<<index<<": Bus Traffic\n"
         << "  BusReadRequests=" << rdc 
         << ", BusRead-exRequests=" << rdxc
         << ", BusInvRequests=" << invc 
         << "\n  BusReplWBRequests=" << repl_wbc
         << ", BusSnoopWBRequests=" << snoop_wbc << "\n";
  }

void MESI_Cache::SNOOP_Stat::dump(int index) {
   if (etos || mtos || stoi || mtoi || etoi) COUT << "MESI"<<index<<": SnoopCC Transitions\n"
        << "  etos=" << etos << ", mtos=" << mtos
        << ", stoi=" << stoi << ", mtoi=" << mtoi 
        << ", etoi=" << etoi << "\n";
  }      
              
/*
* cycle method
*/
// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
    
void MESI_Cache::cycle() {
    
    cache_memory.set_current_frame(0);    
    if (!Topology::cache_read_request(index)) {
         pcc_state=PCC_IDLE;
         return; }
    pcc_state=PCC_INITIAL;
    ++stat.req;                                         // statistics
    Topology::pstats.req++;  
    switch (in.request.type) {                          // select where to go
        case REQUEST_READ   : FIL_action=FIL_READ; goto L_READ;
        case REQUEST_LOCK   : FIL_action=FIL_LOCK; goto L_LOCK;
        case REQUEST_UNLOCK : FIL_action=FIL_UNLOCK; goto L_UNLOCK;
        case REQUEST_WRITE  : FIL_action=FIL_WRITE; goto L_WRITE;
        case REQUEST_PREF   : FIL_action=FIL_PREF; goto L_PREF;
        case REQUEST_PREFEX : FIL_action=FIL_PREFEX; goto L_PREFEX;
        case REQUEST_LPREF  : FIL_action=FIL_LPREF; goto L_LPREF;
        case REQUEST_WRBACK : FIL_action=FIL_WRBACK; goto L_WRBACK;
        default : break;
        }
//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    Topology::pstats.rd_req++;
    if (cache_memory.isvalid(in.request.addr)) {
        stat.read_hit++;
        Topology::pstats.rd_hit++;
        }
    else { // it's a miss
        stat.read_miss++;
        Topology::pstats.rd_miss++;
        Topology::bstats.rdc++;
        bcu_stat.rdc++; //read bus cycle is initiated
        // replacement
        switch(Topology::cache[index]->cache_memory.getstate()) {
          case MESI_Line_State::INVALID:
              {stat.rm_nowb_i++; break;}
          case MESI_Line_State::SHARED:
              {stat.rm_nowb_s++; break;}          
          case MESI_Line_State::EXCLUSIVE:
             {stat.rm_nowb_e++; break;} 
          case MESI_Line_State::MODIFIED:
             {stat.rm_wb++;
              bcu_stat.repl_wbc++;
              Topology::bstats.rwbc++;
              stat.rep++; 
              break;}
        } // end of chack block for replacement
        // snooping
        shared=0;
        for(i=0; i<CPU_COUNT; i++) {
          if (i!=index) {
             if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                shared=1;
                switch(Topology::cache[i]->cache_memory.getstate()) {
                   case MESI_Line_State::EXCLUSIVE:
                       {Topology::cache[i]->snoop.etos++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
                        break;}
                   case MESI_Line_State::MODIFIED:
                       {Topology::cache[i]->snoop.mtos++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
                        bcu_stat.snoop_wbc++;
                        Topology::bstats.swbc++;
                        break;} 
                   default: break;
                } // end of switch                   
             } // end of if
          } // end of if
        } //end-for snooping
        if(shared) 
           cache_memory.setstate(MESI_Line_State::SHARED); //set new state   
        else
           cache_memory.setstate(MESI_Line_State::EXCLUSIVE); //set new state   
       // simulate write of the cache line
       while(cache_memory.line_fetch_putnext(&tadr));
    } // end of else
    if(FIL_action==FIL_LOCK) goto L_LOCK_check_value;
    //
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    goto L_END;  
//---------------------------------------------------------------
L_WRITE:
    Topology::pstats.wr_req++;
    if(cache_memory.isvalid(in.request.addr)){ // write hit
       stat.write_hit++;
       //Topology::pstats.wr_hit++;
       if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         Topology::pstats.wr_hit++;
       }
       if(cache_memory.getstate()==MESI_Line_State::MODIFIED) {
         Topology::pstats.wr_hit++;
       }       
       if(cache_memory.getstate()==MESI_Line_State::SHARED) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         bcu_stat.invc++;
         Topology::pstats.wr_inv++;
         Topology::bstats.invc++;
         stat.write_inv++;
         //invalidate bus cycle
         exist_inv=0;
         for(i=0; i<CPU_COUNT; i++) {
           if (i!=index) {
              if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                 if(Topology::cache[i]->cache_memory.getstate()==MESI_Line_State::SHARED)
                   {Topology::cache[i]->snoop.stoi++;
                    Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                    exist_inv=1;                    
                   } 
              } // end of if
           } // end of if
         } //end of for
         if (exist_inv) stat.write_inv_true++;
       } // end of if  
    }// end of if
    else { // write miss
       Topology::pstats.wr_miss++;
       stat.write_miss++;
       bcu_stat.rdxc++;
       Topology::bstats.rdxc++;
       // replacement
       switch(cache_memory.getstate()) {
         case MESI_Line_State::INVALID:
         case MESI_Line_State::SHARED:
         case MESI_Line_State::EXCLUSIVE:
             {stat.wm_nowb++; break;} 
         case MESI_Line_State::MODIFIED:
             {stat.wm_wb++;
              bcu_stat.repl_wbc++;
              Topology::bstats.rwbc++;
              stat.rep++; 
              break;}
       } // end of switch
       // simulate snooping
       for(i=0; i<CPU_COUNT; i++) {
          if (i!=index) {
             if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                switch(Topology::cache[i]->cache_memory.getstate()) {
                   case MESI_Line_State::EXCLUSIVE:
                       {Topology::cache[i]->snoop.etoi++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                        break;}
                   case MESI_Line_State::MODIFIED:
                       {Topology::cache[i]->snoop.mtoi++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                        bcu_stat.snoop_wbc++;
                        Topology::bstats.swbc++;
                        break;}
                   case MESI_Line_State::SHARED:
                       {Topology::cache[i]->snoop.stoi++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                        break;}      
                   default: break;
                } // end of switch                   
             } // end of if
          } // end of if
       } //end of for
       // set new state
       cache_memory.setstate(MESI_Line_State::MODIFIED);
       // simulate line fetch
       while(cache_memory.line_fetch_putnext(&tadr));
    } //end of else
    if(FIL_action==FIL_UNLOCK) goto L_UNLOCK_value;
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    goto L_END;
//---------------------------------------------------------------
L_LOCK:
    stat.lock++;
    if(!cache_memory.isvalid(in.request.addr)) { //miss
       goto L_READ;
    }
L_LOCK_check_value:   
   lock_value = MEMSIM_READ_LOCK(in.request.addr);
   // a dumm one does test&set and let the kernel spin for him
   MEMSIM_WRITE_LOCK(in.request.addr, 1); 
   if (!lock_value) { // lock is free
       if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
       }
       if(cache_memory.getstate()==MESI_Line_State::SHARED) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         bcu_stat.invc++;
         Topology::bstats.invc++;
         stat.lock_inv++;
         //invalidate cycle
         exist_inv=0;
         for(i=0; i<CPU_COUNT; i++) {
           if (i!=index) {
              if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                 if(Topology::cache[i]->cache_memory.getstate()==MESI_Line_State::SHARED)
                       {Topology::cache[i]->snoop.stoi++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                        exist_inv=1;                       
                       } 
              } // end of if
           } // end of if
         } //end of for
         if(exist_inv) stat.lock_inv_true++;
       } // end of if  
   }//end of if(!lock_value)           
   out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
   out.response.STROBE = 1;
   goto L_END;
//---------------------------------------------------------------
L_UNLOCK:
   stat.unlock++;
   if(!cache_memory.isvalid(in.request.addr)) {//it's a miss
      goto L_WRITE;
   }
   //else it's a hit
   if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
     cache_memory.setstate(MESI_Line_State::MODIFIED);
   }
   if(cache_memory.getstate()==MESI_Line_State::SHARED) {
      cache_memory.setstate(MESI_Line_State::MODIFIED);
      bcu_stat.invc++;
      Topology::bstats.invc++;
      stat.unlock_inv++;
      //invalidate cycle
      exist_inv=0;
      for(i=0; i<CPU_COUNT; i++) {
         if(i!=index) {
           if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
              if(Topology::cache[i]->cache_memory.getstate()==MESI_Line_State::SHARED)
                 {Topology::cache[i]->snoop.stoi++;
                  Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                  exist_inv=1;                 
                 } 
              } // end of if
         } // end of if
      } //end of for
      if(exist_inv) stat.unlock_inv_true++;
   } // end if(SHARED)  
L_UNLOCK_value:   
   MEMSIM_WRITE_LOCK( in.request.addr, 0);
   out.response.signal = PROCESSOR_SATISFIED; 
   out.response.STROBE = 1;
   goto L_END;
    
//---------------------------------------------------------------
L_PREF:
    if (cache_memory.isvalid(in.request.addr)) {
        stat.pref_hit++;
        }
    else { // it's a miss
        stat.pref_miss++;
        Topology::bstats.rdc++;
        bcu_stat.rdc++;
        // replacement
        switch(cache_memory.getstate()) {
          case MESI_Line_State::INVALID:
          case MESI_Line_State::SHARED:
          case MESI_Line_State::EXCLUSIVE:
             {stat.pm_nowb++; break;} 
          case MESI_Line_State::MODIFIED:
             {stat.pm_wb++;
              bcu_stat.repl_wbc++;
              Topology::bstats.rwbc++;
              stat.rep++; 
              break;}
        } // end of switch
        // snooping
        shared=0;
        for(i=0; i<CPU_COUNT; i++) {
          if (i!=index) {
             if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                shared=1;
                switch(Topology::cache[i]->cache_memory.getstate()) {
                   case MESI_Line_State::EXCLUSIVE:
                       {Topology::cache[i]->snoop.etos++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
                        break;}
                   case MESI_Line_State::MODIFIED:
                       {Topology::cache[i]->snoop.mtos++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::SHARED);
                        bcu_stat.snoop_wbc++;
                        Topology::bstats.swbc++;
                        break;} 
                   default: break;
                } // end of switch                   
             } // end of if
          } // end of if
        } //end of for
        if(shared)
          cache_memory.setstate(MESI_Line_State::SHARED); //set new state   
        else 
          cache_memory.setstate(MESI_Line_State::EXCLUSIVE);
        // simulate line fetch
        while(cache_memory.line_fetch_putnext(&tadr));
    } // end of else
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    goto L_END;  

//---------------------------------------------------------------
L_PREFEX:
    if (cache_memory.isvalid(in.request.addr)) {
       stat.prefex_hit++;
       if(cache_memory.getstate()==MESI_Line_State::EXCLUSIVE) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
       }
       if(cache_memory.getstate()==MESI_Line_State::SHARED) {
         cache_memory.setstate(MESI_Line_State::MODIFIED);
         bcu_stat.invc++;
         Topology::bstats.invc++;
         stat.prefex_inv++;
         //invalidate bus cycle
         exist_inv=0;
         for(i=0; i<CPU_COUNT; i++) {
           if (i!=index) {
              if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                 if(Topology::cache[i]->cache_memory.getstate()==MESI_Line_State::SHARED)
                   {Topology::cache[i]->snoop.stoi++;
                    Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                    exist_inv=1;                    
                   } 
              } // end of if
           } // end of if
         } //end of for
         if (exist_inv) stat.prefex_inv_true++;
       } // end of if  
    }// end of if
    else { // it's a miss
        stat.prefex_miss++;
        bcu_stat.rdxc++;
        Topology::bstats.rdxc++;
        // replacement
        switch(cache_memory.getstate()) {
          case MESI_Line_State::INVALID:
          case MESI_Line_State::SHARED:
          case MESI_Line_State::EXCLUSIVE:
             {stat.pxm_nowb++; break;} 
          case MESI_Line_State::MODIFIED:
             {stat.pxm_wb++;
              bcu_stat.repl_wbc++;
              Topology::bstats.rwbc++;
              stat.rep++; 
              break;}
        } // end of switch
        // snooping
        for(i=0; i<CPU_COUNT; i++) {
          if (i!=index) {
             if(Topology::cache[i]->cache_memory.isvalid(in.request.addr)) {
                switch(Topology::cache[i]->cache_memory.getstate()) {
                   case MESI_Line_State::EXCLUSIVE:
                       {Topology::cache[i]->snoop.etoi++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                        break;}
                   case MESI_Line_State::MODIFIED:
                       {Topology::cache[i]->snoop.mtoi++;
                        Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                        bcu_stat.snoop_wbc++;
                        Topology::bstats.swbc++;
                        break;}
                   case MESI_Line_State::SHARED:
                        {Topology::cache[i]->snoop.stoi++;
                         Topology::cache[i]->cache_memory.setstate(MESI_Line_State::INVALID);
                         break;}      
                   default: break;
                } // end of switch                   
             } // end of if
          } // end of if
        } //end of for
        cache_memory.setstate(MESI_Line_State::EXCLUSIVE); //set new state   
        // simulate line fetch
        while(cache_memory.line_fetch_putnext(&tadr));  
    } // end of else
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    goto L_END;  

//----------------------------------------------------------------
L_LPREF:

//----------------------------------------------------------------
L_WRBACK:
//-------------------end of PCC----------------------------------------
L_END: 
 return;
 }

// debugging...

void MESI_Cache::dump_state() {
    COUT << "CC" << index << ": PCC_State: ";
        switch (pcc_state) {
            case PCC_INITIAL         : COUT << "PCC_INITIAL"; pcc_state=PCC_IDLE;break;
            case PCC_IDLE	     : COUT << "PCC_IDLE"; break;	
        }
     if(pcc_state==PCC_INITIAL) {
        COUT << "\t\t[";
        switch (in.request.type) {      
            case REQUEST_READ  : COUT << "READ " << hex <<in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << hex <<in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << hex <<in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << hex <<in.request.addr; break;
            case REQUEST_PREF  : COUT << "PREF " << hex <<in.request.addr; break;
            case REQUEST_PREFEX: COUT << "PREFEX"<< hex <<in.request.addr; break;
            case REQUEST_LPREF : COUT << "LPREF" << hex <<in.request.addr; break;
            case REQUEST_WRBACK: COUT << "WRBACK"<< hex <<in.request.addr; break;
            default : break;
        }
       COUT << "]"; 
     }   
     COUT << "\n";    
     cache_memory.dump_state(index);
    }
                

#endif PRAGMA_IMPLEMENTATION_MESI_CACHE
#endif _MESI_CACHE_H_