#ifndef _BUS_H_
#define _BUS_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: bus.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The bus module. This module also arbitrates for locks -- each cache
 * has its LOCK_REQUEST and LOCK_GRANT line.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "bus_signals.h"

/*
 * This bus module is universal: it can be used in other memsims, raw or
 * extended. It only requires Bus_Signals class to be defined (in 
 * "bus_signals.h") with only two mandatory methods: reset() and dump().
 * The internals of Bus_Signals are not of interest for this module.
 * As of the Topology, three are the functions that this module calls:
 * bool Topology::bus_read_REQs();  
 *      // reads REQ signals from all the modules connected to the bus
 *      // returns true if there are any requests. Of course, it also
 *      // copies these modules' REQ signals to the bus. See topology.cc.
 * void Topology::bus_read_contents(int GRANT_holder);  
 *      // copies the output port (that goes to the bus) from the module
 *      // that holds the bus, to bus_signals variable in Bus_Module
 * bool Topology::bus_read_LOCK_REQs();
 *      // reads LOCK requests and returns true if any. If you don't need it
 *      // for your memsim, just do a 'return false' in its body.
 */

class Bus_Module {
    unsigned busytime;          // statistics
    unsigned lockbusytime;
  protected:
    Bus_Signals bus_contents;
    // common bus signals
    bool BUS_REQ[MAX_CPU_COUNT];
    bool BUS_GNT[MAX_CPU_COUNT];
    int GRANT_holder;
    int priority_position;
    enum State {
        BUS_FREE,
        BUS_BUSY
        } state;
    // locking stuff
    bool LOCK_REQ[MAX_CPU_COUNT];
    bool LOCK_GNT[MAX_CPU_COUNT];
    int LOCK_holder;
    int lock_priority_position;
    enum Lock_State {
        LOCK_FREE,
        LOCK_BUSY
        } lock_state;
    void lock_cycle();
  public:
    Bus_Module();
    virtual void cycle();
    virtual void dump_state();
    virtual void dump_stats(Thread_Time total_time);
    TOPOLOGY_FRIENDS
    };

#ifdef PRAGMA_IMPLEMENTATION_BUS_MODULE
///////////////////////////////////////
#include <stdio.h>

Bus_Module::Bus_Module() {
    bus_contents.reset();
    for (int i=0; i<MAX_CPU_COUNT; ++i) 
        BUS_REQ[i] = BUS_GNT[i] = LOCK_REQ[i] = LOCK_GNT[i] = 0;
    GRANT_holder = 0;
    priority_position = 0;
    state = BUS_FREE;
    busytime = 0;
    LOCK_holder = 0;
    lock_priority_position = 0;
    lock_state = LOCK_FREE;
    lockbusytime = 0;
    }
    
void Bus_Module::dump_stats(Thread_Time total_time) {
    if (total_time == 0) return;
    PRINTF("BUS: busy: %d of %d cycles. Utilization: %4.2f%%\n",
        busytime, total_time, 100.00*busytime/total_time );
    PRINTF("LOCK_LINE: busy: %d of %d cycles. Occupation: %4.2f%%\n",
        lockbusytime, total_time, 100.00*lockbusytime/total_time );
    }

// the cycle() method is called by the Topology in each simulated cycle.
// The bus module has two states: busy and free. If free, it checks if 
// anyone is asking for it, by reading BUS_REQ from each module. Using
// a rotating priority scheme, it decides which module of those with BUS_REQ
// high will receive BUS_GNT high. If busy, it checks if the one who currently
// holds the bus has dropped its BUS_REQ, and gives BUS_GNT to someone else
// or goes idle. The locking arbitration is quite similar.

void Bus_Module::cycle() {
    if (state != BUS_FREE) ++busytime;  // stats
    bool any_requests = Topology::bus_read_REQs();
    lock_cycle();
l_check_state:
    switch (state) {
        case BUS_FREE: 
            if (!any_requests) return;  
            // arbitrate if there is more than one request; 
            // use rotating-priorities scheme
            GRANT_holder = priority_position;
            do {
                if (BUS_REQ[GRANT_holder]) {
                    // give the new holder the lowest priority for the next time
                    priority_position = (GRANT_holder + 1) % MAX_CPU_COUNT;
                    break;
                    }
                GRANT_holder = (GRANT_holder + 1) % MAX_CPU_COUNT;
                } while (GRANT_holder != priority_position);
            BUS_GNT[GRANT_holder] = 1;
            state = BUS_BUSY;
            return;
        case BUS_BUSY:
            if (BUS_REQ[GRANT_holder]) {    // if the holder still holds the bus
                Topology::bus_read_contents(GRANT_holder);
                return;
                }
            // the preceding requester has disconnected from the bus
            BUS_GNT[GRANT_holder] = 0;
            // note: bus contents will be reset (i.e. go inactive); but this applies
            // only to those who read from the bus in the NEXT cycle, since the
            // bus is the last to be polled witihin a cycle
            bus_contents.reset();
            state = BUS_FREE;
            goto l_check_state; // give the bus immediately if there are any new requests
        }
    }   

void Bus_Module::lock_cycle() {
    if (lock_state != LOCK_FREE) ++lockbusytime;    // stats
    bool any_requests = Topology::bus_read_LOCK_REQs();
l_check_lock_state:
    switch (lock_state) {
        case LOCK_FREE: 
            if (!any_requests) return;  
            LOCK_holder = lock_priority_position;
            do {
                if (LOCK_REQ[LOCK_holder]) {
                    lock_priority_position = (LOCK_holder + 1) % MAX_CPU_COUNT;
                    break;
                    }
                LOCK_holder = (LOCK_holder + 1) % MAX_CPU_COUNT;
                } while (LOCK_holder != lock_priority_position);
            LOCK_GNT[LOCK_holder] = 1;
            lock_state = LOCK_BUSY;
            return;
        case LOCK_BUSY:
            if (! LOCK_REQ[LOCK_holder]) {
                // the requester has disconnected from the bus
                LOCK_GNT[LOCK_holder] = 0;
                lock_state = LOCK_FREE;
                goto l_check_lock_state;    // give the lock right immediately if there are any new requests
                }
            return;
        }
    }   


void Bus_Module::dump_state() {
    PRINTF("BUS: State: ");
    if (state == BUS_FREE) 
        PRINTF("FREE\n");
    else    {
        PRINTF("BUSY  GNT: %c   ", GRANT_holder+'0');
        bus_contents.dump();
        PRINTF("\n");
        }
    // ubaceni deo
    //PRINTF("Lock_holder %d ", LOCK_holder);
    for(int i=0; i<4; i++) {
        if(LOCK_REQ[i]) PRINTF("\nLr%dY", i);
        else PRINTF("\nLr%dN", i);
        if(LOCK_GNT[i]) PRINTF(" Lg%dY", i);
        else PRINTF(" Lg%dN", i);
        if(BUS_REQ[i]) PRINTF(" Br%dY", i);
        else PRINTF(" Br%dN", i);
        if(BUS_GNT[i]) PRINTF(" Bg%dY", i);
        else PRINTF(" Bg%dN", i);
      }
    PRINTF("\n");    
    }

#endif //PRAGMA_IMPLEMENTATION_BUS

#endif //_BUS_H_