#ifndef _MESI_CACHE_H_
#define _MESI_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: mesi_cache.h
 * Author: Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 * Note: this file is based on the methodology deviced by 
 *       Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The MESI protocol; definition and implementation.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "bus_signals.h"


// states a line in the MESI protocol can be found in
class MESI_Line_State {
  public:
    enum State { INVALID, SHARED, EXCLUSIVE, MODIFIED } state;
    MESI_Line_State()               {state = INVALID;}
    bool  isvalid()		    {return state != INVALID;}
    State get()                     {return state;}
    void  set(State new_state)      {state = new_state; }
    char  symbol() { // for debuging
    	             switch (state) {
    	               case INVALID:   return 'I';
    		       case SHARED:    return 'S';
    		       case EXCLUSIVE: return 'E';
    		       case MODIFIED:  return 'M' ;
    	             }
    		   }
    };

// the MESI cache controler module declaration
class MESI_Cache {
    int index;
    // instantiate Cache Memory with the MESI_Line_State as its Line_State
    Cache_Memory<MESI_Line_State> cache_memory;  // the whole cache memory
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        bool BUS_GNT;                   // bus grant (from the bus module)
        bool LOCK_GNT;                  // lock grant (from the bus module)
        Bus_Signals bus;                // bus contents, to be read from the bus
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        bool BUS_REQ;                   // bus request: sent to the bus module
        bool LOCK_REQ;                  // lock request: sent to the bus module
        Bus_Signals bus;                // bus contents, to be written on the bus
        } out;
    // states this finite state machine called MESI can be found in
    enum State {
        INITIAL,                        // INITIAL -- not doing anything (except maybe in the snoop part)
        // fetch line caused by RD, WR, LOCK
        FETCH_WAIT_BGNT,                // waiting for the bus to do a read
        FETCH_WAIT_SNOOP,                  // waiting one cycle to let the others see the SNOOP signal before we withdraw it from the bus
        FETCH_LINE_RETRY,
        FETCH_WAIT_MEMORY,
        // write back modified line
        WB_WAIT_MEMORY,
        // invalidate caused by write hit in shared cache line
        INV_WAIT_BGNT,
        INV_WAIT_END,
        // lock (version 2) 
        LOCK_WAIT_BGNT,
        LOCK_SLEEP,
        // unlock
        UNLOCK_WAIT_BGNT
       } state;
    
    // the SCC (snoop cache controller part) may also be in one of three states
    enum Snoop_State {
        SNOOP_INITIAL,                  // not doing a snoop
        SNOOP_WAIT_BGNT,
        SNOOP_WR_WAIT_MEMORY
        } snoop_state;
   
         
    // One 'procedure' is called to fetch an invalid line; it can be called
    // upon a read, a write, or a lock. This is sort of replacement for gosub-return sequence
    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK} FIL_action;
    // some internal vars
//   bool WB_flag; 	// write-back flag
//    bitsADDR wb_addr; 	// write-back address
    int wait_state_counter;
    bool shared_line;
    bool dirty_line;
    // delay parameters of the simulated physical memory: the default values
    const DEFAULT_RD_WAIT_STATES     = 3; // time needed to read/read_ex the first word
    const DEFAULT_RD_WAIT_STATES_2   = 0; // time needed for access to subsequent words
    const DEFAULT_WR_WAIT_STATES     = 3;
    const DEFAULT_WR_WAIT_STATES_2   = 0;
    const DEFAULT_INV_WAIT_STATES    = 2;
    // and the real values, which can be changed at startup
    int READ_WAIT_STATES;               // these will become equal to the
    int READ_WAIT_STATES_2;             // defaults if not otherwise specified
    int WRITE_WAIT_STATES;               // in the INI file or command line
    int WRITE_WAIT_STATES_2;             // switches
    int INV_WAIT_STATES;
    const BUSY_LOCK_SLEEP_CYCLE_COUNT   = 1; 
    const FETCH_RETRY_WAIT_STATES 	= 2;
    bitsLOCK lock_value;
    int busy_lock_sleep_cycle_count;
    struct Stat {       // statistics:
        int req;        // requests
        int inv;        // invalidations
        int rep;        // replacements
        Stat()  { req=inv=rep=0;}
        void dump(int index);
        } stat;
  public:
    MESI_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {return (state == INITIAL);}
    void dump_state();
    void dump_stats() {stat.dump(index);}
    TOPOLOGY_FRIENDS
    };


#ifdef PRAGMA_IMPLEMENTATION_MESI_CACHE

// the MESI cache controller module implementation

#include <iostream.h>
    
MESI_Cache::MESI_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    state = INITIAL;                            // PCC will be in the INITIAL state
    snoop_state = SNOOP_INITIAL;                // and so will the SCC part
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    in.BUS_GNT      = 0;
    out.BUS_REQ     = 0;
    in.LOCK_GNT     = 0;
    out.LOCK_REQ    = 0;
    in.bus.reset();
    out.bus.reset();
    // initialize memory delay parameters. If you want to change this,
    // create limes.ini in your app's directory and place in the line
    // r_ws=4 (for example), or use simulator's command line switches,
    // like -dw_ws=4.
    READ_WAIT_STATES    = sys_INI_read("r_ws", DEFAULT_RD_WAIT_STATES);
    READ_WAIT_STATES_2  = sys_INI_read("r_ws2", DEFAULT_RD_WAIT_STATES_2);
    WRITE_WAIT_STATES   = sys_INI_read("w_ws", DEFAULT_WR_WAIT_STATES);
    WRITE_WAIT_STATES_2 = sys_INI_read("w_ws2", DEFAULT_WR_WAIT_STATES_2);
    INV_WAIT_STATES = sys_INI_read("i_ws", DEFAULT_INV_WAIT_STATES);
    };

void MESI_Cache::Introduce() {
    COUT << "Protocol: MESI (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << "LOCK_line aware.";
    #endif
    COUT << "\n";
    COUT << "READ_WS=" << READ_WAIT_STATES << "/" << READ_WAIT_STATES_2 << ", ";
    COUT << "WRITE_WS=" << WRITE_WAIT_STATES << "/" << WRITE_WAIT_STATES_2 << ", ";
    COUT << "LOCK_WS=" << BUSY_LOCK_SLEEP_CYCLE_COUNT << "\n";
    #ifdef DUMMLOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    }

void MESI_Cache::Stat::dump(int index) {
    if (req) COUT << "MESI"<<index<<": " << \
        "requests="<<req << ", invalidations="<<inv << ", replacements="<<rep
        << "\n";
    }

// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
// This method first reads the bus. Then it checks if there is SNOOP signal
// high on it, and, if so, checks if it has the line to invalidate. If true,
// does so and skips this cycle. This reflects the fact that there are two
// controllers in this one, the Processor Cache Controller (PCC) and the
// Snoop Cache Controller (SCC), each of which has its own copy of cache
// tags, but the cache memory itself can be accessed only by one controller
// at a time, and the SCC part has precedence.
    
void MESI_Cache::cycle() {
    
    Topology::cache_read_bus(index);        // read input from the bus

    // check if there is SNOOP signal on the bus, but issued by a cache
    // other than myself:

    //Topology::system_dump_state();
     
    switch (snoop_state) {
        // the SCC may be in initial state, or not. If it is in initial state,
        // it will check if a SNOOP signal appeared on the bus. Otherwise,
        // it will continue what it was doing in the previous cycle -- doing
        // a cache supply for some other cache controller.
        case SNOOP_INITIAL:
            if (in.bus.SNOOP_REQ && !in.BUS_GNT)    {   // !in.BUS_GNT means 'I don't have the bus'
                // there is: select frame #1 of cache_memory's C (current) vars,
                // which keep the current position of the line
                cache_memory.set_current_frame(1);  
                if (cache_memory.isvalid(in.bus.addr))
                    goto L_SNOOP;   // most of the SCC part is there
                else
                    cache_memory.set_current_frame(0);  // frame #0 for the processor part
                    // it was a snoop miss, so proceed as usual (though it might be more realistic to return)
                }
            break;          // activate processor cache controller
        case SNOOP_WAIT_BGNT		: goto L_STATE_SNOOP_WAIT_BGNT;
        case SNOOP_WR_WAIT_MEMORY	: goto L_STATE_SNOOP_WR_WAIT_MEMORY;
        }
    
    // ok, there was no snoop or it was a miss, so we continue
    switch (state) {
        case INITIAL                    : goto L_STATE_INITIAL;
        case FETCH_WAIT_BGNT		: goto L_STATE_FETCH_WAIT_BGNT;
        case FETCH_WAIT_SNOOP		: goto L_STATE_FETCH_WAIT_SNOOP;
        case FETCH_LINE_RETRY		: goto L_STATE_FETCH_LINE_RETRY;
        case WB_WAIT_MEMORY		: goto L_STATE_WB_WAIT_MEMORY;
        case INV_WAIT_BGNT		: goto L_STATE_INV_WAIT_BGNT;
        case INV_WAIT_END		: goto L_STATE_INV_WAIT_END;
        case LOCK_WAIT_BGNT		: goto L_STATE_LOCK_WAIT_BGNT;
        case LOCK_SLEEP			: goto L_STATE_LOCK_SLEEP;
        case UNLOCK_WAIT_BGNT		: goto L_STATE_UNLOCK_WAIT_BGNT;
        case FETCH_WAIT_MEMORY		: goto L_STATE_FETCH_WAIT_MEMORY;
        }
        
L_STATE_INITIAL:
    COUT << "--P" << index << "(initial) " << Topology::getCurrentTime() << "\n";
    if (! Topology::cache_read_request(index)) return;  // return if there was none
    ++stat.req;                                         // statistics
    switch (in.request.type) {                          // select where to go
        case REQUEST_READ   : goto L_READ;
        case REQUEST_LOCK   : goto L_LOCK;
        case REQUEST_UNLOCK : goto L_UNLOCK;
        case REQUEST_WRITE  : goto L_WRITE;
        default : break;
        }
    return;
    

//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    if (cache_memory.isvalid(in.request.addr)) {
        // a hit, excellent! Say that the request is satisfied, and return
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_READ_end;
        }
    FIL_action = FIL_READ;      // these three lines are an equivalent for
    goto L_FETCH_INVALID_LINE;  // "gosub fetch_invalid_line"; but there can be
    L_FIL_READ_ret:             // no gosub when F.I.L. spans more than one cycle
L_READ_end:
    state = INITIAL;
    return;

//---------------------------------------------------------------
L_WRITE:
    // this is a WRITE. We'll say immediately that it is satisfied.
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    // and check whether it is in the cache, and if so, is it valid.
    // The isvalid() method will do all the work for us, and even pick
    // a line for replacement if necessary.
    if (cache_memory.isvalid(in.request.addr))
        goto L_WRITE_valid;
L_WRITE_invalid:
    FIL_action = FIL_WRITE;         // gosub L_FETCH_INVALID_LINE
    goto L_FETCH_INVALID_LINE;
    L_FIL_WRITE_ret:                // return point
    goto L_WRITE_end;               // that's all there is for a WRITE
L_WRITE_valid:
    // in the MESI protocol, write actions depend on the state of the line
    // if the line is already modified, and exclusive, no further actions
    // are necessary
    if (cache_memory.getstate() == MESI_Line_State::MODIFIED)
        goto L_WRITE_end;  
    else if (cache_memory.getstate() == MESI_Line_State::EXCLUSIVE)
        { cache_memory.setstate(MESI_Line_State::MODIFIED);
          goto L_WRITE_end;}   
    // if not, line is SHARED and we'll have to invalidate the others. But we'll have to get
    // the bus first.
    out.BUS_REQ = 1;                // I want the bus, please
    state = INV_WAIT_BGNT;        // and I'm waiting for it
    // the fact that there is no "return" here means that BUS_GNT will
    // be waited for only if not high already (it could be high if we are
    // doing a WRITE for a LOCK, which may have gained bus before jumping here)
L_STATE_INV_WAIT_BGNT:
    COUT << "--P" << index << "(InvWaitBgnt) " << Topology::getCurrentTime() << "\n";
    if (! in.BUS_GNT) return;       // wait as long as it takes
    // Now, we've got the bus; but the line may have gotten invalidated 
    // meanwhile by our snoopy cache controller part!
    if (cache_memory.getstate() == MESI_Line_State::INVALID) 
        goto L_WRITE_invalid;       // if became invalid meanwhile
    // OK, the line is still valid. Now we'll declare it exclusively ours.
    out.bus.reset();
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.addr     = in.request.addr;     // this addr will go to the bus
    out.bus.size     = in.request.size;     // not really needed
    out.bus.SNOOP_REQ    = 1;               // SNOOP_REQ goes high
    out.bus.INV      = 1;                   // INV too
    wait_state_counter = INV_WAIT_STATES;
    state = INV_WAIT_END;                   // and we'll wait one cycle to let
    return;                                 // the others see these signals
L_STATE_INV_WAIT_END:
    COUT << "--P" << index << "(InvWaitEnd) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter-- >0) return;
    cache_memory.setstate(MESI_Line_State::MODIFIED); // modified
L_WRITE_end:                                // the write is over:
    out.BUS_REQ = out.LOCK_REQ = 0;         // drop the bus, if we had it
    out.bus.INV =0;			    // drop INV line
    state = INITIAL;                        // and go to the INITIAL state
    return;

//---------------------------------------------------------------
L_LOCK:
    // LOCKing uses bus as a hardware binary semafore. There are two possible
    // schemes: the first calls the bus arbitrator, thereby blocking the bus,
    // and the second, more efficient, calls the special lock arbitrator,
    // thus leaving the bus free for other operations. The actual scheme
    // is selected at compile time (check out snoopy.h):
    #ifdef LOCK_LINE_AWARE
        out.LOCK_REQ = 1;       // raise LOCK request high (the second scheme)
    #else
        out.BUS_REQ = 1;        // raise BUS request high (the first scheme)
    #endif
    // we are waiting for the bus to tell us we can proceed
    state = LOCK_WAIT_BGNT;
L_STATE_LOCK_WAIT_BGNT:
    COUT << "--P" << index << "(LockWaitBgnt) " << Topology::getCurrentTime() << "\n";
    #ifdef LOCK_LINE_AWARE
        if (! in.LOCK_GNT) return;
    #else
        if (! in.BUS_GNT) return;
    #endif
    // now check if the line is in the cache (and whether it is valid):
    if (cache_memory.isvalid(in.request.addr)) 
        goto L_LOCK_check_value;                // yes, it is.
    FIL_action = FIL_LOCK;                      // no, do
    goto L_FETCH_INVALID_LINE;                  // gosub F.I.L.
    L_FIL_LOCK_ret:
L_LOCK_check_value:
    // the memsim has to actually read the lock from the real memory, since
    // cache keeps only tags, not memory contents. The real memory is
    // actually totally consistent (PRAM), everything here is about modelling
    // delays. This action would not have been necessary
    // if we had really moved the data from the memory to the simulated 
    // cache and vice versa, but this is an extreme effort (though this
    // controller was originally written that way and crippled later).
    lock_value = MEMSIM_READ_LOCK(in.request.addr);
    #ifndef DUMMLOCK
        // non-dumm lock does test&test&set and spins it self
        if (lock_value == 0) {
            // Do the actual "write 1" to the location for the lock.
            MEMSIM_WRITE_LOCK( in.request.addr, 1); 
            goto L_WRITE;
            }
        out.BUS_REQ = out.LOCK_REQ = 0;
        busy_lock_sleep_cycle_count = BUSY_LOCK_SLEEP_CYCLE_COUNT;
        state = LOCK_SLEEP;
        return;
    #else
        // a dumm one does test&set and let the kernel spin for him
        MEMSIM_WRITE_LOCK( in.request.addr, 1); 
        out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_WRITE_valid;
    #endif  
L_STATE_LOCK_SLEEP:
    COUT << "--P" << index << "(LockSleep) " << Topology::getCurrentTime() << "\n";
    if (busy_lock_sleep_cycle_count--) return;
    goto L_LOCK;

//---------------------------------------------------------------
L_UNLOCK:
    // this is an UNLOCK (lock release) request. Basically it is a write,
    // only that it has to gain the bus first.
    out.LOCK_REQ = 1;
    state = UNLOCK_WAIT_BGNT;
L_STATE_UNLOCK_WAIT_BGNT:     
    COUT << "--P" << index << "(UnlockWaitBgnt) " << Topology::getCurrentTime() << "\n";
    #ifdef LOCK_LINE_AWARE
        if (!in.LOCK_GNT) return;
    #endif
    // 
    // do the actual "write 0" for the lock. Note: it would be a mistake
    // to do that BEFORE this point! Actually, it should be done somewhere
    // at L_WRITE_end -- that is where (when) it would happen in reality
    //
    MEMSIM_WRITE_LOCK( in.request.addr, 0);
    goto L_WRITE;
    
//---------------------------------------------------------------
L_FETCH_INVALID_LINE:
    // This is a "procedure" for fetching an invalid line. cache_memory
    // already points to a line suitable for replacement. We'll check
    // its state and see if we have to do a write-back.
    out.BUS_REQ = 1;                // bus is what we need first
    state = FETCH_WAIT_BGNT;
L_STATE_FETCH_WAIT_BGNT:            // we'll spin here, cycle after cycle,
    COUT << "--P" << index << "(FetchWaitBgnt) " << Topology::getCurrentTime() << "\n";
    if (! in.BUS_GNT) return;       // until the bus arbitrator sends BUS_GNT high
    switch (cache_memory.getstate()) {
        case MESI_Line_State::INVALID:
        case MESI_Line_State::SHARED:
        case MESI_Line_State::EXCLUSIVE: 
            goto L_FETCH_fetch_line;
        case MESI_Line_State::MODIFIED:   // if the line is MODIFIED, it has to be written back
              break;
     }
    ++stat.rep;
    // save address of the cache line which must be replaced
    // sta treba ovde uraditi    
    cache_memory.line_write_init(&out.bus.addr);
    // write back address is wb_addr
    //out.bus.addr = wb_addr;
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP_REQ  = 0;
    wait_state_counter = WRITE_WAIT_STATES;
    state = WB_WAIT_MEMORY;
    return;  // da li ovde treba staviti return
L_STATE_WB_WAIT_MEMORY:
    COUT << "--P" << index << "(WbWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
        }
    // write-back is finished 

L_FETCH_fetch_line:
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP_REQ   = 1;
    if (FIL_action == FIL_WRITE) out.bus.INV = 1; //for read-exclusive bus cycle
    out.bus.memcmd   = REQUEST_READ;
    state = FETCH_WAIT_SNOOP;
    wait_state_counter = 1;
    return;
L_STATE_FETCH_WAIT_SNOOP:
    COUT << "--P" << index << "(FetchWaitSnoop) " << Topology::getCurrentTime() << "\n";
    out.bus.SNOOP_REQ = 0; // ?zasto se ovde ukida odmah SNOOP  
    if (wait_state_counter--) return;
    // now, we have effectively waited two cycles. Why? The first one was
    // to let the other controllers see our signals. The second was to let
    // us see their responses, if any.
    
    // check shared and dirty lines
    shared_line = in.bus.SHARED;
    dirty_line = in.bus.DIRTY;
    if (shared_line && dirty_line) 
       { out.BUS_REQ = 0; out.LOCK_REQ = 0;
         out.bus.INV =0;
         wait_state_counter = FETCH_RETRY_WAIT_STATES;
         state = FETCH_LINE_RETRY;
         goto L_STATE_FETCH_LINE_RETRY;
        }
    wait_state_counter = READ_WAIT_STATES;
    state = FETCH_WAIT_MEMORY;            
L_STATE_FETCH_WAIT_MEMORY:
    COUT << "--P" << index << "(FetchWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter-- >0 ) return;
    if (out.bus.addr <= in.request.addr && 
        in.request.addr+in.request.size <= out.bus.addr + DATA_WIDTH) {
        // here we come if data for read/writing that made us do all this
        // belongs to this word
        switch (FIL_action) {
            case FIL_READ:
                out.response.signal = PROCESSOR_SATISFIED;
                out.response.STROBE = 1;
                break;
            case FIL_LOCK:
                break;
            case FIL_WRITE:
                break;
            }               
        }
    // "write" the data, get next addr and find out if there's more to do
    if (cache_memory.line_fetch_putnext(&out.bus.addr))
       { wait_state_counter = READ_WAIT_STATES_2; 
         return;
         }
    cache_memory.touch();   // force this line MRU (most recently used)
    // counting finished, the whole line fetched
    //out.bus.SUPPLY = 0;     // if it was a cache supply, it's over now
    // now we have to set the state for the new line and to return.
    switch( FIL_action ) {
        case FIL_READ:
            //out.BUS_REQ = 0;
            if (shared_line) 
               cache_memory.setstate(MESI_Line_State::SHARED);
            else 
               cache_memory.setstate(MESI_Line_State::EXCLUSIVE);
            break;
        case FIL_WRITE:
            //out.BUS_REQ = 0;
            cache_memory.setstate(MESI_Line_State::MODIFIED);
            break;
        case FIL_LOCK: // lock strategy
            cache_memory.setstate(MESI_Line_State::SHARED); //????
        }
 
//------------------------------------write-back part
/* 
    // chech write-back flag
    if (! WB_flag)
       goto L_FETCH_LINE_end;
    // Ok, write-back will be performed    
    out.bus.reset();
    // initialize the line for write-back. The cache memory will return
    // the address of the first word in the line, and that's what will
    // go to the bus.
    // this is dummy line_write_init method call (aimed to reset counter)
    cache_memory.line_write_init(&out.bus.addr);
    // write back address is wb_addr
    out.bus.addr = wb_addr;
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP_REQ  = 0;
    wait_state_counter = WRITE_WAIT_STATES;
    state = WB_WAIT_MEMORY;
    return;  // da li ovde treba staviti return
L_STATE_WB_WAIT_MEMORY:
    COUT << "--P" << index << "(WbWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
        }
    // write-back is finished 
    WB_flag = 0;
*/
//--------------------------------------------------------------------
L_FETCH_LINE_end:    
    switch (FIL_action) {
      case FIL_READ: out.BUS_REQ = 0; goto L_FIL_READ_ret;
      case FIL_WRITE: goto L_FIL_WRITE_ret;
      case FIL_LOCK: goto L_FIL_LOCK_ret;
    }   
L_STATE_FETCH_LINE_RETRY:
    COUT << "--P" << "(FetchLineRetry) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter--) return;
    goto L_FETCH_INVALID_LINE;    
    
//---------------------------------------------------------------
// SCC - Snoopy Cache Controler
L_SNOOP:    
    // here we come on a snoop hit. There are two reasons why someone wants
    // us to do snooping:
    // 1) he is writing something into his line and wants us to invalidate ours
    // 2) he wants to read the line and asks us to do the supply, if we are
    //    the owner. In addition, he may be having a write miss: he will get
    //    a copy of the line from us, and then ask us to invalidate ours!
    //snoop_bus_INV = in.bus.INV;
    //bus_request = in.bus.memmcmd; // save bus cycle type
    // REQUEST_READ + INV = REQUEST-EX
    // REQUEST_WRITE + INV = INVALIDATE  
    if ((in.bus.memcmd = REQUEST_READ) && (in.bus.INV)) //it's read-ex
       if (cache_memory.getstate() == MESI_Line_State::MODIFIED) 
          { cache_memory.setstate(MESI_Line_State::INVALID);
            goto L_SNOOP_write_back;
            }
       else 
          goto L_SNOOP_invalidate;
    else if ((in.bus.memcmd == REQUEST_WRITE) && (in.bus.INV))
       { if (cache_memory.getstate() == MESI_Line_State::SHARED)
          goto L_SNOOP_invalidate; }
    else if ((in.bus.memcmd == REQUEST_READ) && (! in.bus.INV))
       switch (cache_memory.getstate()) {
          case MESI_Line_State::SHARED: 
                out.bus.SHARED = 1;
                break;
          case MESI_Line_State::EXCLUSIVE:
                out.bus.SHARED = 1;
                cache_memory.setstate(MESI_Line_State::SHARED);
                break;
          case MESI_Line_State::MODIFIED:
                cache_memory.setstate(MESI_Line_State::SHARED);
                goto L_SNOOP_write_back;
       }
    goto L_SNOOP_end;
L_SNOOP_write_back:
    out.bus.SHARED = 1;
    out.bus.DIRTY = 1;    
    // init the line for writing *to* someone (as in the write-back case)
    // If we had simulated data movement, the Topology would have read
    // the data from our output port and sent it to the bus, and the following
    // line would have looked like cache_memory.line_write_init(
    //                             &out.bus.addr, &out.bus.data);
    out.BUS_REQ = 1;
    snoop_state = SNOOP_WAIT_BGNT;
    return;
L_STATE_SNOOP_WAIT_BGNT:
    COUT << "--P" << index << "(SnoopWaitBgnt) " << Topology::getCurrentTime() << "\n";
    if (in.BUS_GNT) return;
    out.bus.reset();
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.memcmd = REQUEST_WRITE;
    out.bus.size = DATA_WIDTH;
    out.bus.SNOOP_REQ = 0;
    wait_state_counter = WRITE_WAIT_STATES;
    snoop_state = SNOOP_WR_WAIT_MEMORY;
L_STATE_SNOOP_WR_WAIT_MEMORY:
    COUT << "--P" << index << "(SnoopWrWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter--) return;    
    if (cache_memory.line_write_getnext(&out.bus.addr)) 
        { wait_state_counter = WRITE_WAIT_STATES_2;
          return; // return if there is more, it will be continued in the next cycle
          }
    out.BUS_REQ=0;
    goto L_SNOOP_end;
L_SNOOP_invalidate:
    // invalidate the line if the INV signal was high
    cache_memory.setstate(MESI_Line_State::INVALID);
    cache_memory.untouch(); // force LRU
    ++stat.inv;
L_SNOOP_end:    
    // the snoop job is done,set cache_memory's Current Frame to #0 for the PCC
    cache_memory.set_current_frame(0);
    snoop_state = SNOOP_INITIAL;
    return;   
    }

// debugging...

void MESI_Cache::dump_state() {
    COUT << "CC" << index << ": State: ";
    if (snoop_state != SNOOP_INITIAL) { 
        if (snoop_state == SNOOP_WAIT_BGNT) 
            COUT << "SNOOP_WAIT_BGNT";
        else if (snoop_state == SNOOP_WR_WAIT_MEMORY)
            COUT << "SNOOP_WR_WAIT_MEMORY";
        }
    else    {
        switch (state) {
            case INITIAL             : COUT << "INITIAL"; break;
            case FETCH_WAIT_BGNT     : COUT << "FETCH_WAIT_BGNT"; break;
            case FETCH_WAIT_MEMORY   : COUT << "FETCH_WAIT_MEMORY"; break;
            case FETCH_WAIT_SNOOP    : COUT << "FETCH_WAIT_SNOOP"; break;
            case FETCH_LINE_RETRY    : COUT << "FETCH_LINE_RETRY"; break;
            case LOCK_WAIT_BGNT      : COUT << "LOCK_WAIT_BGNT"; break;
            case LOCK_SLEEP          : COUT << "LOCK_SLEEP (" << busy_lock_sleep_cycle_count << ")"; break;
            case WB_WAIT_MEMORY      : COUT << "WB_WAIT_MEMORY (" << wait_state_counter << ")"; break;
            case UNLOCK_WAIT_BGNT    : COUT << "UNLOCK_WAIT_BGNT"; break;
            case INV_WAIT_END        : COUT << "INV_WAIT_END (" << wait_state_counter << ")"; break;
            case INV_WAIT_BGNT	     : COUT << "INV_WAIT_BGNT (" << wait_state_counter << ")"; break;
            }
        }

    COUT << "\t\t[";
    if (state != INITIAL) {
        switch (in.request.type) {      
            case REQUEST_READ  : COUT << "READ " << in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << in.request.addr; break;
            default : break;
            }
        }       
    COUT << "]\n";
    cache_memory.dump_state(index);
    }
                

#endif PRAGMA_IMPLEMENTATION_MESI_CACHE
#endif _MESI_CACHE_H_