#ifndef _CACHE_MEMORY_H_
#define _CACHE_MEMORY_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: cache_memory.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Defines the template for generic cache memory.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/*
 * This is a template for generic cache memory. All the snoopy controllers
 * use it. Everything in the module can be configured:
 * - line state: since this class is a template, the module who uses it
 *   should specify a class that will represent states of each line.
 *   Templates allow that cache memories with different types of line states
 *   coexist in the same simulation (for example, L1 cache may have one type
 *   of the cache memory, and L2 another).
 * - cache geometry: size (in Kbytes), associativity (whether it is 1-way, 
 *   2-way, 4-way, or N-way), and line size. These can be specified either
 *   by the module who instantiates this class, or upon startup, from the
 *   INI file or command line switches.
 *
 * Usage: for a class representing a controller and its associated cache
 * memory, you should first define Line_State class, which describes the
 * states a line can be in, and has the form as in the Berkeley example:
 *
 *  class Ber_Line_State {
 *    public:
 *      enum State { INVALID, UNMOD_SHD, MOD_SHD, MOD_EXC } state;
 *      Ber_Line_State()            { state = INVALID; }
 *      bool  isvalid()             { return state != INVALID; }
 *      State get()                 { return state; }
 *      void  set(State new_state)  { state = new_state; }
 *      char  symbol() {            // for debugging
 *          switch (state) {
 *              case INVALID  : return 'I';
 *              case UNMOD_SHD: return 'U';
 *              case MOD_SHD  : return 'M';
 *              case MOD_EXC  : return 'E';
 *              }
 *          }
 *      };
 *
 * After this class is defined, do 
 * #include "cache_memory.h"
 * 
 * Later, somewhere in your controller class declaration, say
 *  class Ber_Cache {
 *      ...
 *      Cache_Memory<Ber_Line_State> cache_memory;
 *      ...
 * Assume that all this was placed in "ber_cache.h". In the controller
 * implementation ("ber_cache.cc"), before you do #include "ber_cache.h",
 * you must 
 * #define PRAGMA_IMPLEMENTATION_CACHE_MEMORY
 * so that the template implementation becomes visible.
 *
 * In the constructor of your controller class, you must explicitly initialize
 * the cache memory by calling it's Initialize() method. Optionally, you can
 * supply values for the cache geometry. If you dont, the defaults will be used,
 * which can be overriden by INI settings. ("cache_size", "cache_line", 
 * "cache_way").
 *
 * Cache_Memory is intended to provide complete encapsulation of
 * cache memory properties, such as the number of lines/sets, cache line size,
 * associativity... Cache controller that uses it simply calls the isvalid()
 * method with the address of the word it wants to access as the argument,
 * and all the subsequent operations over the word or part of the word,
 * such as read and write, or over the line, such as reading the whole line
 * for writing it into memory or giving it to another cache, or setting the
 * state of the line, are done implicitly, based on the address supplied
 * at the time isvalid() is called. Please refer to the documentation for more
 * info.
 */ 

#include "snoopy.h"

// default cache geometry
const DEFAULT_CACHE_LINE_SIZE = 32;     // in *bytes*, not in words
const DEFAULT_SIZE_KB = 32;
const DEFAULT_WAY = 1;

template <class Line_State>
class Cache_Memory {
  protected:
    // cache geometry, adjustable at init time
    int lines, sets, entries_per_set;   // lines = sets * entries_per_set;
    int cache_line_size;                
    int cache_line_word_count;          // == cache_line_size / DATA_WIDTH
    // cache memory storage pointer
    struct Line {
        bitsADDR tag;
        Line_State state;
        unsigned time_stamp;
        Line() { tag = (bitsADDR)-1; time_stamp = 0; }
        } *line_matrix;
    Line & line(int row, int col) {return line_matrix[row*entries_per_set+col];}
    unsigned time_stamp;
    // speedup to avoid address splitup for each reference.
    const C_MAXFRAMES = 4; 
    struct Current_Word_Position {
        bitsADDR    tag;
        bool        valid;
        int         set;        // 0..sets-1
        int         line;       // 0..entries_per_set - 1 
        int         offset;     // 0..line_size-1
        } C, Cframe[C_MAXFRAMES];   // C == Cframe[Cframe_number]
    int Cframe_number;          // number of the current frame
    // counter for reading and writing a line upon fetch/writeback
    int counter;                // 0..line_size-1 by DATA_WIDTH
    // masks etc. are computed at init time and remain unchanged thereafter
    bitsADDR offset_mask, set_mask;
    int set_shift, tag_shift;
    void parse_addr(bitsADDR addr);
  public:
    void Initialize(                        // this must be explicitly called
        int size=DEFAULT_SIZE_KB, 
        int line_size=DEFAULT_CACHE_LINE_SIZE, 
        int way=DEFAULT_WAY
        );
    void Introduce();
    bool isvalid(bitsADDR addr);
    // the following operate over CURRENT line (C.*) formed by isvalid()
    Line_State::State getstate()
        { return line(C.set,C.line).state.get(); }
    void setstate(Line_State::State new_state)
        { line(C.set,C.line).state.set(new_state); }
    void touch()    { line(C.set,C.line).time_stamp = time_stamp; } // force MRU
    void untouch()  { line(C.set,C.line).time_stamp = 0;     }      // force LRU
    void set_current_frame(int framenumber);
    void push_current() {set_current_frame(1);}     // for compatibility
    void pop_current()  {set_current_frame(0);}     // with earlier simulators
    void dump_state(int index); // dumps the whole cache contents on stdout!
    void get_current_line(bitsADDR *addr);
    void line_fetch_init(bitsADDR *first_addr); 
    bool line_fetch_putnext(bitsADDR *next_addr);
    void line_write_init(bitsADDR *first_addr); 
    bool line_write_getnext(bitsADDR *next_addr);
    };


#ifdef PRAGMA_IMPLEMENTATION_CACHE_MEMORY
/////////////////////////////////////////////////////--- IMPLEMENTATION
#include <stdio.h>      // needed for the dump

// Initialize method must be called before the cache memory is used.

template <class Line_State>
void Cache_Memory<Line_State>::Initialize(int default_size_kb, 
    int default_line_size, int default_way)     {   

    // initialize cache geometry
    int size_kb     = sys_INI_read("cache_size",      default_size_kb);
    cache_line_size = sys_INI_read("cache_line_size", default_line_size);
    entries_per_set = sys_INI_read("cache_way",       default_way);
    lines = size_kb * 1024 / cache_line_size;
    sets  = lines / entries_per_set;
    cache_line_word_count = cache_line_size / DATA_WIDTH;

    // instantiate new lines; they will be automagically initialized 
    line_matrix = new Line[lines];                  // by their ctors

    // set initial time_stamp to be used for individual time_stamp vars
    time_stamp = 0;

    // precompute offset_mask, set_shift, set_mask, tag_shift
    // each addr is [<tag><set><offset within the set>]
    offset_mask = cache_line_size -1;
    int i;
    for( i=cache_line_size, set_shift = 0; i > 1; i=i>>1) ++set_shift;
    set_mask = sets -1; 
    for( i=sets, tag_shift = set_shift; i > 1; i=i>>1) ++tag_shift;
    Cframe_number = 0;
    }       

// print cache geometry to stdout

template <class Line_State>
void Cache_Memory<Line_State>::Introduce() {
    PRINTF("Cache memory: size: %d KB (%dx%d KB), line size: %d bytes (%d words)\n",
        cache_line_size * sets * entries_per_set / 1024,
        entries_per_set, cache_line_size * sets  / 1024,
        cache_line_size, cache_line_word_count);
    }

// based on the given address, locate the line in the cache memory: (internal)

template <class Line_State>
inline void Cache_Memory<Line_State>::parse_addr(bitsADDR addr) {
    C.offset = addr & offset_mask;
    C.set    = (addr >> set_shift) & set_mask;
    C.tag    = addr >> tag_shift;
    }

// cache controllers call the following function first before they do anything 
// with a line. For example: cache_memory.isvalid(some_addr) will return true
// or false, but will also locate the line that some_addr belongs to.
// Further methods will operate on that line, like 
// cache_memory.getstate() : it will return the state of the line,
// as defined by Line_State. The position of the line is remembered in 'C'
// struct ('C' stands for 'Current').
// If isvalid() determines that the line is invalid, or that there is no
// such line in the cache, it will pick a line suitable for replacement.
// Otherwise, it will 'touch' the line and make it MRU (most recently used).

template <class Line_State>
bool Cache_Memory<Line_State>::isvalid(bitsADDR addr) {
    // increment global time_stamp that will be used for LRU algorithm
    ++time_stamp;
    parse_addr(addr);
    int i; 
    for(i=0, C.valid=false; i < entries_per_set; ++i) 
        if (line(C.set,i).tag == C.tag && line(C.set,i).state.isvalid()) {
            C.valid = true;
            C.line = i; 
            line(C.set,i).time_stamp = time_stamp;  // touch
            break;
            }
    if (!C.valid) { // line miss. Sets C.* to the line eligible for removal
        unsigned min_time_stamp = line(C.set,0).time_stamp;
        C.line = 0;
        for(i=1; i < entries_per_set; ++i) 
            if (line(C.set,i).time_stamp  < min_time_stamp) {
                min_time_stamp = line(C.set,i).time_stamp;
                C.line = i;
                }
        }
    return C.valid;
    }   

// set_current_frame method is used for selecting the set of Current variables
// the cache memory will operate with. For example, your processor cache contro-
// ller part will always use frame#0, and your snoopy cache controller part will 
// always use frame#1. Why is this needed? Suppose the PCC has called isvalid() 
// and decided to wait for the bus. His current vars are set. Now, the SCC sees
// an invalidation request, and has to check whether the line exists in the
// cache. He has to call isvalid() as well. But if there existed only one set
// of Current variables, this isvalid() call would spoil those Current vars
// set by the PCC. So the snoopy cache controller always selects frame#1,
// and the processor cache controller always selects frame#0. This frames
// do not exist in the real world, they are only used to support encapsulation
// of the cache memory details from the cache controllers.

template <class Line_State>
void Cache_Memory<Line_State>::set_current_frame(int frame_number) {
    if (frame_number == Cframe_number) return;
    Cframe[Cframe_number] = C;
    Cframe_number = frame_number;
    C = Cframe[frame_number];
    }

// the next two pairs of methods are used for line fetching and writing back.
// Fetch is used for writing the new line to the cache, and Write is used
// for reading the line from the cache. :-) OK, the terms are confusing,
// but the names reflect the controller's point of view: 
// 
// when it wants to fetch the line and put it in the cache, it will do
// cache_memory.line_fetch_init(&addr); the addr will contain the address
// of the first word in the line, and it will be the address that the 
// controller will send on the bus, or give it to the L2 cache or whatever.
// For subsequent words, it will call 
// cache_memory.line_fetch_putnext(&addr)) and the method will put
// the next address (the address of the next word) in the addr and return
// false if there are no more words in the line. 
//
// Similarily, for sending a line from the cache memory to the outside
// (bus, or L2 cache) it will first call line_write_init, and then
// repeatedly call line_write_getnext, (in appropriate cycles), until
// it returns false. This hides the line size from the controller.
//
// Once again, the names seem (and possibly are) inappropriate, but they
// reflect the actions of the owner of this cache memory.
template <class Line_State>
void Cache_Memory<Line_State>::get_current_line(bitsADDR *addr) {
    *addr = (C.tag << tag_shift) | (C.set << set_shift);
    }


template <class Line_State>
void Cache_Memory<Line_State>::line_fetch_init(bitsADDR *first_addr) {
    counter = 0;
    *first_addr = (C.tag << tag_shift) | (C.set << set_shift);
    }

template <class Line_State>
void Cache_Memory<Line_State>::line_write_init(bitsADDR *first_addr) {
    counter = 0;
    *first_addr = (line(C.set,C.line).tag << tag_shift) | (C.set << set_shift);
    }
    
template <class Line_State>
bool Cache_Memory<Line_State>::line_fetch_putnext(bitsADDR *next_addr) {
    counter += DATA_WIDTH;
    *next_addr += DATA_WIDTH;
    if ( counter >= cache_line_size ) {
        line(C.set,C.line).tag = C.tag;
        touch();    // force MRU
        return 0;
        }
    else
        return 1;
    }

template <class Line_State>
bool Cache_Memory<Line_State>::line_write_getnext(bitsADDR *next_addr) {
    counter += DATA_WIDTH;
    if ( counter >= cache_line_size ) 
        return 0;
    else    {
        *next_addr += DATA_WIDTH;
        return 1;
        }
    }

// this final method is used for debugging.

template <class Line_State>
void Cache_Memory<Line_State>::dump_state(int index) {
    int iset, iline;
    /*
    for(iset=0; iset < sets; ++iset) {
        bool anyvalid = false;
        for(int i=0; i<entries_per_set; ++i) 
            if (line(iset,i).state.isvalid()) {anyvalid = true; break;}
        if (anyvalid) {
            PRINTF("  CM%d: S%3d:  ", index, iset);
            for(iline=0; iline<entries_per_set; ++iline) {
                PRINTF("[%c,%d]%x ", 
                    line(iset,iline).state.symbol(),
                    line(iset,iline).tag, 
                    line(iset,iline).time_stamp);
                if (iline==0) PRINTF("  ");
                }
            PRINTF("\n");
            }
        }
    */
    }


#endif PRAGMA_IMPLEMENTATION_CACHE_MEMORY
#endif _CACHE_MEMORY_H_
