#ifndef _MESI_CACHE_H_
#define _MESI_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: mesi_cache.h
 * Author: Aleksandar Milenkovic, emilenka@etf.bg.ac.yu
 * Note: Base Cache Controller is extended to support 
 *       prefetch, prefetch-ex, lazy prefetch and write-back
 *
 * The MESI protocol; definition and implementation.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "bus_signals.h"


// states a line in the MESI protocol can be found in
class MESI_Line_State {
  public:
    enum State { INVALID, SHARED, EXCLUSIVE, MODIFIED } state;
    MESI_Line_State()               {state = INVALID;}
    bool  isvalid()		    {return state != INVALID;}
    State get()                     {return state;}
    void  set(State new_state)      {state = new_state; }
    char  symbol() { // for debuging
    	             switch (state) {
    	               case INVALID:   return 'I';
    		       case SHARED:    return 'S';
    		       case EXCLUSIVE: return 'E';
    		       case MODIFIED:  return 'M' ;
    	             }
    		   }
    };

// the MESI cache controler module declaration
class MESI_Cache {
    int index;
    // instantiate Cache Memory with the MESI_Line_State as its Line_State
    Cache_Memory<MESI_Line_State> cache_memory;  // the whole cache memory
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        bool BUS_GNT;                   // bus grant (from the bus module)
        bool LOCK_GNT;                  // lock grant (from the bus module)
        Bus_Signals bus;                // bus contents, to be read from the bus
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        bool BUS_REQ;                   // bus request: sent to the bus module
        bool LOCK_REQ;                  // lock request: sent to the bus module
        Bus_Signals bus;                // bus contents, to be written on the bus
        } out;
    // Processor Cache Conntoller finite state 
    enum PCC_State {     // processor cache controller
        PCC_INITIAL,                        // INITIAL -- not doing anything (except maybe in the snoop part)
        BLOCKED,
        // lock (version 2) 
        LOCK_WAIT_BGNT,
        LOCK_SLEEP,
        // unlock
        UNLOCK_WAIT_BGNT,
        // wait for BCU (all this states could be changed with a state,
        // but it requests PCC_action parameters
        RM_BCU_WAIT, 		//Read-Miss wait for BCU (Bus Control Unit)
        WM_BCU_WAIT,		//Write-Miss wait for BCU
        WI_BCU_WAIT,		//Write_invalidate wait for BCU
        LOCK_BCU_WAIT,
        UNLOCK_BCU_WAIT,
        PREF_BCU_WAIT,		//Prefetch wait for BCU
        PREFEX_BCU_WAIT,
        WRBACK_BCU_WAIT
       } pcc_state;

    // Bus Control Unit finite states
    enum BCU_State {     
        BCU_INITIAL,                        // INITIAL -- not doing anything (except maybe in the snoop part)
        // fetch line caused by RD, WR, LOCK, PREF, PREF-EX
        FETCH_WAIT_BGNT,                // waiting for the bus to do a read
        FETCH_WAIT_SNOOP,                  // waiting one cycle to let the others see the SNOOP signal before we withdraw it from the bus
        FETCH_LINE_RETRY,
        FETCH_WAIT_MEMORY,
        // write back modified line
        WB_WAIT_MEMORY,
        // write-back instruction
        WRBACK_WAIT_BGNT,
        // invalidate caused by write hit in shared cache line
        INV_WAIT_BGNT,
        INV_WAIT_END,
       } bcu_state;
    
    // the SCC (snoop cache controller part) may also be in one of three states
    enum Snoop_State {
        SNOOP_INITIAL,                  // not doing a snoop
        SNOOP_WAIT_BGNT,
        SNOOP_WR_WAIT_MEMORY
        } snoop_state;
   
         
    // One 'procedure' is called to fetch an invalid line; it can be called
    // upon a read, a write, or a lock. This is sort of replacement for gosub-return sequence
    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK, FIL_PREF, FIL_PREFEX} FIL_action;
    // some internal vars
    enum WB_ACTIONS {NOWB, WB_REPL, WB_INSTR} WB_action;
    bool WB_flag; 	// write-back flag
    //bool WB_instr;      // true for write-back instructions
    bitsADDR wb_addr; 	// write-back address
    bitsADDR tmp_addr;  // to show frame(2) state
    enum INV_ACTIONS {INV_WR, INV_PREFEX} INV_action;
    //bool PREFEX_inv_flag;  //ivalidate cycle initiated by PREFEX
    struct BCU_REQUEST {
      bitsADDR addr;
      int size;
    } bcu_request;
    int wait_state_counter;
    int snoop_wait_state_counter;
    bool shared_line;
    bool dirty_line;
    // delay parameters of the simulated physical memory: the default values
    const DEFAULT_RD_WAIT_STATES     = 5; // time needed to read/read_ex the first word
    const DEFAULT_RD_WAIT_STATES_2   = 0; // time needed for access to subsequent words
    const DEFAULT_WR_WAIT_STATES     = 5;
    const DEFAULT_WR_WAIT_STATES_2   = 0;
    const DEFAULT_INV_WAIT_STATES    = 4;
    // and the real values, which can be changed at startup
    int READ_WAIT_STATES;               // these will become equal to the
    int READ_WAIT_STATES_2;             // defaults if not otherwise specified
    int WRITE_WAIT_STATES;               // in the INI file or command line
    int WRITE_WAIT_STATES_2;             // switches
    int INV_WAIT_STATES;
    const BUSY_LOCK_SLEEP_CYCLE_COUNT   = 1; 
    const FETCH_RETRY_WAIT_STATES 	= 2;
    bitsLOCK lock_value;
    int busy_lock_sleep_cycle_count;
    struct Stat {       // statistics:
        int req;        // requests
        int inv;        // invalidations
        int rep;        // replacements
        int read_hit;   // read hits
        int read_miss;  // read misses
        int dread_hit;  // delayed read hits
        int write_hit;  // write hits
        int write_miss; // write misses
        int dwrite_hit;  // delayed write hits
        int write_inv;   // write invalidate
        int write_dinv;  // write invalidate discarded
        int lock;	 // locks
        int unlock;	 // unlocks
        int pref_hit;    // prefetch hits
        int pref_miss;   // prefetch misses
        int dpref_hit;   // delayed prefetch hits
        int prefex_hit;  // prefetch-ex hits
        int prefex_missf;   	// prefetch-ex fetch misses
        int prefex_missi;  	// prefetch-ex invalidate misses
        int dprefex_hit; 	// delayed prefetch-ex hits
        int wrback_hitm;  	// write-back modified mits
        int wrback_hitse;    	// write-back shared or exclusive hits
        int wrback_miss;	// write-back miss
        int dwrback_miss;	// delayed write back
        Stat()  { req=inv=rep=read_hit=read_miss=dread_hit=0;
                  write_hit=write_miss=dwrite_hit=write_inv=write_dinv=0;
                  lock=unlock=pref_hit=pref_miss=dpref_hit=0;
                  prefex_hit=prefex_missf=prefex_missi=dprefex_hit=0;
                  wrback_hitm=wrback_hitse=wrback_miss=dwrback_miss=0;}
        void dump(int index);
        } stat;
  public:
    MESI_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {
        return ((pcc_state == PCC_INITIAL) && (bcu_state == BCU_INITIAL));
    }
    void dump_state();
    void dump_stats() {stat.dump(index);}
    TOPOLOGY_FRIENDS
    };


#ifdef PRAGMA_IMPLEMENTATION_MESI_CACHE

// the MESI cache controller module implementation

#include <iostream.h>
    
MESI_Cache::MESI_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    pcc_state = PCC_INITIAL;                            // PCC will be in the INITIAL state
    snoop_state = SNOOP_INITIAL;                // and so will the SCC part
    bcu_state = BCU_INITIAL;
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    in.BUS_GNT      = 0;
    out.BUS_REQ     = 0;
    in.LOCK_GNT     = 0;
    out.LOCK_REQ    = 0;
    in.bus.reset();
    out.bus.reset();
    // initialize memory delay parameters. If you want to change this,
    // create limes.ini in your app's directory and place in the line
    // r_ws=4 (for example), or use simulator's command line switches,
    // like -dw_ws=4.
    READ_WAIT_STATES    = sys_INI_read("r_ws", DEFAULT_RD_WAIT_STATES);
    READ_WAIT_STATES_2  = sys_INI_read("r_ws2", DEFAULT_RD_WAIT_STATES_2);
    WRITE_WAIT_STATES   = sys_INI_read("w_ws", DEFAULT_WR_WAIT_STATES);
    WRITE_WAIT_STATES_2 = sys_INI_read("w_ws2", DEFAULT_WR_WAIT_STATES_2);
    INV_WAIT_STATES = sys_INI_read("i_ws", DEFAULT_INV_WAIT_STATES);
    };

void MESI_Cache::Introduce() {
    COUT << "Protocol: MESI (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << "LOCK_line aware.";
    #endif
    COUT << "\n";
    COUT << "READ_WS=" << READ_WAIT_STATES << "/" << READ_WAIT_STATES_2 << ", ";
    COUT << "WRITE_WS=" << WRITE_WAIT_STATES << "/" << WRITE_WAIT_STATES_2 << ", ";
    COUT << "LOCK_WS=" << BUSY_LOCK_SLEEP_CYCLE_COUNT << "\n";
    #ifdef DUMMLOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    }

void MESI_Cache::Stat::dump(int index) {
    if (req) COUT << "MESI"<<index<<": " << \
        "requests="<<req << ", invalidations="<<inv << ", replacements="<<rep
        << "\n       rh=" << read_hit << ", rm=" << read_miss 
        << ", drh=" << dread_hit 
        << "\n       wh=" << write_hit << ", wm=" << write_miss << ", dwh=" << dwrite_hit 
        << ", wi=" << write_inv << ", wid=" << write_dinv 
        << "\n       lock=" << lock << ", unlock=" << unlock 
        << "\n       prh=" << pref_hit << ", prm=" << pref_miss 
        << ", dprh=" << dpref_hit 
        << "\n       pxh=" << prefex_hit << ", pxmf=" << prefex_missf 
        << ", pxmi=" << prefex_missi << ", pxdh=" << dprefex_hit 
        << "\n       wbhm=" << wrback_hitm << ", wbhse=" << wrback_hitse 
        << ", wbm=" << wrback_miss << ", dwbm=" << dwrback_miss << "\n";
        
    }
/*
* cycle method
*/
// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
// This method first reads the bus. Then it checks if there is SNOOP signal
// high on it, and, if so, checks if it has the line to invalidate. If true,
// does so and skips this cycle. This reflects the fact that there are two
// controllers in this one, the Processor Cache Controller (PCC) and the
// Snoop Cache Controller (SCC), each of which has its own copy of cache
// tags, but the cache memory itself can be accessed only by one controller
// at a time, and the SCC part has precedence.
    
void MESI_Cache::cycle() {
    
    Topology::cache_read_bus(index);        // read input from the bus

    // check if there is SNOOP signal on the bus, but issued by a cache
    // other than myself:
    switch (snoop_state) {
        // the SCC may be in initial state, or not. If it is in initial state,
        // it will check if a SNOOP signal appeared on the bus. Otherwise,
        // it will continue what it was doing in the previous cycle -- doing
        // a cache supply for some other cache controller.
        case SNOOP_INITIAL:
            if (in.bus.SNOOP_REQ && !in.BUS_GNT) { // !in.BUS_GNT means 'I don't have the bus'
                // there is: select frame #1 of cache_memory's C (current) vars,
                // which keep the current position of the line
                cache_memory.set_current_frame(1);  
                if (cache_memory.isvalid(in.bus.addr))
                    goto L_SNOOP;   // most of the SCC part is there
                else
                    cache_memory.set_current_frame(0);  // frame #0 for the processor part
                    // it was a snoop miss, so proceed as usual (though it might be more realistic to return)
                }
            break;          // activate processor cache controller
        case SNOOP_WAIT_BGNT		: goto L_STATE_SNOOP_WAIT_BGNT;
        case SNOOP_WR_WAIT_MEMORY	: goto L_STATE_SNOOP_WR_WAIT_MEMORY;
        }
    
//-------------------------------------------------------------------------
//  PCC
//-------------------------------------------------------------------------
    // ok, there was no snoop or it was a miss, so we continue
    cache_memory.set_current_frame(0);    
    switch (pcc_state) {
        case PCC_INITIAL                : goto L_STATE_PCC_INITIAL;
        case BLOCKED      		: goto L_BCU_start;
        case LOCK_WAIT_BGNT		: goto L_STATE_LOCK_WAIT_BGNT;
        case LOCK_SLEEP			: goto L_STATE_LOCK_SLEEP;
        case UNLOCK_WAIT_BGNT		: goto L_STATE_UNLOCK_WAIT_BGNT;
        case RM_BCU_WAIT 		: goto L_STATE_RM_BCU_WAIT;
        case WM_BCU_WAIT		: goto L_STATE_WM_BCU_WAIT;
        case WI_BCU_WAIT		: goto L_STATE_WI_BCU_WAIT;
        case LOCK_BCU_WAIT		: goto L_STATE_LOCK_BCU_WAIT;
        case UNLOCK_BCU_WAIT            : goto L_STATE_UNLOCK_BCU_WAIT;
        case PREF_BCU_WAIT		: goto L_STATE_PREF_BCU_WAIT;
        case PREFEX_BCU_WAIT            : goto L_STATE_PREFEX_BCU_WAIT; 
        case WRBACK_BCU_WAIT		: goto L_STATE_WRBACK_BCU_WAIT;
        }
        
L_STATE_PCC_INITIAL:
    //COUT << "--P" << index << "(initial) " << Topology::getCurrentTime() << "\n";
    if (! Topology::cache_read_request(index)) goto L_BCU_start;  // return if there was none
    ++stat.req;                                         // statistics
    switch (in.request.type) {                          // select where to go
        case REQUEST_READ   : goto L_READ;
        case REQUEST_LOCK   : stat.lock++;
        		      goto L_LOCK;
        case REQUEST_UNLOCK : goto L_UNLOCK;
        case REQUEST_WRITE  : goto L_WRITE;
        case REQUEST_PREF   : goto L_PREF;
        case REQUEST_PREFEX : goto L_PREFEX;
        case REQUEST_LPREF  : goto L_LPREF;
        case REQUEST_WRBACK : goto L_WRBACK;
        default : break;
        }
    goto L_BCU_start;

//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    if (cache_memory.isvalid(in.request.addr)) {
        // a hit, excellent! Say that the request is satisfied, and return
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        stat.read_hit++;
        goto L_READ_end;
        }
    //COUT << "P" << index << " RM: addr=" << hex << in.request.addr << "\n";
    // if BCU is free 
    if (bcu_state == BCU_INITIAL) goto L_INIT_READ_MISS;
    // BCU is busy; wait for BCU 
    pcc_state = RM_BCU_WAIT;
    goto L_BCU_start;
L_STATE_RM_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;  
    // if BCU unit was busy, it may be a consequence of WM at the this cache line  
    // check if it is the case
    if (cache_memory.isvalid(in.request.addr)) {
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        stat.dread_hit++;
        goto L_READ_end;
        }     
L_INIT_READ_MISS:
    stat.read_miss++;
    //initialize frame(2)
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);     
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = BLOCKED;
    FIL_action = FIL_READ;      // these three lines are an equivalent for
    goto L_FETCH_INVALID_LINE;  // "gosub fetch_invalid_line"; but there can be
                               // no gosub when F.I.L. spans more than one cycle
L_READ_end:
    pcc_state = PCC_INITIAL;
    goto L_BCU_start;

//---------------------------------------------------------------
L_WRITE:
    // this is a WRITE. We'll say immediately that it is satisfied.
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    // and check whether it is in the cache, and if so, is it valid.
    // The isvalid() method will do all the work for us, and even pick
    // a line for replacement if necessary.
    if (cache_memory.isvalid(in.request.addr)){ 
        stat.write_hit++;
        goto L_WRITE_valid;}
    //COUT << "P" << index << " WM: addr=" << hex << in.request.addr << "\n"; 
    if (bcu_state == BCU_INITIAL) goto L_INIT_WRITE_MISS;
    pcc_state = WM_BCU_WAIT;
    goto L_BCU_start;    
L_STATE_WM_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
    // if the the requested block was fetched during the previous bus cycle
    if (cache_memory.isvalid(in.request.addr)){ 
        stat.dwrite_hit++;
        goto L_WRITE_valid;}    
L_INIT_WRITE_MISS:     
    stat.write_miss++;
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);  // force init of frame 2
    FIL_action = FIL_WRITE;         // gosub L_FETCH_INVALID_LINE
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = PCC_INITIAL;
    goto L_FETCH_INVALID_LINE;
    goto L_WRITE_end;               // that's all there is for a WRITE
L_WRITE_valid:
    // in the MESI protocol, write actions depend on the state of the line
    // if the line is already modified, and exclusive, no further actions
    // are necessary
    if (cache_memory.getstate() == MESI_Line_State::MODIFIED) {
        goto L_WRITE_end; }      
    else if (cache_memory.getstate() == MESI_Line_State::EXCLUSIVE)
        { cache_memory.setstate(MESI_Line_State::MODIFIED);
          goto L_WRITE_end;}   
    //COUT << "P"<< index << " INV addr:" << hex << in.request.addr << "\n";
    if (bcu_state == BCU_INITIAL) goto L_INIT_WRITE_INV;
    pcc_state = WI_BCU_WAIT;
    goto L_BCU_start;    
L_STATE_WI_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
    // check for delayed invalidate
    if (cache_memory.getstate() == MESI_Line_State::MODIFIED) {
        stat.write_dinv++;
        goto L_WRITE_end;} 
    else if (cache_memory.getstate() == MESI_Line_State::EXCLUSIVE)
        { cache_memory.setstate(MESI_Line_State::MODIFIED);
          stat.write_dinv++;
          goto L_WRITE_end;}   
L_INIT_WRITE_INV:     
    stat.write_inv++;
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);  // force init of frame 2    
    INV_action = INV_WR;
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = PCC_INITIAL;	    // PCC is ready for new requests
    goto L_INV_start;                // start invalidate bus cycle
L_WRITE_end:                                // the write is over:
    if (out.LOCK_REQ) out.LOCK_REQ = 0;
    pcc_state = PCC_INITIAL;                        // and go to the INITIAL state
    goto L_BCU_start;
//---------------------------------------------------------------
L_LOCK:
    //stat.lock++;
    // wait until bcu become free
    if (bcu_state == BCU_INITIAL) goto L_LOCK_INIT;
    pcc_state = LOCK_BCU_WAIT;
    goto L_BCU_start;
L_STATE_LOCK_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
L_LOCK_INIT:     
    // LOCKing uses bus as a hardware binary semafore. There are two possible
    // schemes: the first calls the bus arbitrator, thereby blocking the bus,
    // and the second, more efficient, calls the special lock arbitrator,
    // thus leaving the bus free for other operations. The actual scheme
    // is selected at compile time (check out snoopy.h):
    #ifdef LOCK_LINE_AWARE
        out.LOCK_REQ = 1;       // raise LOCK request high (the second scheme)
    #else
        out.BUS_REQ = 1;        // raise BUS request high (the first scheme)
    #endif
    // we are waiting for the bus to tell us we can proceed
    pcc_state = LOCK_WAIT_BGNT;
L_STATE_LOCK_WAIT_BGNT:
    //COUT << "--P" << index << "(LockWaitBgnt) " << Topology::getCurrentTime() << "\n";
    #ifdef LOCK_LINE_AWARE
        if (! out.LOCK_REQ) out.LOCK_REQ = 1;
        if (! in.LOCK_GNT) return;
    #else
        if (! out.BUS_REQ) out.BUS_REQ = 1;
        if (! in.BUS_GNT) return;
    #endif
    // now check if the line is in the cache (and whether it is valid):
    if (cache_memory.isvalid(in.request.addr)) 
        goto L_LOCK_check_value;                // yes, it is.
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);
    FIL_action = FIL_LOCK;                      // no, do
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = BLOCKED;
    goto L_FETCH_INVALID_LINE;                  // gosub F.I.L.
    L_FIL_LOCK_ret:
L_LOCK_check_value:
    // the memsim has to actually read the lock from the real memory, since
    // cache keeps only tags, not memory contents. The real memory is
    // actually totally consistent (PRAM), everything here is about modelling
    // delays. This action would not have been necessary
    // if we had really moved the data from the memory to the simulated 
    // cache and vice versa, but this is an extreme effort (though this
    // controller was originally written that way and crippled later).
    lock_value = MEMSIM_READ_LOCK(in.request.addr);
    //COUT << "before dummlock: lock_value=" << lock_value << "\n";
    #ifndef DUMMLOCK
        // non-dumm lock does test&test&set and spins it self
        if (lock_value == 0) {
            // Do the actual "write 1" to the location for the lock.
            MEMSIM_WRITE_LOCK( in.request.addr, 1);
            lock_value = MEMSIM_READ_LOCK(in.request.addr);
            goto L_WRITE; 
            }
        out.BUS_REQ = out.LOCK_REQ = 0;
        busy_lock_sleep_cycle_count = BUSY_LOCK_SLEEP_CYCLE_COUNT;
        pcc_state = LOCK_SLEEP;
        return;
    #else
        // a dumm one does test&set and let the kernel spin for him
        MEMSIM_WRITE_LOCK( in.request.addr, 1); 
        out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_WRITE_valid;
    #endif  
L_STATE_LOCK_SLEEP:
    if (busy_lock_sleep_cycle_count--) return;
    goto L_LOCK;

//---------------------------------------------------------------
L_UNLOCK:
    stat.unlock++;
    if (bcu_state == BCU_INITIAL) goto L_UNLOCK_INIT;
    pcc_state = UNLOCK_BCU_WAIT;
    goto L_BCU_start;
L_STATE_UNLOCK_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
    // this is an UNLOCK (lock release) request. Basically it is a write,
    // only that it has to gain the bus first.
L_UNLOCK_INIT:
    out.LOCK_REQ = 1;
    pcc_state = UNLOCK_WAIT_BGNT;
L_STATE_UNLOCK_WAIT_BGNT:     
    #ifdef LOCK_LINE_AWARE
        if (!in.LOCK_GNT) return;
    #endif
    // 
    // do the actual "write 0" for the lock. Note: it would be a mistake
    // to do that BEFORE this point! Actually, it should be done somewhere
    // at L_WRITE_end -- that is where (when) it would happen in reality
    //
    MEMSIM_WRITE_LOCK( in.request.addr, 0);
    goto L_WRITE;
    
//---------------------------------------------------------------
// this part of the code covers both PREF and PREFEX
L_PREF:
    //COUT << "P" << index << "addr=" << hex << in.request.addr << "\n";
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    if(cache_memory.isvalid(in.request.addr)) {
      //eventualno osvezi tagove
      stat.pref_hit++;
      goto L_PREF_end;
    }
    if (bcu_state == BCU_INITIAL) goto L_PREF_INIT;
    pcc_state = PREF_BCU_WAIT;
    goto L_BCU_start;
L_STATE_PREF_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
    // delayed pref-hit
    if(cache_memory.isvalid(in.request.addr)) {
      //eventualno osvezi tagove
      stat.dpref_hit++;
      goto L_PREF_end;
    }    
L_PREF_INIT:
    stat.pref_miss++;
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);
    FIL_action = FIL_PREF;
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = PCC_INITIAL;
    goto L_FETCH_INVALID_LINE;
L_PREF_end:
    pcc_state = PCC_INITIAL;
    goto L_BCU_start;    
//---------------------------------------------------------------
L_PREFEX:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    if(cache_memory.isvalid(in.request.addr)) {
      switch(cache_memory.getstate()) {
         case MESI_Line_State::INVALID:  
         case MESI_Line_State::SHARED: break;
         case MESI_Line_State::EXCLUSIVE:
         case MESI_Line_State::MODIFIED:
             stat.prefex_hit++;
             goto L_PREFEX_end;
      }
    }
    if (bcu_state == BCU_INITIAL) goto L_PREFEX_INIT;
    pcc_state = PREFEX_BCU_WAIT;
    goto L_BCU_start;
L_STATE_PREFEX_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
    // delayed pref-ex
    if(cache_memory.isvalid(in.request.addr)) {
      switch(cache_memory.getstate()) {
         case MESI_Line_State::INVALID:  
         case MESI_Line_State::SHARED: break;
         case MESI_Line_State::EXCLUSIVE:
         case MESI_Line_State::MODIFIED:
             stat.dprefex_hit++;
             goto L_PREFEX_end;
      }
    }
L_PREFEX_INIT:
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = PCC_INITIAL;
    if (cache_memory.getstate()==MESI_Line_State::INVALID)
       { 
        stat.prefex_missf++;
        FIL_action = FIL_PREFEX;
        goto L_FETCH_INVALID_LINE;
       }
    else 
       {
        stat.prefex_missi++;
        INV_action = INV_PREFEX;
        goto L_INV_start;
       }
L_PREFEX_end:
    pcc_state = PCC_INITIAL;
    goto L_BCU_start;
//----------------------------------------------------------------
L_LPREF:
//----------------------------------------------------------------
L_WRBACK:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    if (cache_memory.isvalid(in.request.addr)) {
       if (cache_memory.getstate() == MESI_Line_State::MODIFIED) 
           goto L_WRBACK_valid;
       else 
           stat.wrback_hitse++;
       }
    else 
      stat.wrback_miss++;
    pcc_state = PCC_INITIAL;
    goto L_BCU_start;
L_WRBACK_valid:
    if (bcu_state == BCU_INITIAL) goto L_WRBACK_INIT;
    pcc_state = WRBACK_BCU_WAIT;
    goto L_BCU_start;
L_STATE_WRBACK_BCU_WAIT:
    if (bcu_state != BCU_INITIAL) goto L_BCU_start;
    // delayed wrback-miss
    if (!cache_memory.isvalid(in.request.addr) ||
          (cache_memory.isvalid(in.request.addr) && 
          (cache_memory.getstate() != MESI_Line_State::MODIFIED))) { 
         stat.dwrback_miss++;
         pcc_state = PCC_INITIAL;
         goto L_BCU_start;
    }
L_WRBACK_INIT:    
    stat.wrback_hitm++;
    cache_memory.set_current_frame(2);
    cache_memory.isvalid(in.request.addr);
    WB_action = WB_INSTR;
    bcu_request.addr = in.request.addr;
    bcu_request.size = in.request.size;
    pcc_state = PCC_INITIAL;
    goto L_WB_start;
//-------------------end of PCC----------------------------------------

L_BCU_start:
    cache_memory.set_current_frame(2);
    switch (bcu_state) {
        case BCU_INITIAL                    : return;
        case FETCH_WAIT_BGNT		: goto L_STATE_FETCH_WAIT_BGNT;
        case FETCH_WAIT_SNOOP		: goto L_STATE_FETCH_WAIT_SNOOP;
        case FETCH_LINE_RETRY		: goto L_STATE_FETCH_LINE_RETRY;
        case WB_WAIT_MEMORY		: goto L_STATE_WB_WAIT_MEMORY;
        case INV_WAIT_BGNT		: goto L_STATE_INV_WAIT_BGNT;
        case INV_WAIT_END		: goto L_STATE_INV_WAIT_END;
        case FETCH_WAIT_MEMORY		: goto L_STATE_FETCH_WAIT_MEMORY;
        case WRBACK_WAIT_BGNT	        : goto L_STATE_WRBACK_WAIT_BGNT;
        }
               
//------------------------------------------------------------------------
// 	INValidate bys cycle
//------------------------------------------------------------------------
L_INV_start:
    cache_memory.set_current_frame(2);
    out.BUS_REQ = 1;                // I want the bus, please
    bcu_state = INV_WAIT_BGNT;        // and I'm waiting for it
    // the fact that there is no "return" here means that BUS_GNT will
    // be waited for only if not high already (it could be high if we are
    // doing a WRITE for a LOCK, which may have gained bus before jumping here)
L_STATE_INV_WAIT_BGNT:
    //COUT << "--P" << index << "(InvWaitBgnt) " << Topology::getCurrentTime() << "\n";
    if (! out.BUS_REQ) out.BUS_REQ = 1;
    if (! in.BUS_GNT) return;       // wait as long as it takes
    // Now, we've got the bus; but the line may have gotten invalidated 
    // meanwhile by our snoopy cache controller part!
    if (cache_memory.getstate() == MESI_Line_State::INVALID) {
        //out.BUS_REQ = 0;
        //bcu_state = BCU_INITIAL; 
        if (INV_action == INV_WR) {  
           FIL_action = FIL_WRITE;
           goto L_FETCH_INVALID_LINE;
           }        
        else if (INV_action == INV_PREFEX) {
           FIL_action = FIL_PREFEX;
           goto L_FETCH_INVALID_LINE;
          }
    }
    // OK, the line is still valid. Now we'll declare it exclusively ours.
    out.bus.reset();
    out.bus.memcmd   = REQUEST_WRITE;
    cache_memory.line_fetch_init(&out.bus.addr); // the same as out.bus.addr = bcu_request.addr
    out.bus.size     = in.request.size;     // not really needed
    out.bus.SNOOP_REQ    = 1;               // SNOOP_REQ goes high
    out.bus.INV      = 1;                   // INV too
    wait_state_counter = INV_WAIT_STATES;
    bcu_state = INV_WAIT_END;                   // and we'll wait one cycle to let
    return;                                 // the others see these signals
L_STATE_INV_WAIT_END:
    //COUT << "--P" << index << "(InvWaitEnd) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter-- >0) return;
    out.BUS_REQ = 0;
    out.bus.INV = 0;
    // nemam objasnjenje
    out.LOCK_REQ = 0;
    if (INV_action == INV_WR) {
       cache_memory.setstate(MESI_Line_State::MODIFIED); // modified
       //COUT << "P"<< index << "line becames MOD\n";
    }
    else if (INV_action == INV_PREFEX)
       cache_memory.setstate(MESI_Line_State::EXCLUSIVE);
    
    // na kraju svakog ret
    bcu_state = BCU_INITIAL;
    return;
   
    /*
    *if (PREFEX_inv_flag)
    *  {
    *   out.BUS_REQ = 0;
    *   PREFEX_inv_flag = 0;
    *   out.bus.INV = 0;
    *   cache_memory.setstate(MESI_Line_State::EXCLUSIVE);
    *   goto L_FIL_PREFEX_ret;
    *  }
    *  //ubaciti povratak negde L_WRITE_VALID
    */
//--------------------------------------------------------------------------
//  Fetch Invalid Line
//--------------------------------------------------------------------------
L_FETCH_INVALID_LINE:
    //cache_memory.set_current_frame(2);
    // This is a "procedure" for fetching an invalid line. cache_memory
    // already points to a line suitable for replacement. We'll check
    // its state and see if we have to do a write-back.
    out.BUS_REQ = 1;                // bus is what we need first
    bcu_state = FETCH_WAIT_BGNT;
L_STATE_FETCH_WAIT_BGNT:            // we'll spin here, cycle after cycle,
    //COUT << "--P" << index << "(FetchWaitBgnt) " << Topology::getCurrentTime() << "\n";
    //raise BUS_REQ if it is disabled; 
    // it solves problem when PCC is interrupted by SCC 
    if (!out.BUS_REQ) out.BUS_REQ = 1;
    if (! in.BUS_GNT) return;       // until the bus arbitrator sends BUS_GNT high
    switch (cache_memory.getstate()) {
        case MESI_Line_State::INVALID:
        case MESI_Line_State::SHARED:
        case MESI_Line_State::EXCLUSIVE: 
            { WB_flag =0;  break;}
        case MESI_Line_State::MODIFIED:   // if the line is MODIFIED, it has to be written back
            { WB_flag = 1;
              WB_action = WB_REPL; 
              cache_memory.line_write_init(&wb_addr);
              break;
            }
     }
    ++stat.rep;
    // save address of the cache line which must be replaced
    // sta treba ovde uraditi    
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP_REQ   = 1;
    if ((FIL_action == FIL_WRITE) || (FIL_action == FIL_PREFEX))
             out.bus.INV = 1; //for read-exclusive bus cycle
    out.bus.memcmd   = REQUEST_READ;
    bcu_state = FETCH_WAIT_SNOOP;
    wait_state_counter = 1;
    return;
L_STATE_FETCH_WAIT_SNOOP:
    //COUT << "--P" << index << "(FetchWaitSnoop) " << Topology::getCurrentTime() << "\n";
    out.bus.SNOOP_REQ = 0; // ?zasto se ovde ukida odmah SNOOP  
    if (wait_state_counter--) return;
    // now, we have effectively waited two cycles. Why? The first one was
    // to let the other controllers see our signals. The second was to let
    // us see their responses, if any.
    
    // check shared and dirty lines
    shared_line = in.bus.SHARED;
    dirty_line = in.bus.DIRTY;
    if (shared_line && dirty_line) 
       { out.BUS_REQ = 0; out.LOCK_REQ = 0;
         out.bus.INV =0;
         wait_state_counter = FETCH_RETRY_WAIT_STATES;
         bcu_state = FETCH_LINE_RETRY;
         goto L_STATE_FETCH_LINE_RETRY;
        }
    wait_state_counter = READ_WAIT_STATES;
    bcu_state = FETCH_WAIT_MEMORY;            
L_STATE_FETCH_WAIT_MEMORY:
    //COUT << "--P" << index << "(FetchWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter-- >0 ) return;
    if (out.bus.addr <= bcu_request.addr && 
        bcu_request.addr+bcu_request.size <= out.bus.addr + DATA_WIDTH) {
        // here we come if data for read/writing that made us do all this
        // belongs to this word
        switch (FIL_action) {
            case FIL_READ:
                out.response.signal = PROCESSOR_SATISFIED;
                out.response.STROBE = 1;
                pcc_state = PCC_INITIAL; 
                break;
            case FIL_LOCK:
                break;
            case FIL_WRITE:
                break;
            case FIL_PREF:
            case FIL_PREFEX:
                break;    
            }               
        }
    // "write" the data, get next addr and find out if there's more to do
    if (cache_memory.line_fetch_putnext(&out.bus.addr))
       { wait_state_counter = READ_WAIT_STATES_2; 
         return;
         }
    cache_memory.touch();   // force this line MRU (most recently used)
    // counting finished, the whole line fetched
    //out.bus.SUPPLY = 0;     // if it was a cache supply, it's over now
    // now we have to set the state for the new line and to return.
    switch( FIL_action ) {
        case FIL_READ:
        case FIL_PREF:
            if (shared_line) 
               cache_memory.setstate(MESI_Line_State::SHARED);
            else 
               cache_memory.setstate(MESI_Line_State::EXCLUSIVE);
            break;
        case FIL_PREFEX:
            cache_memory.setstate(MESI_Line_State::EXCLUSIVE);        
            break;
        case FIL_WRITE:
            cache_memory.setstate(MESI_Line_State::MODIFIED);
            break;
        case FIL_LOCK: // lock strategy
            cache_memory.setstate(MESI_Line_State::SHARED); //????
            break;
        }
    // check write-back flag
    if (! WB_flag) {
       goto L_FETCH_LINE_end;
    }
    // Ok, write-back will be performed    
    out.bus.reset();
    //COUT << "after reset: BUS_GNT:" << in.BUS_GNT << "\n";
    // initialize the line for write-back. The cache memory will return
    // the address of the first word in the line, and that's what will
    // go to the bus.
    // this is dummy line_write_init method call (aimed to reset counter)
    cache_memory.line_write_init(&out.bus.addr);
    // write back address is wb_addr
    out.bus.addr = wb_addr;
    goto L_WRITE_BACK_start;
//--------------------------------------------------------------------
// Write-Back bus cycle
//---------------------------------------------------------------------
L_WB_start:
    if (WB_action == WB_INSTR) {
        //out.bus.reset();
        out.BUS_REQ = 1;
    }
    bcu_state = WRBACK_WAIT_BGNT;
L_STATE_WRBACK_WAIT_BGNT:
    if (!out.BUS_REQ) out.BUS_REQ = 1;
    if (!in.BUS_GNT) return;
    cache_memory.line_write_init(&out.bus.addr);
    goto L_WRITE_BACK_start;
L_WRITE_BACK_start:    
    //if (!in.BUS_GNT) COUT << "Uzas!!!! Izlazak na bus bez dozvole! \n";
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP_REQ  = 0;
    wait_state_counter = WRITE_WAIT_STATES;
    bcu_state = WB_WAIT_MEMORY;
    return;  // da li ovde treba staviti return
L_STATE_WB_WAIT_MEMORY:
    //COUT << "--P" << index << "(WbWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
        }
    // WRBACK instruction is finished 
    if (WB_action == WB_INSTR) {
       WB_action = NOWB; 
       // WRBACK is issued only if state was MODIFIED
       // WRBACK_INV will force invalid state
       cache_memory.setstate(MESI_Line_State::EXCLUSIVE);
       return;
    }
    // write-back is finished
    if (WB_flag) WB_flag = 0;
L_FETCH_LINE_end:    
    bcu_state = BCU_INITIAL;
    switch (FIL_action) {
      case FIL_READ:
      case FIL_WRITE:
      case FIL_PREF:
      case FIL_PREFEX:
           out.BUS_REQ = out.LOCK_REQ = 0; 
           out.bus.INV = 0;
           bcu_state = BCU_INITIAL;
           return;           
      case FIL_LOCK: 
           bcu_state = BCU_INITIAL;
           goto L_FIL_LOCK_ret;
    }   
L_STATE_FETCH_LINE_RETRY:
    //COUT << "--P" << "(FetchLineRetry) " << Topology::getCurrentTime() << "\n";
    if (wait_state_counter--) return;
    goto L_FETCH_INVALID_LINE;    
    
//-------------------------------------------------------------------------
// SCC - Snoopy Cache Controler
//--------------------------------------------------------------------------
L_SNOOP:    
    // here we come on a snoop hit. There are two reasons why someone wants
    // us to do snooping:
    // 1) he is writing something into his line and wants us to invalidate ours
    // 2) he wants to read the line and asks us to do the supply, if we are
    //    the owner. In addition, he may be having a write miss: he will get
    //    a copy of the line from us, and then ask us to invalidate ours!
    //snoop_bus_INV = in.bus.INV;
    //bus_request = in.bus.memmcmd; // save bus cycle type
    // REQUEST_READ + INV = REQUEST-EX
    // REQUEST_WRITE + INV = INVALIDATE  
    if ((in.bus.memcmd = REQUEST_READ) && (in.bus.INV)) //it's read-ex
       if (cache_memory.getstate() == MESI_Line_State::MODIFIED) 
          { cache_memory.setstate(MESI_Line_State::INVALID);
            goto L_SNOOP_write_back;
            }
       else 
          goto L_SNOOP_invalidate;
    else if ((in.bus.memcmd == REQUEST_WRITE) && (in.bus.INV))
       { if (cache_memory.getstate() == MESI_Line_State::SHARED)
          goto L_SNOOP_invalidate; }
    else if ((in.bus.memcmd == REQUEST_READ) && (! in.bus.INV))
       switch (cache_memory.getstate()) {
          case MESI_Line_State::INVALID:
                break;
          case MESI_Line_State::SHARED: 
                out.bus.SHARED = 1;
                break;
          case MESI_Line_State::EXCLUSIVE:
                out.bus.SHARED = 1;
                cache_memory.setstate(MESI_Line_State::SHARED);
                break;
          case MESI_Line_State::MODIFIED:
                cache_memory.setstate(MESI_Line_State::SHARED);
                goto L_SNOOP_write_back;
       }
    goto L_SNOOP_end;
L_SNOOP_write_back:
    out.bus.SHARED = 1;
    out.bus.DIRTY = 1;    
    // init the line for writing *to* someone (as in the write-back case)
    // If we had simulated data movement, the Topology would have read
    // the data from our output port and sent it to the bus, and the following
    // line would have looked like cache_memory.line_write_init(
    //                             &out.bus.addr, &out.bus.data);
    out.BUS_REQ = 1;
    snoop_state = SNOOP_WAIT_BGNT;
    return;
L_STATE_SNOOP_WAIT_BGNT:
    //COUT << "--P" << index << "(SnoopWaitBgnt) " << Topology::getCurrentTime() << "\n";
    if (! in.BUS_GNT) return;
    out.bus.reset();
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.memcmd = REQUEST_WRITE;
    out.bus.size = DATA_WIDTH;
    out.bus.SNOOP_REQ = 0;
    snoop_wait_state_counter = WRITE_WAIT_STATES;
    snoop_state = SNOOP_WR_WAIT_MEMORY;
L_STATE_SNOOP_WR_WAIT_MEMORY:
    //COUT << "--P" << index << "(SnoopWrWaitMemory) " << Topology::getCurrentTime() << "\n";
    if (snoop_wait_state_counter--) return;    
    if (cache_memory.line_write_getnext(&out.bus.addr)) 
        { snoop_wait_state_counter = WRITE_WAIT_STATES_2;
          return; // return if there is more, it will be continued in the next cycle
          }
    out.BUS_REQ=0;
    goto L_SNOOP_end;
L_SNOOP_invalidate:
    // invalidate the line if the INV signal was high
    cache_memory.setstate(MESI_Line_State::INVALID);
    cache_memory.untouch(); // force LRU
    ++stat.inv;
L_SNOOP_end:    
    // the snoop job is done,set cache_memory's Current Frame to #0 for the PCC
    cache_memory.set_current_frame(0);
    snoop_state = SNOOP_INITIAL;
    return;   
  }

// debugging...

void MESI_Cache::dump_state() {
    COUT << "CC" << index << ": PCC_State: ";
    if (snoop_state != SNOOP_INITIAL) { 
        if (snoop_state == SNOOP_WAIT_BGNT) 
            COUT << "SNOOP_WAIT_BGNT";
        else if (snoop_state == SNOOP_WR_WAIT_MEMORY)
            COUT << "SNOOP_WR_WAIT_MEMORY";
        }
    else    {
        switch (pcc_state) {
            case PCC_INITIAL         : COUT << "PCC_INITIAL"; break;
            case BLOCKED	     : COUT << "BLOCKED"; break;
            case LOCK_WAIT_BGNT      : COUT << "LOCK_WAIT_BGNT"; break;
            case LOCK_SLEEP          : COUT << "LOCK_SLEEP (" << busy_lock_sleep_cycle_count << ")"; break;
            case UNLOCK_WAIT_BGNT    : COUT << "UNLOCK_WAIT_BGNT"; break;
            case RM_BCU_WAIT         : COUT << "RM_BCU_WAIT"; break;
            case WM_BCU_WAIT         : COUT << "WM_BCU_WAIT"; break;
            case WI_BCU_WAIT	     : COUT << "WI_BCU_WAIT"; break;
            case LOCK_BCU_WAIT	     : COUT << "LOCK_BCU_WAIT"; break;
            case UNLOCK_BCU_WAIT     : COUT << "UNLOCK_BCU_WAIT"; break;
            case PREF_BCU_WAIT	     : COUT << "PREF_BCU_WAIT"; break;
            case PREFEX_BCU_WAIT     : COUT << "PREFEX_BCU_WAIT"; break; 
            case WRBACK_BCU_WAIT     : COUT << "WRBACK_BCU_WAIT"; break;
               }
        } 
    COUT << "\t\t[";
    if (pcc_state != PCC_INITIAL) {
        switch (in.request.type) {      
            case REQUEST_READ  : COUT << "READ " << hex <<in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << hex <<in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << hex <<in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << hex <<in.request.addr; break;
            case REQUEST_PREF  : COUT << "PREF " << hex <<in.request.addr; break;
            case REQUEST_PREFEX: COUT << "PREFEX"<< hex <<in.request.addr; break;
            case REQUEST_LPREF : COUT << "LPREF" << hex <<in.request.addr; break;
            case REQUEST_WRBACK: COUT << "WRBACK"<< hex <<in.request.addr; break;
            default :  break;
            }
        }       
    COUT << "]\n";    
    cache_memory.dump_state(index);
    
    COUT << "     BCU_State: ";
            switch (bcu_state) {
            case BCU_INITIAL            : COUT << "BCU_INITIAL"; break; 
            case FETCH_WAIT_BGNT        : COUT << "FETCH_WAIT_BGNT"; break;
            case FETCH_WAIT_SNOOP       : COUT << "FETCH_WAIT_SNOOP"; break;
            case FETCH_LINE_RETRY	: COUT << "FETCH_LINE_RETRY"; break;
            case WB_WAIT_MEMORY		: COUT << "WB_WAIT_MEMORY"; break;
            case INV_WAIT_BGNT		: COUT << "INV_WAIT_BGNT"; 
                                          break;
            case INV_WAIT_END		: COUT << "INV_WAIT_END"; break;
            case FETCH_WAIT_MEMORY	: COUT << "FETCH_WAIT_MEMORY"; break;
            case WRBACK_WAIT_BGNT	: COUT << "WRBACK_WAIT_BGNT"; break;
        }
    if (bcu_state != BCU_INITIAL) {
        cache_memory.set_current_frame(2);
        cache_memory.get_current_line(&tmp_addr);
        COUT << "\t[" << hex << tmp_addr << "]\n";
    } 
    COUT << "\n";    

    }
                

#endif PRAGMA_IMPLEMENTATION_MESI_CACHE
#endif _MESI_CACHE_H_