#ifndef _MESI_CACHE_H_
#define _MESI_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module : MESI_cache.h
 * Authors: Igor Ikodinovic   ( igi@eunet.yu )
 *          Dusko Krsmanovic  ( dusko@orao.etf.bg.ac.yu )
 *
 * The MESI protocol - definition and implementation.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "bus_signals.h"

// states a line in the MESI protocol can be found in
class MESI_Line_State {
  public:
    enum State {INVALID,SHARED,EXCLUSIVE,MODIFIED} state;
    MESI_Line_State()           { state = INVALID; }
    bool  isvalid()             { return state != INVALID; }
    State get()                 { return state; }
    void  set(State new_state)  { state = new_state; }
    char  symbol() {            // for debugging
        switch (state) {
            case MODIFIED  : return 'M';     // Modified
            case EXCLUSIVE : return 'E';     // Exclusive
            case SHARED    : return 'S';     // Shared
            case INVALID   : return 'I';     // Invalid
        }
    }
};

// the MESI cache controler module declaration
class MESI_Cache {
    int index;
    // instantiate Cache Memory with the MESI_Line_State as its Line_State
    Cache_Memory<MESI_Line_State> cache_memory;  // the whole cache memory
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        bool BUS_GNT;                   // bus grant (from the bus module)
        bool LOCK_GNT;                  // lock grant (from the bus module)
        Bus_Signals bus;                // bus contents, to be read from the bus 
    } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to IO_Buf: PROCESSOR_SATISFIED,PROCESSOR_STALLED..
            bool STROBE;                // set to high when 'signal' is formed
        } response;
        bool BUS_REQ;                   // bus request: sent to the bus module
        bool LOCK_REQ;                  // lock request: sent to the bus module
        Bus_Signals bus;                // bus contents, to be written on the bus
    } out;
    // the SCC (Snoop Cache Controller part) may be in one of three states
    enum SCC_State {
        SNOOP_INITIAL,
        SNOOP_PUT_INIT,
        SNOOP_PUT_NEXT
    } SCC_state;
    // states that a PCC (Processor Cache Controller part) can be found in
    enum PCC_State {
        INITIAL, 
        READ_WAIT_BGNT,
        READ_WB_WAIT_MEMORY,
        READ_WAIT_1,
        READ_GET_NEXT,
        READ_FETCH_WAIT_MEMORY,
        WRITE_WAIT_BGNT,
        WRITE_WB_WAIT_MEMORY,
        WRITE_WAIT_1,
        WRITE_GET_NEXT,
        WRITE_FETCH_WAIT_MEMORY,
        WRITE_INV_WAIT_BGNT,
        WRITE_INV_WAIT_1,
        LOCK_WAIT_BGNT,
        LOCK_WB_WAIT_MEMORY,
        LOCK_WAIT_1,
        LOCK_GET_NEXT,
        LOCK_FETCH_WAIT_MEMORY,
        LOCK_SLEEP
    } PCC_state;
    int wait_state_counter;
    bool snoop_bus_INV;
    MESI_Line_State::State next_line_state;
    // delay parameters of the simulated physical memory: the default values
    const DEFAULT_READ_WAIT_STATES      = 3; // time needed to read the first word
    const DEFAULT_READ_WAIT_STATES_2    = 0; // time needed for access to subsequent words
    const DEFAULT_WRITE_WAIT_STATES     = 3; // time needed to write the first word
    const DEFAULT_WRITE_WAIT_STATES_2   = 0; // time needed for writing of subsequent words
    // and the real values, which can be changed at startup
    int READ_WAIT_STATES;               // these will become equal to the
    int READ_WAIT_STATES_2;             // defaults if not otherwise specified
    int WRITE_WAIT_STATES;              // in the INI file or command line
    int WRITE_WAIT_STATES_2;            // switches
    const BUSY_LOCK_SLEEP_CYCLE_COUNT   = 1; 
    bitsLOCK lock_value;
    int busy_lock_sleep_cycle_count;
    struct Stat {                       // statistics:
        int rhit;                       // read hits
        int rmis;                       // read misses
        int whit;                       // write hits
        int wmis;                       // write misses
        int wb;                         // writebacks
        int req;                        // read+write+lock+unlock requests
        int inv;                        // invalidations
        int cs;                         // cache supplies
        Stat() {rhit=rmis=whit=wmis=wb=req=inv=cs=0;}
        void dump(int index);
    } stat;
  public:
    MESI_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {return (PCC_state == INITIAL);}
    void dump_state();
    void dump_stats() {stat.dump(index);}
    TOPOLOGY_FRIENDS
};


#ifdef PRAGMA_IMPLEMENTATION_MESI_CACHE

// the MESI cache controller module implementation

#include <iostream.h>

MESI_Cache::MESI_Cache() {                      // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    SCC_state = SNOOP_INITIAL;                  // and so will the SCC part
    PCC_state = INITIAL;                        // PCC will be in the INITIAL state
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    in.BUS_GNT      = 0;
    out.BUS_REQ     = 0;
    in.LOCK_GNT     = 0;
    out.LOCK_REQ    = 0;
    in.bus.reset();
    out.bus.reset();
    // initialize memory delay parameters. If you want to change this create limes.ini in your
    // app's directory and place in the line r_ws=4 (for example), or use simulator's command
    // line Switches, like -dw_ws=4.
    READ_WAIT_STATES    = sys_INI_read("r_ws", DEFAULT_READ_WAIT_STATES);
    READ_WAIT_STATES_2  = sys_INI_read("r_ws2", DEFAULT_READ_WAIT_STATES_2);
    WRITE_WAIT_STATES   = sys_INI_read("w_ws", DEFAULT_WRITE_WAIT_STATES);
    WRITE_WAIT_STATES_2 = sys_INI_read("w_ws2", DEFAULT_WRITE_WAIT_STATES_2);
};

void MESI_Cache::Introduce() {
    COUT << "Protocol: MESI (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << " ";
    #endif
    COUT << "\n";
    COUT << "READ_WS=" << READ_WAIT_STATES << "/" << READ_WAIT_STATES_2 << ", ";
    COUT << "WRITE_WS=" << WRITE_WAIT_STATES << "/" << WRITE_WAIT_STATES_2 << ", ";
    COUT << "LOCK_WS=" << BUSY_LOCK_SLEEP_CYCLE_COUNT << "\n";
    #ifdef DUMMLOCK
        COUT << " ";
    #endif
    cache_memory.Introduce();
}

void MESI_Cache::Stat::dump(int index) {
    if (req) COUT << "MESI"<<index<<": " << ",rhit="<<rhit << ",rmis="<<rmis << ",whit="<<whit
    << ",wmis="<<wmis << ",wb="<<wb << "req="<<req << ",inv="<<inv << ",cs="<<cs << "\n";
}

// the main method: Topology calls it in each simulated cycle. If it was not in initial state, it 
// will resume from the point where it left off in the previous cycle.
// This method first reads the bus. Then it checks if there is SNOOP signal high on it, and, if so,
// checks if it has the line to invalidate. If true, does so and skips this cycle. This reflects 
// the fact that there are two controllers in this one, the Processor Cache Controller (PCC) and 
// the Snoop Cache Controller (SCC), each of which has its own copy of cache tags, but the cache 
// memory itself can be accessed only by one controller at a time, and the SCC part has precedence.

void MESI_Cache::cycle() {
    out.bus.DIRTY=0;
    out.bus.SHARE=0;
    out.LOCK_REQ=0;
    Topology::cache_read_bus(index);    // read input from the bus

    // check if there is SNOOP signal on the bus, but issued by a cache other than myself:
    
    switch (SCC_state) {
          // the SCC may be in initial state, or not. If it is in initial state, it will check if 
          // a SNOOP signal appeared on the bus. Otherwise, it will continue what it was doing in 
          // the previous cycle -- doing a cache supply for some other cache controller.
          
          case SNOOP_INITIAL:
               if (in.bus.SNOOP && !in.BUS_GNT) {
                   // !in.BUS_GNT means 'I don't have the bus'
                   cache_memory.set_current_frame(1);   // frame #1 for the snoop part
                   snoop_bus_INV = in.bus.INV;
                   if (cache_memory.isvalid(in.bus.addr)) {
                       out.bus.SHARE=1;
                       if (cache_memory.getstate()==MESI_Line_State::MODIFIED) out.bus.DIRTY=1;
                       if (in.bus.INV!=1)
                           switch(cache_memory.getstate()) {
                                 case MESI_Line_State::SHARED:
                                 case MESI_Line_State::EXCLUSIVE:
                                      cache_memory.setstate(MESI_Line_State::SHARED);
                                      cache_memory.set_current_frame(0);
                                      SCC_state = SNOOP_INITIAL;
                                      return;
                                 case MESI_Line_State::MODIFIED:
                                      SCC_state = SNOOP_PUT_INIT;
                                      return;
                           }
                       else
                           switch(cache_memory.getstate()) {
                                 case MESI_Line_State::SHARED:
                                 case MESI_Line_State::EXCLUSIVE:
                                      ++stat.inv;
                                      cache_memory.setstate(MESI_Line_State::INVALID);
                                      cache_memory.untouch();   // force LRU
                                      cache_memory.set_current_frame(0);
                                      SCC_state = SNOOP_INITIAL;
                                      return;
                                 case MESI_Line_State::MODIFIED:
                                      SCC_state = SNOOP_PUT_INIT;
                                      return;
                           }
                   }
                   else 
                       cache_memory.set_current_frame(0);  // frame #0 for the processor part
               }
               break;  // activate processor cache controller

          case SNOOP_PUT_INIT:
               ++stat.cs;
               cache_memory.line_write_init(&out.bus.addr);
               wait_state_counter = WRITE_WAIT_STATES;
               SCC_state = SNOOP_PUT_NEXT;
               return;

          case SNOOP_PUT_NEXT:
               if (wait_state_counter--) return;
               if (cache_memory.line_write_getnext(&out.bus.addr)) {
                   wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
     	           return; 
               }

               if (snoop_bus_INV) {
                   // invalidate the line if the INV signal was high
                   ++stat.inv; 
                   cache_memory.setstate(MESI_Line_State::INVALID);
                   cache_memory.untouch(); // force LRU
               }
               else 
                   cache_memory.setstate(MESI_Line_State::SHARED);

               cache_memory.set_current_frame(0);
               SCC_state = SNOOP_INITIAL;
               return;
    }

    // Ok, there was no snoop or it was a miss, so PCC part follows now

    switch (PCC_state) {
          case INITIAL                 : goto L_STATE_INITIAL;
          case READ_WAIT_BGNT          : goto L_STATE_READ_WAIT_BGNT;
          case READ_WB_WAIT_MEMORY     : goto L_STATE_READ_WB_WAIT_MEMORY;
          case READ_WAIT_1             : goto L_STATE_READ_WAIT_1;
          case READ_GET_NEXT           : goto L_STATE_READ_GET_NEXT;
          case READ_FETCH_WAIT_MEMORY  : goto L_STATE_READ_FETCH_WAIT_MEMORY;
          case WRITE_WAIT_BGNT         : goto L_STATE_WRITE_WAIT_BGNT;
          case WRITE_WB_WAIT_MEMORY    : goto L_STATE_WRITE_WB_WAIT_MEMORY;
          case WRITE_WAIT_1            : goto L_STATE_WRITE_WAIT_1;
          case WRITE_GET_NEXT          : goto L_STATE_WRITE_GET_NEXT;
          case WRITE_FETCH_WAIT_MEMORY : goto L_STATE_WRITE_FETCH_WAIT_MEMORY;
          case WRITE_INV_WAIT_BGNT     : goto L_STATE_WRITE_INV_WAIT_BGNT;
          case WRITE_INV_WAIT_1        : goto L_STATE_WRITE_INV_WAIT_1;
          case LOCK_WAIT_BGNT          : goto L_STATE_LOCK_WAIT_BGNT;
          case LOCK_WB_WAIT_MEMORY     : goto L_STATE_LOCK_WB_WAIT_MEMORY;
          case LOCK_WAIT_1             : goto L_STATE_LOCK_WAIT_1;
          case LOCK_GET_NEXT           : goto L_STATE_LOCK_GET_NEXT;
          case LOCK_FETCH_WAIT_MEMORY  : goto L_STATE_LOCK_FETCH_WAIT_MEMORY;  
          case LOCK_SLEEP              : goto L_STATE_LOCK_SLEEP;
    }

L_STATE_INITIAL:
    if (! Topology::cache_read_request(index)) return;  // return if there was no request
    ++stat.req;                                         // statistics
    switch (in.request.type) {                          // see what request it was and go there
        case REQUEST_READ   : goto L_READ;
        case REQUEST_WRITE  : goto L_WRITE;
        case REQUEST_LOCK   : goto L_LOCK;
        case REQUEST_UNLOCK : goto L_UNLOCK;
        default : break;
    }
    return;

 
//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    if (cache_memory.isvalid(in.request.addr)) {
        // a hit, excellent!
        ++stat.rhit;
        goto L_READ_end;
    }
    // a miss, we ask for bus to read the line from the one who has it
    ++stat.rmis;
    out.BUS_REQ = 1;
    PCC_state = READ_WAIT_BGNT;

L_STATE_READ_WAIT_BGNT:
    if (! in.BUS_GNT) return;

    if (cache_memory.getstate()!=MESI_Line_State::MODIFIED) goto L_READ_READ_LINE;
 
    // READ_WRITEBACK 
    ++stat.wb;      
    out.bus.reset(); 
    // initialize the line for write-back. The cache memory will return the address of the first 
    // word in the line, and that's what will go to the bus.
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.memcmd = REQUEST_WRITE;
    out.bus.size = DATA_WIDTH;
    wait_state_counter = WRITE_WAIT_STATES;
    PCC_state = READ_WB_WAIT_MEMORY;
    return;

L_STATE_READ_WB_WAIT_MEMORY:
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
    }
    // END READ_WRITEBACK

L_READ_READ_LINE:
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = in.request.size;
    out.bus.SNOOP = 1;
    out.bus.memcmd = REQUEST_READ;
    wait_state_counter = 1;    // we'll wait one cycle to let
    PCC_state = READ_WAIT_1;   // the others see these signals
    return; 

L_STATE_READ_WAIT_1:
    if (wait_state_counter--) {
        out.bus.SNOOP = 0;
        return;
    }

    if (in.bus.SHARE) 
        next_line_state = MESI_Line_State::SHARED;
    else
        next_line_state = MESI_Line_State::EXCLUSIVE;

    if (in.bus.DIRTY!=1) goto L_READ_FETCH_NEW_LINE; 

    // DIRTY=1, so we tranfer the line from the cache which has that modified line
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_WRITE;
    wait_state_counter = WRITE_WAIT_STATES;
    PCC_state = READ_GET_NEXT;
    return;

L_STATE_READ_GET_NEXT:
    if (wait_state_counter--) return;
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
    }

goto L_READ_FETCHED; 

L_READ_FETCH_NEW_LINE:
    // DIRTY=0, so we fetch the line from memory
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_READ;
    wait_state_counter = READ_WAIT_STATES - 2;   // 2 is the number of cycles passed already
    PCC_state = READ_FETCH_WAIT_MEMORY;   // this will loop <word count> times
    return;

L_STATE_READ_FETCH_WAIT_MEMORY:
    if (wait_state_counter-- >0 ) return;
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = READ_WAIT_STATES_2;   // there are more words
        return;
    }

L_READ_FETCHED:
    cache_memory.touch();  // force this line MRU (most recently used)
    cache_memory.setstate(next_line_state); 

L_READ_end:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    PCC_state = INITIAL;
    out.BUS_REQ = 0;
    return;


//---------------------------------------------------------------
L_WRITE:
    // This is a WRITE.
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    // Check the state of the line
    if (cache_memory.isvalid(in.request.addr)) { 
        ++stat.whit;
        // if the line is already modified, or exclusive, no further actions are necessary
        // if not, we'll have to invalidate the others.
        if ((cache_memory.getstate() == MESI_Line_State::MODIFIED) ||
            (cache_memory.getstate() == MESI_Line_State::EXCLUSIVE))
            goto L_WRITE_end;
        // in an unoptimized version put L_WRITE_INVALID_STATE instead of L_WRITE_SHARED_STATE
        if (cache_memory.getstate() == MESI_Line_State::SHARED) goto L_WRITE_SHARED_STATE;
    }

L_WRITE_INVALID_STATE:
    // a miss, we do a BusRdX
    ++stat.wmis;
    out.BUS_REQ = 1;
    PCC_state = WRITE_WAIT_BGNT;

L_STATE_WRITE_WAIT_BGNT:
    if (! in.BUS_GNT) return;

    if (cache_memory.getstate()!=MESI_Line_State::MODIFIED) goto L_WRITE_READ_LINE;

    // WRITE_WRITEBACK   
    ++stat.wb;      
    out.bus.reset(); 
    // initialize the line for write-back. The cache memory will return the address of the first 
    // word in the line, and that's what will go to the bus.
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_WRITE;
    wait_state_counter = WRITE_WAIT_STATES;
    PCC_state = WRITE_WB_WAIT_MEMORY;
    return;

L_STATE_WRITE_WB_WAIT_MEMORY:
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
    }
    // END WRITE_WRITEBACK
    
L_WRITE_READ_LINE:
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = in.request.size;
    out.bus.SNOOP = 1;
    out.bus.INV = 1;
    out.bus.memcmd = REQUEST_READ;
    wait_state_counter = 1;     // we'll wait one cycle to let
    PCC_state = WRITE_WAIT_1;   // the others see these signals
    return; 

L_STATE_WRITE_WAIT_1:
    if (wait_state_counter--) {
        out.bus.SNOOP = 0;
        out.bus.INV = 0;    
        return;
    }
    
    if (in.bus.DIRTY!=1) goto L_WRITE_FETCH_NEW_LINE; 
    
    // DIRTY=1, so we tranfer the line from the cache which has that modified line
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_WRITE;
    wait_state_counter = WRITE_WAIT_STATES;
    PCC_state = WRITE_GET_NEXT;
    return;

L_STATE_WRITE_GET_NEXT:
    if (wait_state_counter--) return;
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
    }

    cache_memory.touch();   // force this line MRU (most recently used)
    goto L_WRITE_end;

L_WRITE_FETCH_NEW_LINE:
    // DIRTY=0, so we fetch the line from memory
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_READ;
    wait_state_counter = READ_WAIT_STATES - 2;   // 2 is the number of cycles passed already
    PCC_state = WRITE_FETCH_WAIT_MEMORY;   // this will loop <word count> times
    return;

L_STATE_WRITE_FETCH_WAIT_MEMORY:
    if (wait_state_counter-- >0 ) return;
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = READ_WAIT_STATES_2;   // there are more words
        return;
    }

    cache_memory.touch();   // force this line MRU (most recently used)
    goto L_WRITE_end;

// in an unoptimized version you can delete all lines until L_WRITE_end
L_WRITE_SHARED_STATE:
    out.BUS_REQ = 1;
    PCC_state = WRITE_INV_WAIT_BGNT;

L_STATE_WRITE_INV_WAIT_BGNT:
    // check the line state to see if it has been invalidated while we were waiting for the bus
    if (cache_memory.getstate() == MESI_Line_State::INVALID) goto L_WRITE_INVALID_STATE;
    if (! in.BUS_GNT) return;   // wait as long as it takes

    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = in.request.size;
    out.bus.SNOOP = 1;
    out.bus.INV = 1;
    wait_state_counter = 1;         // we'll wait one cycle to let
    PCC_state = WRITE_INV_WAIT_1;   // the others see these signals
    return;                                 

L_STATE_WRITE_INV_WAIT_1:
    if (wait_state_counter--) {
        out.bus.SNOOP = 0;
        out.bus.INV = 0;
        return;
    }

    cache_memory.touch();   // force this line MRU (most recently used)

L_WRITE_end:
    cache_memory.setstate(MESI_Line_State::MODIFIED);
    out.BUS_REQ = 0;
    PCC_state = INITIAL;
    return;


//---------------------------------------------------------------
L_LOCK:
    // It is a LOCK. Check if the line is in the cache (and whether it is valid):
    if (cache_memory.isvalid(in.request.addr)) {
        ++stat.rhit;                    // yes, it is.
        goto L_LOCK_check_value;
    }
    // a miss, we ask for bus to read the line from the one who has it
    ++stat.rmis;
    out.BUS_REQ = 1;                    // raise BUS request high
    PCC_state = LOCK_WAIT_BGNT;

L_STATE_LOCK_WAIT_BGNT:                 // we are waiting to get the bus
    if (! in.BUS_GNT) return;

    if (cache_memory.getstate()!=MESI_Line_State::MODIFIED) goto L_LOCK_READ_LINE;

    // LOCK_WRITEBACK
    ++stat.wb;
    out.bus.reset(); 
    // initialize the line for write-back. The cache memory will return the address of the first
    // word in the line, and that's what will go to the bus.
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.memcmd = REQUEST_WRITE;
    out.bus.size = DATA_WIDTH;
    wait_state_counter = WRITE_WAIT_STATES;
    PCC_state = LOCK_WB_WAIT_MEMORY;
    return;

L_STATE_LOCK_WB_WAIT_MEMORY:
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
    }
    // END LOCK_WRITEBACK 

L_LOCK_READ_LINE:
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = in.request.size;
    out.bus.SNOOP = 1;
    out.bus.memcmd = REQUEST_READ;
    wait_state_counter = 1;    // we'll wait one cycle to let
    PCC_state = LOCK_WAIT_1;   // the others see these signals
    return;

L_STATE_LOCK_WAIT_1:
    if (wait_state_counter--) {
        out.bus.SNOOP = 0;
        return;
    }

    if (in.bus.SHARE) 
           next_line_state = MESI_Line_State::SHARED;
    else
           next_line_state = MESI_Line_State::EXCLUSIVE;

    if (in.bus.DIRTY!=1) goto L_LOCK_FETCH_NEW_LINE; 

    // DIRTY=1, so we tranfer the line from the cache which has that modified line
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_WRITE;
    wait_state_counter = WRITE_WAIT_STATES;
    PCC_state = LOCK_GET_NEXT;
    return;

L_STATE_LOCK_GET_NEXT:
    if (wait_state_counter--) return;
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
    }

goto L_LOCK_FETCHED;

L_LOCK_FETCH_NEW_LINE:
    // DIRTY=0, so we fetch the line from memory
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size = DATA_WIDTH;
    out.bus.memcmd = REQUEST_READ;
    wait_state_counter = READ_WAIT_STATES - 2;  // 2 is the number of cycles passed already
    PCC_state = LOCK_FETCH_WAIT_MEMORY;   // this will loop <word count> times
    return;
    
L_STATE_LOCK_FETCH_WAIT_MEMORY:
    if (wait_state_counter-- >0 ) return;
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = READ_WAIT_STATES_2;   // there are more words
        return;
    }

L_LOCK_FETCHED:
    cache_memory.touch();   // force this line MRU (most recently used)
    cache_memory.setstate(next_line_state);

L_LOCK_check_value:
    // the memsim has to actually read the lock from the real memory, since cache keeps only tags,
    // not memory contents. The real memory is actually totally consistent (PRAM), everything here
    // is about modelling delays. This action would not have been necessary if we had really moved
    // the data from the memory to the simulated cache and vice versa, but this is an extreme 
    // effort. It only writes and reads lock values.
    lock_value = MEMSIM_READ_LOCK(in.request.addr);
    if (lock_value == 0) {
        // Do the actual "write 1" to the location for the lock.
        MEMSIM_WRITE_LOCK( in.request.addr, 1); 
        goto L_WRITE;
    }

    out.BUS_REQ = 0;
    busy_lock_sleep_cycle_count = BUSY_LOCK_SLEEP_CYCLE_COUNT;
    PCC_state = LOCK_SLEEP;

L_STATE_LOCK_SLEEP:
    if (busy_lock_sleep_cycle_count--) return;
    goto L_LOCK;


//---------------------------------------------------------------
L_UNLOCK:
    MEMSIM_WRITE_LOCK( in.request.addr, 0);
    goto L_WRITE;
}   


// debugging...

void MESI_Cache::dump_state() {
    COUT << "CC" << index << ": State: ";
    if (SCC_state != SNOOP_INITIAL)  
        switch (SCC_state) {
              case SNOOP_PUT_INIT  : COUT << "SNOOP_PUT_INIT"; break;
              case SNOOP_PUT_NEXT  : COUT << "SNOOP_PUT_NEXT"; break;
              default              : COUT << "SNOOP_?"; break;
        }
    else 
        switch (PCC_state) {
              case INITIAL                 : COUT << "INITIAL"; break;
              case READ_WAIT_BGNT          : COUT << "READ_WAIT_BGNT"; break;
              case READ_WB_WAIT_MEMORY     : COUT << "READ_WB_WAIT_MEMORY"; break;
              case READ_WAIT_1             : COUT << "READ_WAIT_1"; break;
              case READ_GET_NEXT           : COUT << "READ_GET_NEXT"; break;
              case READ_FETCH_WAIT_MEMORY  : COUT << "READ_FETCH_WAIT_MEMORY"; break;
              case WRITE_WAIT_BGNT         : COUT << "WRITE_WAIT_BGNT"; break;
              case WRITE_WB_WAIT_MEMORY    : COUT << "WRITE_WB_WAIT_MEMORY"; break;
              case WRITE_WAIT_1            : COUT << "WRITE_WAIT_1"; break;
              case WRITE_GET_NEXT          : COUT << "WRITE_GET_NEXT"; break;
              case WRITE_FETCH_WAIT_MEMORY : COUT << "WRITE_FETCH_WAIT_MEMORY"; break;
              case WRITE_INV_WAIT_BGNT     : COUT << "WRITE_INV_WAIT_BGNT"; break;
              case WRITE_INV_WAIT_1        : COUT << "WRITE_INV_WAIT_1"; break;
              case LOCK_WAIT_BGNT          : COUT << "LOCK_WAIT_BGNT"; break;
              case LOCK_WB_WAIT_MEMORY     : COUT << "LOCK_WB_WAIT_MEMORY"; break;
              case LOCK_WAIT_1             : COUT << "LOCK_WAIT_1"; break;
              case LOCK_GET_NEXT           : COUT << "LOCK_GET_NEXT"; break;
              case LOCK_FETCH_WAIT_MEMORY  : COUT << "LOCK_FETCH_WAIT_MEMORY"; break;
              case LOCK_SLEEP              : COUT << "LOCK_SLEEP (" << busy_lock_sleep_cycle_count << ")";break;
              default                      : COUT << "CACHE_?"; break;
        }
    COUT << "\t\t[";
    if (PCC_state != INITIAL) 
        switch (in.request.type) {
              case REQUEST_READ   : COUT << "READ "   << in.request.addr; break;
              case REQUEST_WRITE  : COUT << "WRITE "  << in.request.addr; break;
              case REQUEST_LOCK   : COUT << "LOCK "   << in.request.addr; break;
              case REQUEST_UNLOCK : COUT << "UNLOCK " << in.request.addr; break;
              default             : break;
        }
    COUT << "]\n";
    cache_memory.dump_state(index);
}

#endif PRAGMA_IMPLEMENTATION_MESI_CACHE
#endif _MESI_CACHE_H_
