/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: topology.cc
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Implementation of topology methods.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include <stdio.h>
#include "topology.h"

#include CACHE_HDR      // include header files with definitions
#include BUS_HDR        // of the classes that represent
#include IO_BUF_HDR     // the hardware modules

// these are the pointers for module instances
IO_BUF_CLASS            * Base_Topology::io_buf[MAX_CPU_COUNT];
CACHE_CLASS             * Base_Topology::cache[MAX_CPU_COUNT];
BUS_CLASS               * Base_Topology::bus;

// these routines are called from the memsim

void Base_Topology::system_init() { 
    bus = new BUS_CLASS;
    for(int i=0; i<MAX_CPU_COUNT; ++i)  {
        io_buf[i] = new IO_Buf;
        cache[i]  = new CACHE_CLASS;        
        io_buf[i]->assign_index(i);
        cache[i]->assign_index(i);
    }
    cache[0]->Introduce();
}
void Base_Topology::system_cleanup(Thread_Time last_time) { 
    int rhit=0;            // read hits
    int rmis=0;            // read misses
    int whit=0;            // write hits
    int wmis=0;            // write misses
    int wb=0;              // write-backs
    int req=0;             // read+write+lock+unlock requests
    int inv=0;             // invalidations
    int cs=0;              // cache supplies
    FILE *izlaz;

    PRINTF("TOTAL simulation time: %d cycles.\n", last_time);
    for(int i=0; i<MAX_CPU_COUNT; ++i) { 
        cache[i]->dump_stats();     // should be doing DELETE in addition...
        rhit+=cache[i]->stat.rhit;
        rmis+=cache[i]->stat.rmis;
        whit+=cache[i]->stat.whit;
        wmis+=cache[i]->stat.wmis;
        wb+=cache[i]->stat.wb;
        req+=cache[i]->stat.req;
        inv+=cache[i]->stat.inv;
        cs+=cache[i]->stat.cs;
    }
    bus->dump_stats(last_time);
    izlaz=fopen("TABELA.TXT","a+");
    fprintf(izlaz,"\n%d, %d, %d, %d, %d, %d, %d, %d, %d, %d",
            rhit,rmis,whit,wmis,wb,req,inv,cs,last_time,bus->busytime);
    fclose(izlaz);
}

// returns true if all the modules in the system are idle
bool Base_Topology::system_isidle() {
    for (int i=0; i<CPU_COUNT; ++i) {
        if (!io_buf[i]->isidle()) return false;
        if (!cache[i]->isidle()) return false;
    }
    return true;
}
void Base_Topology::system_dump_state() {
    for(int i=0; i<CPU_COUNT; ++i)  
        cache[i]->dump_state();
    bus->dump_state();      
}

// Called at times when a set of request *could* exist

void Base_Topology::system_collect_requests() {
    for(int i=0; i<CPU_COUNT; ++i) 
        io_buf[i]->start_cycle();
}

// The main function: polls all the modules to do their share in this cycle.

void Base_Topology::system_cycle() {
    for(int i=0; i<CPU_COUNT; ++i) 
        cache[i]->cycle();
    bus->cycle();
    for(i=0; i<CPU_COUNT; ++i) 
        io_buf[i]->end_cycle();
}

// Topology for your own simulator will need all the methods listed 
// above; only implementations of these method will differ.

// These methods are called from various modules; their names and
// implementations depend on the structure of your system.

// ******************* write in IO_buf's input port *******************

// each IO_Buf calles this method (passing its index) upon the end
// of each simulated cycle. This method checks if the cache controller
// placed immediately below the io_buf module in question has set its
// STROBE bit high. If so, the controller has an information to communicate
// to the io_buf module. The topology will read this information from the
// controller's output port, write it into the io_buf module's input port,
// and clear the STROBE bit.

bool Base_Topology::io_buf_input(int index) {
    if (cache[index]->out.response.STROBE) {
        io_buf[index]->in.response.signal = 
            cache[index]->out.response.signal;
        cache[index]->out.response.STROBE = 0;  // clear
        return true;    // tell the io_buf that a response appeared
        }       
    else
        return false;
}

// ********************* cache modules services *********************

// The similar procedure is followed for reading io_buf's output
// (request description) and writing it into cache controller's 
// input port. The STROBE bit is also used. However, the controller
// module will not call this method at every cycle, but only when
// it's idle. If it is already processing previous request (such as
// read that was a miss), it will not call this method and the STROBE
// bit will not be cleared. Only when the controller gets idle again
// (returns to its INITIAL state) it will call this method. By inspecting
// its STROBE bit, the IO_Buf can know that the controller has read the
// request.

bool Base_Topology::cache_read_request(int index)   {
    if (io_buf[index]->out.request.STROBE) {
        cache[index]->in.request = 
            * io_buf[index]->out.request.request_ptr;
        // this is where virtual address translates to physical.
        // In this case, just physcial = virtual.
        cache[index]->in.request.addr = 
            (cache[index]->in.request.virtual_addr);
        io_buf[index]->out.request.STROBE = 0;          // clear
        return true;    // yes, a request exists
        }
    else
        return false;   // no request this time
}

// ****************** BUS-related Topology functions ******************

// The bus module calls these to read bus and lock requests
bool Base_Topology::bus_read_REQs() {
    bool any=0;
    for (int i=0; i<CPU_COUNT; ++i) 
        any = any | (bus->BUS_REQ[i] = cache[i]->out.BUS_REQ);
    return any;
}

bool Base_Topology::bus_read_LOCK_REQs() {
    bool any=0;
    for (int i=0; i<CPU_COUNT; ++i) 
        any = any | (bus->LOCK_REQ[i] = cache[i]->out.LOCK_REQ);
    return any;
}

// this method is called by the bus when it wants to read the contents
// of the controller's output port that should appear on the bus,
// for the module that currently holds the bus

void Base_Topology::bus_read_contents(int master_index) {
    bus->bus_contents = cache[master_index]->out.bus;
    bool share = 0;
    bool dirty = 0;
    for (int i=0; i<CPU_COUNT; ++i) {
        share |= cache[i]->out.bus.SHARE;   // SHARE is wired OR line
        dirty |= cache[i]->out.bus.DIRTY;   // DIRTY is also wired OR line
    }
    for (i=0; i<CPU_COUNT; ++i) {
        cache[i]->in.bus.SHARE = share;
        cache[i]->in.bus.DIRTY = dirty;
    }
    bus->bus_contents.SHARE = share;
    bus->bus_contents.DIRTY = dirty;
}

// this method is called by each cache controller module when it wants
// to read what is on the bus. Of course, each cache controller calls
// this method in each simulated cycle.

void Base_Topology::cache_read_bus(int index)   {
    cache[index]->in.BUS_GNT  = bus->BUS_GNT[index];
    cache[index]->in.LOCK_GNT = bus->LOCK_GNT[index];
    cache[index]->in.bus      = bus->bus_contents;
}
