#ifndef _DRA_CACHE_H_
#define _DRA_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: dra_cache.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The Dragon protocol; definition and implementation.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "bus_signals.h"

// states a line in the Dragon protocol can be found in
class Dra_Line_State {
  public:
    enum State { UNMOD_SHD, UNMOD_EXC, MOD_SHD, MOD_EXC } state;
    Dra_Line_State()            { state = UNMOD_SHD; }
    bool  isvalid()             { return 1; }
    State get()                 { return state; }
    void  set(State new_state)  { state = new_state; }
    char  symbol() {
        switch (state) {
            case UNMOD_SHD: return 's';     // Unmodified, Shared
            case UNMOD_EXC: return 'e';     // Unmodifed, Exclusive
            case MOD_SHD  : return 'S';     // Modified, Shared
            case MOD_EXC  : return 'E';     // Modified, Exclusive
            }
        }
    };

// the Dragon cache controler module declaration
class Dra_Cache {
    int index;
    // instantiate Cache Memory with the Dra_Line_State as its Line_State
    Cache_Memory<Dra_Line_State> cache_memory;
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        bool BUS_GNT;                   // bus grant (from the bus module)
        bool LOCK_GNT;                  // lock grant (from the bus module)
        Bus_Signals bus;                // bus contents, to be read from the bus
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        bool BUS_REQ;                   // bus request: sent to the bus module
        bool LOCK_REQ;                  // lock request: sent to the bus module
        Bus_Signals bus;                // bus contents, to be written on the bus
        } out;
    // states this finite state machine called Dragon can be found in
    enum State {
        INITIAL,                        // INITIAL -- not doing anything (except maybe in the snoop part)
        WRITE_WAIT_BGNT,                // waiting for the bus to do a write
        WRITE_WAIT_2,                   // waiting one cycle to let the others see the SNOOP signal before we withdraw it from the bus
        UNLOCK_WAIT_BGNT,               // waiting for the bus to do an unlock
        LOCK_WAIT_BGNT,                 // waiting for the bus to do a lock
        LOCK_SLEEP,                     // sleeping before reattempting lock_acquire
        FETCH_WAIT_BGNT,                // waiting for the bus to fetch a line
        WB_WAIT_MEMORY,                 // waiting for the memory to complete a write
        FETCH_WAIT_SUPPLY_2,            // waiting to see if it will be a cache supply or a memory supply
        FETCH_WAIT_MEMORY               // waiting for a memory to complete a read
        } state;
    // the SCC (snoop cache controller part) may also be in one of two states
    enum Snoop_State {
        SNOOP_INITIAL,                  // not doing a snoop
        SNOOP_WAIT_1                    // doing a supply for someone else
        } snoop_state;
    // One 'procedure' is called to fetch an invalid line; it can be called
    // upon a read, a write, or a lock. This is a sort of replacement for gosub-return sequence
    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK} FIL_action;
    // some internal vars
    int wait_state_counter;
    bool cache_supply;
    bool bus_wr_miss;
    bool bus_sh_line;
    // delay parameters of the simulated physical memory: the default values
    const DEFAULT_READ_WAIT_STATES      = 3; // time needed to read the first word
    const DEFAULT_READ_WAIT_STATES_2    = 0; // time needed for access to subsequent words
    const DEFAULT_WRITE_WAIT_STATES     = 3;
    const DEFAULT_WRITE_WAIT_STATES_2   = 0;
    // and the real values, which can be changed at startup time by means of INI definitions
    int READ_WAIT_STATES;               // these will become equal to the
    int READ_WAIT_STATES_2;             // defaults if not otherwise specified
    int WRITE_WAIT_STATES;              // in the INI file or command line
    int WRITE_WAIT_STATES_2;            // switches
    const BUSY_LOCK_SLEEP_CYCLE_COUNT   = 1;
    bitsLOCK lock_value;
    int busy_lock_sleep_cycle_count;
    // statistics
    struct Stat {
        int req, reql;      // requests (locks)
        int rep;            // replacements
        int wbrd;           // write broadcasts
        int wupg;           // write updates
        Stat()  { req=reql=rep=wbrd=wupg=0;}
        void dump(int index);
        } stat;
  public:
    Dra_Cache();
    void assign_index(int ind) {index = ind;}
    void Introduce();
    void cycle();   
    bool isidle() {return (state == INITIAL);}
    void dump_state();
    void dump_stats() {stat.dump(index);}
    TOPOLOGY_FRIENDS
    };


#ifdef PRAGMA_IMPLEMENTATION_DRA_CACHE

// the Dragon cache controller module implementation

#include <iostream.h>
    
Dra_Cache::Dra_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    state = INITIAL;                            // PCC will be in the INITIAL state
    snoop_state = SNOOP_INITIAL;                // and so will the SCC part
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    in.BUS_GNT      = 0;
    out.BUS_REQ     = 0;
    in.LOCK_GNT     = 0;
    out.LOCK_REQ    = 0;
    in.bus.reset();
    out.bus.reset();
    // initialize memory delay parameters. If you want to change this,
    // create limes.ini in your app's directory and place in the line
    // r_ws=4 (for example), or use simulator's command line switches,
    // like -dw_ws=4.
    READ_WAIT_STATES    = sys_INI_read("r_ws", DEFAULT_READ_WAIT_STATES);
    READ_WAIT_STATES_2  = sys_INI_read("r_ws2", DEFAULT_READ_WAIT_STATES_2);
    WRITE_WAIT_STATES   = sys_INI_read("w_ws", DEFAULT_WRITE_WAIT_STATES);
    WRITE_WAIT_STATES_2 = sys_INI_read("w_ws2", DEFAULT_WRITE_WAIT_STATES_2);
    };

void Dra_Cache::Introduce() {
    COUT << "Protocol: Dragon (v1.0). ";
    #ifdef LOCK_LINE_AWARE
        COUT << "LOCK_line aware.";
    #endif
    COUT << "\n";
    COUT << "READ_WS=" << READ_WAIT_STATES << "/" << READ_WAIT_STATES_2 << ", ";
    COUT << "WRITE_WS=" << WRITE_WAIT_STATES << "/" << WRITE_WAIT_STATES_2 << ", ";
    COUT << "LOCK_WS=" << BUSY_LOCK_SLEEP_CYCLE_COUNT << "\n";
    #ifdef DUMMLOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    }

void Dra_Cache::Stat::dump(int index) {
    if (req) COUT << "Dra"<<index<<": " <<                                      \
        "req="<<req << "/"<<reql << ",rep="<<rep << 
        ",wbrd="<<wbrd << ",wupg="<<wupg 
        << "\n";
    }

// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
// This method first reads the bus. Then it checks if there is SNOOP signal
// high on it, and, if so, checks if it has the line to invalidate. If true,
// does so and skips this cycle. This reflects the fact that there are two
// controllers in this one, the Processor Cache Controller (PCC) and the
// Snoop Cache Controller (SCC), each of which has its own copy of cache
// tags, but the cache memory itself can be accessed only by one controller
// at a time, and the SCC part has precedence.

// For details about the Dragon protocol operation, please refer to the
// Limes documentation. Also, analyzing the WTI and Berkeley protocol 
// implementations before analyzing this code is suggested.

// Note that there are many similar elements for the Berkeley and Dragon
// controller classes; both could be derived from a superclass, but the code
// is easier to understand this way
    
void Dra_Cache::cycle() {

    Topology::cache_read_bus(index);                // read input from the bus

    out.bus.SH_LINE = 0;                            // clear SH_LINE
    switch (snoop_state) {
        // the SCC may be in initial state, or not. If it is in initial state,
        // it will check if a SNOOP signal appeared on the bus. Otherwise,
        // it will continue what it was doing in the previous cycle -- doing
        // a cache supply for some other cache controller.
        case SNOOP_INITIAL:
            if (in.bus.SNOOP && !in.BUS_GNT)    {   // !in.BUS_GNT means 'I don't have the bus'
                // there is: select frame #1 of cache_memory's C (current) vars,
                // which keep the current position of the line
                cache_memory.set_current_frame(1);  
                if (cache_memory.isvalid(in.bus.addr))
                    goto L_SNOOP;   // most of the SCC part is there
                else
                    cache_memory.set_current_frame(0);  // frame #0 for the processor part
                    // it was a snoop miss, so proceed as usual (though it might be more realistic to return)
                }
            break;          // activate processor cache controller
        case SNOOP_WAIT_1:
            goto L_STATE_SNOOP_WAIT_1;
        }
    
    // there was no snoop or it was a miss, so we go on
    switch (state) {
        case INITIAL                        : goto L_STATE_INITIAL;
        case WRITE_WAIT_BGNT                : goto L_STATE_WRITE_WAIT_BGNT;
        case WRITE_WAIT_2                   : goto L_STATE_WRITE_WAIT_2;
        case UNLOCK_WAIT_BGNT               : goto L_STATE_UNLOCK_WAIT_BGNT;
        case LOCK_WAIT_BGNT                 : goto L_STATE_LOCK_WAIT_BGNT;
        case LOCK_SLEEP                     : goto L_STATE_LOCK_SLEEP;
        case FETCH_WAIT_BGNT                : goto L_STATE_FETCH_WAIT_BGNT;
        case WB_WAIT_MEMORY                 : goto L_STATE_WB_WAIT_MEMORY;
        case FETCH_WAIT_SUPPLY_2            : goto L_STATE_FETCH_WAIT_SUPPLY_2;
        case FETCH_WAIT_MEMORY              : goto L_STATE_FETCH_WAIT_MEMORY;
        }
        
L_STATE_INITIAL:
    if (! Topology::cache_read_request(index)) return;  // return if there was none
    ++stat.req;                                         // statistics
    switch (in.request.type) {                          // select where to go
        case REQUEST_READ   : goto L_READ;
        case REQUEST_LOCK   : goto L_LOCK;
        case REQUEST_UNLOCK : goto L_UNLOCK;
        case REQUEST_WRITE  : goto L_WRITE;
        default : break;
        }
    return;
    

//---------------------------------------------------------------
L_READ:
    // OK, so we've got a READ. It will be a hit or a miss.
    if (cache_memory.isvalid(in.request.addr)) {
        // a hit, great! Say that the request is satisfied, and return
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_READ_end;
        }
    FIL_action = FIL_READ;      // these three lines are an equivalent for
    goto L_FETCH_INVALID_LINE;  // "gosub fetch_invalid_line"; but there can be
    L_FIL_READ_ret:             // no gosub when F.I.L. spans more than one cycle
L_READ_end:
    state = INITIAL;
    return;

//---------------------------------------------------------------
L_WRITE:
    // this is a WRITE. We'll say immediately that it is satisfied.
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    // and check whether it is in the cache, and if so, is it valid.
    // The isvalid() method will do all the work for us, and even pick
    // a line for replacement if necessary.
    if (cache_memory.isvalid(in.request.addr))
        goto L_WRITE_hit;
    // write miss
    FIL_action = FIL_WRITE;         // gosub L_FETCH_INVALID_LINE
    goto L_FETCH_INVALID_LINE;
    L_FIL_WRITE_ret:                // return point
    // now, the missing line is fetched. If someone else has it,
    // we'll have to broadcast the word. If not, we'll put the line
    // in the Modified, Exclusive state and leave
    if (bus_sh_line)
        goto L_WRITE_broadcast_word;
    else    {
        cache_memory.setstate(Dra_Line_State::MOD_EXC);
        goto L_WRITE_end;
        }
L_WRITE_hit:
    // Write hit is not that simple: if the line is in one of 'shared'
    // states, the new word will have to be broadcasted to the others
    switch (cache_memory.getstate()) {
        case Dra_Line_State::MOD_EXC    :
        case Dra_Line_State::UNMOD_EXC  :
            cache_memory.setstate(Dra_Line_State::MOD_EXC);
            goto L_WRITE_end;
        case Dra_Line_State::MOD_SHD    :
        case Dra_Line_State::UNMOD_SHD  :
            break;
        }
    out.BUS_REQ = 1;
    state = WRITE_WAIT_BGNT;
L_STATE_WRITE_WAIT_BGNT:
    if (! in.BUS_GNT) return;
L_WRITE_broadcast_word:
    // we were waiting for the bus, but the line couldn't be invalidated
    // (the Dragon does not have INVALID state), nor upgraded to Exclusive
    // meanwhile.
    ++stat.wbrd;
    out.bus.reset();
    out.bus.addr     = in.request.addr & (bitsADDR)~(DATA_WIDTH-1); // align addr
    cache_memory.isvalid(out.bus.addr);     // touch for read on the aligned address
    out.bus.size     = DATA_WIDTH;          
    out.bus.SNOOP    = 1;
    out.bus.memcmd   = REQUEST_WRITE;
    state = WRITE_WAIT_2;
    wait_state_counter = 1;
    return;
L_STATE_WRITE_WAIT_2:   
    out.bus.SNOOP    = 0;   // but I have to wait one more cycle to see SH
    if (wait_state_counter--) return;
    // the story is similar to the one for the Berekeley: first we waited
    // to get the bus, then we sent WRITE command onto it and waited one
    // cycle so that the others can see it, and than again one more cycle
    // so that we can see if any of them raised SH_LINE, indicating that
    // the line is shared. The new state will depend on the status of this
    // line
    if (in.bus.SH_LINE)
        cache_memory.setstate(Dra_Line_State::MOD_SHD);
    else
        cache_memory.setstate(Dra_Line_State::MOD_EXC);
L_WRITE_end:
    // That's all there is for a write. Drop the bus request(s) and return
    out.BUS_REQ = out.LOCK_REQ = 0;
    state = INITIAL;
    return;

//---------------------------------------------------------------
L_LOCK:
    // The lock story is exactly the same as the one for the Berkeley.
    ++stat.reql;
    #ifdef LOCK_LINE_AWARE
        out.LOCK_REQ = 1;
    #else
        out.BUS_REQ = 1;
    #endif
    state = LOCK_WAIT_BGNT;
L_STATE_LOCK_WAIT_BGNT:
    #ifdef LOCK_LINE_AWARE
        if (! in.LOCK_GNT) return;
    #else
        if (! in.BUS_GNT) return;
    #endif
    if (cache_memory.isvalid(in.request.addr)) goto L_LOCK_check_value;
    FIL_action = FIL_LOCK;
    goto L_FETCH_INVALID_LINE;
    L_FIL_LOCK_ret:
L_LOCK_check_value:
    lock_value = MEMSIM_READ_LOCK(in.request.addr);
    #ifndef DUMMLOCK
        if (lock_value == 0) {
            // Do the actual "write 1" to the location for the lock.
            MEMSIM_WRITE_LOCK( in.request.addr, 1); 
            goto L_WRITE;
            }
        out.BUS_REQ = out.LOCK_REQ = 0;
        busy_lock_sleep_cycle_count = BUSY_LOCK_SLEEP_CYCLE_COUNT;
        state = LOCK_SLEEP;
        return;
    #else
        MEMSIM_WRITE_LOCK( in.request.addr, 1); 
        out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_WRITE_hit;
    #endif  
L_STATE_LOCK_SLEEP:
    if (busy_lock_sleep_cycle_count--) return;
    goto L_LOCK;

//---------------------------------------------------------------
L_UNLOCK:
    // this is an UNLOCK (lock release) request. Basically it is a write,
    // only that it has to gain the bus first.
    out.LOCK_REQ = 1;
    state = UNLOCK_WAIT_BGNT;
L_STATE_UNLOCK_WAIT_BGNT:     
    #ifdef LOCK_LINE_AWARE
        if (!in.LOCK_GNT) return;
    #endif
    // 
    // do the actual "write 0" for the lock. Note: it would be a mistake
    // to do that BEFORE this point! Actually, it should be done somewhere
    // at L_WRITE_end -- that is where (when) it would happen in reality
    //
    MEMSIM_WRITE_LOCK( in.request.addr, 0);
    goto L_WRITE;


//---------------------------------------------------------------
L_FETCH_INVALID_LINE:
    // This routine is also very similar to the Berkeley's. Please refer to it.
    out.BUS_REQ = 1;
    state = FETCH_WAIT_BGNT;
L_STATE_FETCH_WAIT_BGNT:
    if (! in.BUS_GNT) return;
    switch (cache_memory.getstate()) {
        case Dra_Line_State::UNMOD_EXC:
        case Dra_Line_State::UNMOD_SHD: 
            goto L_FETCH_fetch_line;
        case Dra_Line_State::MOD_SHD:       // MODified lines
        case Dra_Line_State::MOD_EXC:       // must be written back
            break;
        }
    ++stat.rep;
    out.bus.reset();
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP    = 0;
    wait_state_counter = WRITE_WAIT_STATES;
    state = WB_WAIT_MEMORY;
    return;
L_STATE_WB_WAIT_MEMORY:
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
        }
L_FETCH_fetch_line:
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP    = 1;
    if (FIL_action == FIL_WRITE) out.bus.WR_MISS = 1;
    out.bus.memcmd   = REQUEST_READ;
    state = FETCH_WAIT_SUPPLY_2;
    wait_state_counter = 1;
    return;
L_STATE_FETCH_WAIT_SUPPLY_2:
    out.bus.SNOOP = 0; out.bus.WR_MISS = 0; 
    if (wait_state_counter--) return;
    state = FETCH_WAIT_MEMORY;  
    bus_sh_line  = in.bus.SH_LINE;      // remember for later
    cache_supply = in.bus.SUPPLY;
    if (!cache_supply)                  // no owner
        wait_state_counter = READ_WAIT_STATES - 2;  //-2 if memory is not inihibited
    else
        wait_state_counter = 0; 
L_STATE_FETCH_WAIT_MEMORY:
    if (wait_state_counter-- >0 ) return;
    if (out.bus.addr <= in.request.addr && 
        in.request.addr+in.request.size <= out.bus.addr + DATA_WIDTH) {
        // here we come if data for read/writing that made us do all this
        // belongs to this word
        switch (FIL_action) {
            case FIL_READ:
                out.response.signal = PROCESSOR_SATISFIED;
                out.response.STROBE = 1;
                break;
            case FIL_LOCK:
                break;
            case FIL_WRITE:
                break;
            }               
        }
    // write the data, get next addr and find out if there's more to do
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        if (cache_supply)
            wait_state_counter = 0;
        else
            wait_state_counter = READ_WAIT_STATES_2;
        return;
        }
    cache_memory.touch();   // force this line MRU
    // counting finished, the whole line fetched.
    // The new state will depend both on the action (read/write)
    // and on the value of SH_LINE (remembered at the beginnging).
    switch( FIL_action ) {
        case FIL_READ:
            out.BUS_REQ = 0;
            cache_memory.setstate(
                bus_sh_line ? Dra_Line_State::UNMOD_SHD 
                            : Dra_Line_State::UNMOD_EXC );
            goto L_FIL_READ_ret;
        case FIL_WRITE:
            goto L_FIL_WRITE_ret;
        case FIL_LOCK:
            cache_memory.setstate(
                bus_sh_line ? Dra_Line_State::UNMOD_SHD 
                            : Dra_Line_State::UNMOD_EXC );
            goto L_FIL_LOCK_ret;
        }
    return; // can't happen
    

//---------------------------------------------------------------
L_SNOOP:    // here we come on a snoop hit
    // here we come on a snoop hit. There are two reasons why someone wants
    // us to do snooping:
    // 1) he is writing something into his line and wants us to UPGRADE ours
    // 2) he wants to read the line and asks us to do the supply, if we are
    //    the owner. In addition, he may be having a write miss: he will get
    //    a copy of the line from us, and then ask us later to update ours.
    out.bus.SH_LINE = 1;                // say that you have this line
    if (in.bus.memcmd != REQUEST_READ) {
        // this was a write update. A real hardware would, at this point,
        // read the data from the bus and write it in its own cache
        // The state changes to Unmodified, Shared (Unmodified because we
        // loose the ownership, which is good, because we won't have to
        // write the line back upon a replacement)
        cache_memory.setstate(Dra_Line_State::UNMOD_SHD);
        ++stat.wupg;
        goto L_SNOOP_end;   
        }
    // this is a READ request. If we own the line (Modified), we'll do the supply
    switch (cache_memory.getstate()) {
        case Dra_Line_State::UNMOD_EXC:
        case Dra_Line_State::UNMOD_SHD: 
            cache_memory.setstate(Dra_Line_State::UNMOD_SHD);
            goto L_SNOOP_end;
        case Dra_Line_State::MOD_SHD:
        case Dra_Line_State::MOD_EXC:
            break;
        }
    bus_wr_miss = in.bus.WR_MISS;   // remember whether it is a write miss (for the requester)
    out.bus.SUPPLY = 1;             // we're the owner and we do the supply
    cache_memory.line_write_init(&out.bus.addr);
    snoop_state = SNOOP_WAIT_1;
    return;
    // what follows is the cache supply process
L_STATE_SNOOP_WAIT_1:
    if (cache_memory.line_write_getnext(&out.bus.addr)) 
        return;                     // return if not finished
    out.bus.SUPPLY = 0;             // supply finished
    cache_memory.setstate(
        bus_wr_miss ? Dra_Line_State::UNMOD_SHD 
                    : Dra_Line_State::MOD_SHD   );
L_SNOOP_end:
    // the snoop job is done, set cache_memory's Current Frame to #0 for the PCC
    cache_memory.set_current_frame(0);
    snoop_state = SNOOP_INITIAL;
    return;
    }

// debugging...

void Dra_Cache::dump_state() {
    COUT << "CC" << index << ": State: ";
    if (snoop_state != SNOOP_INITIAL) { 
        if (snoop_state == SNOOP_WAIT_1) 
            COUT << "SNOOP_WAIT_1";
        }
    else    {
        switch (state) {
            case INITIAL                : COUT << "INITIAL"; break;
            case WRITE_WAIT_BGNT        : COUT << "WRITE_WAIT_BGNT"; break;
            case WRITE_WAIT_2           : COUT << "WRITE_WAIT_2"; break;
            case UNLOCK_WAIT_BGNT       : COUT << "UNLOCK_WAIT_BGNT"; break;
            case LOCK_WAIT_BGNT         : COUT << "LOCK_WAIT_BGNT"; break;
            case LOCK_SLEEP             : COUT << "LOCK_SLEEP (" << busy_lock_sleep_cycle_count << ")"; break;
            case FETCH_WAIT_BGNT        : COUT << "FETCH_WAIT_BGNT"; break;
            case WB_WAIT_MEMORY         : COUT << "WB_WAIT_MEMORY (" << wait_state_counter << ")"; break;
            case FETCH_WAIT_SUPPLY_2    : COUT << "FETCH_WAIT_SUPPLY_2"; break;
            case FETCH_WAIT_MEMORY      : COUT << "FETCH_WAIT_MEMORY (" << wait_state_counter << ")"; break;
            }
        }

    COUT << "\t\t[";
    if (state != INITIAL) {
        switch (in.request.type) {  
            case REQUEST_READ  : COUT << "READ " << in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << in.request.addr; break;
            default : break;
            }
        }       
    COUT << "]\n";
    cache_memory.dump_state(index);
    }
                

#endif PRAGMA_IMPLEMENTATION_DRA_CACHE
#endif _DRA_CACHE_H_