#ifndef _IO_BUF_H_
#define _IO_BUF_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: io_buf.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This module isolates the other modules from the Limes kernel.
 * It reads the requests from the kernel, and leaves them on its
 * output ports. It reads the responses from its input ports and
 * transmits them to the kernel.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR

class IO_Buf {
    int index;                  // index of the belonging PE (processing element)
    bool request_pending;       // 1 (true) if yes, 0 (false) if not
    // output signals
    struct Out {
        // to the cache
        struct Request {
            struct Memory_Request *request_ptr;     // request description
            int STROBE;         // set by IO_Buf and cleared by Topology 
            } request;
        } out;
    struct In {
        // input signals: from the cache
        struct Response {
            enum PE_response signal;    // filled by topology when called
            } response;
        } in;
  public:
    IO_Buf();
    void assign_index(int ind)  { index = ind; }
    void start_cycle();             
    void end_cycle();
    bool isidle() {return !request_pending;}
    TOPOLOGY_FRIENDS
    };

#ifdef PRAGMA_IMPLEMENTATION_IO_BUF
//---------------------------------

#include <iostream.h>

IO_Buf::IO_Buf() {
    request_pending = false;
    index = 0;
    out.request.STROBE = 0;
    }

// each IO_Buf module has two main methods: start_cycle() and end_cycle().
// The Topology decides when to call each: at the beginning of the simulated
// cycle, start_cycle() is called for each IO_Buf instance if there is an
// indication that one or more requests exist at this cycle. Each IO_Buf
// checks it, reads the request from the kernel and leaves it on its output
// port, waiting for the cache to collect it. At the end of each simulated
// cycle, end_cycle() is called, which checks if there is any response from
// the underlying cache controller. The response is forwarded to the kernel.
    

void IO_Buf::start_cycle() {
    if (REQUEST_EXISTS(index)) {
        if (request_pending) {
            cerr << "\aIO_Buf[" << index << "] overrun error!\n"; exit(1); }
        REQUEST_FILL(index, out.request.request_ptr);

        // misalignment not allowed! GCC never makes one anyway.
        bitsADDR addr = out.request.request_ptr->virtual_addr;
        int size = out.request.request_ptr->size;
        int mask = ~(DATA_WIDTH-1);
        if ( (addr & mask) != ( (addr+size-1) & mask ) )  {
            cerr << "\aIO_Buf[" << index << "] alignment error!\n"; exit(1); }

        out.request.STROBE  = 1;
        request_pending = true;
        }
    }

void IO_Buf::end_cycle() {
    if (request_pending) {
        if (!Topology::io_buf_input(index))         // if there is no response
            REQUEST_RESPOND(index, PROCESSOR_STALLED);
        else    {
            // if there was a read from the CPU, it had waited till now
            REQUEST_RESPOND(index, in.response.signal);
            request_pending = false;
            }
        }
    else    // if there was no request, there will be no response
        REQUEST_RESPOND(index, PROCESSOR_NORESPONSE);       
    }   


#endif // PRAGMA_IMPLEMENTATION_IO_BUFF_
#endif //_IO_BUF_H_