#ifndef _WIN_CACHE_H_
#define _WIN_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: win_cache.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The Word INvalidate protocol; definition and implementation.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "bus_signals.h"

// states a line in the WIN protocol can be found in
class WIN_Line_State {
  public:
    enum State { 
        INVALID,                // Invalid (more than two words invalid)
        UNMOD_SHD,              // Unmodified, Shared
        UNMOD_EXC,              // Unmodified, Exclusive
        MOD_SHD,                // Modified, Shared
        MOD_EXC,                // Modified, Exclusive
        IW1,                    // One word invalid
        IW2                     // Two words invalid
        } state;
    // the WIN protocol maintains one 'valid' bit for each word in the line
    bits32 invalidity;          // only up to 32-word sized lines!
    WIN_Line_State()            { state = INVALID; }
    bool  isvalid()             { return state != INVALID; }
    State get()                 { return state; }
    void  set(State new_state)  { state = new_state; }
    char  symbol() {
        switch (state) {
            case INVALID  : return 'I';
            case UNMOD_SHD: return 's';
            case UNMOD_EXC: return 'e';
            case MOD_SHD  : return 'S';
            case MOD_EXC  : return 'E';
            case IW1      : return '1';
            case IW2      : return '2';
            }
        }
    };

// extended cache memory so that supports WIN 'word valid' bits in a line

class WIN_Cache_Memory : public Cache_Memory<WIN_Line_State> {
    unsigned int Cword() { return C.offset/DATA_WIDTH; }
  public:
    void setstate(WIN_Line_State::State new_state)  {
        Cache_Memory<WIN_Line_State>::setstate(new_state);
        if ( (new_state != WIN_Line_State::IW1) &&
             (new_state != WIN_Line_State::IW2) ) 
            line(C.set,C.line).state.invalidity = 0;    // clear invalidity
        }
    bool iswordvalid() {
        return ! ( (line(C.set,C.line).state.invalidity >> Cword()) & 1);
        }
    void validateword() {
        line(C.set,C.line).state.invalidity &= ~(1<<Cword());
        }
    void invalidateword() {
        line(C.set,C.line).state.invalidity |= (1<<Cword());
        }
    };           
    

// the WIN cache controler module declaration
class WIN_Cache {
    int index;
    WIN_Cache_Memory cache_memory;      // the whole cache memory
    struct In {
        struct Memory_Request request;  // from the CPU
        bool BUS_GNT;
        bool LOCK_GNT;
        Bus_Signals bus;        
        } in;
    struct Out {
        struct Response {               // to the CPU
            enum PE_response signal;
            bool STROBE;
            } response;
        bool BUS_REQ;
        bool LOCK_REQ;
        Bus_Signals bus;
        } out;
    enum State {
        INITIAL,
        READ_WAIT_BGNT,
        WRITE_WAIT_BGNT,
        WRITE_WAIT_2,
        UNLOCK_WAIT_BGNT,
        LOCK_WAIT_BGNT,
        LOCK_WAIT_BUS_GNT,
        LOCK_SLEEP,
        FETCH_WAIT_BGNT,
        WB_WAIT_MEMORY,
        FETCH_WAIT_SUPPLY_2,
        FETCH_WAIT_MEMORY,
        FIW_WAIT_SUPPLY_2,
        FIW_WAIT_MEMORY
        } state;
    enum Snoop_State {
        SNOOP_INITIAL,
        SNOOP_WAIT_1
        } snoop_state;
    // possible actions when fetching invalid LINE
    enum FIL_ACTIONS {FIL_READ, FIL_WRITE, FIL_LOCK} FIL_action;
    // possible actions when fetching invalid WORD
    enum FIW_ACTIONS {FIW_READ,            FIW_LOCK} FIW_action;
    const DEFAULT_READ_WAIT_STATES      = 3; // time needed to read the first word
    const DEFAULT_READ_WAIT_STATES_2    = 0; // time needed for access to subsequent words
    const DEFAULT_WRITE_WAIT_STATES     = 3;
    const DEFAULT_WRITE_WAIT_STATES_2   = 0;
    int READ_WAIT_STATES;               // these will become equal to the
    int READ_WAIT_STATES_2;             // defaults if not otherwise specified
    int WRITE_WAIT_STATES;              // in the INI file or command line
    int WRITE_WAIT_STATES_2;            // switches
    const BUSY_LOCK_SLEEP_CYCLE_COUNT   = 1;
    // auxillary vars
    bitsLOCK lock_value;
    int busy_lock_sleep_cycle_count;
    int wait_state_counter;
    bool cache_supply;
    bool bus_inv, bus_word, bus_sh_line;
    WIN_Line_State::State line_state;
    // statistics
    struct Stat {
        int req;        // requests
        int wh_IW1;     // write hits at IW1 line
        int wm_IW1;     // write miss at IW1 line
        int wms_IW1;    // write miss at IW1 line but for less then a word
        int upg2;       // upgrades from IW2 to IW1
        int upg1;       // upgrades from IW1 to clean
        int inv;        // invalidations
        int rep;        // replacements
        Stat()  { req=wh_IW1=wm_IW1=wms_IW1=upg2=upg1=inv=rep=0;}
        void dump(int index);
        } stat;
  public:
    WIN_Cache();
    void assign_index(int ind) {index = ind;}
    void cycle();   
    void dump_state();
    void Introduce();
    bool isidle() {return (state == INITIAL);}
    void dump_stats() {stat.dump(index);}
    TOPOLOGY_FRIENDS
    };


#ifdef PRAGMA_IMPLEMENTATION_WIN_CACHE

// the WIN cache controller module implementation

#include <iostream.h>
    
WIN_Cache::WIN_Cache() {
    index = 0;
    cache_memory.Initialize();
    state = INITIAL;
    snoop_state = SNOOP_INITIAL;
    out.response.signal = PROCESSOR_NORESPONSE;
    out.response.STROBE = 0;
    in.BUS_GNT      = 0;
    out.BUS_REQ     = 0;
    in.bus.reset();
    out.bus.reset();
    // initialize memory delay parameters
    READ_WAIT_STATES    = sys_INI_read("r_ws", DEFAULT_READ_WAIT_STATES);
    READ_WAIT_STATES_2  = sys_INI_read("r_ws2", DEFAULT_READ_WAIT_STATES_2);
    WRITE_WAIT_STATES   = sys_INI_read("w_ws", DEFAULT_WRITE_WAIT_STATES);
    WRITE_WAIT_STATES_2 = sys_INI_read("w_ws2", DEFAULT_WRITE_WAIT_STATES_2);
    };


void WIN_Cache::Introduce() {
    COUT << "Protocol: WIN (v1.0). ";
    #ifdef LOCK_LINE_AWARE  // lock_line
        COUT << "LOCK_line aware.";
    #endif
    COUT << "\n";
    COUT << "READ_WS=" << READ_WAIT_STATES << "/" << READ_WAIT_STATES_2 << ", ";
    COUT << "WRITE_WS=" << WRITE_WAIT_STATES << "/" << WRITE_WAIT_STATES_2 << ", ";
    COUT << "LOCK_WS=" << BUSY_LOCK_SLEEP_CYCLE_COUNT << "\n";
    #ifdef DUMMLOCK
        COUT << "Dumm locks: test&set instead of test&test&set.\n";
    #endif  
    cache_memory.Introduce();
    }


void WIN_Cache::Stat::dump(int index) {
    if (req) COUT << "WIN"<<index<<": " <<                                      \
        "req="<<req << ",wh_IW1="<<wh_IW1 << ",wm_IW1="<<wm_IW1 <<  
        ",wms_IW1="<<wms_IW1 << ",upg2="<<upg2 << ",upg1="<<upg1 << 
        ",inv="<<inv << ",rep="<<rep                                
        << "\n";
    }

// For details about the WIN protocol operation, please refer to 
// [ToM92] Tomasevic, M., Milutinovic, V., "A Simulation Study of Snoopy Cache 
// Coherence Protocols", Proceedings of the Hawaii International Conference 
// on System Sciences, pp. 426-436, Koloa, Hawaii, USA, January 1992
// Also, analyzing the WTI, Berkeley and Dragon protocol 
// implementations before analyzing this code is suggested.

void WIN_Cache::cycle() {
    
    Topology::cache_read_bus(index);                // read the bus inputs first

    out.bus.SH_LINE = 0;                            // clear SH_LINE
    switch (snoop_state) {
        case SNOOP_INITIAL:
            if (in.bus.SNOOP && !in.BUS_GNT)    {   // if a snoop and not mine
                cache_memory.set_current_frame(1);  // frame#1 for the snoopy part
                if (cache_memory.isvalid(in.bus.addr))
                    goto L_SNOOP;
                else
                    cache_memory.set_current_frame(0);  // frame #0 for the processor part
                }
            break;  // jump to PCC - processor cache controller
        case SNOOP_WAIT_1:
            goto L_STATE_SNOOP_WAIT_1;
        }
    
    // there was no snoop or it was a miss, so we go on
    switch (state) {
        case INITIAL                        : goto L_STATE_INITIAL;
        case READ_WAIT_BGNT                 : goto L_STATE_READ_WAIT_BGNT;
        case WRITE_WAIT_BGNT                : goto L_STATE_WRITE_WAIT_BGNT;
        case WRITE_WAIT_2                   : goto L_STATE_WRITE_WAIT_2;
        case UNLOCK_WAIT_BGNT               : goto L_STATE_UNLOCK_WAIT_BGNT;
        case LOCK_WAIT_BGNT                 : goto L_STATE_LOCK_WAIT_BGNT;
        case LOCK_WAIT_BUS_GNT              : goto L_STATE_LOCK_WAIT_BUS_GNT;
        case LOCK_SLEEP                     : goto L_STATE_LOCK_SLEEP;
        case FETCH_WAIT_BGNT                : goto L_STATE_FETCH_WAIT_BGNT;
        case WB_WAIT_MEMORY                 : goto L_STATE_WB_WAIT_MEMORY;
        case FETCH_WAIT_SUPPLY_2            : goto L_STATE_FETCH_WAIT_SUPPLY_2;
        case FETCH_WAIT_MEMORY              : goto L_STATE_FETCH_WAIT_MEMORY;
        case FIW_WAIT_SUPPLY_2              : goto L_STATE_FIW_WAIT_SUPPLY_2;
        case FIW_WAIT_MEMORY                : goto L_STATE_FIW_WAIT_MEMORY;
        }
        
L_STATE_INITIAL:
    if (! Topology::cache_read_request(index)) return;
    ++stat.req;
    switch (in.request.type) {
        case REQUEST_READ   : goto L_READ;
        case REQUEST_LOCK   : goto L_LOCK;
        case REQUEST_UNLOCK : goto L_UNLOCK;
        case REQUEST_WRITE  : goto L_WRITE;
        default : break;
        }
    return;


//---------------------------------------------------------------------
L_READ:
    if (! cache_memory.isvalid(in.request.addr)) {
        // line miss
        L_READ_call_FIL:
        FIL_action = FIL_READ;
        goto L_FETCH_INVALID_LINE;
        L_FIL_READ_ret:
        goto L_READ_end;
        }
    // line hit
    if (cache_memory.iswordvalid()) {
        // read hit
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_READ_end;
        }
    // line hit, but the read missed. Get the bus first.
    out.BUS_REQ = 1;
    state = READ_WAIT_BGNT;
L_STATE_READ_WAIT_BGNT:
    if (! in.BUS_GNT) return;
    if (cache_memory.getstate() == WIN_Line_State::INVALID)
        goto L_READ_call_FIL;   // the line's got invalidated completely meanwhile
    FIW_action = FIW_READ;
    goto L_FETCH_INVALID_WORD;
    L_FIW_READ_ret:
    out.BUS_REQ = 0;
L_READ_end:
    state = INITIAL;
    return;


//---------------------------------------------------------------
L_UNLOCK:
    out.LOCK_REQ = 1;
    state = UNLOCK_WAIT_BGNT;
L_STATE_UNLOCK_WAIT_BGNT:     
    #ifdef LOCK_LINE_AWARE
        if (!in.LOCK_GNT) return;
    #endif
    // do the actual write to the global coherent memory, and continue to WRITE
    MEMSIM_WRITE_LOCK( in.request.addr, 0);

//---------------------------------------------------------------
L_WRITE:
    out.response.signal = PROCESSOR_SATISFIED;
    out.response.STROBE = 1;
    if (! cache_memory.isvalid(in.request.addr))    {
        // line miss
        L_WRITE_call_FIL:
        FIL_action = FIL_WRITE;
        goto L_FETCH_INVALID_LINE;
        L_FIL_WRITE_ret:    // return point
        if (bus_sh_line)
            goto L_WRITE_invalidate_word;
        else    {
            cache_memory.setstate(WIN_Line_State::MOD_EXC);
            goto L_WRITE_end;
            }
        }
    // line hit. But more checks are to come.
L_WRITE_line_hit:
    switch (cache_memory.getstate()) {
        case WIN_Line_State::MOD_EXC    :
        case WIN_Line_State::UNMOD_EXC  :
            // the line is exclusively owned. It is safe to write to it.
            cache_memory.setstate(WIN_Line_State::MOD_EXC);
            goto L_WRITE_end;
        default:
            break;
        }
    out.BUS_REQ = 1;
    state = WRITE_WAIT_BGNT;
L_STATE_WRITE_WAIT_BGNT:
    if (! in.BUS_GNT) return;
    // first we have to check what is the state of the line.
    line_state = cache_memory.getstate();
    if (line_state == WIN_Line_State::INVALID)
        goto L_WRITE_call_FIL;  // got invalidated in the bgnt-waiting phase.
    // check the line's partial validity
    if (cache_memory.iswordvalid()) {
        // this word is valid. If some others are not, fetch the whole line.
        switch (line_state) {
            case WIN_Line_State::IW1:
                ++stat.wh_IW1;
            case WIN_Line_State::IW2:
                goto L_WRITE_call_FIL;
            default:
                // ok, the line is fully valid.
                break;
            }
        }
    else    {
        // this word is not valid. Is it the only one ?
        if (line_state == WIN_Line_State::IW1) {
            // yes, but are we writing a full word ?
            if (in.request.size < DATA_WIDTH)   {
                ++stat.wms_IW1;
                goto L_WRITE_call_FIL;  // no. Go and fetch the line.
                }
            ++stat.wm_IW1;
            }
        else
            goto L_WRITE_call_FIL;  // not the only one. Get the line.
        }
L_WRITE_invalidate_word:
    out.bus.reset();
    out.bus.addr     = in.request.addr; 
    out.bus.SNOOP    = 1;
    out.bus.INV      = 1;
    out.bus.WORD     = 1;
    state = WRITE_WAIT_2;
    wait_state_counter = 1;
    return;
L_STATE_WRITE_WAIT_2:   
    out.bus.SNOOP    = 0;   // but I have to wait one more cycle to check SH
    if (wait_state_counter--) return;
    if (in.bus.SH_LINE)
        cache_memory.setstate(WIN_Line_State::MOD_SHD);
    else
        cache_memory.setstate(WIN_Line_State::MOD_EXC);
L_WRITE_end:
    out.BUS_REQ = out.LOCK_REQ = 0;
    state = INITIAL;
    return;


//---------------------------------------------------------------
L_LOCK:
    #ifdef LOCK_LINE_AWARE
        out.LOCK_REQ = 1;
    #else
        out.BUS_REQ = 1;
    #endif
    state = LOCK_WAIT_BGNT;
L_STATE_LOCK_WAIT_BGNT:
    #ifdef LOCK_LINE_AWARE
        if (! in.LOCK_GNT) return;
    #else
        if (! in.BUS_GNT) return;
    #endif
    if (cache_memory.isvalid(in.request.addr)) {
        if (cache_memory.iswordvalid()) {
            goto L_LOCK_check_value;
            }
        else    
            goto lock_fetch_word;
        }
    else    
        goto lock_fetch_line;
lock_fetch_word:
    // the line is valid, but the word we need is not.
    // if LOCK_LINE_AWARE, we don't have the bus yet
    out.BUS_REQ = 1;
    state = LOCK_WAIT_BUS_GNT;
L_STATE_LOCK_WAIT_BUS_GNT:
    if (! in.BUS_GNT) return;
    // now we are sure we have the bus. But the line may have become invalid meanwhile.
    if (cache_memory.isvalid(in.request.addr)) {    // still valid
        FIW_action = FIW_LOCK;
        goto L_FETCH_INVALID_WORD;
        L_FIW_LOCK_ret:
        goto L_LOCK_check_value;
        }
    // sorry, we have lost the line.
lock_fetch_line:    // the whole line is invalid and needs to be fetched
    FIL_action = FIL_LOCK;
    goto L_FETCH_INVALID_LINE;
    L_FIL_LOCK_ret:
    
L_LOCK_check_value:
    lock_value = MEMSIM_READ_LOCK(in.request.addr);
    #ifndef DUMMLOCK
        if (lock_value == 0) {
            /* Do the actual "write 1" to the location for the lock. */
            MEMSIM_WRITE_LOCK( in.request.addr, 1); 
            goto L_WRITE;
            }
        out.BUS_REQ = out.LOCK_REQ = 0;
        busy_lock_sleep_cycle_count = BUSY_LOCK_SLEEP_CYCLE_COUNT;
        state = LOCK_SLEEP;
        return;
    #else
        MEMSIM_WRITE_LOCK( in.request.addr, 1); 
        out.response.signal = lock_value ? PROCESSOR_LOCK_BUSY : PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        goto L_WRITE_line_hit;
    #endif  
L_STATE_LOCK_SLEEP:
    if (busy_lock_sleep_cycle_count--) return;
    goto L_LOCK;


//---------------------------------------------------------------
L_FETCH_INVALID_LINE:
    out.BUS_REQ = 1;
    state = FETCH_WAIT_BGNT;
L_STATE_FETCH_WAIT_BGNT:
    if (! in.BUS_GNT) return;
    switch (cache_memory.getstate()) {
        case WIN_Line_State::MOD_SHD:
        case WIN_Line_State::MOD_EXC:
            break;
        default:
            goto L_FETCH_fetch_line;
        }
    ++stat.rep;
    out.bus.reset();
    cache_memory.line_write_init(&out.bus.addr);
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP    = 0;
    wait_state_counter = WRITE_WAIT_STATES;
    state = WB_WAIT_MEMORY;
    return;
L_STATE_WB_WAIT_MEMORY:
    if (wait_state_counter--) return;
    if (cache_memory.line_write_getnext(&out.bus.addr)) {
        wait_state_counter = WRITE_WAIT_STATES_2;   // there are more words
        return;
        }
    // ok, the whole line is written back. let's see now who can supply me
    // with the one I need.
L_FETCH_fetch_line:
    out.bus.reset();
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP    = 1;
    out.bus.memcmd   = REQUEST_READ;
    state = FETCH_WAIT_SUPPLY_2;
    wait_state_counter = 1;
    return;
L_STATE_FETCH_WAIT_SUPPLY_2:
    out.bus.SNOOP = 0;
    if (wait_state_counter--) return;
    state = FETCH_WAIT_MEMORY;      // this will loop <number of words> times
    bus_sh_line  = in.bus.SH_LINE;  // remember for later
    cache_supply = in.bus.SUPPLY;
    if (!cache_supply)              // no owner
        wait_state_counter = READ_WAIT_STATES - 2;  //-2 if memory's not inhibited
    else    
        wait_state_counter = 0; // no wait: i'll be taking data cycle after cycle
L_STATE_FETCH_WAIT_MEMORY:
    if (wait_state_counter-- >0 ) return;
    // now the data is in in.data, whether it was mem or cache supply
    if (out.bus.addr <= in.request.addr && 
        in.request.addr+in.request.size <= out.bus.addr + DATA_WIDTH) {
        // here we come if data for read/writing that made us do all this
        // belongs to this word
        switch (FIL_action) {
            case FIL_READ:
                out.response.signal = PROCESSOR_SATISFIED;
                out.response.STROBE = 1;
                break;
            case FIL_LOCK:
                break;
            case FIL_WRITE: 
                break;
            }               
        }
    // write the data, get next addr and find out if there's more to do
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        if (cache_supply)
            wait_state_counter = 0;
        else
            wait_state_counter = READ_WAIT_STATES_2;
        return;
        }
    cache_memory.touch();   // force this line MRU
    // counting finished, the whole line fetched
    switch( FIL_action ) {
        case FIL_READ:
            out.BUS_REQ = 0;
            cache_memory.setstate(
                bus_sh_line ? WIN_Line_State::UNMOD_SHD 
                            : WIN_Line_State::UNMOD_EXC );
            goto L_FIL_READ_ret;
        case FIL_WRITE:
            goto L_FIL_WRITE_ret;
        case FIL_LOCK:
            cache_memory.setstate(
                bus_sh_line ? WIN_Line_State::UNMOD_SHD 
                            : WIN_Line_State::UNMOD_EXC );
            goto L_FIL_LOCK_ret;
        }
    return; // can't happen


//---------------------------------------------------------------
L_FETCH_INVALID_WORD:
    out.bus.reset();
    out.bus.addr     = in.request.addr & (bitsADDR)~(DATA_WIDTH-1); // align addr
    cache_memory.isvalid(out.bus.addr);      // touch for the write to alligned addr
    out.bus.memcmd   = REQUEST_READ;
    out.bus.size     = DATA_WIDTH;
    out.bus.SNOOP    = 1;
    out.bus.WORD     = 1;
    state = FIW_WAIT_SUPPLY_2;
    wait_state_counter = 1;
    return;
L_STATE_FIW_WAIT_SUPPLY_2:
    out.bus.SNOOP = 0; 
    if (wait_state_counter--) return;
    bus_sh_line      = in.bus.SH_LINE;     
    cache_supply     = in.bus.SUPPLY;
    if (!cache_supply) {       
        state = FIW_WAIT_MEMORY;
        wait_state_counter = READ_WAIT_STATES - 2;      //-2 if mem is not inhibited
        }
    else
        wait_state_counter = 0; // no waiting: the data is here already
L_STATE_FIW_WAIT_MEMORY:
    if (wait_state_counter-- >0 ) return;
    switch (cache_memory.getstate()) {
        case WIN_Line_State::IW2:
            cache_memory.validateword();
            cache_memory.setstate(WIN_Line_State::IW1); 
            ++stat.upg2;
            break;
        case WIN_Line_State::IW1:
            ++stat.upg1;
        default:
            cache_memory.setstate(bus_sh_line ?
                WIN_Line_State::UNMOD_SHD :
                WIN_Line_State::UNMOD_EXC);
            break;
        }
    // now satisfy the caller
    switch (FIW_action) {
        case FIW_READ:
            out.response.signal = PROCESSOR_SATISFIED;
            out.response.STROBE = 1;
            goto L_FIW_READ_ret;
        case FIW_LOCK:
            goto L_FIW_LOCK_ret;
        }
    return;     // can't happen


//--------------------------------------------------------------------------
L_SNOOP:    // here we come on a snoop hit
    bus_inv  = in.bus.INV;
    bus_word = in.bus.WORD;
    out.bus.SH_LINE = 1;
    line_state = cache_memory.getstate();   // temp var, to speed up checking
    if (bus_inv) {
        // a word is being invalidated
        if (! cache_memory.iswordvalid())
            goto L_SNOOP_end;   // we don't have it, so it's of no concern to us
        if (line_state == WIN_Line_State::IW2) {    // try ::IW1 here for 1-word-limit
            // this one's got too polluted. Kill it.
            cache_memory.setstate(WIN_Line_State::INVALID); ++stat.inv;
            out.bus.SH_LINE = 0;        // this line is no longer shared
            goto L_SNOOP_end;
            }
        cache_memory.invalidateword();  // current word, of course
        // degrade the state of this line
        if (line_state == WIN_Line_State::IW1)
            cache_memory.setstate(WIN_Line_State::IW2);
        else
            cache_memory.setstate(WIN_Line_State::IW1);
        goto L_SNOOP_end;
        }
    // ok, we're here, this is not an invalidation, but a read request (v1.0)
    switch (line_state) {
        case WIN_Line_State::UNMOD_SHD:
        case WIN_Line_State::UNMOD_EXC:
            // the line is fully valid, but not owned, so set it to UNMOD_SHD
            cache_memory.setstate(WIN_Line_State::UNMOD_SHD);
            goto L_SNOOP_end;
        case WIN_Line_State::MOD_SHD:
        case WIN_Line_State::MOD_EXC:
            // or if the line is owned, we'll have to do the supply job
            break;
        default:
            // otherwise it's dirty, so we don't change state nor do the supply
            goto L_SNOOP_end;
        }
    out.bus.SUPPLY = 1;
    if (!bus_word)  // he wants the whole line
        cache_memory.line_write_init(&out.bus.addr);
    snoop_state = SNOOP_WAIT_1;
    return;
L_STATE_SNOOP_WAIT_1:
    if (!bus_word && 
        cache_memory.line_write_getnext(&out.bus.addr)) 
        return; // i have fetched the next data and i return if not finished
    out.bus.SUPPLY = 0; // supply finished
    cache_memory.setstate(WIN_Line_State::MOD_SHD); //!!! what if was inv...?
L_SNOOP_end:
    cache_memory.set_current_frame(0);  // restore Current's for the processor part
    snoop_state = SNOOP_INITIAL;
    return;
    }
            


void WIN_Cache::dump_state() {
    COUT << "CC" << index << ": State: ";
    if (snoop_state != SNOOP_INITIAL) { 
        if (snoop_state == SNOOP_WAIT_1) 
            COUT << "SNOOP_WAIT_1";
        }
    else    {
        switch (state) {
            case INITIAL            : COUT << "INITIAL"; break;
            case WRITE_WAIT_BGNT    : COUT << "WRITE_WAIT_BGNT"; break;
            case READ_WAIT_BGNT     : COUT << "READ_WAIT_BGNT"; break;
            case WRITE_WAIT_2       : COUT << "WRITE_WAIT_2"; break;
            case LOCK_WAIT_BGNT     : COUT << "LOCK_WAIT_BGNT"; break;
            case LOCK_WAIT_BUS_GNT  : COUT << "LOCK_WAIT_BUS_GNT"; break;
            case UNLOCK_WAIT_BGNT   : COUT << "UNLOCK_WAIT_BGNT"; break;
            case LOCK_SLEEP         : COUT << "LOCK_SLEEP (" << busy_lock_sleep_cycle_count << ")"; break;
            case FETCH_WAIT_BGNT    : COUT << "FETCH_WAIT_BGNT"; break;
            case WB_WAIT_MEMORY     : COUT << "WB_WAIT_MEMORY (" << wait_state_counter << ")"; break;
            case FETCH_WAIT_SUPPLY_2: COUT << "FETCH_WAIT_SUPPLY_2"; break;
            case FETCH_WAIT_MEMORY  : COUT << "FETCH_WAIT_MEMORY (" << wait_state_counter << ")"; break;
            case FIW_WAIT_SUPPLY_2  : COUT << "FIW_WAIT_SUPPLY_2"; break;
            case FIW_WAIT_MEMORY    : COUT << "FIW_WAIT_MEMORY (" << wait_state_counter << ")"; break;
            }
        }

    // print what request you are processing
    COUT << "\t\t[";
    if (state != INITIAL) {
        switch (in.request.type) {  
            case REQUEST_READ  : COUT << "READ " << in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << in.request.addr; break;
            default : break;
            }
        }       
    COUT << "]\n";
    cache_memory.dump_state(index);
    }
                

#endif // PRAGMA_IMPLEMENTATION_WIN_CACHE
#endif // _WIN_CACHE_H_