#ifndef _WTI_CACHE_H_
#define _WTI_CACHE_H_

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: wti_cache.h
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * The WTI (Write-through Invalidate) protocol. Hint: the simplest of all.
 * It is included mostly for educational purposes.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "snoopy.h"
#include TOPOLOGY_HDR
#include "cache_memory.h"
#include "bus_signals.h"

// states a line in the WTI protocol can be found in
class WTI_Line_State {
  public:
    enum State { INVALID, VALID } state;
    WTI_Line_State()                {state = INVALID;}
    bool  isvalid()                 {return state == VALID;}
    State get()                     {return state;}
    void  set(State new_state)      {state = new_state; }
    char  symbol()                  {return state == VALID ? 'V' : 'I';}
    };

// the WTI cache controler module declaration
class WTI_Cache {
    int index;                          // index of the belonging processor
    // instantiate cache memory with WTI_Line_State as its Line_State
    Cache_Memory<WTI_Line_State> cache_memory;  
    // port definitions
    struct In {                         // Input port:
        struct Memory_Request request;  // from the CPU
        bool BUS_GNT;                   // bus grant (from the bus module)
        bool LOCK_GNT;                  // ignored (for compatibility with other protocols)
        Bus_Signals bus;                // bus contents, to be read from the bus
        } in;
    struct Out {                        // Output port
        struct Response {               // to the CPU
            enum PE_response signal;    // signal to the IO_Buf: PROCESSOR_SATISFIED, PROCESSOR_STALLED etc.
            bool STROBE;                // set to high when 'signal' is formed
            } response;
        bool BUS_REQ;                   // bus request: sent to the bus module
        bool LOCK_REQ;                  // ignored 
        Bus_Signals bus;                // bus contents, to be written on the bus
        } out;
    // states this finite state machine called WTI can be found in
    enum State {
        INITIAL,                        // INITIAL -- not doing anything
        READ_WAIT_BGNT,                 // waiting for bus to do a READ
        READ_WAIT_MEMORY,               // waiting for the memory to complete
        WRITE_WAIT_BGNT,                // waiting for bus to do a WRITE
        WRITE_WAIT_MEMORY,              // waiting for the memory to complete
        LOCK_WAIT_BGNT,                 // waiting for bus to do a LOCL
        LOCK_READ_WAIT_MEMORY,          // waiting for bus do read lock value
        LOCK_SLEEP                      // sleeping, simulates spin delay
        } state;
    // some internal vars
    bool snoop_hit;                     // true for a snoop hit
    int wait_state_counter;
    const READ_WAIT_STATES              = 2;
    const WRITE_WAIT_STATES             = 2;
    const BUSY_LOCK_SLEEP_CYCLE_COUNT   = 1;
    int busy_lock_sleep_cycle_count;
    // statistics
    struct Stat {
        int req;                        // requests
        int inv;                        // invalidations
        Stat()  { req=inv=0;}
        void dump(int index);
        } stat;
  public:                               // declaration of methods
    WTI_Cache();                        // ctor
    void assign_index(int ind) {        // index is assigned by Topology
        index = ind; }  
    void Introduce();                   // called at startup
    void cycle();                       // called at each simulated cycle
    bool isidle() {                     // false if not in INITIAL state
        return (state == INITIAL); }    
    void dump_state();                  // for debugging
    void dump_stats() {                 // dump statistics at end
        stat.dump(index); }     
    TOPOLOGY_FRIENDS                    // Topology class is WTI's friend
    };


#ifdef PRAGMA_IMPLEMENTATION_WTI_CACHE

// the WTI cache controller module implementation

#include <iostream.h>
    
WTI_Cache::WTI_Cache() {                        // ctor
    index = 0;                                  // default, but will be changed
    cache_memory.Initialize();                  // initialize cache memory
    state = INITIAL;                            // go to the INITIAL state
    out.response.signal = PROCESSOR_NORESPONSE; // clear input and output ports
    out.response.STROBE = 0;
    in.BUS_GNT = 0;
    out.BUS_REQ = 0;
    in.bus.reset();
    out.bus.reset();
    };

void WTI_Cache::Introduce() {
    COUT << "Protocol: WTI (v1.0). ";
    COUT << "\n";
    COUT << "READ_WS=" << READ_WAIT_STATES << ", ";
    COUT << "WRITE_WS=" << WRITE_WAIT_STATES << ", ";
    COUT << "LOCK_WS=" << BUSY_LOCK_SLEEP_CYCLE_COUNT << "\n";
    cache_memory.Introduce();
    }

void WTI_Cache::Stat::dump(int index) {
    if (req) COUT << "WTI"<<index<<": " <<
        "req="<<req << ",inv="<<inv << "\n";
    }

// the main method: Topology calls it in each simulated cycle. If it was not
// in initial state, it will resume from the point where it left off in the
// previous cycle.
// This method first reads the bus. Then it checks if there is SNOOP signal
// high on it, and, if so, checks if it has the line to invalidate. If true,
// does so and skips this cycle. This reflects the fact that there are two
// controllers in this one, the Processor Cache Controller (PCC) and the
// Snoop Cache Controller (SCC), each of which has its own copy of cache
// tags, but the cache memory itself can be accessed only by one controller
// at a time, and the SCC part has precedence.

// A few words about this protocol: WTI means 'write-through, invalidate'.
// That means that each write is write-through, i.e. straight to the memory
// (and also to the cache if we have the line in it), while instructing
// all the other caches to invalidate their own copy of the line, if they
// have one. On a read miss, the whole line will be fetched from the memory.
    
void WTI_Cache::cycle() {
    
    Topology::cache_read_bus(index);        // read the bus contents first

    // check if there is SNOOP signal on the bus, but issued by a cache
    // other than myself:

    if (in.bus.SNOOP && !in.BUS_GNT) {      // !in.BUS_GNT means 'I don't have the bus'
        // there is: select frame #1 of cache_memory's C (current) vars,
        // which keep the current position of the line
        cache_memory.set_current_frame(1);  
        // check if the address is 'valid' -- there is a line, and is not INVALID
        snoop_hit = cache_memory.isvalid(in.bus.addr);
        if (snoop_hit) {
            // if so, force this line INVALID:
            cache_memory.setstate(WTI_Line_State::INVALID);
            cache_memory.untouch();         // force invalid line MRU
            ++stat.inv;                     // statistics: increment the invalidation count
            }
        // now restore Current frame #0 for use with the PCC part
        cache_memory.set_current_frame(0);  
        if (snoop_hit) return; // if this was invalidate HIT, skip this cycle
        // (actually, it should skip this cycle anyway; try it to see how
        // it affects the performance.)
        }
        
    // OK, there was no snoop or there was a snoop miss, so we continue.
    // We might or might not be in INITIAL state. If we are, we'll check
    // if there's a request for us from our processor. If we are not, that
    // means we continue from where we stopped in the previous cycle, and
    // we'll use 'goto' to resume.
    
    switch (state) {
        case INITIAL                : goto STATE_INITIAL;
        case READ_WAIT_BGNT         : goto STATE_READ_WAIT_BGNT;
        case READ_WAIT_MEMORY       : goto STATE_READ_WAIT_MEMORY;
        case WRITE_WAIT_BGNT        : goto STATE_WRITE_WAIT_BGNT;
        case WRITE_WAIT_MEMORY      : goto STATE_WRITE_WAIT_MEMORY;
        case LOCK_WAIT_BGNT         : goto STATE_LOCK_WAIT_BGNT;
        case LOCK_READ_WAIT_MEMORY  : goto STATE_LOCK_READ_WAIT_MEMORY;
        case LOCK_SLEEP             : goto STATE_LOCK_SLEEP;
        default                     : break;
        }
    
STATE_INITIAL:  // check if there is a request
    if (! Topology::cache_read_request(index)) return;  // not as such
    ++stat.req;                                         // statistics
    // much of the following should be self-explanatory:
    switch (in.request.type) {
        case REQUEST_READ : 
            // first, check if there is a line in the cache
            if (cache_memory.isvalid(in.request.addr))  {
                // yes. There is nothing else to be done, so tell the processor
                // that its request was satisfied
                out.response.signal = PROCESSOR_SATISFIED;
                out.response.STROBE = 1;
                state = INITIAL;    // and set state to INITIAL (was already)
                return;             // nothing else to be done in this cycle
                }
            // no, the line is not in the cache, or is invalid
            out.BUS_REQ = 1;        // tell the bus that we need it and set
            state = READ_WAIT_BGNT; // 'state' so that in the next cycle we'll
            return;                 // continue from STATE_READ_WAIT_BGNT
        case REQUEST_UNLOCK:        // will fall down to the 'WRITE' case
            MEMSIM_WRITE_LOCK( in.request.addr, 0); // do the actual write to the memory
        case REQUEST_WRITE :
            // a write is always satisfied immediately
            out.response.signal = PROCESSOR_SATISFIED;
            out.response.STROBE = 1;
            out.BUS_REQ = 1;        // we need the bus to write the word through
            state = WRITE_WAIT_BGNT;
            return;
        case REQUEST_LOCK :
            out.BUS_REQ = 1;        // the bus is used as a binary semafore
            state = LOCK_WAIT_BGNT; // for accessing lock variables
            return;
        default: break;
        }
    return;         
                

STATE_READ_WAIT_BGNT:   // here we come if we are in state READ_WAIT_BGNT
    // don't forget that all the inputs from the bus have been read already
    if (!in.BUS_GNT) return;    // return if the grant signal has not come yet
    // initialize cache memory line for 'fetching' from the memory: this
    // method operates on the current line; by calling isvalid(), we also
    // instructed the cache memory module to select an appropriate line for
    // replacement, which did so by inspecting in.request.addr's tag and
    // checking LRU info in the belonging set. So we initialize the cache
    // memory for this operation, and it will return the addr of the first
    // word in the line. Therefore, we don't have to know anything at all 
    // about the cache geometry.
    cache_memory.line_fetch_init(&out.bus.addr);
    out.bus.memcmd = REQUEST_READ;  // a 'READ' cmd will be seen on the bus
    out.bus.size   = DATA_WIDTH;    // this isn't really necessary, but...
    out.bus.SNOOP  = 0;             // no snooping and no INV; we just want 
    out.bus.INV    = 0;             // to fetch the line from the memory
    // now prepare to wait READ_WAIT_STATES cycles for the memory to finish.
    wait_state_counter = READ_WAIT_STATES;
    state = READ_WAIT_MEMORY;
    return;
    
STATE_READ_WAIT_MEMORY:
    if (wait_state_counter--) return;
    // ok, the slow memory has completed. Note that other protocols use
    // more sophisticated scheme, where the # of wait states for subsequent 
    // words may differ from the number of wait states for the first word

    // check if the current address matched the one from the request
    if (out.bus.addr <= in.request.addr && 
        in.request.addr+in.request.size <= out.bus.addr + DATA_WIDTH) {
        // yes, send this word that has just been read to the processor
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;
        }
    // cache memory hides the line size from us: so, we'll repeatedly call
    // .line_fetch_putnext() method, which will return the address of the
    // next word, and false if there are no more words in the line
    if (cache_memory.line_fetch_putnext(&out.bus.addr)) {
        wait_state_counter = READ_WAIT_STATES;  // more words to come
        return;
        }
    // counting done, whole line fetched
    // set new state to VALID
    cache_memory.setstate(WTI_Line_State::VALID);
    // drop the bus request. The bus module will in turn drop our 
    // BUS_GNT input signal immediately
    out.BUS_REQ = 0;
    state = INITIAL;    // return to the initial state.
    return;
        
STATE_WRITE_WAIT_BGNT:  
    if (!in.BUS_GNT) return;
    // a real cache would now do a "write" in its memory
    // send the bus the information for both the main memory and other 
    // cache controllers:
    out.bus.memcmd   = REQUEST_WRITE;
    out.bus.addr     = in.request.addr;
    out.bus.size     = in.request.size;
    out.bus.SNOOP    = 1;                       // SNOOP = high
    out.bus.INV      = 1;                       // INVALIDATE = high
    wait_state_counter = WRITE_WAIT_STATES;     // now wait for the memory
    state = WRITE_WAIT_MEMORY;                  // to complete
    return;
                    
STATE_WRITE_WAIT_MEMORY:
    out.bus.SNOOP    = 0;   // do not activate snoop controllers
    if (wait_state_counter--) return;
    out.BUS_REQ = 0;
    state = INITIAL;
    return;

    // This cache controller is implemented so that it knows about locks,
    // and performs them as TEST&TEST&SET instead of just TEST&SET. 
    // TEST&SET could be accomplished by gaining exclusive access to the
    // bus, reading the location, writing non-zero to it, freeing the bus,
    // and returning the old value of the given location. TEST&TEST&SET
    // immits the third step (writing) if it founds a non-zero value in
    // the location, thereby saving possibly a lot of invalidations.
    // (Berkeley and Dragon use special arbitrator for locks so that the
    // bus can be used for ordinary read and writes unobstructedly.
    // If the controller finds a non-zero value on the lock address, it
    // could either return PROCESSOR_LOCK_BUSY response, and let the simulator
    // spin by calling the cache again and again, or it can spin itself
    // thereby making the simulation a bit faster. On the system where
    // each process resides on a different processor, there is no difference.
    // between the two approaches, so the other one is used, though it does
    // not reflect the reality. But the results will be the same.
    
STATE_LOCK_WAIT_BGNT:   
    if (!in.BUS_GNT) return;
    // check if it is in the cache
    if (cache_memory.isvalid(in.request.addr)) 
        goto check_lock_value;
    // ok, there is not, read it from the memory
    out.bus.memcmd   = REQUEST_READ;
    out.bus.addr     = in.request.addr;
    out.bus.size     = sizeof(bitsLOCK);
    out.bus.SNOOP    = 0;
    out.bus.INV      = 0;
    wait_state_counter = READ_WAIT_STATES;
    state = LOCK_READ_WAIT_MEMORY;
    return;
        
STATE_LOCK_READ_WAIT_MEMORY:
    if (wait_state_counter--) return;

check_lock_value:
    if (MEMSIM_READ_LOCK(in.request.addr) == 0) {   // lock free
        out.response.signal = PROCESSOR_SATISFIED;
        out.response.STROBE = 1;        
        // Do the actual "write 1" to the lock location in 
        // real (application's) memory
        MEMSIM_WRITE_LOCK( in.request.addr, 1); 
        // go to the regular WRITE to do the rest of the job
        goto STATE_WRITE_WAIT_BGNT;     // [we do have GNT already]
        }
    // the lock is busy; there is nothing to be written. drop the bus and sleep
    out.BUS_REQ = 0;
    busy_lock_sleep_cycle_count = BUSY_LOCK_SLEEP_CYCLE_COUNT;
    state = LOCK_SLEEP;
    return;
    
STATE_LOCK_SLEEP:
    if (busy_lock_sleep_cycle_count--) return;
    out.BUS_REQ = 1;
    state = LOCK_WAIT_BGNT;         // start again
    return;
    }   

// This method is used for debugging. It is used primarily when the simulator
// is first developed as DETACHED.

void WTI_Cache::dump_state() {
    COUT << "CC" << index << ": State: ";
    switch (state) {
        case INITIAL                : COUT << "INITIAL"; break;
        case READ_WAIT_BGNT         : COUT << "READ_WAIT_BGNT"; break;
        case READ_WAIT_MEMORY       : COUT << "READ_WAIT_MEMORY (" << wait_state_counter << ")"; break;
        case WRITE_WAIT_BGNT        : COUT << "WRITE_WAIT_BGNT"; break;
        case WRITE_WAIT_MEMORY      : COUT << "WRITE_WAIT_MEMORY (" << wait_state_counter << ")"; break;
        case LOCK_WAIT_BGNT         : COUT << "LOCK_WAIT_BGNT"; break;
        case LOCK_READ_WAIT_MEMORY  : COUT << "LOCK_READ_WAIT_MEMORY (" << wait_state_counter << ")"; break;
        case LOCK_SLEEP             : COUT << "LOCK_SLEEP (" << busy_lock_sleep_cycle_count << ")"; break;
        }

    COUT << "\t\t[";
    if (state != INITIAL) {
        switch (in.request.type) {      
            case REQUEST_READ  : COUT << "READ " << in.request.addr; break;
            case REQUEST_WRITE : COUT << "WRIT " << in.request.addr; break;
            case REQUEST_LOCK  : COUT << "LOCK " << in.request.addr; break;
            case REQUEST_UNLOCK: COUT << "UNLK " << in.request.addr; break;
            default : break;
            }
        }       
    COUT << "]\n";
    cache_memory.dump_state(index);
    }
                

#endif PRAGMA_IMPLEMENTATION_WTI_CACHE
#endif _WTI_CACHE_H_