%option noyywrap

%{
/*
 * A scanner for analyzing processor events described in a plain text
 * file and converting it to MEMSIM requests.
 */

#include <ctype.h>
#include "memsim_interface.h"

static int *time;
static int curproc;
static struct Memory_Request *request;

static void initreqs() {
    int i;
    for(i=0; i<MAX_CPU_COUNT; ++i) {
        request[i].type = REQUEST_NONE;
        request[i].requested_now = 0;
        request[i].pid = i;
        request[i].data = -1;
        request[i].isshared = 1;
        request[i].satisfied = PROCESSOR_NORESPONSE;
        }
	curproc = 0;
    }

static void resetreqs() {
    int i;
    for(i=0; i<MAX_CPU_COUNT; ++i) {
        request[i].requested_now = 0;
        request[i].satisfied = PROCESSOR_NORESPONSE;
        }
	curproc = 0;
    }

%}

DIGIT       [0-9]
NUMBER      {DIGIT}+
PROC        P{NUMBER}
CMD         (R|W|L|U|P|X|F|B|O|H|C)[wb]?
WS          [ \t]*

%%

WS

^{WS}#.*\n		/* eat up comments */

^{WS}{NUMBER}:  {
	int current_time = *time + 1;
	int new_time = atoi(yytext);
	if (new_time > current_time) {
		/* check first if a processor is stuck in the 'past' i.e. current_time */
		int i;
		for (i=0; i<CPU_COUNT; ++i) {
			if (request[i].satisfied == PROCESSOR_STALLED) {
				yyless(0); yy_set_bol(1);
				*time = current_time;
				return;
				}
			}
		}
	*time = new_time;
	resetreqs();
    }

^{WS}:?	{
	++*time;
	resetreqs();			/* [ws][:] */
	}

{PROC}  {
    curproc = atoi(yytext+1);
    request[curproc].requested_now = 1;
    }

{CMD}   {
    switch (toupper(yytext[0])) {
        case 'R': request[curproc].type = REQUEST_READ; break;
        case 'W': request[curproc].type = REQUEST_WRITE; break;
        case 'L': request[curproc].type = REQUEST_LOCK; break;
        case 'U': request[curproc].type = REQUEST_UNLOCK; break;
        case 'P': request[curproc].type = REQUEST_PREF; break;
	case 'X': request[curproc].type = REQUEST_PREFEX; break;
	case 'F': request[curproc].type = REQUEST_LPREF; break;
	case 'B': request[curproc].type = REQUEST_WRBACK; break;  
        case 'O': request[curproc].type = REQUEST_OWL; break;
        case 'H': request[curproc].type = REQUEST_OWH; break;
        case 'C': request[curproc].type = REQUEST_CWL; break;
        }
    switch (yytext[1]) {
        case 'w': request[curproc].size = 2; break;
        case 'b': request[curproc].size = 1; break;
        default : request[curproc].size = 4; break;
        }
    }

{NUMBER}        request[curproc].virtual_addr = (bitsMAX)atol(yytext);

,{WS}{NUMBER}   request[curproc].data = atol(yytext+1);

\n		return;

.		

%%

void yytrace_init(struct Memory_Request *req_array, int *time_ptr, FILE*trace) {
	request = req_array;
	time = time_ptr;
	yyin = trace;
	initreqs();
	}

