/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: sync/abstract/sync.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This is where "abstract" synchronization primitives are implemented.
 * Rewrite them if you need - perhaps to add more statistics,
 * or to modify their behavior.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "kernel.h"	
#include "funcdecl.h"
#include "sync.h"

/*
 * "Abstract" synchronization primitives are not seen by the memory 
 * simulator: it never sees a LOCK or an UNLOCK request. When a thread
 * enters sys_lock_acquire() (an alias for LOCK()), for example, if
 * a lock is busy, it will block -- set its state to BLOCKED by calling
 * the sys_thread_sleep() kernel routine. From now on, that thread will
 * be ignored by the kernel and the processor that the thread belongs to
 * will not see any activities from the thread, until another thread
 * (the one that enters sys_lock_release()) wakes him up by setting its
 * state to READY through the sys_thread_wakeup(thread_ID) kernel call,
 * provided, of course, that this thread is the first on the waiting queue
 * for that lock.
 * The net effect of this approach is as if there were some perfect sync
 * network separated from the bus or whatever interconnection network
 * the memory simulator simulates. Basically, this is an example of how
 * one could implement his own sync primitives in a case when some
 * abstraction is needed, since some researches are directed to investigating
 * different means of communication.
 *
 * ENTER_KERNEL_MODE updates the thread's private timer by adding sys_IC 
 * (number of cycles elapsed during the last trip to user mode) to it, while
 * EXIT_KERNEL_MODE resets sys_IC upon return to user mode.
 * All the rutines called directly from the application should be enclosed
 * in this pair of macros.
 */

/*
 * Locks
 */
void sys_lock_init(struct LockStruct *L)	{
	ENTER_KERNEL_MODE
	L->value = 0;
	L->queue_top = L->queue_length = 0;
	EXIT_KERNEL_MODE
	}

void sys_lock_acquire(struct LockStruct *L)	{
	Thread_Time time_now = 0;
	ENTER_KERNEL_MODE
	STATS( time_now = sys_current_thread->thread_time;)
	sys_scheduler();	/* first reschedule someone else if he is earler */
	if (L->value)	{	/* busy_lock */
		int tail = (L->queue_top + L->queue_length) % MAX_THREAD_COUNT;
		++L->queue_length;
		L->blocked_threads_queue[tail] = sys_thread_ID;
		sys_thread_sleep();	/* will return when READY and the lock is free */
		}
	L->value = 1;		/* set the lock to busy */
	STATS( 
		sys_current_thread->stat.lock_time += 
		sys_current_thread->thread_time - time_now; 
		sys_current_thread->stat.locks++;
		)
	EXIT_KERNEL_MODE
	}

void sys_lock_release(struct LockStruct *L)	{
	ENTER_KERNEL_MODE
	sys_scheduler();	/* first reschedule someone else if he is earler */
	if (L->queue_length >0)	{
		/* release the first thread in the queue */
		sys_thread_wakeup(L->blocked_threads_queue[L->queue_top]);
		L->queue_top = (L->queue_top +1) % MAX_THREAD_COUNT;
		--L->queue_length;
		}
	L->value = 0;	/* this L->value is actually redundant, L->queue_length is enough */
	EXIT_KERNEL_MODE
	}

/*
 * Barriers
 */
void sys_barrier_init(struct BarrierStruct *B) {
	ENTER_KERNEL_MODE
	B->count = 0;	/* the number of those waiting for the barrier to break */
	EXIT_KERNEL_MODE
	}

void sys_barrier(struct BarrierStruct *B, int N) {
	Thread_Time time_now = 0;
	ENTER_KERNEL_MODE
	STATS( time_now = sys_current_thread->thread_time; )
	sys_scheduler();	/* first reschedule someone else if he is earler */
	if (++B->count < N) {
		B->blocked_threads_list[B->count-1] = sys_thread_ID;
		sys_current_thread->thread_state = THREAD_BLOCKED;
		sys_scheduler();	/* will return when it becomes READY */
		}
	else { 	/* the last thread to enter the barrier */
		int i;
		for (i=0; i < B->count -1; ++i)
			sys_thread_wakeup(B->blocked_threads_list[i]);
		B->count = 0;
		}
	STATS( 
		sys_current_thread->stat.barrier_time += 
		sys_current_thread->thread_time - time_now; 
		sys_current_thread->stat.barriers++;
		)
	EXIT_KERNEL_MODE
	}
