/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: sync/realistic/sync.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * This is where synchronization primitives are implemented.
 * Rewrite them if you need - perhaps to add more statistics,
 * or to modify their behavior.
 * This implementation implies "realistic" sync primitives: 
 * LOCK(lockname) issued by the application is passed to the memory
 * simulator as REQUEST_LOCK request, and similarily for the UNLOCK.
 * Such a memory simulator must be aware of these two types of requests
 * in addition to ordinary READs and WRITEs. This implementation is
 * suitable for SMP systems, and this model is the default.
 * Barriers are transformed in a series of LOCK, UNLOCK, READ and WRITE
 * requests -- as they would be on a real SMP machine.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "kernel.h"	
#include "funcdecl.h"
#include "sync.h"

/*
 * The following sync primitives implementations use
 * SYS_LOAD, SYS_STORE, SYS_LOCK, SYS_UNLOCK macros
 * defined in kernel.h.
 * You can change them to suit your needs.
 * Basically, since this file is not augmented as application files,
 * you can use the following kernel.h primitives to force the data 
 * pass through the memsim:
 * void SYS_LOCK(lock_variable), 	where lock_variable is a simple integer.
 * void SYS_UNLOCK(lock_variable), 
 * void SYS_STORE(variable, value),			does variable = value
 * (typeof variable) SYS_LOAD(variable),	reads variable.
 *
 * ENTER_KERNEL_MODE updates the thread's private timer by adding sys_IC 
 * (number of cycles elapsed during the last trip to user mode) to it, while
 * EXIT_KERNEL_MODE resets sys_IC upon return to user mode.
 * All the rutines called directly from the application should be enclosed
 * in this pair of macros.
 */

/*
 * Locks
 */
void sys_lock_init(struct LockStruct *L)  {
	ENTER_KERNEL_MODE
	SYS_STORE(L->value,0);
	EXIT_KERNEL_MODE
}

void sys_lock_acquire(struct LockStruct *L)  {
	Thread_Time time_now=0;
	ENTER_KERNEL_MODE
	STATS( time_now = sys_current_thread->thread_time;)
	SYS_LOCK(L->value);
	STATS( 
		sys_current_thread->stat.lock_time += 
		sys_current_thread->thread_time - time_now; 
		sys_current_thread->stat.locks++;
	)
	EXIT_KERNEL_MODE
}

void sys_lock_release(struct LockStruct *L)  {
	ENTER_KERNEL_MODE
	SYS_UNLOCK(L->value);
	EXIT_KERNEL_MODE
}

/*
*
* added instructions
*/
/*
void sys_pref(bitsADDR A)  {
     ENTER_KERNEL_MODE
     SYS_PREF(A);
     EXIT_KERNEL_MODE;
}
*/
/*******end of inserted part*****/


/*
 * Barriers
 */
void sys_barrier_init(struct BarrierStruct *B)  {
	ENTER_KERNEL_MODE
	SYS_STORE(B->lock.value,0);
	SYS_STORE(B->sleeplock.value,0);
	SYS_STORE(B->sleepers,0);
	SYS_LOCK(B->sleeplock.value);
	EXIT_KERNEL_MODE
}

void sys_barrier(struct BarrierStruct *B, int N)  {
        Thread_Time time_now=0;
        ENTER_KERNEL_MODE
        STATS( time_now = sys_current_thread->thread_time; )
        SYS_BARRIER_ENTER(B->sleeplock.value,N)
        SYS_LOCK(B->lock.value)
        SYS_BARRIER_IN(B->sleepers+1)
	SYS_STORE(B->sleepers, SYS_LOAD(B->sleepers)+1);
        if ( SYS_LOAD(B->sleepers) < N ) {	
                SYS_UNLOCK(B->lock.value)
                SYS_BARRIER_WAIT
                SYS_LOCK(B->sleeplock.value)
                SYS_BARRIER_OUT(N-B->sleepers+1)
                SYS_STORE(B->sleepers, SYS_LOAD(B->sleepers)-1);
                if (SYS_LOAD(B->sleepers) > 0) {
                        SYS_BARRIER_EXIT
                        SYS_UNLOCK(B->sleeplock.value)
                }
                else {
                        SYS_BARRIER_EXIT
                        SYS_UNLOCK(B->lock.value)
                }
        }								
        else {
                SYS_BARRIER_OUT(N-B->sleepers+1)
                SYS_STORE(B->sleepers, SYS_LOAD(B->sleepers)-1);
                if (SYS_LOAD(B->sleepers) > 0) {
                        SYS_BARRIER_EXIT
                        SYS_UNLOCK(B->sleeplock.value)
                }
                else {
                        SYS_BARRIER_EXIT
                        SYS_UNLOCK(B->lock.value)
                }
        }
	STATS( 
                sys_current_thread->stat.barrier_time += 
                sys_current_thread->thread_time - time_now; 
                sys_current_thread->stat.barriers++;
        )
        EXIT_KERNEL_MODE
}
