/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Module: scanf.c
 * Author: Davor Magdic, magdic00390d@buef31.etf.bg.ac.yu
 *
 * Here is a replacement for scanf, sscanf, fscanf.
 * The heart of scanf(), do_scan, is not really implemented;
 * In fact, real fscanf() is called, but in such way that all the
 * data pass through the simulator as it normally would.
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdarg.h>
#include <stdio.h>
#include "sys_stdlib.h"

/*
 *  +------------------------------------------------------------------+
 *  |                             SCANFs                               |
 *  +------------------------------------------------------------------+
 */ 

static int strlen(char *s) {
    register int i=0;
    while (*s++) ++i;
    return i;
    }

#define MAX_SCANF_ARGS 20
#define MAX_INSTRING_SIZE 500
static int sys_doscan(FILE *fstream, char * format, va_list ap) {
    struct Arg {
        void * original_ptr;
        void * temp_ptr;
        int size;
        int buffer[3];              /* max data size */
        int isstring;
        } arg[MAX_SCANF_ARGS];
    int argc = 0;
    char *f; int i;
    int inconv = 0, islong=0, isshort=0, size_modifier=0;
    int donecount;

    /* scan the format string */
    for(f = format; *f; ++f) {
        if (!inconv) {
            if (*f == '%')  {       /* a new argument */
                inconv = 1;     
                arg[argc].original_ptr = va_arg(ap, void *);
                arg[argc].isstring = 0;
                islong = isshort = size_modifier = 0;
                if (argc >= MAX_SCANF_ARGS) {
                    fprintf(stderr,"sys_stdlib error: too many arguments for scanf()\n");
                    exit(1);
                    }
                }
            }               
        else    {                   /* we are in a conversion */
            switch(*f) {
                case 'l': 
                case 'L': islong  = 1; break;
                case 'h': isshort = 1; break;
                case 's':
                    arg[argc].isstring = 1;
                    goto arg_setup;
                case 'e':
                case 'f':
                case 'g':
                    if (islong) arg[argc].size = sizeof(double);
                    else        arg[argc].size = sizeof(float);
                    goto arg_setup;
                case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
                    size_modifier = 10 * size_modifier + (*f - '0');
                    break;
                case 'c':
                    if (size_modifier == 0)
                        arg[argc].size = sizeof(char);
                    else
                        arg[argc].isstring = 1;
                    goto arg_setup;
                case 'd':
                case 'i':
                case 'o':
                case 'u':
                case 'x':
                    if (islong)         arg[argc].size = sizeof(long int);
                    else if (isshort)   arg[argc].size = sizeof(short int);
                    else                arg[argc].size = sizeof(int);
                    goto arg_setup;
                case '%':
                    arg[argc].size = sizeof(int);   /* unexpected %: go default */
                    --f; 
                arg_setup:
                    if (!arg[argc].isstring) 
                        arg[argc].temp_ptr = &arg[argc].buffer;
                    else    {
                        arg[argc].temp_ptr = (void *) malloc(MAX_INSTRING_SIZE);
                        ((char *)arg[argc].temp_ptr)[size_modifier] = '\0';
                        }
                    ++argc;
                    inconv = 0;
                    break;
                default: 
                    break;
                }
            }
        }   
    va_end(ap);

    /* 
     * OK, now the argument table is set up. Push all temp_ptrs, in
     * reverse order, and do the scanf.
     */
    for (i=argc-1; i>=0; --i) 
        asm __volatile__ ("pushl %0" : "=g" (arg[i].temp_ptr));
    donecount = fscanf(fstream, format);
    /* update esp */
    for (i=0; i<argc; ++i)
        asm __volatile__ ("addl $4, %esp");
    
    /* scanf-ing is finished, now copy the results to the proper places */
    for (i=0; i<donecount && i<argc; ++i) {
        if (! arg[i].isstring) {
            /* not a string, but a char, a short int, an int, a float, or a double */
            /* copy the result to the buffer */
            sys_memcpy(arg[i].original_ptr, arg[i].temp_ptr, arg[i].size);
            }
        else    {
            int j, len = strlen(arg[i].temp_ptr);
            for (j=0; j<len; ++j) 
                *(char *)(arg[i].original_ptr+j) = *(char *)(arg[i].temp_ptr+j);
            free(arg[i].temp_ptr);
            }
        }
    return donecount;
    }

int sys_fscanf(FILE * fstream, char * format, ...) {
    va_list ap;
    va_start(ap, format);
    return sys_doscan(fstream, format, ap);
    }

int sys_scanf(char * format, ...) {
    va_list ap;
    va_start(ap, format);
    return sys_doscan(stdin, format, ap);
    }

int sys_sscanf(char *string, char * format, ...) {
    va_list ap; int i; int slen = strlen(string);
    int retval;
        
    va_start(ap, format);
    /* push back source string into stdin */
    ungetc('\n', stdin);        /* but first one newline */
    for (i=slen-1; i>=0; --i) 
        ungetc(string[i], stdin);
    retval = sys_doscan(stdin, format, ap);
    getchar();                  /* eat up the newline */
    return retval;
    }

int sys_sprintf( char *str, const char *format, ...) {
    char temp[1000];    /* C programmer's dissease. Hope it suffices. :-) */
    int retval; register int i=0;
    va_list ap;
    va_start(ap, format);
    retval = vsprintf(temp, format, ap);
    while((str[i] = temp[i])) ++i;
    return retval;
    }
    
