%option noyywrap

%{

/*
 * This is a scanner for analyzing processor trace file and making a separate
 * trace file for each processor. It is used within tracer.c module.
 *
 * Author:  Igor Ikodinovic   ( igi@eunet.yu )
 *          Department of Computer Engineering
 *          Faculty of Electrotechnical Engineering
 *          University of Belgrade
 *          Serbia, Yugoslavia
 *          1998
 * 
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include "kernel.h"
#include "tracer_types.h"

static char progress_meter = 1;
static int progress_counter = 0;
static long int total_length;
static time_t parsing_start_time;
extern int PROGRESS_STEP;

static enum Memory_Request_Type prev_request [MAX_CPU_COUNT];
static bitsADDR prev_address [MAX_CPU_COUNT];
static int      same_lock_request [MAX_CPU_COUNT];
static Thread_Time lock_time_start [MAX_CPU_COUNT];
static Thread_Time lock_time [MAX_CPU_COUNT];
static Thread_Time lock_busy_time [MAX_CPU_COUNT];
static Thread_Time trace_time;

static int proc;

extern int    P_TRACEfile [MAX_CPU_COUNT];
extern char   P_TRACEfilename [MAX_CPU_COUNT] [MAX_FILENAME_LENGTH];
static struct P_event event;
static union  P_global global;

extern type_table_of_lists barrier_data [MAX_CPU_COUNT];

void write_barrier_ENTER_data (int proc, bitsADDR addr, int total, Thread_Time ENTER_time, long int ENTER_pos) {
    add_new_to_table_of_lists (barrier_data, proc, addr, total, ENTER_time, ENTER_pos);
}
void write_barrier_IN_data (int proc, int IN_order, Thread_Time IN_time, Thread_Time original_IN_time) {
    update_IN_data_last_table_of_lists (barrier_data, proc, IN_order, IN_time, original_IN_time);
}
void write_barrier_WAIT_data (int proc, Thread_Time WAIT_time, long int WAIT_pos) {
    update_WAIT_data_last_table_of_lists (barrier_data, proc, WAIT_time, WAIT_pos);
}
void write_barrier_OUT_data (int proc, int OUT_order, Thread_Time OUT_time, Thread_Time original_OUT_time) {
    update_OUT_data_last_table_of_lists (barrier_data, proc, OUT_order, OUT_time, original_OUT_time);
}
void write_barrier_EXIT_data (int proc, Thread_Time EXIT_time, long int EXIT_pos) {
    update_EXIT_data_last_table_of_lists (barrier_data, proc, EXIT_time, EXIT_pos);
}

bitsADDR to_addr(char *q) {
    int i,j;
    char w[2] = " \0";
    bitsADDR x = 0;
    for (i=0; (q[i]==32) || (q[i]==9); i++);
    for (j=i; (q[j]>=48) && (q[j]<=57); j++) {
        w[0]=q[j];
        x=10*x+atoi(w);
    }
    return(x);
}

%}

SPACE              [ \t]*
DIGIT              [0-9]
NUMBER             {DIGIT}+
TIME               T={NUMBER}
PROC               P{NUMBER}:
REQUEST            R|W|L|U|P|X|F|B|O|H|C
AREA               c|s|d

%s REQUEST_Type REQUEST_Area REQUEST_Address REQUEST_Size 
%s CREATED_Parent CREATED_Time
%s BARRIER_ENTER_Address BARRIER_ENTER_Total
%s BARRIER_IN_Address BARRIER_IN_Order BARRIER_IN_Total
%s BARRIER_WAIT_Address BARRIER_WAIT_Total
%s BARRIER_OUT_Address BARRIER_OUT_Order BARRIER_OUT_Total
%s BARRIER_EXIT_Address BARRIER_EXIT_Total


%%


{SPACE}        /* skip blanks and tabs */

^{SPACE}\n     /* skip blank lines */

^{SPACE}#.*\n  /* skip lines with comments (they begin with #) */

{TIME}  {
        trace_time = (Thread_Time) atoi(yytext+2);
}

{PROC}  {
        proc = atoi(yytext+1);
        BEGIN(REQUEST_Type);
}

<REQUEST_Type>{REQUEST}  {
        event.global = (char) GLOBAL_NONE;
        switch (yytext[0]) {
            case 'R': event.request = (char) REQUEST_READ; break;
            case 'W': event.request = (char) REQUEST_WRITE; break;
            case 'L': event.request = (char) REQUEST_LOCK; break;
            case 'U': event.request = (char) REQUEST_UNLOCK; break;
            case 'P': event.request = (char) REQUEST_PREF; break;
            case 'X': event.request = (char) REQUEST_PREFEX; break;
            case 'F': event.request = (char) REQUEST_LPREF; break;
            case 'B': event.request = (char) REQUEST_WRBACK; break;
            case 'O': event.request = (char) REQUEST_OWL; break;
            case 'H': event.request = (char) REQUEST_OWH; break;
            case 'C': event.request = (char) REQUEST_CWL; break;
            default : break;
        }
        BEGIN(REQUEST_Area);
}

<REQUEST_Area>{AREA}  {
        event.area = yytext[0];
        BEGIN(REQUEST_Address);
}

<REQUEST_Address>{SPACE}{NUMBER}  {
        event.address = to_addr(yytext);
        BEGIN(REQUEST_Size);
}

<REQUEST_Size>,{NUMBER}  {
        event.size = (char)atoi(yytext+1);
        BEGIN(INITIAL);
}

CREATED {
        event.global = (char) GLOBAL_CREATED;
        BEGIN(CREATED_Parent);
}

<CREATED_Parent>"("P{NUMBER} {
        global.created.parent = atoi(yytext+2);
        BEGIN(CREATED_Time);
}

<CREATED_Time>,T={NUMBER}")" {
        global.created.time = atoi(yytext+3);
        BEGIN(INITIAL);
}

BARRIER_ENTER  {
        event.global = (char) GLOBAL_BARRIER_ENTER;
        BEGIN(BARRIER_ENTER_Address);
}

<BARRIER_ENTER_Address>"("{NUMBER}  {
        global.barrier.address = to_addr(yytext+1);
        BEGIN(BARRIER_ENTER_Total);
}

<BARRIER_ENTER_Total>,{NUMBER}")"  {
        global.barrier.total = atoi(yytext+1);
        BEGIN(INITIAL);
}

BARRIER_IN  {
        event.global = (char) GLOBAL_BARRIER_IN;
        BEGIN(BARRIER_IN_Address);
}

<BARRIER_IN_Address>"("{NUMBER}  {
        global.barrier.address = to_addr(yytext+1);
        BEGIN(BARRIER_IN_Order);
}

<BARRIER_IN_Order>,{NUMBER}  {
        global.barrier.order = atoi(yytext+1);
        BEGIN(BARRIER_IN_Total);
}

<BARRIER_IN_Total>"/"{NUMBER}")"  {
        global.barrier.total = atoi(yytext+1);
        BEGIN(INITIAL);
}

BARRIER_WAIT  {
        event.global = (char) GLOBAL_BARRIER_WAIT;
        BEGIN(BARRIER_WAIT_Address);
}

<BARRIER_WAIT_Address>"("{NUMBER}  {
        global.barrier.address = to_addr(yytext+1);
        BEGIN(BARRIER_WAIT_Total);
}

<BARRIER_WAIT_Total>,{NUMBER}")"  {
        global.barrier.total = atoi(yytext+1);
        BEGIN(INITIAL);
}

BARRIER_OUT  {
        event.global = (char) GLOBAL_BARRIER_OUT;
        BEGIN(BARRIER_OUT_Address);
}

<BARRIER_OUT_Address>"("{NUMBER}  {
        global.barrier.address = to_addr(yytext+1);
        BEGIN(BARRIER_OUT_Order);
}

<BARRIER_OUT_Order>,{NUMBER}  {
        global.barrier.order = atoi(yytext+1);
        BEGIN(BARRIER_OUT_Total);
}

<BARRIER_OUT_Total>"/"{NUMBER}")"  {
        global.barrier.total = atoi(yytext+1);
        BEGIN(INITIAL);
}

BARRIER_EXIT  {
        event.global = (char) GLOBAL_BARRIER_EXIT;
        BEGIN(BARRIER_EXIT_Address);
}

<BARRIER_EXIT_Address>"("{NUMBER}  {
        global.barrier.address = to_addr(yytext+1);
        BEGIN(BARRIER_EXIT_Total);
}

<BARRIER_EXIT_Total>,{NUMBER}")"  {
        global.barrier.total = atoi(yytext+1);
        BEGIN(INITIAL);
}

\n  {
        /* 
         * Write request in file, but don't write repeated lock requests 
         */
        if ((prev_request[proc] == REQUEST_LOCK) &&
            (event.request == REQUEST_LOCK) &&
            (event.address == prev_address[proc]))
        {
            same_lock_request[proc] = 1;
            lock_time[proc] = trace_time;
        }
        else {
            if (same_lock_request[proc]) {
                lock_busy_time[proc] += lock_time[proc] - lock_time_start[proc];
                same_lock_request[proc] = 0;
            }
            if (event.request == REQUEST_LOCK) lock_time_start[proc] = trace_time;
            event.time = trace_time - lock_busy_time[proc];
            if (write(P_TRACEfile[proc], &event, sizeof(event)) == -1) {
                fprintf (stderr, "\nError writing to trace file %s\n", P_TRACEfilename[proc]);
                exit (1);
            }
            if (event.global != GLOBAL_NONE) {
                if (write(P_TRACEfile[proc], &global, sizeof(global)) == -1) {
                    fprintf (stderr, "\nError writing to trace file %s\n", P_TRACEfilename[proc]);
                    exit (1);
                }
                /* Fill in barrier data structures */
                switch (event.global) {
                    case GLOBAL_CREATED:
                         break;
                    case GLOBAL_BARRIER_ENTER:
                         write_barrier_ENTER_data (proc, global.barrier.address, global.barrier.total, event.time, tell(P_TRACEfile[proc]));
                         break;
                    case GLOBAL_BARRIER_IN:
                         write_barrier_IN_data (proc, global.barrier.order, event.time, event.time+lock_busy_time[proc]);
                         break;
                    case GLOBAL_BARRIER_WAIT:
                         write_barrier_WAIT_data (proc, event.time, tell(P_TRACEfile[proc]));
                         break;
                    case GLOBAL_BARRIER_OUT:
                         write_barrier_OUT_data (proc, global.barrier.order, event.time, event.time+lock_busy_time[proc]);
                         break;
                    case GLOBAL_BARRIER_EXIT:
                         write_barrier_EXIT_data (proc, event.time, tell(P_TRACEfile[proc]));
                         break;
                    default:
                         fprintf (stderr, "\nError - Unknown global operation type: %d\n", event.global);
                         exit (1);
                }
            }
            prev_request[proc] = event.request;
            prev_address[proc] = event.address;
        }
        /*
         *  Update progress meter and timer
         */
        if (progress_meter) {
            ++progress_counter;            
            if ((progress_counter%PROGRESS_STEP == 0) ||
                (ftell(yyin) == total_length))
            {
                printf ("Parsing trace file  -  Progress: %5.1f%%  Time: %lu seconds\r", 
                        ((float)ftell(yyin))/total_length*100,
                        time(NULL)-parsing_start_time);
                fflush(stdout);
            }
        }
}

.       /* skip every other character */


%%


void yyproctraces_init(FILE *trace) {
    int i;
    /* Initialize progress meter parmeters */
    fseek (trace,0,2); /* goto end of file */
    total_length = ftell (trace);
    fseek (trace,0,0); /* goto beginning of file */
    parsing_start_time = time(NULL);
    /* Initialize trace file */
    yyin = trace;
    /* Initialize barrier table and other variables */
    for (i=0; i<MAX_CPU_COUNT; i++) {
        barrier_data[i].list_head = barrier_data[i].list_tail = NULL;
        same_lock_request[i] = 0;
        prev_request[i] = REQUEST_NONE;
        lock_busy_time[i] = 0;
    }
}
