/* 
 * This is main module of a trace-driven multiprocessor simulator.
 * It's job is to perform scheduling and to take care about locks and barriers.
 * It can also produce traces.
 *
 * Author:  Igor Ikodinovic  ( igi@eunet.yu )
 *          Department of Computer Engineering
 *          Faculty of Electrotechnical Engineering
 *          University of Belgrade
 *          Serbia, Yugoslavia
 *          1998
 *
 */

#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#define extern
#include "kernel.h"
#undef extern
#include "tracer_types.h"

char *INIfile = "limes.ini";  /* default ini file name */
char **startup_argv = NULL;   /* the list of limes args */
#include "utilities.h"

static type_hash_table locks_hash_table[HASHTABLE_SIZE];
static type_hash_table barriers_hash_table[HASHTABLE_SIZE];
type_table_of_lists barrier_data[MAX_CPU_COUNT];
#include "tracer_funcs.h"

static void make_proctraces_from_bin_file(int Tracefile);

/* Trace file */
static int trace_file_type = 0;  /* 1 for text, 2 for binary trace file */
static union Trace_file TRACEfile;
static char *TRACEfilename;

/* Temporary trace files */
char *P_TRACEfilename_root = "TRACE_P"; /* default temporary trace filename */
int P_TRACEfile [MAX_CPU_COUNT];
char P_TRACEfilename [MAX_CPU_COUNT] [MAX_FILENAME_LENGTH];

/* Produced trace file */
int Produced_trace_file_type = 0;  /* 1 for text, 2 for binary */
static union Trace_file Produced_TRACEfile;
static char *Produced_TRACEfilename;

static struct P_event event;
static union P_global global;

int PROGRESS_STEP = 200;

main (int argc, char *argv[]) {
    int n = 1, i, j;
    int any_STALLED;
    int finished_thread_count = 0;
    int ind;
    int progress_counter = 0;
    int thread [MAX_CPU_COUNT];
    long int current_pos_in_barrier [MAX_CPU_COUNT];
    long int current_trace_pos [MAX_CPU_COUNT];
    long int current_length;
    long int total_length = 0;
    time_t tracer_start_time;
    time_t simulation_start_time;
    Thread_Time last_simulation_time = 0;
    Thread_Time time_now = 0;
    Thread_Time startup_time [MAX_CPU_COUNT];
    Thread_Time busy_time [MAX_CPU_COUNT];
    Thread_Time time_diff [MAX_CPU_COUNT];
    Thread_Time current_lock_busy_time [MAX_CPU_COUNT];
    char thread_running [MAX_CPU_COUNT];
    char thread_in_barrier [MAX_CPU_COUNT];
    char finished_threads [MAX_CPU_COUNT];
    char repeating_global [MAX_CPU_COUNT];
    char progress_meter = 1;
    char beep_at_end = 1;
    char repeat_unsatisfied_locks = 0;
    struct list_element_2 *barrier [MAX_CPU_COUNT];
    struct list_element_2 *match_barrier_IN [MAX_CPU_COUNT];
    struct list_element_2 *match_barrier_OUT [MAX_CPU_COUNT];

    /* Start */
    tracer_start_time = time(NULL);
    startup_argv = &argv[1];

    printf("\n************************************ TRACER ************************************\n");
    printf("Command line: ");
    for(i=0; argv[i]; ++i) printf("%s ",argv[i]); 
    printf("\nSimulator predefinitions: ");
    printf("MAX_CPU_COUNT=%d, BUSY_LOCK_TIME=%d, BUSY_LOCK_OPTIMIZATION=%s\n",
        MAX_CPU_COUNT, BUSY_LOCK_SLEEP_TIME, BUSY_LOCK_OPTIMIZATION ? "YES" : "NO ");

    printf("--------------------------------------------------------------------------------\n");
    printf("                                   Simulation                                   \n\n");
    memsim_init();
    printf("Preparing TRACER for simulation ...\r");
    fflush(stdout);

    /* Analyze command line options */
    while(n < argc) {
        if ((argv[n][0] == '-') && (argv[n][1] == 'd')) /* initialization values in form -dITEM=VALUE */
            ;
        else if (strcmp (argv[n],(char*)&("-fx")) == 0) { /* input is text trace file */
            trace_file_type = 1;
            if ((n < (argc-1)) && (argv[n+1][0] != '-'))
                TRACEfilename = argv[n+++1];
            else 
                TRACEfilename = "TRACE.txt";
            if ((TRACEfile.txt = fopen (TRACEfilename,"r")) == NULL) {
                fprintf (stderr, "\nError opening trace file %s\n", TRACEfilename);
                exit (1);
            }
        }
        else if (strcmp (argv[n],(char*)&("-fb")) == 0) { /* input is binary trace file */
            trace_file_type = 2;
            if ((n < (argc-1)) && (argv[n+1][0] != '-'))
                TRACEfilename = argv[n+++1];
            else 
                TRACEfilename = "TRACE.bin";
            if ((TRACEfile.bin = open (TRACEfilename,0)) == -1) {
                fprintf (stderr, "\nError opening trace file %s\n", TRACEfilename);
                exit (1);
            }
        }
        else if ((strcmp (argv[n],(char*)&("-h")) == 0) ||
                 (strcmp (argv[n],(char*)&("--help")) == 0))
            goto PRINT_OPTIONS;
        else if (strcmp (argv[n],(char*)&("-ly")) == 0) /* trace will have repeated locks */
            repeat_unsatisfied_locks = 1;
        else if (strcmp (argv[n],(char*)&("-ln")) == 0) /* trace will not have repeated locks */
            repeat_unsatisfied_locks = 0;
        else if (strcmp (argv[n],(char*)&("-t")) == 0) { /* output text trace to stdout */
            Produced_trace_file_type = 1;
            Produced_TRACEfilename = "";
            Produced_TRACEfile.txt = stdout;
        }
        else if (strcmp (argv[n],(char*)&("-tx")) == 0) { /* produce text trace file */
            Produced_trace_file_type = 1;
            if ((n < (argc-1)) && (argv[n+1][0] != '-'))
                Produced_TRACEfilename = argv[n+++1];
            else 
                Produced_TRACEfilename = "PRODUCED.txt";
            if ((Produced_TRACEfile.txt = fopen (Produced_TRACEfilename,"w")) == NULL) {
                fprintf (stderr, "\nError opening file %s for producing\n", Produced_TRACEfilename);
                exit (1);
            }
        }
        else if (strcmp (argv[n],(char*)&("-tb")) == 0) { /* produce binary trace file */
            Produced_trace_file_type = 2;
            if ((n < (argc-1)) && (argv[n+1][0] != '-'))
                Produced_TRACEfilename = argv[n+++1];
            else 
                Produced_TRACEfilename = "PRODUCED.bin";
            creat (Produced_TRACEfilename,0666); /* brrrr */
            if ((Produced_TRACEfile.bin = open (Produced_TRACEfilename,1)) == -1) {
                fprintf (stderr, "\nError opening file %s for producing\n", Produced_TRACEfilename);
                exit (1);
            }
        }
        else {
            fprintf(stderr, "\nError - Unknown TRACER option: %s", argv[n]);
            PRINT_OPTIONS :
            fprintf(stdout, "\n");
            fprintf(stdout, "Syntax:\n");
            fprintf(stdout, "<SIM>tracer [-f[x|b] file] [-t]|[-t[x|b] outfile] [-l[y|n]] [-ditem=value] [-h]|[--help]\n");
            fprintf(stdout, "Options:\n");
            fprintf(stdout, "\t-d   specifies item's value (overrides one in .ini file)\n");
            fprintf(stdout, "\t-fx  run simulation with input file in text format\n");
            fprintf(stdout, "\t-fb  run simulation with input file in binary format\n");
            fprintf(stdout, "\t-h   print this help message\n");
            fprintf(stdout, "\t-ly  trace will have repeated locks if memmory simulator allows it\n");
            fprintf(stdout, "\t-ln  trace will not have repeated locks\n");
            fprintf(stdout, "\t-t   produce trace in text format and send it to stdout\n");
            fprintf(stdout, "\t-tx  produce trace file in text format\n");
            fprintf(stdout, "\t-tb  produce trace file in binary format\n");
            fprintf(stdout, "\t--help  print this help message\n");
            exit (1);
        }
        ++n;
    }

    /* Open temporary trace files */
    for (i=0; i<MAX_CPU_COUNT; i++) {
        for (j=0; j<strlen(P_TRACEfilename_root); j++) 
            P_TRACEfilename[i][j] = P_TRACEfilename_root[j];
        P_TRACEfilename[i][j++] = (char)(i/10+48);
        P_TRACEfilename[i][j++] = (char)(i%10+48);
        creat (P_TRACEfilename[i],0666); /* brrrr */
        if ((P_TRACEfile[i] = open (P_TRACEfilename[i],2)) == -1) {
            fprintf (stderr, "\nError opening temporary trace file %s\n", P_TRACEfilename[i]);
            exit (1);
        }
    }    

    printf("                                                                                \r");

    /* Extract traces for each processor from global trace file */
    switch (trace_file_type) {
        case 1  :  /* text trace file */
            yyproctraces_init(TRACEfile.txt);
            while (yylex());
            break;
        case 2  :  /* binary trace file */
            make_proctraces_from_bin_file(TRACEfile.bin);
            break;
        default :  /* should never come here */
            break; 
    }
    printf("\n");

    for (i=0; i<MAX_CPU_COUNT; i++) {
        /* Remember sum of lengths of each trace file */
        total_length += tell(P_TRACEfile[i]);
        /* Rewind each processor's temporary trace file to the beginning */
        lseek(P_TRACEfile[i], 0, 0);
    }


    /* This is the core of the tracer. Here we go through each processor's 
       trace file and schedule their requests for execution */

    simulation_start_time = time(NULL);

    /* Initialize hash tables */
    for (i=0; i<HASHTABLE_SIZE; i++) {
        locks_hash_table[i].list_head = NULL;
        barriers_hash_table[i].list_head = NULL;
    }

    /* Thread initialization */
    for (i=0; i<MAX_CPU_COUNT; i++) {
        sys_threads[i].thread_state = THREAD_READY;
        sys_threads[i].thread_ID = i;
        sys_threads[i].proc_ID = i;
        sys_threads[i].thread_time = 0;
    }

    /* Other necessary initializations */
    for (i=0; i<MAX_CPU_COUNT; i++) {
        startup_time[i] = 0;
        busy_time[i] = 0;
        time_diff[i] = 0;
        thread_running[i] = 0;
        finished_threads[i] = 0;
        thread[i] = i;
        thread_in_barrier[i] = 0;
        repeating_global[i] = 0;
        finished_threads[i] = 0;
    }
    
    /* Main scheduler */

    /* Read first request from each thread and mark FINISHED threads if any */
    sys_thread_count = MAX_CPU_COUNT;
    for (i=0; i<MAX_CPU_COUNT; i++)
        if ((sys_threads[i].thread_state == THREAD_READY) &&
            (sys_threads[i].thread_time <= time_now))
        {
            ind = read(P_TRACEfile[thread[i]], &event, sizeof(event));
            if (ind == -1) {
                fprintf (stderr, "\nError reading temporary trace file %s\n", P_TRACEfilename[i]);
                exit (1);
            }
            if (ind == 0) {
                --sys_thread_count;
                finished_threads[i] = 1;
                ++finished_thread_count;
            }
            else {
                sys_threads[i].thread_time = event.time;
                sys_threads[i].memory_request.requested_now = 0;
                sys_threads[i].memory_request.type = (enum Memory_Request_Type) event.request;
                sys_threads[i].memory_request.virtual_addr = event.address;
                sys_threads[i].memory_request.pid = sys_threads[i].proc_ID;
                sys_threads[i].memory_request.size = event.size;
                sys_threads[i].memory_request.area = event.area;
                sys_threads[i].memory_request.isshared = (event.area == 'd');
                if (event.global != GLOBAL_NONE) {
                    ind = read(P_TRACEfile[i], &global, sizeof(global));
                    if (ind == -1) {
                        fprintf (stderr, "\nError reading temporary trace file %s\n", P_TRACEfilename[i]);
                        exit (1);
                    }
                    if (ind == 0) {
                        --sys_thread_count;
                        finished_threads[i] = 1;
                        ++finished_thread_count;
                    }
                    else {
                        sys_threads[i].global_operation.type = (enum Global_Synchro_Type) event.global;
                        switch (sys_threads[i].global_operation.type) {
                            case GLOBAL_CREATED:
                                 sys_threads[i].global_operation.params.created.parent = global.created.parent;
                                 sys_threads[i].global_operation.params.created.time = global.created.time;
                            case GLOBAL_BARRIER_ENTER:
                                 sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                 sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                 break;
                            case GLOBAL_BARRIER_IN:
                                 sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                 sys_threads[i].global_operation.params.barrier.order = global.barrier.order;
                                 sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                 break;
                            case GLOBAL_BARRIER_WAIT:
                                 sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                 sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                 break;
                            case GLOBAL_BARRIER_OUT:
                                 sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                 sys_threads[i].global_operation.params.barrier.order = global.barrier.order;
                                 sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                 break;
                            case GLOBAL_BARRIER_EXIT:
                                 sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                 sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                 break;
                            default:
                                 fprintf (stderr, "\nError - Unknown global operation type %d in temporary trace file %s\n", 
                                          sys_threads[i].global_operation.type, P_TRACEfilename[thread[i]]);
                                 exit (1);
                        }
                    }
                }
                else 
                    sys_threads[i].global_operation.type = GLOBAL_NONE;
            }
        }

    /* Find the first thread and let it run */
    j = 0;
    for (i=1; i<sys_thread_count; i++)
        if (sys_threads[i].thread_time < sys_threads[j].thread_time) j = i;
    thread_running[j] = 1;

    /* Scheduling loop */
    while (finished_thread_count<MAX_CPU_COUNT) { 

        /* Run threads that are created now */
        for (i=0; i<sys_thread_count; i++)
            if ((sys_threads[i].global_operation.type == GLOBAL_CREATED) &&
                (sys_threads[i].global_operation.params.created.time <= sys_threads[j=sys_threads[i].global_operation.params.created.parent].thread_time) &&
                !thread_running[i])
            {
                startup_time[i] = startup_time[j]+time_diff[j]+busy_time[j];
                sys_threads[i].global_operation.params.created.time = startup_time[i]+sys_threads[i].thread_time;
                thread_running[i] = 1;
            }

        any_STALLED = 0;
        for (i=0; i<sys_thread_count; i++)
            if (sys_threads[i].thread_state == THREAD_STALLED) {
                any_STALLED = 1;
                break;
            }

        /* Determine time_now */
        if (any_STALLED)
            ++time_now;
        else {
            /* 
             * If there are no STALLED threads determine time_now as the 
             * earliest possible event 
             */
            /* Find the time of the first READY thread */
            for (i=0; i<sys_thread_count; i++)
                if (((sys_threads[i].thread_state == THREAD_READY) ||
                     (sys_threads[i].thread_state == THREAD_IO)) &&
                    (sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i] >= time_now) &&
                    thread_running[i])
                {
                    time_now = sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i];
                    break;
                }
            /* And now find the earliest request among all READY threads */
            for (j=i+1; j<sys_thread_count; j++)
                if (((sys_threads[j].thread_state == THREAD_READY) ||
                     (sys_threads[j].thread_state == THREAD_IO)) &&
                    (sys_threads[j].thread_time+time_diff[j]+busy_time[j]+startup_time[j] < time_now) &&
                    thread_running[j])
                {
                    time_now = sys_threads[j].thread_time+time_diff[j]+busy_time[j]+startup_time[j];
                }
        }

        /* Schedule requests that arrive at time_now if there are any */
        for (i=0; i<sys_thread_count; i++)
            if (((sys_threads[i].thread_state == THREAD_READY) ||
                 (sys_threads[i].thread_state == THREAD_IO)) &&
                (sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i] == time_now) &&
                thread_running[i])
            {
                sys_threads[i].memory_request.requested_now = 1;
            }

        /* Produce trace file if such option was given */
        if (Produced_trace_file_type)
            if (repeat_unsatisfied_locks) {
                /* 
                 * If unsatisfied lock requests should appear repeatedly in
                 * trace, then we need to avoid repeating of global operations, 
                 * as all operations that need to be repeated are already left 
                 * untouched by the parser. Repeated lock requests themselves
                 * are visible to the trace producer.
                 * This pertains only to the case when lock requests are left 
                 * to the scheduler to handle them - it is realized through 
                 * the use of PROCESSOR_LOCK_BUSY processor responses and
                 * THREAD_IO thread states.
                 * If memory simulator handles lock requests itself then there 
                 * can be no repeated locks in trace!
                 */
                enum Global_Synchro_Type global_op [MAX_CPU_COUNT];
                for (i=0; i<sys_thread_count; i++)
                    if ((sys_threads[i].thread_state == THREAD_IO) && 
                        (sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i] == time_now) &&
                        repeating_global[i])
                    {
                        global_op[i] = sys_threads[i].global_operation.type;
                        sys_threads[i].global_operation.type = GLOBAL_NONE;
                    }

                produce_trace(
                    Produced_TRACEfilename,
                    Produced_TRACEfile,
                    Produced_trace_file_type,
                    time_now
                ); /* from utilities.h */

                /*
                 * Yet make sure that global operations are visible to the memory 
                 * simulator.
                 * This pertains only to the case when lock requests are left 
                 * to the scheduler to handle them, which is explained above.
                 */
                for (i=0; i<sys_thread_count; i++)
                    if ((sys_threads[i].thread_state == THREAD_IO) && 
                        (sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i] == time_now) &&
                        repeating_global[i])
                        sys_threads[i].global_operation.type = global_op[i];
            }
            else {
                /* 
                 * If unsatisfied lock requests should not appear repeatedly
                 * in trace then make repeated locks invisible to the trace 
                 * producer.
                 * This pertains only to the case when lock requests are left 
                 * to the scheduler to handle them - it is realized through 
                 * the use of PROCESSOR_LOCK_BUSY processor responses and 
                 * THREAD_IO thread states.
                 * If memory simulator handles lock requests itself then there 
                 * can be no repeated locks in trace anyway!
                 */
                for (i=0; i<sys_thread_count; i++)
                    if ((sys_threads[i].thread_state == THREAD_IO) && 
                        (sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i] == time_now))
                        sys_threads[i].memory_request.requested_now = 0;

                produce_trace(
                    Produced_TRACEfilename,
                    Produced_TRACEfile,
                    Produced_trace_file_type,
                    time_now
                ); /* from utilities.h */

                /* 
                 * Yet make sure that requests are visible to the memory 
                 * simulator.
                 * This pertains only to the case when lock requests are left 
                 * to the scheduler to handle them, which is explained above.
                 */
                for (i=0; i<sys_thread_count; i++)
                    if ((sys_threads[i].thread_state == THREAD_IO) &&
                        (sys_threads[i].thread_time+time_diff[i]+busy_time[i]+startup_time[i] == time_now))
                        sys_threads[i].memory_request.requested_now = 1;
            }

        /* Simulate events */
        memory_simulate(last_simulation_time,time_now);
        last_simulation_time = time_now+1;

        /* Handle barrier */
        for (i=0; i<sys_thread_count; i++)
            switch (sys_threads[i].global_operation.type) {
                case GLOBAL_CREATED:
                     break;
                case GLOBAL_BARRIER_ENTER:
                     /* See if thread has gained a lock to report itself to a barrier */
                     if (sys_threads[i].memory_request.satisfied == PROCESSOR_SATISFIED) {
                         /* Report thread IN */
                         thread_in_barrier[i] = 1;
                         /* Get barrier data */
                         get_barrier (i, sys_threads[i].thread_time, &barrier[i]);
                         /* Find matching thread and barrier */
                         find_match_IN (i,
                             sys_threads[i].global_operation.params.barrier.virtual_addr,
                             read_barrier(sys_threads[i].global_operation.params.barrier.virtual_addr)+1,
                             barrier[i]->ENTER_time,
                             sys_thread_count,
                             &thread[i],
                             &match_barrier_IN[i]);
                         current_pos_in_barrier[i] = match_barrier_IN[i]->ENTER_pos;
                         /* Calculate time difference */
                         time_diff[i] += barrier[i]->IN_time - match_barrier_IN[i]->IN_time;
                     }
                     break;
                case GLOBAL_BARRIER_IN:
                     if (sys_threads[i].memory_request.satisfied == PROCESSOR_SATISFIED) {
                         /* Number of threads in barrier is increased */
                         write_barrier (sys_threads[i].global_operation.params.barrier.virtual_addr, 
                             read_barrier(sys_threads[i].global_operation.params.barrier.virtual_addr)+1);
                     }
                     break;
                case GLOBAL_BARRIER_WAIT:
                     /* See if thread has gained a lock to go out */
                     if (sys_threads[i].memory_request.satisfied == PROCESSOR_SATISFIED) {
                         /* Find matching thread and position in matching thread */
                         find_match_OUT (i,
                             sys_threads[i].global_operation.params.barrier.virtual_addr,
                             sys_threads[i].global_operation.params.barrier.total-
                                 read_barrier(sys_threads[i].global_operation.params.barrier.virtual_addr)+1,
                             barrier[i]->ENTER_time,
                             sys_thread_count,
                             &thread[i],
                             &match_barrier_OUT[i]);
                         current_pos_in_barrier[i] = match_barrier_OUT[i]->WAIT_pos;
                         /* Recalculate time difference */
                         time_diff[i] += match_barrier_IN[i]->OUT_time - match_barrier_OUT[i]->OUT_time;
                     }
                     break;
                case GLOBAL_BARRIER_OUT:
                     if (sys_threads[i].memory_request.satisfied == PROCESSOR_SATISFIED) {
                         /* Number of threads in barrier is decreased */
                         write_barrier (sys_threads[i].global_operation.params.barrier.virtual_addr, 
                             read_barrier(sys_threads[i].global_operation.params.barrier.virtual_addr)-1);
                     }
                     break;
                case GLOBAL_BARRIER_EXIT:
                     if ((sys_threads[i].memory_request.satisfied == PROCESSOR_SATISFIED) &&
                         (thread_in_barrier[i])) 
                     {
                         /* Report exiting barrier */
                         thread_in_barrier[i] = 0;
                         /* Recalculate time difference */
                         if (match_barrier_IN[i]->OUT_order == 1)
                             time_diff[i] += match_barrier_IN[i]->EXIT_time - barrier[i]->EXIT_time;
                         else
                             time_diff[i] += match_barrier_OUT[i]->EXIT_time - barrier[i]->EXIT_time; 
                         /* Continue execution of requests after barrier */
                         thread[i] = i;
                         /* Move to the position after barrier exit */
                         lseek (P_TRACEfile[i], barrier[i]->EXIT_pos, 0);
                     }
                     break;
                case GLOBAL_NONE:
                     break;
                default:
                     fprintf (stderr, "\nError - Unknown global operation type: %d\n", sys_threads[i].global_operation.type);
                     exit (1);
            }

        /* Read next request. Also mark FINISHED threads if any */
        for (i=0; i<sys_thread_count; i++)
            if (sys_threads[i].memory_request.satisfied == PROCESSOR_SATISFIED) {
                if (thread_in_barrier[i]) {
                    current_trace_pos[thread[i]] = tell(P_TRACEfile[thread[i]]);
                    lseek (P_TRACEfile[thread[i]], current_pos_in_barrier[i], 0);
                }
                ind = read(P_TRACEfile[thread[i]], &event, sizeof(event));
                if (ind == -1) {
                    fprintf (stderr, "\nError reading temporary trace file %s\n", P_TRACEfilename[thread[i]]);
                    exit (1);
                }
                if (ind == 0) {
                    finished_threads[i] = 1;
                    ++finished_thread_count;
                }
                else {
                    sys_threads[i].thread_time = event.time;
                    sys_threads[i].memory_request.requested_now = 0;
                    sys_threads[i].memory_request.type = (enum Memory_Request_Type) event.request;
                    sys_threads[i].memory_request.virtual_addr = event.address;
                    sys_threads[i].memory_request.pid = sys_threads[i].proc_ID;
                    sys_threads[i].memory_request.size = event.size;
                    sys_threads[i].memory_request.area = event.area;
                    sys_threads[i].memory_request.isshared = (event.area == 'd');
                    if (event.global != GLOBAL_NONE) {
                        ind = read(P_TRACEfile[thread[i]], &global, sizeof(global));
                        if (ind == -1) {
                            fprintf (stderr, "\nError reading temporary trace file %s\n", P_TRACEfilename[thread[i]]);
                            exit (1);
                        }
                        if (ind == 0) {
                            finished_threads[i] = 1;
                            ++finished_thread_count;
                        }
                        else {
                            sys_threads[i].global_operation.type = (enum Global_Synchro_Type) event.global;
                            switch (sys_threads[i].global_operation.type) {
                                case GLOBAL_CREATED:
                                     sys_threads[i].global_operation.params.created.parent = global.created.parent;
                                     sys_threads[i].global_operation.params.created.time = global.created.time;
                                case GLOBAL_BARRIER_ENTER:
                                     sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                     sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                     break;
                                case GLOBAL_BARRIER_IN:
                                     sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                     sys_threads[i].global_operation.params.barrier.order = global.barrier.order;
                                     sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                     break;
                                case GLOBAL_BARRIER_WAIT:
                                     sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                     sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                     break;
                                case GLOBAL_BARRIER_OUT:
                                     sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                     sys_threads[i].global_operation.params.barrier.order = global.barrier.order;
                                     sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                     break;
                                case GLOBAL_BARRIER_EXIT:
                                     sys_threads[i].global_operation.params.barrier.virtual_addr = global.barrier.address;
                                     sys_threads[i].global_operation.params.barrier.total = global.barrier.total;
                                     break;
                                default:
                                     fprintf (stderr, "\nError - Unknown global operation type %d in temporary trace file %s\n", 
                                              sys_threads[i].global_operation.type, P_TRACEfilename[thread[i]]);
                                     exit (1);
                            }
                        }
                    }
                    else 
                        sys_threads[i].global_operation.type = GLOBAL_NONE;
                }
                if (thread_in_barrier[i]) {
                    current_pos_in_barrier[i] = tell(P_TRACEfile[thread[i]]);
                    lseek (P_TRACEfile[thread[i]], current_trace_pos[thread[i]], 0);
                }                
            }

        /* 
         * Now we collect informations which of the requests were satisfied 
         * (the candidates are either those who issued a request in this cycle,
         * or who were stalled, or wait for lock). 
         */
        for (i=0; i<sys_thread_count; i++)
            if (sys_threads[i].memory_request.requested_now ||
                (sys_threads[i].thread_state == THREAD_STALLED) ||
                (sys_threads[i].thread_state == THREAD_IO)) 
            {
                sys_threads[i].memory_request.requested_now = 0;
                if (sys_threads[i].thread_state == THREAD_IO)
                    ++current_lock_busy_time[i];
                switch (sys_threads[i].memory_request.satisfied) {
                    case PROCESSOR_SATISFIED:
                         if (sys_threads[i].thread_state == THREAD_IO) {
                             if (current_lock_busy_time[i] > 1)
                                 busy_time[i] += -(BUSY_LOCK_SLEEP_TIME+1)+current_lock_busy_time[i];
                             repeating_global[i] = 0;
                         }
                         if (finished_threads[i])
                             sys_threads[i].thread_state = THREAD_FINISHED;
                         else
                             sys_threads[i].thread_state = THREAD_READY;
                         break;
                    case PROCESSOR_LOCK_BUSY:
                         sys_threads[i].thread_state = THREAD_IO;
                         repeating_global[i] =1;
                         current_lock_busy_time[i] = 0;
                         busy_time[i] += BUSY_LOCK_SLEEP_TIME+1;
                         break;
                    case PROCESSOR_STALLED:
                         sys_threads[i].thread_state = THREAD_STALLED;
                         ++busy_time[i];
                         break;
                    default:
                         break;
                }
            }
            
        /*
         *  Update progress meter and timer
         */
        if (progress_meter) {
            ++progress_counter;
            if (progress_counter%PROGRESS_STEP == 0) {
                current_length = 0;
                for (i=0; i<MAX_CPU_COUNT; i++)
                    current_length += tell(P_TRACEfile[i]);
                printf ("Running simulation  -  Progress: %5.1f%%  Time: %lu seconds\r", 
                        ((float)current_length)/total_length*100,
                        time(NULL)-simulation_start_time);
                fflush(stdout);
            }
        }

    }

    /* Update final report */
    if (progress_meter) {
        printf ("Running simulation  -  Progress: %5.1f%%  Time: %lu seconds\n", 
                100.0,
                time(NULL)-simulation_start_time);
        fflush(stdout);
    }

    /* Close and remove temporary 'per processor' trace files */
    for (i=0; i<MAX_CPU_COUNT; i++) {
        if (close (P_TRACEfile[i]) == -1) {
            fprintf (stderr, "\nError closing temporary trace file %s\n", P_TRACEfilename[i]);
            exit (1);
        }
        if (unlink (P_TRACEfilename[i])) {
            fprintf (stderr, "\nError deleting temporary trace file %s\n", P_TRACEfilename[i]);
            exit (1);
        }
    }
    /* Close produced trace file if any */
    if (Produced_trace_file_type)
        switch (Produced_trace_file_type) {
            case 1 :
                 if (fclose (Produced_TRACEfile.txt)) {
                     fprintf (stderr, "\nError closing produced trace file %s\n", Produced_TRACEfilename);
                     exit (1);
                 }
                 break;
            case 2 : 
                 if (close (Produced_TRACEfile.bin) == -1) {
                     fprintf (stderr, "\nError closing produced trace file %s\n", Produced_TRACEfilename);
                     exit (1);
                 }
                 break;
        }

    /* End */
    printf("--------------------------------------------------------------------------------\n");
    printf("                                 Simulation end                                 \n");
    memsim_cleanup();
    printf("\nOverall time = %lu seconds\n", time(NULL)-tracer_start_time);
    printf("********************************************************************************\n");

}


/*
 *  This procedure does the same with binary trace files as trace.lex does 
 *  with text traces - makes a separate trace file for each processor.
 */
static void make_proctraces_from_bin_file(int Tracefile) {
    int i;
    char proc;
    short int next, current;
    Thread_Time trace_time;
    enum Memory_Request_Type prev_request [MAX_CPU_COUNT];
    bitsADDR prev_address [MAX_CPU_COUNT];
    int      same_lock_request [MAX_CPU_COUNT];
    Thread_Time lock_time_start [MAX_CPU_COUNT];
    Thread_Time lock_time [MAX_CPU_COUNT];
    Thread_Time lock_busy_time [MAX_CPU_COUNT];    
    /* Progress meter params */
    char progress_meter = 1;
    int progress_counter = 0;
    long int total_length;
    time_t parsing_start_time;

    /* Initialize progress meter parmeters */
    lseek (Tracefile,0,2); /* goto end of file */
    total_length = tell (Tracefile);
    lseek (Tracefile,0,0); /* goto beginning of file */
    parsing_start_time = time(NULL);
 
    /* Initialize barrier table and other variables */
    for (i=0; i<MAX_CPU_COUNT; i++) {
        barrier_data[i].list_head = barrier_data[i].list_tail = NULL;
        same_lock_request[i] = 0;
        prev_request[i] = REQUEST_NONE;
        lock_busy_time[i] = 0;
    }

    while ((read(Tracefile, &next, sizeof(next)) != 0) &&
           (read(Tracefile, &trace_time, sizeof(trace_time)) != 0))
    {
        current = sizeof(next)+sizeof(trace_time);
        while (current<next) {
            if ((read(Tracefile, &proc, sizeof(proc)) == -1) ||
                (read(Tracefile, &(event.request), sizeof(event.request)) == -1) ||
                (read(Tracefile, &(event.area), sizeof(event.area)) == -1) ||
                (read(Tracefile, &(event.address), sizeof(event.address)) == -1) ||
                (read(Tracefile, &(event.size), sizeof(event.size)) == -1) ||
                (read(Tracefile, &(event.global), sizeof(event.global)) == -1))
            {   
                fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                exit (1);
            }
            current = current+sizeof(proc)+sizeof(event.request)+
                              sizeof(event.area)+sizeof(event.address)+
                              sizeof(event.size)+sizeof(event.global);
            switch ((enum Global_Operation_Type) (event.global)) {
                case GLOBAL_CREATED:
                     if ((read(Tracefile, &(global.created.parent), sizeof(global.created.parent)) == -1) ||
                         (read(Tracefile, &(global.created.time), sizeof(global.created.time)) == -1))
                     {
                         fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                         exit (1);
                     }
                     current = current+sizeof(global.created.parent)+
                                       sizeof(global.created.time);
                     break;                
                case GLOBAL_BARRIER_ENTER:
                     if ((read(Tracefile, &(global.barrier.address), sizeof(global.barrier.address)) == -1) ||
                         (read(Tracefile, &(global.barrier.total), sizeof(global.barrier.total)) == -1))
                     {   
                         fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                         exit (1);
                     }
                     current = current+sizeof(global.barrier.address)+
                                       sizeof(global.barrier.total);
                     break;
                case GLOBAL_BARRIER_IN:
                     if ((read(Tracefile, &(global.barrier.address), sizeof(global.barrier.address)) == -1) ||
                         (read(Tracefile, &(global.barrier.order), sizeof(global.barrier.order)) == -1) ||
                         (read(Tracefile, &(global.barrier.total), sizeof(global.barrier.total)) == -1))
                     {   
                         fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                         exit (1);
                     }
                     current = current+sizeof(global.barrier.address)+
                                       sizeof(global.barrier.order)+
                                       sizeof(global.barrier.total);
                     break;
                case GLOBAL_BARRIER_WAIT:
                     if ((read(Tracefile, &(global.barrier.address), sizeof(global.barrier.address)) == -1) ||
                         (read(Tracefile, &(global.barrier.total), sizeof(global.barrier.total)) == -1))
                     {   
                         fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                         exit (1);
                     }
                     current = current+sizeof(global.barrier.address)+
                                       sizeof(global.barrier.total);
                     break;
                case GLOBAL_BARRIER_OUT:
                     if ((read(Tracefile, &(global.barrier.address), sizeof(global.barrier.address)) == -1) ||
                         (read(Tracefile, &(global.barrier.order), sizeof(global.barrier.order)) == -1) ||
                         (read(Tracefile, &(global.barrier.total), sizeof(global.barrier.total)) == -1))
                     {   
                         fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                         exit (1);
                     }
                     current = current+sizeof(global.barrier.address)+
                                       sizeof(global.barrier.order)+
                                       sizeof(global.barrier.total);
                     break;
                case GLOBAL_BARRIER_EXIT:
                     if ((read(Tracefile, &(global.barrier.address), sizeof(global.barrier.address)) == -1) ||
                         (read(Tracefile, &(global.barrier.total), sizeof(global.barrier.total)) == -1))
                     {   
                         fprintf(stderr, "\nError reading from trace file %s\n", TRACEfilename);
                         exit (1);
                     }
                     current = current+sizeof(global.barrier.address)+
                                       sizeof(global.barrier.total);
                     break;
                case GLOBAL_NONE:
                     break;
                default:
                     fprintf (stderr, "\nError in trace - Unknown global operation type: %d\n",event.global);
                     exit (1);
            }

            /* 
             * Write request in file, but don't write repeated lock requests 
             */
            if ((prev_request[proc] == REQUEST_LOCK) &&
                (event.request == REQUEST_LOCK) &&
                (event.address == prev_address[proc]))
            {
                same_lock_request[proc] = 1;
                lock_time[proc] = trace_time;
            }
            else {
                if (same_lock_request[proc]) {
                    lock_busy_time[proc] += lock_time[proc] - lock_time_start[proc];
                    same_lock_request[proc] = 0;
                }
                if (event.request == REQUEST_LOCK) lock_time_start[proc] = trace_time;
                event.time = trace_time - lock_busy_time[proc];
                if (write(P_TRACEfile[proc], &event, sizeof(event)) == -1) {
                    fprintf (stderr, "\nError writing to trace file %s\n", P_TRACEfilename[proc]);
                    exit (1);
                }
                if (event.global != GLOBAL_NONE) {
                    if (write(P_TRACEfile[proc], &global, sizeof(global)) == -1) {
                        fprintf (stderr, "\nError writing to trace file %s\n", P_TRACEfilename[proc]);
                        exit (1);
                    }
                    switch (event.global) {
                        case GLOBAL_CREATED:
                             break;
                        case GLOBAL_BARRIER_ENTER:
                             write_barrier_ENTER_data (proc, global.barrier.address, global.barrier.total, event.time, tell(P_TRACEfile[proc]));
                             break;
                        case GLOBAL_BARRIER_IN:
                             write_barrier_IN_data (proc, global.barrier.order, event.time, event.time+lock_busy_time[proc]);
                             break;
                        case GLOBAL_BARRIER_WAIT:
                             write_barrier_WAIT_data (proc, event.time, tell(P_TRACEfile[proc]));
                             break;
                        case GLOBAL_BARRIER_OUT:
                             write_barrier_OUT_data (proc, global.barrier.order, event.time, event.time+lock_busy_time[proc]);
                             break;
                        case GLOBAL_BARRIER_EXIT:
                             write_barrier_EXIT_data (proc, event.time, tell(P_TRACEfile[proc]));
                             break;
                        default:
                             fprintf (stderr, "\nError - Unknown global operation type: %d\n", event.global);
                             exit (1);
                    }
                }
                prev_request[proc] = event.request;
                prev_address[proc] = event.address;
            }

            /*
             *  Update progress meter and timer
             */
            if (progress_meter) {
                ++progress_counter;            
                if ((progress_counter%PROGRESS_STEP == 0) ||
                    (tell(Tracefile) == total_length))
                {
                    printf ("Parsing trace file  -  Progress: %5.1f%%  Time: %lu seconds\r", 
                            ((float)tell(Tracefile))/total_length*100,
                            time(NULL)-parsing_start_time);
                    fflush(stdout);
                }
            }
        }
    }
}

void sys_simulation_end (int exitcode) {
    exit(exitcode);
}
